import sys

sys.path.insert(0, "../../azure")
from azure.storage.v2.sharedaccesssignature import BlobSharedAccessSignature

_ERROR_VALUE_NONE = '{0} should not be None.'
	
def _validate_not_none(param_name, param):
    if param is None:
        raise ValueError(_ERROR_VALUE_NONE.format(param_name))

def generate_blob_shared_access_signature(
    account_name, account_key, container_name, blob_name, snapshot=None, permission=None,
    expiry=None, start=None, id=None, ip=None, protocol=None,
    cache_control=None, content_disposition=None,
    content_encoding=None, content_language=None,
    content_type=None):
    '''
    Generates a shared access signature for the blob or one of its snapshots.
    Use the returned signature with the sas_token parameter of any BlobService.

    :param str account_name:
        Name of the storage account.
    :param str account_key:
        Key for the storage account.
    :param str container_name:
        Name of container.
    :param str blob_name:
        Name of blob.
    :param str snapshot:
        The snapshot parameter is an opaque DateTime value that,
        when present, specifies the blob snapshot to grant permission.
    :param BlobPermissions permission:
        The permissions associated with the shared access signature. The 
        user is restricted to operations allowed by the permissions.
        Permissions must be ordered read, write, delete, list.
        Required unless an id is given referencing a stored access policy 
        which contains this field. This field must be omitted if it has been 
        specified in an associated stored access policy.
    :param expiry:
        The time at which the shared access signature becomes invalid. 
        Required unless an id is given referencing a stored access policy 
        which contains this field. This field must be omitted if it has 
        been specified in an associated stored access policy. Azure will always 
        convert values to UTC. If a date is passed in without timezone info, it 
        is assumed to be UTC.
    :type expiry: datetime or str
    :param start:
        The time at which the shared access signature becomes valid. If 
        omitted, start time for this call is assumed to be the time when the 
        storage service receives the request. Azure will always convert values 
        to UTC. If a date is passed in without timezone info, it is assumed to 
        be UTC.
    :type start: datetime or str
    :param str id:
        A unique value up to 64 characters in length that correlates to a 
        stored access policy. To create a stored access policy, use :func:`~set_container_acl`.
    :param str ip:
        Specifies an IP address or a range of IP addresses from which to accept requests.
        If the IP address from which the request originates does not match the IP address
        or address range specified on the SAS token, the request is not authenticated.
        For example, specifying sip=168.1.5.65 or sip=168.1.5.60-168.1.5.70 on the SAS
        restricts the request to those IP addresses.
    :param str protocol:
        Specifies the protocol permitted for a request made. The default value
        is https,http. See :class:`~azure.storage.common.models.Protocol` for possible values.
    :param str cache_control:
        Response header value for Cache-Control when resource is accessed
        using this shared access signature.
    :param str content_disposition:
        Response header value for Content-Disposition when resource is accessed
        using this shared access signature.
    :param str content_encoding:
        Response header value for Content-Encoding when resource is accessed
        using this shared access signature.
    :param str content_language:
        Response header value for Content-Language when resource is accessed
        using this shared access signature.
    :param str content_type:
        Response header value for Content-Type when resource is accessed
        using this shared access signature.
    :param ~azure.storage.blob.models.UserDelegationKey user_delegation_key:
        Instead of an account key, the user could pass in a user delegation key.
        A user delegation key can be obtained from the service by authenticating with an AAD identity;
        this can be accomplished by calling get_user_delegation_key.
        When present, the SAS is signed with the user delegation key instead.
    :return: A Shared Access Signature (sas) token.
    :rtype: str
    '''
    _validate_not_none('container_name', container_name)
    _validate_not_none('blob_name', blob_name)
    _validate_not_none('account_name', account_name)
    _validate_not_none('account_key', account_key)
    sas = BlobSharedAccessSignature(account_name=account_name, account_key=account_key)

    return sas.generate_blob(
        container_name=container_name,
        blob_name=blob_name,
        snapshot=snapshot,
        permission=permission,
        expiry=expiry,
        start=start,
        id=id,
        ip=ip,
        protocol=protocol,
        cache_control=cache_control,
        content_disposition=content_disposition,
        content_encoding=content_encoding,
        content_language=content_language,
        content_type=content_type,
    )