#!/usr/bin/env python


#-------------------------------------------------------------------------
# Copyright (c) Microsoft.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#--------------------------------------------------------------------------
MYPY = False
if MYPY:
    from typing import Dict
import logging
import os
import sys
import traceback
import getopt
import shutil
import time
from datetime import datetime as dt
import json
from datetime import timedelta


sys.path.insert(0, os.path.abspath('.'))
sys.path.insert(1, 'lib')
logger = logging.getLogger()

from azure.storage.v2 import blobsharedaccesssignature as blobsas
import distro
from azure.storage import TableService, BlobService

from util import autoupdate
from util import common
from util.SystemUtil import CpuUtil, SystemUtil, DiskUtil, MemoryUtil
from ruleengine import RuleEngine
from ruleengine import DataCollector

from util import storage
from util.storage import RunJson, FindingsJson, ImpactedResourcesJson
from lib import applicationinsights

SHARE_DIAGNOSTICS_QUESTION = "Agree to share diagnostic information? [https://azure.microsoft.com/en-gb/support/trust-center/privacy/support-diagnostic-information-collection/]"
global_info_json_dict = {}  # type: Dict[str, str]
global_perf_insight_init = False

def python_version():
    if len(sys.version_info) < 1:
        version = '0.0'
    elif len(sys.version_info) < 2:
        version = str(sys.version_info[0])
    else:
        version = str(sys.version_info[0]) + '.' + str(sys.version_info[1])

    return version

def send_telemetry(name=None, custom_properties=None, error_info=None, isUnhandled=None):
    logger.debug("Entering send_telemetry")

    # If running from Python 2.7+ (application insights not available for Python < 2.7)
    if float(python_version()) >= common.MINIMUM_SUPPORTED_PYTHON_VERSION_FOR_TELEMETRY:
        from lib.applicationinsights import TelemetryClient, channel

        if common.thisazurevm is not None and common.thisazurevm.get('Compute') is not None:
            telemetry_key, telemetry_endpoint_uri = common.get_applicationinsights_info(common.thisazurevm['Compute']['azEnvironment'])

            sender = channel.AsynchronousSender(service_endpoint_uri=telemetry_endpoint_uri)
            queue = channel.AsynchronousQueue(sender)
            channel = channel.TelemetryChannel(None, queue)

            tc = TelemetryClient(telemetry_key, channel)
            properties = {
                'SubscriptionId': common.thisazurevm['Compute']['subscriptionId'],
                'Publisher': common.thisazurevm['Compute']['publisher'],
                'VMId': common.thisazurevm['Compute']['vmId'],
                'Location': common.thisazurevm['Compute']['location'],
                'LinuxDistroInfo': ' '.join(distro.linux_distribution()),
                'PerfInsightsVersion': common.SCRIPT_VERSION,
                'IsLinux': True,
                'ReportId': common.REPORT_ID,
                'IsUnhandled': "True" if isUnhandled else "False"
            }

            if error_info is None and name is None:
                logger.error("Cannot send telemetry for type 'event' because name was not passed")
                return

            if custom_properties is not None:
                properties.update(custom_properties)

            if error_info is not None:
                # errorinfo: { type, value, traceback }
                try:
                    #In the case Azure operation failed, sometimes the exception message is not in utf-8, track_exception use str() to convert
                    #exception message to string will cause UnicodeEncodeError
                    exception_value = str(error_info[1])
                except UnicodeEncodeError:
                    exception_value = repr(error_info[1]).encode('utf-8')

                tc.track_exception(error_info[0], exception_value, error_info[2], properties)
            else:
                tc.track_event(name, properties)

            tc.flush()
            #Need to wait a bit for event/exception to be sent, otherwise the script might exit before telemetry is sent
            time.sleep(1)
            logger.debug("Telemetry sent.")
    else:
        logger.warning("Telemetry can't be sent as running from Python < 2.7")


def log_unhandled_exception(exc_type, exc_value, exc_traceback):
    if issubclass(exc_type, KeyboardInterrupt):
        sys.__excepthook__(exc_type, exc_value, exc_traceback)
        return

    try:
        logger.error("Uncaught exception", exc_info=(exc_type, exc_value, exc_traceback))
        traceback.print_exc()
        if global_perf_insight_init == False:
            # Get Azure VM information for send_telemetry
            dc = DataCollector.InstanceMetadataInfoDC()
            common.thisazurevm = dc.to_json()
            if common.thisazurevm:
                common.thisazurevm = common.thisazurevm['CollectedData'][0]['Data']

        exception_info = (exc_type, exc_value, exc_traceback)
        send_telemetry(name=None, custom_properties=None, error_info=exception_info, isUnhandled=True)
    except:
        sys.__excepthook__(exc_type, exc_value, exc_traceback)
        return

def display_disclaimer():
    if common.bypass_agreements is False:
        print(common.GREEN)
        print("This tool will collect useful diagnostic information (such as configuration and log files etc.) for troubleshooting IaaS performance scenarios in Microsoft Azure." )
        print("This tool will not be performing any downloads and installations of any modules/packages on this machine.")
        print("This tool will create a tarfile that includes the collected log files and findings that may be uploaded to Microsoft support automatically or by you directly.")
        print("The data uploaded to Microsoft will be stored by Microsoft for a maximum of 90 days after your issue is resolved and will be used and retained consistent with the standards set forth at the Microsoft Trust Center(https://www.microsoft.com/en-us/trustcenter/cloudservices/azure).")
        print(common.RESET)


def ask_for_consent(question):
    answer = ''
    try:
        while not(answer == "y" or answer == "yes" or \
            answer == "n" or answer == "no"):
            if float(python_version()) < 3:
                answer = raw_input(question + " (y/n): ").lower().strip()
            else:
                answer = input(question + " (y/n): ").lower().strip()
    except EOFError as e:
        logger.error("Received EOFError. Defaulting to no consent and continuing: " + str(e))
        answer="n"
    if (answer[0] == "y"):
        return True
    else:
        return False


def check_requirements():

    logger.debug("Entering check_requirements")

    # Python version
    if float(python_version()) < common.MINIMUM_SUPPORTED_PYTHON_VERSION:
        logger.error("Python version: " + python_version() + " is not supported with this version of PerfInsights")
        sys.exit(1)
    else:
        logger.info("Python version: " + python_version()  + " is supported.")

    # Get guest OS distribution information
    dc = DataCollector.MachineMapInfoDC()
    os_info = dc.to_json()
    if os_info:
        common.linux_distribution = os_info['CollectedData'][0]['Data']['os']["name"]
        common.linux_version = os_info['CollectedData'][0]['Data']['os']["version"]
        common.linux_id_like = os_info['CollectedData'][0]['Data']['os'].get("id_like", " ")
        common.linux_id = os_info['CollectedData'][0]['Data']['os'].get("id", " ".join(distro.linux_distribution()))
        common.linux_version_id = os_info['CollectedData'][0]['Data']['os'].get('version_id', distro.linux_distribution()[1])

    if common.linux_distribution not in common.PIL_SUPPORTED_LINUX_DISTRIBUTIONS:
        logger.error("Linux distribution: " + common.linux_distribution + " is not supported by PerfInsights.")
        # sys.exit(1)
    else:
        logger.info("Linux distribution: " + common.linux_distribution + " is supported.")


def parse_script_arguments(arg_args):
    """
    Function
    =========
    parse_script_arguments

    Description
    ==============
    Parses the parameters provided to the script in order to verify they are valid and update global variables.

    Arguments
    ==========
        __in <dic> arg_args : script argument dictionnay

    Return Values
    ==============

        N/A
    """

    commandline = ""

    try:
        _options, arg_args = getopt.getopt(sys.argv[1:], "hs:r:o:d:aqyz:t:k:i:u", ["help", "support-request=", "run=", "output-folder=", "duration=", "accept-disclaimer-and-share-diagnostics", "bypass-questions", "debug", "feature-debug=", "storage-account=", "storage-account-key=", "resource-id=", "skip-auto-update"])
    except getopt.GetoptError:
        print("One or more command line options passed are invalid. Review help, fix the issue and try again.")
        common.show_help()
        sys.exit(2)

    # Read values from environment that may have been passed by the VM Extension
    common.storage_account_key = os.environ.get("SAKEY","")
    common.user_agrees_to_share_data = os.environ.get("USER_AGREES_TO_SHARE_DATA", False)

    if len(_options) > 0:
        for _option, _argument in _options:
            if _option in ("-h", "--help"):
                common.show_help()
                sys.exit(0)
            elif _option in ("-s", "--support-request"):
                common.support_request_number = _argument
                # Verify input matches MSSolve case number
                commandline = commandline + _option + " " + common.support_request_number + " "
            elif _option in ("-r", "--run"):
                common.perfinsights_scenario = _argument.lower()
                if common.perfinsights_scenario not in [scenario["Name"] for scenario in common.POSSIBLE_SCENARIOS if scenario["Implemented"] == "Available"]:
                    print("\nPlease specify a valid scenario. e.g. quick or vmslow")
                    common.show_help()
                    sys.exit(1)
                commandline = commandline + _option + " " + common.perfinsights_scenario + " "
            elif _option in ("-o", "--outputFolder"):
                if _argument == './':
                    common.output_folder = os.path.dirname(os.path.realpath(__file__))
                else:
                    common.output_folder = _argument
                commandline = commandline + _option + " " + common.output_folder + " "
            elif _option in ("-d", "--duration"):
                common.capture_duration_sec = _argument
                if _argument[len(_argument)-1:len(_argument)].upper() == 'H':
                    common.capture_duration_sec = int(_argument[0:(len(_argument) - 1)]) * 60 * 60
                elif _argument[len(_argument)-1:len(_argument)].upper() == 'M':
                    common.capture_duration_sec = int(_argument[0:(len(_argument) - 1)]) * 60
                elif _argument[len(_argument)-1:len(_argument)].upper() == 'S':
                    common.capture_duration_sec = int(_argument[0:(len(_argument) - 1)])
                else:
                    print("Please specify a valid time to run value in hour(s) (H), minute(s) (M) or second(s) (S) with -d/--duration.")
                    common.show_help()
                    sys.exit(1)
                commandline = commandline + _option + " " + str(common.capture_duration_sec) + "s "
            elif _option in ("-a", "--accept-disclaimer-and-share-diagnostics"):
                common.bypass_agreements = True
                common.user_agrees_to_share_data = True
                commandline = commandline + _option + " "
            elif _option in ("-q", "--bypass-questions"):
                common.bypass_questions = 'y'
                commandline = commandline + _option + " " + common.bypass_questions + " "
            elif _option in ("-y", "--debug"):
                common.debug_enabled = True
                commandline = commandline + _option + " "
            elif _option in ("-t", "--storage-account"):
                common.storage_account = _argument
                commandline = commandline + _option + ' ' + _argument + ' '
            elif _option in ("-k", "--storage-account-key"):
                common.storage_account_key = _argument.encode('unicode_escape').decode('utf-8')
                commandline = commandline + _option + ' ' + '**REDACTED**' + ' '
            elif _option in ("-i", "--resource-id"):
                common.resource_id = _argument.encode('unicode_escape').decode('utf-8')
                commandline = commandline + _option + ' ' + common.resource_id + ' '
            elif _option in ("-u", "--skip-auto-update"):
                common.auto_update = False
                commandline = commandline + _option + " "
        if (common.storage_account != "" or common.storage_account_key != "" or common.resource_id != "") \
            and (common.storage_account == "" or common.storage_account_key == "" or common.resource_id == ""):
            print("One or more of the parameters (storage-account, storage-account-key or resource-id) are not supplied. When any one of these parameters are supplied, all of them must be provided.")
            common.show_help()
            sys.exit(1)

        if common.storage_account != "" and common.storage_account_key != "" and common.resource_id != "":
            common.upload_to_storage = True

    else:
        common.show_help()
        sys.exit(1)

    common.script_arguments = commandline.lstrip().rstrip()    
    logger.info("Script arguments: " + common.script_arguments)


def create_output_folders():
    # Create the root output folder
    if not os.path.exists(common.output_folder):
        try:
            os.mkdir(common.output_folder)
        except OSError:  
            print("Creation of the output folder %s failed" % common.output_folder)
            sys.exit(1)
        else:  
            print("Successfully created the output folder %s " % common.output_folder)

    #overwrite the output folder value to reflect the Performance Diagnostics report folder
    common.output_folder = common.output_folder + "/PerformanceDiagnostics_" + common.get_current_datetime("file")
    if not os.path.exists(common.output_folder):
        os.makedirs(common.output_folder)

    # create diagnostics output and guest logs folders
    common.diag_output_folder_path = common.output_folder + common.PIFL_DIAG_OUTPUT_FOLDER_NAME
    if not os.path.exists(common.diag_output_folder_path):
        os.makedirs(common.diag_output_folder_path)
        common.shared_guestfiles_output_folder_path = common.diag_output_folder_path + common.PIFL_GUEST_LOG_FOLDER_NAME
        os.makedirs(common.shared_guestfiles_output_folder_path)

    # create perfinsights log folder
    common.log_folder_path = common.output_folder + common.PIFL_LOG_FOLDER_NAME
    if not os.path.exists(common.log_folder_path):
        os.makedirs(common.log_folder_path)
    common.pifl_log_file_path = common.log_folder_path + "/" + common.PIFL_LOG_FILE_NAME


def init_storage():
    """
    Function
    =========
    init_storage

    Description
    ==============
    Initialize Storage: set up the TableService account to be used with the CosmosDB API. Additionally, make sure that the tables & containers
        that we need to use do exist.

    Arguments
    ==========
        N/A

    Return Values
    ==============

        N/A
    """

    if common.upload_to_storage:
        az_env = '[unknown]'
        if common.thisazurevm is not None and common.thisazurevm.get('Compute') is not None:
            az_env = common.thisazurevm['Compute'].get('azEnvironment')
        storage_account_name, storage_endpoint = common.get_storage_endpoint_info(common.storage_account, az_env)
        blob_service_endpoint = ".blob" + storage_endpoint
        table_service_endpoint = ".table" + storage_endpoint
        logger.info("StorageAccount=" + storage_account_name + ", BlobEndpoint=" + blob_service_endpoint + ", TableEndpoint=" + table_service_endpoint)
        try:
            common.table_service = TableService(account_name=storage_account_name, account_key=common.storage_account_key, host_base=table_service_endpoint)
        except Exception as e:
            logger.error("Unable to connect to table service for storage account: " + common.storage_account + table_service_endpoint)
            logger.error(type(e).__name__)
            logger.error(e.message)
            sys.exit(2)

        try:
            common.block_blob_service = BlobService(account_name=storage_account_name, account_key=common.storage_account_key, host_base=blob_service_endpoint) 
        except Exception as e:
            logger.error("Unable to connect to block blob service for storage account: " + common.storage_account + blob_service_endpoint)
            logger.error(type(e).__name__)
            logger.error(e.message)
            sys.exit(2)

        # Create necessary containers if they do not exist
        common.block_blob_service.create_container(common.EXTENSION_RESULT_CONTAINER_NAME, fail_on_exist=False)

        # Create necessary tables if they do not exist
        common.table_service.create_table(common.EXTENSION_RESULT_RUNS_TABLE_NAME, fail_on_exist=False)
        common.table_service.create_table(common.EXTENSION_RESULT_FINDINGS_TABLE_NAME, fail_on_exist=False)
        common.table_service.create_table(common.EXTENSION_RESULT_IMPACTED_RESOURCES_TABLE_NAME, fail_on_exist=False)


def init_perfInsights():
    """
    Function
    =========
    init_perfInsights

    Description
    ==============
    Initialize PerfInsights: define global variables (RootFolder_Path, common.Output_Folder_Path, common.Log_File) and create output folders.

    Arguments
    ==========
        N/A

    Return Values
    ==============

        N/A
    """

    logger.debug("Entering init_perfInsights")

    # Get Azure VM information
    dc = DataCollector.InstanceMetadataInfoDC()
    common.thisazurevm = dc.to_json()
    if common.thisazurevm:
        common.thisazurevm = common.thisazurevm['CollectedData'][0]['Data']

    # DataCollector.MachineMapInfoDC().get_system_physicalmemory_info()
    common.total_physical_memory = MemoryUtil.total_physical_memory_kbytes()
    common.available_physical_memory = MemoryUtil.available_physical_memory_kbytes()
    common.free_physical_memory = MemoryUtil.free_physical_memory_kbytes()
    common.temp_drive = DiskUtil.get_azure_temporary_drive()
    common.physical_processors_count = CpuUtil.cpu_count_logical()

    # Load rules configuration
    if common.perfinsights_scenario in [scenario["Name"] for scenario in common.POSSIBLE_SCENARIOS if scenario["Implemented"] == "Available"]:
        rule_config_file = common.PIFL_RULE_CONFIG_FILE
    else:
        raise RuntimeError("Unrecognized Scenario: {0}".format(common.perfinsights_scenario))

    with open(os.path.join(os.path.dirname(RuleEngine.__file__), rule_config_file)) as f:
        common.RULEENGINE_RULES = json.load(f)

    if not common.RULEENGINE_RULES:
        logger.error("Rules configuration were not loaded. PerfInsights for Linux will exit.")
        sys.exit(2)

    # Initalize run ReportId
    import uuid
    common.REPORT_ID = str(uuid.uuid4())
    global_perf_insight_init = True


def display_vm_info():
    """
    Function
    =========
    display_vmInfo

    Description
    ==============
    Display virtual information information

    Arguments
    ==========
        N/A

    Return Values
    ==============

        N/A
    """

    if common.thisazurevm:
        logger.info('')
        logger.info("Virtual machine information")
        logger.info("---------------------------")
        logger.info("\t-Name: " + common.thisazurevm['Compute']['Name'] )
        logger.info("\t-Size: " + common.thisazurevm['Compute']['vmSize'] )
        logger.info("\t-Max IOPS: " + str(common.thisazurevm['VmConfig']['MaxIops']) )
        logger.info("\t-Max throughput (MB/s): " + str(common.thisazurevm['VmConfig']['MaxThroughputMBPS']) )
        logger.info("\t-Storage tier: " + str(common.thisazurevm['VmConfig']['StorageTier']) )
    else:
        logger.warning("Instance metadata for this virtual machine is not available.")

    logger.info("\nGuest OS information")
    logger.info("--------------------")
    logger.info("\t-Linux distribution: " + common.linux_distribution )
    logger.info("\t-Distribution version: " + common.linux_version )
    logger.info("\t-Distribution version id: " + common.linux_version_id )
    logger.info("\t-Processor count: " + str(common.physical_processors_count) )
    logger.info("\t-Physical memory size (GB): " + str(float("{0:.2f}".format(float(common.total_physical_memory)/1024/1024))) )
    logger.info("\t-Available memory size (GB): " + str(float("{0:.2f}".format(float(common.available_physical_memory)/1024/1024))) )
    logger.info("\t-Temporary disk: " + ('' if common.temp_drive is None else common.temp_drive) )
    logger.info('')


def get_script_execution_information(argscriptexecutiontime, argscriptarguments):
    """
    Function
    =========
    get_script_execution_information

    Description
    ==============
    Gather script execution information in JSON-like format

    Arguments
    ==========
        N/A
    
    Return Values
    ==============

        N/A
    """

    vmoverviewobj=[]
    
    logger.debug("Collecting PerfInsights execution context...")

    kernelinfo = SystemUtil.get_kernel_details()
    
    import time

    vmoverviewobj.append({"CommandLine":common.script_arguments,
            "ExecutionTime":str(argscriptexecutiontime),
            "Distribution":common.linux_id_like + " " + common.linux_id,
            "OSVersion":common.linux_distribution + " " + common.linux_version,
            "KernelVersion":kernelinfo["current"],
            "Scenario":common.get_scenario_displayname(common.perfinsights_scenario),
            "ScriptVersion":common.SCRIPT_VERSION,
            "SRNumber":common.support_request_number,
            "TempDrive":common.temp_drive,
            "Hostname":os.uname()[1],
            "StartTimeUTC":common.execution_start_time,
            "EndTimeUTC":common.execution_end_time,
            "TimeZone": time.tzname,
            "ResourceId": common.resource_id,
            "ReportId": common.REPORT_ID})

    if common.thisazurevm:
        vmoverviewobj.append({"VMName":common.thisazurevm['Compute']['Name'], "VMSize":common.thisazurevm['Compute']['vmSize']})
    else:
        # Azure VM is not set
        vmoverviewobj.append({ "VMName":"N/A", "VMSize":"N/A"})

    return vmoverviewobj


def run_scenario():
    """
    Function
    =========
    run_scenario

    Description
    ==============
    Executes scenarios which are used to collect all system configuration (network, system, storage, ...).

    Arguments
    ==========
        N/A    
    Return Values
    ==============
        N/A
    """
    
    global global_info_json_dict

    logger.debug("Entering run_scenario_{0}_analysis".format(common.perfinsights_scenario))

    # Init Rule Engine
    if common.perfinsights_scenario == "vmslow":
        rule_engine = RuleEngine.RuleEngine.build_for_vmslow_analysis()
    elif common.perfinsights_scenario == "quick":
        rule_engine = RuleEngine.RuleEngine.build_for_quick_analysis()
    elif common.perfinsights_scenario == "hpc":
        rule_engine = RuleEngine.RuleEngine.build_for_hpc_scenario()
    else:
        raise RuntimeError("Unrecognized Scenario: {0}".format(common.perfinsights_scenario))

    rule_engine.start(common.RULEENGINE_RULES)
    
    # Stop Rule Engine
    rule_engine.stop()

    # Append results from DC/DMs and rules to final JSON file
    common.create_json(global_info_json_dict, "DataCollectors", rule_engine.get_dc_results())
    common.create_json(global_info_json_dict, "RuleResults", rule_engine.get_rule_results())
    common.create_json(global_info_json_dict, "Configuration", rule_engine.get_rule_configuration())

    logger.info("Performance Analysis completed.")
    logger.info("")


def init_logger():
    log_level = logging.INFO
    if common.debug_enabled:
        log_level = logging.DEBUG

    logging.basicConfig(filename=common.pifl_log_file_path, level=log_level, 
                        format='[%(asctime)s]:[%(levelname)s][%(name)s] %(message)s')

    consoleFormatString = '%(message)s'
    consoleFormatter = logging.Formatter(consoleFormatString)
    consoleHandler = logging.StreamHandler(sys.stdout)
    consoleHandler.setLevel(log_level)
    consoleHandler.setFormatter(consoleFormatter)
    logger.addHandler(consoleHandler)
    logging.Formatter.converter = time.gmtime


def init_file_paths():
    common.current_working_folder = os.getcwd()
    common.output_folder = common.current_working_folder
    common.perfinsights_script_folder = os.path.dirname(__file__)
    if common.perfinsights_script_folder == '':
        common.perfinsights_script_folder = os.getcwd()
    common.vm_sizes_lookup_file_path = os.path.join(os.path.dirname(common.__file__), common.PIFL_VM_SIZE_LOOKUP_FILE_NAME)


def compress_report():
    # Display compressed file location
    
    output_folder = common.output_folder
    output_filename = output_folder + ".tar.gz"
    
    logger.info("")
    logger.info("Please review the report inside " + output_filename + " file and share with Microsoft support engineer if you are working on an active support ticket.")
    common.compress_file(output_folder , output_filename, "")
    if os.path.exists(output_filename):
        if common.debug_enabled != True:
            shutil.rmtree(output_folder, ignore_errors=True)

    return output_filename


def create_report(scriptexecutiontime):
    # Get script execution information and add to common.PIFL_PERFDIAG_FILENAME
    executionjsoncontent = get_script_execution_information(scriptexecutiontime, common.script_arguments)
    for item in executionjsoncontent:
        for key in item.keys():
            common.create_json(global_info_json_dict, key, item[key])
    
    # To fix: jsoncontent = jsoncontent[1:-1]
    jsoncontent = common.create_final_json(global_info_json_dict)
    jsoncontent = jsoncontent[1:-1]
    
    # Uncompress report files
    common.uncompress_file(common.perfinsights_script_folder + '/' + common.PIFL_REPORT_COMPRESSED_FILENAME, common.output_folder , "PerfInsights report files have been extracted")

    # Add information to common.PIFL_PERFDIAG_FILENAME, adding header
    common.output_findings_json(jsoncontent , common.diag_output_folder_path + "/" + common.PIFL_PERFDIAG_FILENAME, True, False)
    
    # Footer to common.PIFL_PERFDIAG_FILENAME
    common.output_findings_json("" , common.diag_output_folder_path + "/" + common.PIFL_PERFDIAG_FILENAME, False, True)


def upload_report_to_storage():
    if common.upload_to_storage == True:
        logger.info("Uploading the PerfInsights report and results to the storage account provided.")
        logger.debug('Uploading result to blob storage')
        filepath = common.output_folder + ".tar.gz"
        split_filepath = filepath.split('/')
        file_to_upload = split_filepath[len(split_filepath) - 1]
        common.result_blob_filename = file_to_upload
        common.block_blob_service.put_block_blob_from_path(common.EXTENSION_RESULT_CONTAINER_NAME, common.result_blob_filename, filepath)        
        blob_url = common.block_blob_service.make_blob_url(common.EXTENSION_RESULT_CONTAINER_NAME, common.result_blob_filename)
        global_info_json_dict['BlobUrl'] = blob_url
        generate_sas_token_for_blob()

        logger.debug('Inserting RunEntity to table storage')
        run_json = RunJson(global_info_json_dict, common.support_request_number)
        entity = run_json.as_entity()
        common.table_service.insert_entity(common.EXTENSION_RESULT_RUNS_TABLE_NAME, entity)

        logger.debug('Inserting FindingEntities to table storage')
        findings_json = FindingsJson(global_info_json_dict)
        entities = findings_json.as_entity_list()
        storage.batch_insert_storage_entities(common.table_service, common.EXTENSION_RESULT_FINDINGS_TABLE_NAME, entities)

        logger.debug('Inserting ImpactedResourceEntities to table storage')
        impacted_resources_json = ImpactedResourcesJson(global_info_json_dict)    
        entities = impacted_resources_json.as_entity_list()
        storage.batch_insert_storage_entities(common.table_service, common.EXTENSION_RESULT_IMPACTED_RESOURCES_TABLE_NAME, entities)

def generate_sas_token_for_blob():
    if common.user_agrees_to_share_data == True and common.upload_to_storage == True:
        logger.debug('Generating SAS token for the report...')

        sas_token = ""
        try:
            cwd = os.getcwd()
            blob_name = os.path.basename(global_info_json_dict['BlobUrl'])

            sas_token = blobsas.generate_blob_shared_access_signature(account_name=common.storage_account, 
                account_key=common.storage_account_key, container_name=common.EXTENSION_RESULT_CONTAINER_NAME, blob_name=blob_name,
                permission='r', start=dt.utcnow(), expiry=(dt.utcnow() + timedelta(days=30)))

            global_info_json_dict['BlobUrlWithSASToken'] = global_info_json_dict['BlobUrl'] + "?" + sas_token

            #write sas link to a file
            try:
                f= open(os.path.join(cwd,blob_name) + ".saslink","w+")
                f.write(global_info_json_dict['BlobUrlWithSASToken'])
            finally:
                if f:
                    f.close()

        except:
            logger.error('Unable to generate SAS token for PerfInsights report')
            traceback.print_exc()

        logger.debug('SAS URL for PerfInsights report: ' + global_info_json_dict['BlobUrlWithSASToken'])

# MAIN  
def main(argv=None):
    executionStart = dt.now()
    common.execution_start_time = common.get_current_datetime('display')

    #Exception hook for unhandled exception
    sys.excepthook = log_unhandled_exception

    # Init working and output folder location
    init_file_paths()

    # Parse script arguments
    parse_script_arguments(argv)

    # Create all the output folders
    create_output_folders()

    # Start logging
    init_logger()

    # Run auto update, if applicable
    # If auto update runs successfully, the rest of this script will not run since it will have been replaced with the latest version of perfinsights.py
    if common.auto_update:
        update_success = autoupdate.update_perf_insights()
        if update_success:
            autoupdate.replace_perfinsights_version()
            logger.info('Exit since the latest instance just ran.')
            return
        else:
            logger.info('PerfInsights was not updated. Running the previously installed version %s' % common.SCRIPT_VERSION)
    else:
        logger.info('Skipping auto update')

    logger.info("PerfInsights for Linux")
    logger.info("")
    logger.info("Script version: " + common.SCRIPT_VERSION)
    logger.info("Start time: " + common.get_current_datetime("display"))
    logger.info("Selected analysis: " + common.get_scenario_displayname(common.perfinsights_scenario))
    logger.info("PerfInsights log file: " + common.pifl_log_file_path)
    logger.info("")

    # Check script requirements
    check_requirements()

    # Ask for consent to share diagnostics data
    if not (common.bypass_agreements):
        # Disclaimer
        display_disclaimer()
        common.user_agrees_to_share_data = ask_for_consent(SHARE_DIAGNOSTICS_QUESTION)
    else:
        common.user_agrees_to_share_data = True

    # Wrap all following calls in a try/catch to emit failures to telemetry
    try:
        # Initialize PerfInsights
        init_perfInsights()
        
        # Initialize Storage Account
        init_storage()
        
        # Display VM information
        display_vm_info()
    
        # Execute Scenario
        run_scenario()
        
        # Get script execution end time    
        common.execution_end_time = common.get_current_datetime('display')
        executionEnd = dt.now()

        deltatime = executionEnd-executionStart

        scriptexecutiontime = ''
        try:
            scriptexecutiontime = (deltatime).total_seconds()
        except:
            # Fix for Python version < 2.7
            scriptexecutiontime = deltatime.days * 1440 + deltatime.seconds

        # Create perfinsights report
        create_report(scriptexecutiontime)

        # Send telemetry
        custom_props = {
            'StartTime': str(executionStart),
            'DurationInSeconds': str(scriptexecutiontime),
            'AcceptDisclaimerAndShareDiagnostics': common.user_agrees_to_share_data,
            'IsDebug': common.debug_enabled,
            'ScenarioName': common.perfinsights_scenario,
        }

        send_telemetry('PerfInsightsRun', custom_props, isUnhandled=False)

        logger.info("End time: " + common.get_current_datetime("display"))
        logger.info("Total execution time: " + str(scriptexecutiontime) + " seconds.")
        logger.info("PerfInsights for Linux - Execution completed.")

        # Compress all collected data and report files
        results_filename = compress_report()
        logger.debug("Results file name:" + results_filename)

        # Upload results and generate SAS token
        upload_report_to_storage()

    except Exception as e:
        logger.error("Encountered exception during run, failing and sending telemetry information.")
        traceback.print_exc()

        exception_info = sys.exc_info()
        send_telemetry(type(e).__name__, custom_properties=None, error_info=exception_info, isUnhandled=False)
        sys.exit(1)

if __name__ == "__main__":
    
    main(sys.argv[1:])
