# -------------------------------------------------------------------------
# Copyright (c) Microsoft.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

from __future__ import division

import os
import platform
import time
import shutil
from datetime import datetime
import threading
import time
import logging

import util.common as common
import ruleengine.DataCollector
from ruleengine import HighCpuRule, HighMemoryRule, HighDiskRule, HighResolutionDiskRule
from ruleengine.RuleEngineCls import IAnalyticRule, RuleUtil

logger = logging.getLogger(__name__)


class HighCpuRuleConfiguration(IAnalyticRule):

    def init(self, rule_config):
        self.rule_definition = rule_config
        self.rule_configuration = self.rule_definition.get('Configuration')
        self.capture_interval_seconds = int(self.rule_configuration['CpuStatCaptureIntervalInSeconds'])
        # Profiling duration_seconds in seconds
        # (adds 1 interval to the duration_seconds as the
        # first sample will be discarded, thus we need
        # to capture duration + interval to reflect provided
        # parameter)
        self.capture_duration_seconds = common.capture_duration_sec + self.capture_interval_seconds
        common.cpu_stat_capture_interval_secs = self.capture_interval_seconds

        self.rule_start_time = None
        self.rule_end_time = None
        self.resource_monitor = None

    def id(self):
        return 'PIFL.AR0001'

    def name(self):
        return self.rule_definition['Name']

    def configuration(self):
        return self.rule_configuration

    def friendly_name(self):
        return self.rule_definition['FriendlyName']

    def category(self):
        return self.rule_definition['Category']

    def on_start(self):
        logger.info(self.friendly_name() + " - Collect statistics...")

        self.rule_start_time = common.get_current_datetime("charting")

        # Init the resource monitor and return it to the RuleEngine caller
        # Resource monitor will be started from the RuleEngine
        self.resource_monitor = HighCpuRule.ResourceMonitor.init(self.capture_duration_seconds, self.capture_interval_seconds, self.friendly_name(), self.rule_definition)

        return self.resource_monitor

    def on_stop(self):

        result_dict = {}

        self.rule_end_time = common.get_current_datetime("charting")

        # Get results from the resource monitor
        result_dict = self.resource_monitor.stop()
        # [...]__init__

        data_collector_json = RuleUtil.create_rule_result(self, **result_dict)

        # Analyze the results and create findings
        high_periods = result_dict['HighProcessorsUsagePeriods']
        if common.perfinsights_scenario == 'hpc':
            logger.info("HPC Scenario may run benchmarks, so suppressing any findings based on high CPU usage.")
        elif high_periods is not None and len(high_periods) > 0:
            findings = []
            for period in high_periods:
                if period['AllProcessorsCpuPercentage'] > self.rule_configuration['CpuPercentageThresholdAllCore']:
                    finding = RuleUtil.create_finding(self.id(), self.rule_definition['FindingTemplates'], 0, [])
                    findings.append(finding)
                else:
                    for processor in period['ProcessorsUsage']:
                        if processor['ProcessorCpuPercentage']:
                            finding = RuleUtil.create_finding(self.id(), self.rule_definition['FindingTemplates'], 0, [])
                            findings.append(finding)
                            break

                # We only need to record a single finding
                if len(findings) > 0:
                    break

            # Tag the processes as impacted resources
            if len(data_collector_json['TopLongCpuSpinners']) > 0:
                processes = list(map(lambda p:
                                     {'ResourceName': 'Process',
                                      'ResourceValue': '%s (%s)' % (p['ProcessName'], p['ProcessId']),
                                      'Configurations': [
                                          {'Start Time': p['CpuSpinPeriods'][0]['StartTime'],
                                           'Duration (ms)': p['CpuSpinPeriods'][0]['CpuUsageMS'],
                                           'Total CPU Usage (ms)': p['TotalResourceUsageMS'],
                                     'CPU Usage %': "{:.2f}".format(p['TotalResourceUsagePercentage'])}]},
                data_collector_json['TopLongCpuSpinners']))
                for finding in findings:
                    finding["ImpactedResources"] = processes
            else:
                del findings[:]

            data_collector_json['RuleFindings'] = findings

        logger.info(self.friendly_name() + " - Statistics have been processed.")

        return data_collector_json

    def get_results(self):
        pass


class HighMemoryRuleConfiguration(IAnalyticRule):

    def init(self, rule_def):
        self.rule_definition = rule_def
        self.rule_configuration = self.rule_definition.get('Configuration')
        self.capture_interval_seconds = int(self.rule_configuration['MemoryStatCaptureIntervalInSeconds'])
        # Profiling duration_seconds in seconds
        # (adds 1 interval to the duration_seconds as the
        # first sample will be discarded, thus we need
        # to capture duration + interval to reflect provided
        # parameter)
        self.capture_duration_seconds = common.capture_duration_sec + self.capture_interval_seconds
        common.memory_stat_capture_interval_secs = self.capture_interval_seconds

        self.rule_start_time = None
        self.rule_end_time = None
        self.resource_monitor = None

    def id(self):
        return 'PIFL.AR0002'

    def name(self):
        return self.rule_definition['Name']

    def configuration(self):
        return self.rule_configuration

    def friendly_name(self):
        return self.rule_definition['FriendlyName']

    def category(self):
        return self.rule_definition['Category']

    def on_start(self):
        logger.debug(self.friendly_name())
        logger.info("Collect Memory statistics...")

        self.rule_start_time = common.get_current_datetime("charting")

        # Init the resource monitor and return it to the RuleEngine caller
        # Resource monitor will be started from the RuleEngine
        self.resource_monitor = HighMemoryRule.ResourceMonitor.init(self.capture_duration_seconds, self.capture_interval_seconds, self.friendly_name(), self.rule_definition)

        return self.resource_monitor

    def on_stop(self):
        result_dict = {}

        # insert the processing code

        self.rule_end_time = common.get_current_datetime("charting")

        # Get results from the resource monitor
        result_dict = self.resource_monitor.stop()
        # [...]

        data_collector_json = RuleUtil.create_rule_result(self, **result_dict)

        # Analyze the results and create findings
        high_periods = result_dict['HighMemoryUsagePeriods']
        if common.perfinsights_scenario == 'hpc':
            logger.info("HPC Scenario may run benchmarks, so suppressing any findings based on high memory usage.")
        elif high_periods is not None and len(high_periods) > 0:
            findings = []
            for period in high_periods:
                if period['AvailablePhysicalMemoryPercentage'] < self.rule_configuration['AvailablePhysicalMemoryPercentageThreshold']:
                    finding = RuleUtil.create_finding(self.id(), self.rule_definition['FindingTemplates'], 0, [])
                    findings.append(finding)
                    # We only need to record a single finding
                    break

            # Tag the processes as impacted resources
            if len(data_collector_json['TopMemorySpinners']) > 0:
                phy_kb = int(data_collector_json['PhysicalMemoryKB'])
                swap_kb = int(data_collector_json['SwapMemoryKB'])
                processes = list(map(lambda p:
                                     {'ResourceName': 'Process',
                                      'ResourceValue': '%s (%s)' % (p['ProcessName'], p['ProcessId']),
                                      'Configurations': [
                                          {'Start Time': p['MemorySpinPeriods'][0]['StartTime'],
                                           'Average Resident Memory (KB)': p['AvgResidentMemoryKB'],
                                           'Average Swap Memory (KB)': p['AvgSwapMemoryKB'],
                                           'Average Resident Memory (%)': 0.0 if phy_kb == 0 else '%0.2f' % (float(p['AvgResidentMemoryKB']) * 100.0 / phy_kb),
                                           'Average Swap Memory (%)': 0.0 if swap_kb == 0 else '%0.2f' % (float(p['AvgSwapMemoryKB']) * 100.0 / swap_kb)}
                                      ]},
                                     data_collector_json['TopMemorySpinners']))
                for finding in findings:
                    finding['ImpactedResources'] = processes
            else:
                del findings[:]

            data_collector_json['RuleFindings'] = findings
        logger.debug(self.friendly_name() + " statistics have been processed.")

        return data_collector_json

    def get_results(self):
        pass
    
class HighDiskRuleConfiguration(IAnalyticRule):

    def init(self, rule_def):
        self.rule_definition = rule_def
        self.rule_configuration = self.rule_definition.get('Configuration')
        self.capture_interval_seconds = int(self.rule_configuration['DiskStatCaptureIntervalInSeconds'])
        # Profiling duration_seconds in seconds
        # (adds 1 interval to the duration_seconds as the
        # first sample will be discarded, thus we need
        # to capture duration + interval to reflect provided
        # parameter)
        self.capture_duration_seconds = common.capture_duration_sec + self.capture_interval_seconds
        common.disk_stat_capture_interval_secs = self.capture_interval_seconds

        self.kernel_version = platform.release().replace('"', '')
        self.kernel_version = float(self.kernel_version.split('.')[0] + '.' + self.kernel_version.split('.')[1])

        self.rule_start_time = None
        self.rule_end_time = None
        self.resource_monitor = None
        self.iostatRawDict = []
        self.iostatResultDict = []

    def id(self):
        return 'PIFL.AR0003'

    def name(self):
        return self.rule_definition['Name']

    def configuration(self):
        return self.rule_configuration

    def friendly_name(self):
        return self.rule_definition['FriendlyName']

    def category(self):
        return self.rule_definition['Category']

    def on_start(self):
        logger.info(self.friendly_name() + " - Collect statistics...")

        self.rule_start_time = common.get_current_datetime("charting")

        self.resource_monitor = HighDiskRule.ResourceMonitor.init(self.capture_duration_seconds, self.capture_interval_seconds, self.friendly_name(), self.rule_definition)
        return self.resource_monitor

        # self.iostatRawDict = self.get_io_stats()

    def on_stop(self):
        result_dict = {}

        self.rule_end_time = common.get_current_datetime("charting")

        # Get results from the resource monitor
        result_dict = self.resource_monitor.stop()

        data_collector_json = RuleUtil.create_rule_result(self, **result_dict)

        # Look for periods of high disk read or write latency
        high_periods = result_dict['HighDisksUsagePeriods']
        rule_config = RuleUtil.get_rule_config(self.rule_definition['FindingTemplates'], 0)
        if common.perfinsights_scenario == 'hpc':
            logger.info("HPC Scenario may run benchmarks, so suppressing any findings based on high disk usage.")
        elif high_periods is not None and len(high_periods) > 0:
            findings = []
            for period in high_periods:
                for disk in period['DisksPerfData']:
                    if disk['ReadAWait'] > rule_config['LatencyThresholdMS'] or \
                            disk['WriteAWait'] > rule_config['LatencyThresholdMS']:
                        # We have a period of high latency, we need to find a matching high disk usage period to
                        # attribute running processes
                        processes = []
                        proc_periods = result_dict['TopLongDiskSpinners']
                        for process in proc_periods:
                            for proc_period in process['DiskSpinPeriods']:
                                if proc_period['EndTime'] < period['StartTime']:
                                    continue
                                elif proc_period['StartTime'] > period['EndTime']:
                                    continue
                                else:
                                    processes.append({'ResourceName': 'Process',
                                        'ResourceValue': '%s (%s)' % (process['ProcessName'], process['ProcessId']),
                                        'Configurations': [
                                            {'Start Time': proc_period['StartTime'],
                                             'AvgDiskIops': "{:.2f}".format(process['AvgDiskIops'])}
                                        ]})
                        finding = RuleUtil.create_finding(self.id(),
                                                          self.rule_definition['FindingTemplates'], 0, processes, finding_parameters=rule_config)
                        findings.append(finding)
                        break
            data_collector_json['RuleFindings'] = findings

        logger.debug(self.friendly_name() + " - Statistics have been processed.")

        return data_collector_json

    def get_results(self):
        pass

    def cleanup(self):
        pass


class HighResolutionDiskRuleConfiguration(IAnalyticRule):

    def init(self, rule_def):
        self.rule_definition = rule_def
        self.rule_configuration = self.rule_definition.get('Configuration')


        # Update DetectWindowMS configuration as determined by StorageTier
        if common.thisazurevm is not None and  common.thisazurevm['VmConfig']['StorageTier']== 'premium':
            self.rule_configuration['DetectWindowMS'] = self.rule_configuration['DetectWindowConfigurationMS']['PremiumVM']
        else: 
            # Default to the standard vm interval if StorageTier is unknown or not premium
            self.rule_configuration['DetectWindowMS'] = self.rule_configuration['DetectWindowConfigurationMS']['StandardVM']

        self.capture_interval_seconds = self.rule_configuration['DetectWindowMS']/1000
        self.capture_duration_seconds = common.capture_duration_sec

        self.rule_start_time = None
        self.rule_end_time = None

    def id(self):
        return 'PIFL.AR0005'

    def name(self):
        return self.rule_definition['Name']

    def configuration(self):
        return self.rule_configuration

    def friendly_name(self):
        return self.rule_definition['FriendlyName']

    def category(self):
        return self.rule_definition['Category']

    def on_start(self):
        logger.info("Collect Disk statistics...")

        self.rule_start_time = common.get_current_datetime("charting")

        self.resource_monitor = HighResolutionDiskRule.ResourceMonitor.init(self.capture_duration_seconds, self.capture_interval_seconds, self.friendly_name(), self.rule_definition)
        return self.resource_monitor

    def on_stop(self):
        result_dict = {}

        self.rule_end_time = common.get_current_datetime("charting")

        result_dict = self.resource_monitor.stop()

        data_collector_json = RuleUtil.create_rule_result(self, **result_dict)

        logger.debug(self.friendly_name() + " statistics have been processed.")

        return data_collector_json

    def get_results(self):
        pass
