#-------------------------------------------------------------------------
# Copyright (c) Microsoft.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#--------------------------------------------------------------------------
import datetime
import glob
import json
import logging
import os
import platform
import re
import shutil
import subprocess
import sys
import xml.etree.ElementTree as ET
from operator import itemgetter

import distro
import six
from six.moves import urllib

import util.common as common
from ruleengine.RuleEngineCls import IDataCollector, IDataMonitor, DataCollectorUtil
from util.SystemUtil import ProcessUtil, DiskUtil, PartitionUtil
from util.common import CachedProperty

logger = logging.getLogger(__name__)

if six.PY3:
    long = int

MYPY = False
if MYPY:
    from typing import List, Dict


class InstanceMetadataInfoDC(IDataCollector):

    VM_MAX_IOPS_DEFAULT = 3200
    VM_MAX_THROUGHPUT_DEFAULT = 24

    def __init__(self):
        self._dc_config = None
        self.metadataComputeObj = {}
        self.metadataNetworkObj = {}

    def __del__(self):
        pass

    def init(self, dc_config):
        self._dc_config = dc_config

    def index(self):
        if self._dc_config:
            return self._dc_config['Index']
        else:
            return None

    def name(self):
        if self._dc_config:
            return self._dc_config['Name']
        else:
            return None

    def friendly_name(self):
        if self._dc_config:
            return self._dc_config['FriendlyName']
        else:
            return None

    def to_json(self):
        if self.name():
            logger.debug(self.name() + ": to_json")

        instancemetadataDict = {}
        data_collection_start_time = common.get_current_datetime("display")
        if self.get_metadata() is True:
            instancemetadataDict['Compute'] = self.compute_map
            instancemetadataDict['Network'] = self.network_map
            instancemetadataDict['VmConfig'] = self.storagetier_map

        data_collection_end_time = common.get_current_datetime("display")
        datacollectorJSON = DataCollectorUtil.create_json(self.name(), data_collection_start_time, data_collection_end_time, instancemetadataDict)

        if self.name():
            logger.debug(self.name() + ": to_json")

        return datacollectorJSON

    def get_metadata(self):

        result = False

        if self.name():
            logger.debug(self.name() + ": get_metadata")

        try:
            aimdurl = "http://169.254.169.254/metadata/instance?api-version=2019-08-15"
            header = {'Metadata': 'True'}
            amdirequest = urllib.request.Request(url=aimdurl, headers=header)
            amdiresponse = urllib.request.urlopen(amdirequest, timeout=15)
            if amdiresponse.getcode() == 200:
                data = amdiresponse.read()
                try:
                    #get whole metadata
                    self.metadataComputeObj = json.loads(data.decode("utf-8"))

                    # Get network only
                    self.metadataNetworkObj = self.metadataComputeObj["network"]

                    result = True

                except ValueError:
                    logger.warning("Exception when reading Azure metadata information.")

            else:
                logger.warning("Can't get Azure metadata information. HTTP status code: " + str(amdiresponse.getcode()))

        except urllib.error.HTTPError as e:
            logger.warning("Exception when gathering Azure metadata information: " + str(e.code) + ".")

        except urllib.error.URLError:
            #if we get here, it would be due to a timeout error meaning this is not an Azure VM.
            # No need to add error code to message
            logger.warning("PerfInsights is only supported on Azure VMs.")

        if result == False:
            logger.warning("Script will continue to run but some information may be missing from the report.")

        return result

    @CachedProperty
    def compute_map(self):

        computeDict = {}
        computeDict['azEnvironment'] = self.metadataComputeObj["compute"]["azEnvironment"]
        computeDict['subscriptionId'] = self.metadataComputeObj["compute"]["subscriptionId"]
        computeDict['vmId'] = self.metadataComputeObj["compute"]["vmId"]
        computeDict['Name'] = self.metadataComputeObj["compute"]["name"]
        computeDict['osType'] = self.metadataComputeObj["compute"]["osType"]
        computeDict['resourceGroupName'] = self.metadataComputeObj["compute"]["resourceGroupName"]
        computeDict['resourceId'] = self.metadataComputeObj["compute"]["resourceId"]
        computeDict['location'] = self.metadataComputeObj["compute"]["location"]
        computeDict['publisher'] = self.metadataComputeObj["compute"]["publisher"]
        computeDict['offer'] = self.metadataComputeObj["compute"]["offer"]
        computeDict['sku'] = self.metadataComputeObj["compute"]["sku"]
        computeDict['version'] = self.metadataComputeObj["compute"]["version"]
        computeDict['vmSize'] = self.metadataComputeObj["compute"]["vmSize"]
        computeDict['tags'] = self.metadataComputeObj["compute"]["tags"]
        computeDict['platformFaultDomain'] = self.metadataComputeObj["compute"]["platformFaultDomain"]
        computeDict['platformUpdateDomain'] = self.metadataComputeObj["compute"]["platformUpdateDomain"]
        computeDict['placementGroupId'] = self.metadataComputeObj["compute"]["placementGroupId"]
        computeDict['zone'] = self.metadataComputeObj["compute"]["zone"]
        computeDict['vmScaleSetName'] = self.metadataComputeObj["compute"]["vmScaleSetName"]

        return computeDict

    @CachedProperty
    def network_map(self):
        interface = self.metadataNetworkObj['interface']
        if type(interface) is not list:
            interface = [interface]
        
        return {'interface' : interface }

    @CachedProperty
    def storagetier_map(self):
        storagetierDict = {}
        vmsizereferencefile = common.open_as_csvfile(common.vm_sizes_lookup_file_path, ",")
        if vmsizereferencefile is not None:
            for _currentvmsizeref in vmsizereferencefile:
                compute_map_dict = self.compute_map
                if _currentvmsizeref[0] == compute_map_dict['vmSize']:  # pylint: disable=E1136
                    # state = "nice" if is_nice else "not nice"
                    storagetierDict['MaxIops'] = (_currentvmsizeref[1] if _currentvmsizeref[1] else 0)
                    storagetierDict['MaxThroughputMBPS'] = (_currentvmsizeref[2] if _currentvmsizeref[2] else 0)
                    storagetierDict['StorageTier'] = (_currentvmsizeref[3] if _currentvmsizeref[3] else 0)
                    break
            if not storagetierDict:
                storagetierDict['MaxIops'] = 0
                storagetierDict['MaxThroughputMBPS'] = 0
                storagetierDict['StorageTier'] = 'Standard'
        else:
            logger.error("VM sizes reference file was not found: %s", common.vm_sizes_lookup_file_path)

        return storagetierDict


class VolumeMapInfoCollector(IDataCollector):
    '''
    Class
    =========
    VolumeMapInfoCollector

    Description
    ==============
    Data collector class used to retrieve
    storage information.
    Inherit from IDataCollector.
    '''

    def __init__(self):
        '''
        Function
        =========
        __init__

        Description
        ==============
        Initialize the data collector class.

        Arguments
        ==========
        N/A

        Return Values
        ==============
        N/A
        '''

        self.data_collector_config = None
        self.lvm_dict = {}
        self.mdadm_dict = {}
        self.block_devices_dict = {}
        self.partitions_dict = {}
        self.mount_points_dict = {}

    def name(self):
        '''
        Function
        =========
        name

        Description
        ==============
        Returns the data collector name.

        Arguments
        ==========
        __in <class> self

        Return Values
        ==============
        __out <string> data collector name or None
        '''

        if self.data_collector_config:
            return self.data_collector_config['Name']
        else:
            return None

    def friendly_name(self):
        '''
        Function
        =========
        name

        Description
        ==============
        Returns the data collector friendly name.

        Arguments
        ==========
        __in <class> self

        Return Values
        ==============
        __out <string> data collector friendly name or None
        '''

        if self.data_collector_config:
            return self.data_collector_config['FriendlyName']
        else:
            return None

    def index(self):
        if self.data_collector_config:
            return self.data_collector_config['Index']
        else:
            return None

    def init(self, data_collector_config):
        '''
        Function
        =========
        init

        Description
        ==============
        Set the data collector configuration.
        This function is called by the RuleEngine
        which provides the data collector
        configuration.

        Arguments
        ==========
        __in <class> self
        __in <object> data_collector_config

        Return Values
        ==============
        N/A
        '''

        self.data_collector_config = data_collector_config

    def to_json(self):
        '''
        Function
        =========
        to_json

        Description
        ==============
        Run all sub data collectors and
        consolidate the results in the
        expected lists/dictionaries.

        Arguments
        ==========
        __in <class> self

        Return Values
        ==============
        __out <dict> final_result_dict: data collector results
        '''

        if self.name():
            logger.debug(self.name() + ": to_json")

        final_result_dict = {}

        data_collection_start_time = common.get_current_datetime("display")

        # Get all storage components information
        self.lvm_dict = self.get_lvm_devices()
        self.mdadm_dict = self.get_mdadm_devices()
        self.partitions_dict = self.get_partitions()
        self.block_devices_dict = self.get_block_devices()
        self.mount_points_dict = self.get_mount_points()

        # Update LVM devices with updated information
        # from other components which were not
        # available at collection time
        if self.lvm_dict:
            self.update_lvm()

        # Update block devices with updated information
        # from other components which were not
        # available at collection time
        if self.block_devices_dict:
            self.update_block_devices()

        # Update partitions information with updated information
        # from other components which were not
        # available at collection time
        if self.partitions_dict:
            self.update_partitions()

        # Update mount points with updated information
        # from other components which were not
        # available at collection time
        if self.mount_points_dict:
            self.update_mount_points()

        # Update MDADM devices with updated information
        # from other components which were not
        # available at collection time
        if self.mdadm_dict:
            self.update_mdadm_devices()

        # Build the result lists and
        # dictionaries from the collected data
        final_result_dict["DiskMap"] = self.block_device_map
        final_result_dict["PartitionMap"] = self.partition_map
        final_result_dict["MountPointMap"] = self.mount_point_map
        final_result_dict["LvmMap"] = self.lvm_map
        final_result_dict["MdadmMap"] = self.mdadm_map

        data_collection_end_time = common.get_current_datetime("display")
        final_result_json = DataCollectorUtil.create_json(self.name(), data_collection_start_time, data_collection_end_time, final_result_dict)

        logger.info("Storage configuration collected.")

        if self.name():
            logger.debug(self.name() + ": to_json")

        return final_result_json

    def get_lvm_devices(self):
        '''
        Function
        =========
        get_lvm_devices

        Description
        ==============
        Get all LVM physical volumes, volume groups,
        logical volumes available on the system.
        Make use of vgs, lvs and pvs command line tools.

        Arguments
        ==========
        __in <class> self

        Return Values
        ==============
        __out <dict> lvm_dict: LVM devices dictionary

        '''

        if self.name():
            logger.debug(self.name() + ": get_lvm_devices")

        lvm_dict = {}
        vg_dict = {}
        lv_dict = {}
        pv_dict = {}

        if os.path.exists('/etc/lvm'):
            # LVM is available on the system

            # First, get volume groups:
            # Run vgs with required fields and
            # parse the output to build the dictionary
            # with LVMVolumeGroup objects
            vgs_command_line = "vgs -o vg_name,vg_uuid,vg_size,vg_free,vg_extent_size,vg_extent_count,vg_free_count,pv_count,lv_count,snap_count,lv_name,Fmt,devices,vg_missing_pv_count --units b --separator ';' --noheadings 2> /dev/null"
            if ('debian' in common.linux_distribution.lower() and common.linux_version_id == '8') or \
                ('centos' in common.linux_distribution.lower() and common.linux_version_id == '6.5') or \
                ('ubuntu' in common.linux_distribution.lower() and common.linux_version_id == '14.04'):
                vgs_command_line = "vgs -o vg_name,vg_uuid,vg_size,vg_free,vg_extent_size,vg_extent_count,vg_free_count,pv_count,lv_count,snap_count,lv_name,Fmt,devices --units b --separator ';' --noheadings 2> /dev/null"

            subproc_output = subprocess.Popen([vgs_command_line], stdout=subprocess.PIPE, shell=True)
            vgroups_output = subproc_output.stdout.read().decode()
            vgroups_output = [line.split(';') for line in vgroups_output.splitlines()]
            for vg in vgroups_output:
                if len(vg) > 1:
                    if vg[0].strip() not in vg_dict:
                        # Volume group is not yet referenced
                        vg_dict[vg[0].strip()] = VolumeMapInfoCollector.LVMVolumeGroup(vg)
                    else:
                        # Volume group is yet referenced
                        # but it has additional devices informations
                        vg_dict[vg[0].strip()].update_vg(vg)
            lvm_dict['VolumeGroups'] = vg_dict

            # Second, get logical volumes:
            # Run lvs with required fields and
            # parse the output to build the dictionary
            # with LVMLogicalVolume objects
            lvs_command_line = "lvs -o lv_name,lv_uuid,vg_name,lv_size,lv_metadata_size,seg_count,lv_time,copy_percent,stripes,stripe_size,region_size,devices,sync_percent,lv_dm_path,lv_parent,lv_layout,cache_policy,cache_total_blocks,seg_le_ranges,data_stripes,data_copies --units b --separator ';' --noheadings 2> /dev/null"
            if 'debian' in common.linux_distribution.lower():
                if common.linux_version_id == '8':
                    lvs_command_line = "lvs -o lv_name,lv_uuid,vg_name,lv_size,lv_metadata_size,seg_count,lv_time,copy_percent,stripes,stripe_size,region_size,devices,sync_percent,lv_dm_path,lv_parent,lv_layout --units b --separator ';' --noheadings 2> /dev/null"
                elif common.linux_version_id == '9' or \
                    'red hat enterprise linux server' in common.linux_distribution.lower() and common.linux_version_id == '6.7':
                    lvs_command_line = "lvs -o lv_name,lv_uuid,vg_name,lv_size,lv_metadata_size,seg_count,lv_time,copy_percent,stripes,stripe_size,region_size,devices,sync_percent,lv_dm_path,lv_parent,lv_layout,cache_policy,cache_total_blocks,seg_le_ranges --units b --separator ';' --noheadings 2> /dev/null"
            if 'centos' in common.linux_distribution.lower() and common.linux_version_id == '6.5' or \
                ('ubuntu' in common.linux_distribution.lower() and common.linux_version_id == '14.04'):
                lvs_command_line = "lvs -o lv_name,lv_uuid,vg_name,lv_size,lv_metadata_size,seg_count,lv_time,copy_percent,stripes,stripe_size,region_size,devices --units b --separator ';' --noheadings 2> /dev/null"
            if ('ubuntu' in common.linux_distribution.lower() and common.linux_version_id == '16.04'):
                lvs_command_line = "lvs -o lv_name,lv_uuid,vg_name,lv_size,lv_metadata_size,seg_count,lv_time,copy_percent,stripes,stripe_size,region_size,devices,sync_percent,lv_dm_path,lv_parent,lv_layout,cache_policy,cache_total_blocks --units b --separator ';' --noheadings 2> /dev/null"

            subproc_output = subprocess.Popen([lvs_command_line], stdout=subprocess.PIPE, shell=True)
            lvolumes_output = subproc_output.stdout.read().decode()
            lvolumes_output = [line.split(';') for line in lvolumes_output.splitlines()]
            for lv in lvolumes_output:
                if len(lv) > 1:
                    if lv[0].strip() not in lv_dict:
                        # Logical volume is not yet referenced
                        lv_dict[lv[0].strip()] = VolumeMapInfoCollector.LVMLogicalVolume(lv)
                    else:
                        # Logical volume is yet referenced
                        # but it has additional devices informations
                        lv_dict[lv[0].strip()].update_lv(lv)
            lvm_dict['LogicalVolumes'] = lv_dict
            subproc_output.wait()

            # Last, get physical volumes:
            # Run pvs with required fields and
            # parse the output to build the dictionary
            # with LVMPhysicalVolume objects
            pvs_command_line = "pvs -o pv_name,pv_uuid,vg_name,lv_name,pv_fmt,pv_size,pv_free,pv_used,dev_size,pv_pe_count,pv_pe_alloc_count,pv_allocatable,pv_exported,pv_missing,pv_major,pv_minor,pv_in_use,pv_duplicate --units b --separator ';' --noheadings 2> /dev/null"
            if 'debian' in common.linux_distribution.lower() and common.linux_version_id == '8' or \
                'red hat enterprise linux server' in common.linux_distribution.lower() and common.linux_version_id == '6.7' or \
                ('ubuntu' in common.linux_distribution.lower() and common.linux_version_id == '16.04'):
                pvs_command_line = "pvs -o pv_name,pv_uuid,vg_name,lv_name,pv_fmt,pv_size,pv_free,pv_used,dev_size,pv_pe_count,pv_pe_alloc_count,pv_allocatable,pv_exported,pv_missing --units b --separator ';' --noheadings 2> /dev/null"
            if 'centos' in common.linux_distribution.lower() and common.linux_version_id == '6.5' or \
                ('ubuntu' in common.linux_distribution.lower() and common.linux_version_id == '14.04'):
                pvs_command_line = "pvs -o pv_name,pv_uuid,vg_name,lv_name,pv_fmt,pv_size,pv_free,pv_used,dev_size,pv_pe_count,pv_pe_alloc_count --units b --separator ';' --noheadings 2> /dev/null"

            subproc_output = subprocess.Popen([pvs_command_line], stdout=subprocess.PIPE, shell=True)
            pvolumes_output = subproc_output.stdout.read().decode()
            pvolumes_output = [line.split(';') for line in pvolumes_output.splitlines()]
            for pv in pvolumes_output:
                if len(pv) > 1:
                    if pv[0].strip() not in pv_dict:
                        # Physical volume is not yet referenced
                        pv_dict[pv[0].strip()] = VolumeMapInfoCollector.LVMPhysicalVolume(pv)
                    else:
                        # Physical volume is yet referenced
                        # but it has additional lv informations
                        pv_dict[pv[0].strip()].update_pv(pv)
            lvm_dict['PhysicalVolumes'] = pv_dict

        logger.debug("LVM information collected.")

        return lvm_dict

    def get_mdadm_devices(self):
        '''
        Function
        =========
        get_mdadm_devices

        Description
        ==============
        Get all MDADM devices and member
        block devices available on the system.

        Make use of /proc/mdstat file to identify
        existing MDADM devices.

        cat /proc/mdstat
            Personalities : [raid0] [linear] [multipath] [raid1] [raid6] [raid5] [raid4] [raid10]
            md2 : active linear sdp1[1] sdq1[0]
                2145120105 blocks super 1.2 0k rounding

            md0 : active raid0 sdc1[0] sdd1[1]
            2145120000 blocks super 1.2 64k chunks

        References:
            https://raid.wiki.kernel.org/index.php/Mdstat

        Arguments
        ==========
        __in <class> self

        Return Values
        ==============
        __out <dict> mdadm_dict: MDADM devices dictionary

        '''

        if self.name():
            logger.debug(self.name() + ": get_mdadm_devices")

        mdadm_dict = {}

        if os.path.exists('/proc/mdstat'):
            # MDADM is available on the system
            with open('/proc/mdstat', 'r') as f:
                mdstat_file = f.readlines()

            line_count = 0
            while line_count < len(mdstat_file):
                if 'md' in mdstat_file[line_count]:
                    # This line contains information about
                    # the device by itself

                    # Collect first line with all major information
                    device = mdstat_file[line_count].split()
                    # Collect second line with additional information
                    device_ex = mdstat_file[line_count + 1].split()

                    mdadm_dict[device[0]] = VolumeMapInfoCollector.MDADMDevice(device, device_ex)
                    line_count += 1
                line_count += 1

        logger.debug("MDADM information collected.")

        return mdadm_dict

    def get_partitions(self):
        '''
        Function
        =========
        get_partitions

        Description
        ==============
        Get all partitions available on
        the system.

        Make use of /proc/partitions file to identify
        existing partitions.

        Major versions
            Block devices:
                8 sd
                9 md
                65 sd
                66 sd
                67 sd
                68 sd
                69 sd
                70 sd
                71 sd
                128 sd
                129 sd
                130 sd
                131 sd
                132 sd
                133 sd
                134 sd
                135 sd
                253 device-mapper

        Minor versions for sd devices
            For block devices, it starts from 0 and
            increments by 15 (limit due to the
            max number of partitions which is 14)
            0 = block device
            1 -> 15 = partitions
            16 = block device
            17 -> 30 = partitions
            and so on...

        Minor versions for major version 9
            For MDADM devices, it starts from 0 and
            increments by 1.
            Only numbered list of metadisk groups.

        Minor versions for major version 253
            For LVM devices, it starts from 0 and
            increments by 1.
            Only numbered list of metadisk groups.

        Reference:
            cat /proc/devices
            https://github.com/torvalds/linux/blob/master/Documentation/admin-guide/devices.txt
            https://learning.oreilly.com/library/view/linux-device-drivers/0596000081/ch03s02.html

        Arguments
        ==========
        __in <class> self

        Return Values
        ==============
        __out <dict> partitions_dict: partitions dictionnary

        '''

        if self.name():
            logger.debug(self.name() + ": get_partitions")

        partitions_dict = {}

        # Read the partition file
        partitions_file = [line.split() for line in open('/proc/partitions', 'r').readlines()][2:]

        # Create a list of tuples that
        # will be used to sort the list
        partitions_tuples = []
        for partition in partitions_file:
            partitions_tuples.append((int(partition[0]), int(partition[1]), long(partition[2]), partition[3]))

        # Sort the list of tuples:
        # this is required to parse the
        # partition file correctly
        partitions_sorted = sorted(partitions_tuples, key = itemgetter(0, 1))

        # Set the list of possible major values:
        # these are the minor values matching
        # block devices we are interested in
        # (hd, md and dm devices are not included)
        partitions_major_indexes = [8, 65, 66, 67, 68, 69, 70, 71, 128, 129, 130, 131, 132, 133, 134, 135]

        # Set the list of possible minor values:
        # these are the minor values matching
        # partitions
        partitions_minor_indexes = partitions_minor_indexes_ref = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15]

        # Parse partitions file for block devices
        for major_version in partitions_major_indexes:
            # Parse all major versions in incremental
            # order

            # Reset partitions_minor_indexes from
            # partitions_minor_indexes_ref which should never
            # change.
            # We reset this list because at each new device,
            # the minor version restarts at 1
            partitions_minor_indexes = partitions_minor_indexes_ref

            for partition in partitions_sorted:
                # Parse sorted partition list of tupples
                if partition[0] == major_version:
                    # This is a block device
                    if partition[1] in partitions_minor_indexes:
                        # This is a partition, add to the
                        # dictionary
                        partitions_dict[partition[3]] = VolumeMapInfoCollector.Partition(partition)
                    elif partition[1] > 0 and (partition[1] % 16 == 0):
                        # Minor version is for block device (modulo 16)
                        # so we need to increase the minor version
                        # increment to the next block device index
                        # by adding 16 to each minor indexes
                        partitions_minor_indexes = [item + 16 for item in partitions_minor_indexes]

        # Parse partitions file for md devices
        for partition in partitions_sorted:
            if partition[0] == 9:
                # Major version is for LVM device (dm)
                for fs in glob.glob('/proc/fs/*'):
                    file_system = os.path.basename(fs)
                    for dev in glob.glob('/proc/fs/' + file_system + '/*'):
                        device = os.path.basename(dev)
                        if partition[3] == device:
                            partitions_dict[partition[3]] = VolumeMapInfoCollector.Partition(partition)
                            break

        # Parse partitions file for dm devices
        for partition in partitions_sorted:
            if partition[0] == 253:
                # Major version is for MDADM device (md)
                for fs in glob.glob('/proc/fs/*'):
                    file_system = os.path.basename(fs)
                    for dev in glob.glob('/proc/fs/' + file_system + '/*'):
                        device = os.path.basename(dev)
                        if partition[3] == device:
                            partitions_dict[partition[3]] = VolumeMapInfoCollector.Partition(partition)
                            break

        logger.debug("Partitions information collected.")

        return partitions_dict

    def get_block_devices(self):
        '''
        Function
        =========
        get_block_devices

        Description
        ==============
        Get all block, md and dm devices on the system.

        Parse /sys/block directory to find
        block devices matching the pattern which
        targets only sd?, md? and DM-? devices.

        Arguments
        ==========
        __in <class> self

        Return Values
        ==============
        __out <dict> block_devices_dict: block devices dictionary

        '''

        if self.name():
            logger.debug(self.name() + ": get_block_devices")

        block_devices_dict = {}

        for device in glob.glob('/sys/block/*'):
            for pattern in DiskUtil.VALID_BLOCK_DEVICES_PATTERN:
                short_device_name = os.path.basename(device)
                if re.compile(pattern).match(short_device_name):
                    block_devices_dict['/dev/' + short_device_name] = VolumeMapInfoCollector.BlockDevice('/dev/' + short_device_name)

        logger.debug("Block devices information collected.")

        return block_devices_dict

    def get_mount_points(self):
        '''
        Function
        =========
        get_mount_points

        Description
        ==============
        Get all mount points available on
        the system.

        Make use of /proc/mounts to identify
        existing mount points.

        Arguments
        ==========
        __in <class> self

        Return Values
        ==============
        __out <dict> mount_points_dict: mount points dictionary

        '''

        if self.name():
            logger.debug(self.name() + ": get_mount_points")

        mount_points_dict = {}

        mounts_file = [line.split() for line in open('/proc/mounts', 'r').readlines()]
        for line in mounts_file:
            if '/dev/' in line[0]:
                # The associated device is a
                # block device so we reference it
                mount_points_dict[line[1]] = VolumeMapInfoCollector.MountPoint(line)

        logger.debug("Mount points information collected.")

        return mount_points_dict

    def update_lvm(self):
        '''
        Function
        =========
        update_lvm

        Description
        ==============
        When LVM information was collected,
        additionnal information were not.

        Update volume groups with member
        physical volumes (from lvm dictionary).

        Update logical volumes with corresponding
        mount points (from mount points dictionary).

        Arguments
        ==========
        __in <class> self

        Return Values
        ==============
        __out <dict> mount_points_dict: mount points dictionary

        '''

        if self.name():
            logger.debug(self.name() + ": update_lvm")

        for vg in self.lvm_dict['VolumeGroups'].values():
            for pv in self.lvm_dict['PhysicalVolumes'].values():
                if pv.vg_name == vg.name:
                    vg.physical_volumes.append(pv.name)

        if self.mount_points_dict:
            for lv in self.lvm_dict['LogicalVolumes'].values():
                for mp in self.mount_points_dict.values():
                    if lv.dev_mapper_path == mp.parent_mapper_device_name:
                        lv.mount_point = mp.path

        logger.debug("LVM information updated.")

    def update_block_devices(self):
        '''
        Function
        =========
        update_block_devices

        Description
        ==============
        When block devices information was collected,
        additionnal information were not.

        Update dm-? devices (LVM) and md? devices (MDADM)
        with member block devices
        (from lvm and mdadm dictionaries).

        Update block devices to inherit from LVM attributes
        and member block devices (from lvm dictionary).

        Update block devices to inherit from MDADM attributes
        and member block devices (from mdadm dictionary).

        Update block devices with associated partitions.

        Arguments
        ==========
        __in <class> self

        Return Values
        ==============
        __out <dict> mount_points_dict: mount points dictionary

        '''

        if self.name():
            logger.debug(self.name() + ": update_block_devices")

        if self.lvm_dict:
            # Update LVM inherited attributes
            for device in self.block_devices_dict.values():
                for lv in self.lvm_dict['LogicalVolumes'].values():
                    # Parse all logical volumes
                    for child_device in lv.physical_volumes:
                        # Parse logical volume for its physical volumes
                        # which should be a sd? or md? device
                        if child_device['device'] == device.name:
                            # Found the matching PV so will update
                            # attributes as a LVM member
                            device.is_lvm_member = True
                            if lv.raid_type == 'linear':
                                device.is_lvm_spanned = True
                            if lv.raid_type == 'mirror':
                                device.is_lvm_mirrored = True
                            if lv.raid_type == 'raid,raid5,raid5_ls':
                                device.is_lvm_raid_5 = True
                            if lv.raid_type == 'striped':
                                device.is_lvm_striped = True
                            break

                    if lv.device_name == device.name:
                        # Found the matching dm-? device so
                        # will update attributes as a LVM device
                        device.is_lvm_device = True
                        if lv.raid_type == 'linear':
                            device.is_lvm_spanned = True
                        if lv.raid_type == 'mirror':
                            device.is_lvm_mirrored = True
                        if lv.raid_type == 'raid,raid5,raid5_ls':
                            device.is_lvm_raid_5 = True
                        if lv.raid_type == 'striped':
                            device.is_lvm_striped = True
                        # Add member block devices from
                        # physical volumes referenced in logical
                        # volume
                        for child_device in lv.physical_volumes:
                            device.member_block_devices.append(child_device['device'])

        if self.mdadm_dict:
            # Update MDADM inherited attributes by
            # listing all MDADM member block devices
            for device in self.block_devices_dict.values():
                for md in self.mdadm_dict.values():
                    # Parse all MDADM devices
                    for bd in md.block_devices:
                        # Parse all MDADM member for its
                        # member block devices
                        if bd == device.name:
                            # Found the matching member device
                            # so will update attributes as
                            # a MDADM member
                            device.is_mdadm_member = True
                            if md.raid_type == 'linear':
                                device.is_mdadm_spanned = True
                            if md.raid_type == 'raid0':
                                device.is_mdadm_striped = True
                            if md.raid_type == 'raid1':
                                device.is_mdadm_mirrored = True
                            if md.raid_type == 'raid4':
                                device.is_mdadm_raid_4 = True
                            if md.raid_type == 'raid5':
                                device.is_mdadm_raid_5 = True
                            if md.raid_type == 'raid6':
                                device.is_mdadm_raid_6 = True
                            break

                    if md.name == device.name:
                        # Found the matching md? device so
                        # will update attributes as a MDADM device
                        # and add member block devices from sd? type

                        device.is_mdadm_device = True
                        if md.raid_type == 'linear':
                            device.is_mdadm_spanned = True
                        if md.raid_type == 'raid0':
                            device.is_mdadm_striped = True
                        if md.raid_type == 'raid1':
                            device.is_mdadm_mirrored = True
                        if md.raid_type == 'raid4':
                            device.is_mdadm_raid_4 = True
                        if md.raid_type == 'raid5':
                            device.is_mdadm_raid_5 = True
                        if md.raid_type == 'raid6':
                            device.is_mdadm_raid_6 = True
                        device.member_block_devices = md.block_devices
                        break

        if self.partitions_dict:
            # Update block devices with existing
            # partitions information
            for device in self.block_devices_dict.values():
                for part in self.partitions_dict.values():
                    if part.parent_block_device_name == device.name:
                        device.partitions.append({
                            'Mountpoint' : part.mount_point,
                            'Partition': part.name,
                            'SectorSize': part.sector_size_bytes,
                            'SizeGB': part.size_gbytes,
                            'UUID': part.uuid,
                            'fsType': part.file_system
                        })

        logger.debug("Block devices information updated.")

    def update_mount_points(self):
        '''
        Function
        =========
        update_mount_points

        Description
        ==============
        When mount points information was collected,
        additionnal information were not.

        Update mount points to inherit from block devices
        attributes.

        Arguments
        ==========
        __in <class> self

        Return Values
        ==============
        __out <dict> mount_points_dict: mount points dictionary

        '''

        if self.name():
            logger.debug(self.name() + ": update_mount_points")

        if self.block_devices_dict:
            for mp in self.mount_points_dict.values():
                for device in self.block_devices_dict.values():
                    if device.name in mp.parent_device:
                        mp.is_os_disk = device.is_os_disk
                        mp.is_data_disk = device.is_data_disk
                        mp.is_temp_disk = device.is_temp_disk
                        mp.is_lvm_device = device.is_lvm_device
                        mp.is_mdadm_device = device.is_mdadm_device
                        mp.is_lvm_member = device.is_lvm_member
                        mp.is_mdadm_member = device.is_mdadm_member
                        mp.is_lvm_striped = device.is_lvm_striped
                        mp.is_lvm_spanned = device.is_lvm_spanned
                        mp.is_lvm_mirrored = device.is_lvm_mirrored
                        mp.is_lvm_raid_4 = device.is_lvm_raid_4
                        mp.is_lvm_raid_5 = device.is_lvm_raid_5
                        mp.is_lvm_raid_6 = device.is_lvm_raid_6
                        mp.is_mdadm_striped = device.is_mdadm_striped
                        mp.is_mdadm_spanned = device.is_mdadm_spanned
                        mp.is_mdadm_mirrored = device.is_mdadm_mirrored
                        mp.is_mdadm_raid_4 = device.is_mdadm_raid_4
                        mp.is_mdadm_raid_5 = device.is_mdadm_raid_5
                        mp.is_mdadm_raid_6 = device.is_mdadm_raid_6

        logger.debug("Mount points information updated.")

    def update_partitions(self):
        '''
        Function
        =========
        update_partitions

        Description
        ==============
        When partitions information was collected,
        additionnal information were not.

        Update partitions to inherit from block devices
        attributes.

        Update partitions with its mount point
        file system.
        If the partition is a MDADM member device,
        update file system as being MDADM.

        Arguments
        ==========
        __in <class> self

        Return Values
        ==============
        __out <dict> mount_points_dict: mount points dictionary

        '''

        if self.name():
            logger.debug(self.name() + ": update_partitions")

        if self.block_devices_dict:
            for part in self.partitions_dict.values():
                for device in self.block_devices_dict.values():
                    if device.name == part.parent_block_device_name:
                        part.is_os_disk = device.is_os_disk
                        part.is_data_disk = device.is_data_disk
                        part.is_temp_disk = device.is_temp_disk
                        part.is_lvm_device = device.is_lvm_device
                        part.is_mdadm_device = device.is_mdadm_device
                        part.is_lvm_member = device.is_lvm_member
                        part.is_mdadm_member = device.is_mdadm_member
                        part.is_lvm_striped = device.is_lvm_striped
                        part.is_lvm_spanned = device.is_lvm_spanned
                        part.is_lvm_mirrored = device.is_lvm_mirrored
                        part.is_lvm_raid_4 = device.is_lvm_raid_4
                        part.is_lvm_raid_5 = device.is_lvm_raid_5
                        part.is_lvm_raid_6 = device.is_lvm_raid_6
                        part.is_mdadm_striped = device.is_mdadm_striped
                        part.is_mdadm_spanned = device.is_mdadm_spanned
                        part.is_mdadm_mirrored = device.is_mdadm_mirrored
                        part.is_mdadm_raid_4 = device.is_mdadm_raid_4
                        part.is_mdadm_raid_5 = device.is_mdadm_raid_5
                        part.is_mdadm_raid_6 = device.is_mdadm_raid_6
                        break

                if self.mount_points_dict:
                    for mp in self.mount_points_dict.values():
                        if part.mount_point == mp.path and part.file_system == '':
                            part.file_system = mp.file_system
                            break

                if self.mdadm_dict:
                    for md in self.mdadm_dict.values():
                        if '/dev/' + part.name in md.partitions and part.file_system == '':
                            part.file_system = 'MDADM device'
                            break

        logger.debug("Partitions information updated.")

    def update_mdadm_devices(self):
        '''
        Function
        =========
        update_mdadm_devices

        Description
        ==============
        When MDADM devices was collected,
        additionnal information were not.

        Update MDADM devices with associated
        mount point.

        Arguments
        ==========
        __in <class> self

        Return Values
        ==============
        __out <dict> mount_points_dict: mount points dictionary

        '''

        if self.name():
            logger.debug(self.name() + ": update_mdadm_devices")

        if self.mount_points_dict:
            for md in self.mdadm_dict.values():
                for mp in self.mount_points_dict.values():
                    if mp.parent_device == md.name:
                        md.mount_points.append({'MountPoint': mp.path})

        logger.debug("MDADM information updated.")

    @CachedProperty
    def block_device_map(self):
        '''
        Function
        =========
        block_device_map

        Description
        ==============
        Parse block_devices_dict and build the
        expected list of device dictionary that
        will be used to output the JSON file.

        Arguments
        ==========
        __in <class> self

        Return Values
        ==============
        __out <dict> block_devices_list: block devices dict list

        '''

        if self.name():
            logger.debug(self.name() + ": block_device_map")

        block_devices_list = []

        if self.block_devices_dict:
            for bd in self.block_devices_dict.values():
                partitions = []
                for part in bd.partitions:
                    partitions.append(part)
                member_devices = []
                for part in bd.member_block_devices:
                    member_devices.append(part)
                block_devices_list.append({
                    'BlockDevice': bd.name,
                    'HealthStatus' : 'N/A',
                    'IsDataDisk': bd.is_data_disk,
                    'IsLvmDevice': bd.is_lvm_device,
                    'IsMdadmDevice': bd.is_mdadm_device,
                    'IsLvmMember': bd.is_lvm_member,
                    'IsMdadmMember': bd.is_mdadm_member,
                    'IsOSDisk': bd.is_os_disk,
                    'IsTempDisk': bd.is_temp_disk,
                    'MemberBlockDevices' : member_devices,
                    'Partitions': partitions,
                    'SCSIBus': bd.scsi_bus,
                    'SCSILun': bd.scsi_lun,
                    'SCSIPort': bd.scsi_port,
                    'SCSITargetId': bd.scsi_target_id,
                    'SCSI_ID': bd.scsi_id,
                    'SizeGB': bd.size_gbytes
                        })

        if self.name():
            logger.debug(self.name() + ": block_device_map")

        return block_devices_list

    @CachedProperty
    def partition_map(self):
        '''
        Function
        =========
        partition_map

        Description
        ==============
        Parse partitions_dict and build the
        expected list of partitions dictionary that
        will be used to output the JSON file.

        Arguments
        ==========
        __in <class> self

        Return Values
        ==============
        __out <dict> partitions_list: partitions dict list

        '''

        if self.name():
            logger.debug(self.name() + ": partition_map")

        partitions_list = []

        if self.partitions_dict:
            for part in self.partitions_dict.values():
                part_dict = {}
                part_dict['PartitionName'] = part.name
                part_dict['UUID'] = part.uuid
                part_dict['FileSystem'] = part.file_system
                part_dict['IsOSDisk'] = part.is_os_disk
                part_dict['IsDataDisk'] = part.is_data_disk
                part_dict['IsTempDisk'] = part.is_temp_disk
                part_dict['IsLvmMember'] = part.is_lvm_member
                part_dict['IsMdadmMember'] = part.is_mdadm_member
                part_dict['LvmMember'] = part.is_lvm_device
                part_dict['MdadmMember'] = part.is_mdadm_device
                part_dict['IsStriped'] = part.is_lvm_striped or part.is_mdadm_striped
                part_dict['IsMirrored'] = part.is_lvm_mirrored or part.is_mdadm_mirrored
                part_dict['IsRaid5'] = part.is_lvm_raid_5 or part.is_mdadm_raid_5
                part_dict['IsSpanned'] = part.is_lvm_spanned or part.is_mdadm_spanned
                part_dict['MemberBlockDevice'] = 'TBD'
                part_dict['Mountpoint'] = part.mount_point
                part_dict['NumberOfFiles'] = part.number_of_files
                part_dict['SectorSize'] = part.sector_size_bytes
                part_dict['SizeGB'] = part.size_gbytes
                part_dict['FreeSpaceBytes'] = part.free_space_bytes
                part_dict['FreeSpaceGB'] = part.free_space_gbytes
                part_dict['Type'] = 'TBD'
                partitions_list.append(part_dict)

        if self.name():
            logger.debug(self.name() + ": partition_map")

        return partitions_list

    @CachedProperty
    def mount_point_map(self):
        '''
        Function
        =========
        mount_point_map

        Description
        ==============
        Parse mount_points_dict and build the
        expected list of mount points dictionary that
        will be used to output the JSON file.

        Arguments
        ==========
        __in <class> self

        Return Values
        ==============
        __out <dict> mount_points_list: mount points dict list

        '''

        if self.name():
            logger.debug(self.name() + ": mount_point_map")

        mount_points_list = []

        if self.mount_points_dict:
            for mp in self.mount_points_dict.values():
                mp_dict = {}
                mp_dict['Mountpoint'] = mp.path
                mp_dict['DeviceName'] = mp.parent_device
                mp_dict['FileSystem'] = mp.file_system
                mp_dict['SizeGB'] = mp.size_gbytes
                mp_dict['FreeSpaceGB'] = mp.free_space_gbytes
                mp_dict['IsOSDisk'] = mp.is_os_disk
                mp_dict['IsTempDisk'] = mp.is_temp_disk
                mp_dict['IsDataDisk'] = mp.is_data_disk
                mp_dict['IsLvm'] = mp.is_lvm_device
                mp_dict['IsMdadm'] = mp.is_mdadm_device
                mp_dict['IsMirrored'] = mp.is_lvm_mirrored or mp.is_mdadm_mirrored
                mp_dict['IsRaid5'] = mp.is_lvm_raid_5 or mp.is_mdadm_raid_5
                mp_dict['IsSpanned'] = mp.is_lvm_spanned or mp.is_mdadm_spanned
                mp_dict['IsStriped'] = mp.is_lvm_striped or mp.is_mdadm_striped
                mp_dict['NumberOfFiles'] = mp.number_of_files
                mount_points_list.append(mp_dict)

        if self.name():
            logger.debug(self.name() + ": mount_point_map")

        return mount_points_list

    @CachedProperty
    def lvm_map(self):
        '''
        Function
        =========
        lvm_map

        Description
        ==============
        Parse lvm_dict and build the
        expected list of LVM vgs dictionary that
        will be used to output the JSON file.

        Arguments
        ==========
        __in <class> self

        Return Values
        ==============
        __out <dict> lvm_list: LVM vg dict list

        '''

        if self.name():
            logger.debug(self.name() + ": lvm_map")

        lvm_list = []

        if self.lvm_dict:
            for vg in self.lvm_dict['VolumeGroups'].values():
                vg_dict = {}
                vg_lvs = []
                for vg_lv_name in vg.logical_volumes:
                    for lv in self.lvm_dict['LogicalVolumes'].values():
                        if vg_lv_name == lv.name:
                            vg_lvs.append({
                                'BlockDevice': lv.device_name,
                                'LogicalVolumeName': lv.name,
                                'LogicalVolumePath': '/dev/' + lv.vg_name + '/' + lv.name,
                                'LogicalVolumeSizeGB': lv.lv_size_gbytes,
                                'MapperPath': lv.dev_mapper_path,
                                'MountPoint': lv.mount_point,
                                'RAIDType': lv.raid_type,
                                'ReadAheadSectors': 0,
                                'SegmentsNumber': lv.segments_count,
                                'UUID': lv.uuid,
                                'PhysicalVolumes': lv.physical_volumes
                            })

                    pass
                vg_dict['Format'] = vg.lvm_format
                vg_dict['LogicalVolumes'] = vg_lvs
                vg_dict['LogicalVolumesCount'] = len(vg_dict['LogicalVolumes'])
                vg_dict['Partitions'] = ''
                vg_dict['PhysicalExtentSize'] = vg.extent_size_bytes
                vg_dict['UUID'] = vg.uuid
                vg_dict['VolumeGroupName'] = vg.name
                vg_dict['PhysicalVolumes'] = []
                for pv_name in vg.physical_volumes:
                    vg_dict['PhysicalVolumes'].append({'BlockDevice': pv_name})
                vg_dict['PhysicalVolumesCount'] = len(vg_dict['PhysicalVolumes'])
                vg_dict['lvmmapSizeGB'] = vg.vg_size_gbytes
                lvm_list.append(vg_dict)

        if self.name():
            logger.debug(self.name() + ": lvm_map")

        return lvm_list

    @CachedProperty
    def mdadm_map(self):
        '''
        Function
        =========
        mdadm_map

        Description
        ==============
        Parse mdadm_dict and build the
        expected list of MDADM devices dictionary that
        will be used to output the JSON file.

        Arguments
        ==========
        __in <class> self

        Return Values
        ==============
        __out <dict> mdadm_list: MDADM device dict list

        '''

        if self.name():
            logger.debug(self.name() + ": mdadm_map")

        mdadm_list  = []

        if self.mdadm_dict:
            for md in self.mdadm_dict.values():
                md_dict = {}
                md_dict['BlockDevice'] = md.name
                md_dict['CreationTime'] = md.creation_time
                md_dict['TotalDevices'] = md.total_devices_count
                md_dict['ActiveDevices'] = md.active_devices_count
                md_dict['RAIDDevices'] = md.raid_devices_count
                md_dict['WorkingDevices'] = md.working_devices_count
                md_dict['FailedDevices'] = md.failed_devices_count
                md_dict['SpareDevices'] = md.spare_devices_count
                md_dict['ArraySizeGB'] = md.size_gbytes
                md_dict['ChunkSize'] = md.chunk_size_kbytes
                md_dict['MemberBlockDevices'] = []
                for device in md.block_devices:
                    md_dict['MemberBlockDevices'].append({'BlockDevice': device})
                md_dict['MountPoints'] = md.mount_points
                md_dict['Partitions'] = []
                for part in md.partitions:
                    md_dict['Partitions'].append({'Partition': part})
                md_dict['Raid'] = md.raid_type
                md_dict['Version'] = md.version
                md_dict['Status'] = md.status
                mdadm_list.append(md_dict)

        if self.name():
            logger.debug(self.name() + ": mdadm_map")

        return mdadm_list


    class BlockDevice():
        '''
        Class
        =========
        BlockDevice

        Description
        ==============
        Class used to store block device
        information.
        '''

        def __init__(self, device_name):
            '''
            Function
            =========
            __init__

            Description
            ==============
            Initialize a block device object instance.

            Arguments
            ==========
            __in <class> self
            __in <string> device_name: block device name

            Return Values
            ==============
            N/A

            '''

            self.name = device_name
            self.device_mapper_name = DiskUtil.get_device_mapper_name_from_device(self.name.replace('/dev/', ''))
            self.scsi_id = DiskUtil.get_device_scsi_id(self.name.replace('/dev/', ''))
            self.scsi_bus = self.scsi_id['SCSIBus']
            self.scsi_port = self.scsi_id['SCSIPort']
            self.scsi_target_id = self.scsi_id['SCSITargetId']
            self.scsi_lun = self.scsi_id['SCSILun']
            self.is_os_disk = DiskUtil.is_azure_os_disk(self.scsi_bus, self.scsi_port, self.scsi_target_id, self.scsi_lun)
            self.is_data_disk = DiskUtil.is_azure_data_disk(self.scsi_bus, self.scsi_port, self.scsi_target_id, self.scsi_lun)
            self.is_temp_disk = DiskUtil.is_azure_temp_disk(self.scsi_bus, self.scsi_port, self.scsi_target_id, self.scsi_lun)
            self.is_lvm_device = False # DiskUtil.is_lvm_device(self.name.replace('/dev/', ''))
            self.is_mdadm_device = False # DiskUtil.is_mdadm_device(self.name.replace('/dev/', ''))
            self.size_gbytes = float("{0:.2f}".format(DiskUtil.get_block_device_size_gigabytes(self.name.replace('/dev/', ''))))
            self.sector_size = DiskUtil.get_block_device_sector_size_bytes(self.name.replace('/dev/', ''))
            self.health_status = ''
            # Following variables will be updated later on
            # when LVM and MDADM devices will be parsed
            self.is_lvm_member = False
            self.is_mdadm_member = False
            self.is_lvm_striped  = False
            self.is_lvm_spanned  = False
            self.is_lvm_mirrored  = False
            self.is_lvm_raid_4  = False
            self.is_lvm_raid_5  = False
            self.is_lvm_raid_6  = False
            self.is_mdadm_striped  = False
            self.is_mdadm_spanned  = False
            self.is_mdadm_mirrored  = False
            self.is_mdadm_raid_4  = False
            self.is_mdadm_raid_5  = False
            self.is_mdadm_raid_6  = False
            self.member_block_devices = []
            # Following variable will be updated later on
            # when partitions will be parsed
            self.partitions = []


    class Partition():
        '''
        Class
        =========
        Partition

        Description
        ==============
        Class used to store partition
        information.
        '''

        def __init__(self, partition):
            '''
            Function
            =========
            __init__

            Description
            ==============
            Initialize a partition object instance.

            Arguments
            ==========
            __in <class> self
            __in <tuple> partition: partition object retrieved
                from /proc/partitions

            Return Values
            ==============
            N/A

            '''

            self.name = partition[3]
            self.major = partition[0]
            self.minor = partition[1]
            # parent_block_device_name should be
            # /dev/sd? or /dev/dm-? or /dev/md?
            self.parent_block_device_name = PartitionUtil.get_partition_parent_block_device_name(self.name)
            # parent_mapper_device_name is provisionned
            # if parent device is a LVM logical volume
            self.parent_mapper_device_name = DiskUtil.get_device_mapper_name_from_device(self.parent_block_device_name.replace('/dev/', ''))
            self.mount_point = PartitionUtil.get_partition_mount_point(self.name, self.parent_mapper_device_name)
            self.sector_size_bytes = PartitionUtil.get_partition_block_size_bytes(self.mount_point)
            self.file_system = PartitionUtil.get_partition_fstype(self.name)
            self.size_bytes = PartitionUtil.get_partition_size_bytes(self.name)
            self.size_gbytes = float("{0:.2f}".format(float(self.size_bytes) / 1024/1024))
            self.free_space_bytes = PartitionUtil.get_partition_free_space_bytes(self.mount_point)
            self.free_space_gbytes = float("{0:.2f}".format(float(self.free_space_bytes) / 1024/1024/1024))
            self.available_space_bytes = PartitionUtil.get_partition_available_space_bytes(self.mount_point)
            self.available_space_gbytes = float("{0:.2f}".format(float(self.available_space_bytes) / 1024/1024/1024))
            self.number_of_files = PartitionUtil.get_partition_number_of_files(self.mount_point)
            self.uuid = PartitionUtil.get_partition_uuid(self.name)
            # Following variables will be updated later
            # when block devices, LVM and MDADM
            # information will be up to date
            self.is_os_disk = False
            self.is_data_disk = False
            self.is_temp_disk = False
            self.is_lvm_device = False
            self.is_mdadm_device = False
            self.is_lvm_member = False
            self.is_mdadm_member = False
            self.is_lvm_striped  = False
            self.is_lvm_spanned  = False
            self.is_lvm_mirrored  = False
            self.is_lvm_raid_4  = False
            self.is_lvm_raid_5  = False
            self.is_lvm_raid_6  = False
            self.is_mdadm_striped  = False
            self.is_mdadm_spanned  = False
            self.is_mdadm_mirrored  = False
            self.is_mdadm_raid_4  = False
            self.is_mdadm_raid_5  = False
            self.is_mdadm_raid_6  = False


    class MountPoint():
        '''
        Class
        =========
        MountPoint

        Description
        ==============
        Class used to store mount point
        information.
        '''

        def __init__(self, mount_point):
            '''
            Function
            =========
            __init__

            Description
            ==============
            Initialize a mount point object instance.

            Arguments
            ==========
            __in <class> self
            __in <tuple> mount_point: mount point object retrieved
                from /proc/mounts

            Return Values
            ==============
            N/A

            '''

            self.path = mount_point[1]
            # By default device associated with
            # a mount point can be a LVM logical
            # volume so parent_mapper_device_name
            # may have a device mapper device
            self.parent_device = self.parent_mapper_device_name = mount_point[0]
            if 'mapper' in self.parent_device:
                # If parent device is a dm device, get
                # the device short name (/dev/dm-?)
                self.parent_device = '/dev/' + DiskUtil.get_device_name_from_device_mapper(self.parent_device.split('/')[3])
            self.file_system = mount_point[2]
            self.size_bytes = PartitionUtil.get_partition_size_bytes(self.parent_device.replace('/dev/', ''))
            self.size_gbytes = float("{0:.2f}".format(float(self.size_bytes) / 1024/1024))
            self.free_space_bytes = PartitionUtil.get_partition_free_space_bytes(self.path)
            self.free_space_gbytes = float("{0:.2f}".format(float(self.free_space_bytes) / 1024/1024/1024))
            self.read_only = True if ('ro' in mount_point[3]) else False
            self.read_write = True if ('rw' in mount_point[3]) else False
            self.number_of_files = PartitionUtil.get_partition_number_of_files(self.path)
            # Following variables will be updated later
            # with block devices, LVM and MDADM
            # information
            self.is_os_disk = False
            self.is_data_disk = False
            self.is_temp_disk = False
            self.is_lvm_device = False
            self.is_mdadm_device = False
            self.is_lvm_member = False
            self.is_mdadm_member = False
            self.is_lvm_striped  = False
            self.is_lvm_spanned  = False
            self.is_lvm_mirrored  = False
            self.is_lvm_raid_4  = False
            self.is_lvm_raid_5  = False
            self.is_lvm_raid_6  = False
            self.is_mdadm_striped  = False
            self.is_mdadm_spanned  = False
            self.is_mdadm_mirrored  = False
            self.is_mdadm_raid_4  = False
            self.is_mdadm_raid_5  = False
            self.is_mdadm_raid_6  = False


    class LVMPhysicalVolume():
        '''
        Class
        =========
        LVMPhysicalVolume

        Description
        ==============
        Class used to store LVM physical
        volume information.

        Fields retrieved with pvs:
            #0 pv_name: Name
            #1 pv_uuid: Unique identifier
            #2 vg_name: Name
            #3 lv_name: Name.  LVs created for internal
                use are enclosed in brackets.
            #4 pv_fmt: Type of metadata
            #5 pv_size: Size of PV in current units
            #6 pv_free: Total amount of unallocated space
                in current units
            #7 pv_used: Total amount of allocated space
                in current units
            #8 dev_size: Size of underlying device in
                current units
            #9 pv_pe_count: Total number of Physical
                Extents
            #10 pv_pe_alloc_count: Total number of allocated
                Physical Extents
            
            Following fields are not available on CentOS 6.5:
            #11 pv_allocatable: Set if this device can be
                used for allocation
            #12 pv_exported: Set if this device is exported
            #13 pv_missing: Set if this device is missing
                in system

            Following fields are not available on Debian 8:
            #14 pv_major: Device major number
            #15 pv_minor: Device minor number
            #16 pv_in_use: Set if PV is used
            #17 pv_duplicate: Set if PV is an unchosen
                duplicate
        '''

        DEBIAN_8_FIELDS_COUNT = 14
        CENTOS_65_FIELDS_COUNT = 11

        def __init__(self, physical_volume):
            '''
            Function
            =========
            __init__

            Description
            ==============
            Initialize a LVM physical volume object
            instance.

            Arguments
            ==========
            __in <class> self
            __in <tuple> physical_volume: pv object
                retrieved from pvs command

            Return Values
            ==============
            N/A

            '''

            self.name = physical_volume[0].strip()
            self.pv_uuid = physical_volume[1]
            self.vg_name = physical_volume[2]
            self.logical_volumes = []
            self.lvm_format = physical_volume[4]
            self.update_pv(physical_volume)
            self.pv_size_bytes = int(physical_volume[5].replace('B', ''))
            self.pv_free_space_bytes = int(physical_volume[6].replace('B', ''))
            self.pv_used_space_bytes = int(physical_volume[7].replace('B', ''))
            self.device_size_bytes = float(physical_volume[8].replace('B', ''))
            self.extents = int(physical_volume[9])
            self.extents_allocatable = int(physical_volume[10])

            # Following fields are not available
            # on CentOS 6.5
            self.allocatable = (True if physical_volume[11] != '' else False) if len(physical_volume) > self.CENTOS_65_FIELDS_COUNT else None
            self.exported = (True if physical_volume[12] != '' else False) if len(physical_volume) > self.CENTOS_65_FIELDS_COUNT else None
            self.missing = (True if physical_volume[13] != '' else False) if len(physical_volume) > self.CENTOS_65_FIELDS_COUNT else None

            # Following fields are not available
            # on Debian 8
            self.major = physical_volume[14] if len(physical_volume) > self.DEBIAN_8_FIELDS_COUNT else None
            self.minor = physical_volume[15] if len(physical_volume) > self.DEBIAN_8_FIELDS_COUNT else None
            self.in_use = (True if physical_volume[16] != '' else False) if len(physical_volume) > self.DEBIAN_8_FIELDS_COUNT else None
            self.duplicate = (True if physical_volume[17] != '' else False) if len(physical_volume) > self.DEBIAN_8_FIELDS_COUNT else None

        def update_pv(self, physical_volume):
            '''
            Function
            =========
            update_pv

            Description
            ==============
            Update a LVM physical device object
            with logical volume membership.

            A physical volume can be used by
            several logical volumes so we add here
            the additional logical volumes the pv
            is part of.

            Arguments
            ==========
            __in <class> self
            __in <tuple> physical_volume: pv object
                retrieved from pvs command

            Return Values
            ==============
            N/A

            '''

            if physical_volume[3] != '':
                self.logical_volumes.append(physical_volume[3])


    class LVMVolumeGroup():
        '''
        Class
        =========
        LVMPhysicalVolume

        Description
        ==============
        Class used to store LVM volume group
        information.

        Fields retrieved with vgs:
            #0 vg_name: Name
            #1 vg_uuid: Unique identifier
            #2 vg_size: Total size of VG in current units
            #3 vg_free: Total amount of free space in current units
            #4 vg_extent_size: Size of Physical Extents in current units
            #5 vg_extent_count: Total number of Physical Extents
            #6 vg_free_count: Total number of unallocated Physical Extents
            #7 pv_count: Number of PVs in VG
            #8 lv_count: Number of LVs
            #9 snap_count: Number of snapshots
            #10 lv_name: Name.  LVs created for internal use are enclosed in brackets
            #11 Fmt: LVM version
            #12 devices

            Following fields are not available on Debian 8 and CentOS 6.5:
            #13 vg_missing_pv_count: Number of PVs in VG which are missing
        '''

        OLD_DISTROS_8_FIELDS_COUNT = 13

        def __init__(self, volume_group):
            '''
            Function
            =========
            __init__

            Description
            ==============
            Initialize a LVM volume group object
            instance.

            Arguments
            ==========
            __in <class> self
            __in <tuple> volume_group: vg object
                retrieved from vgs command

            Return Values
            ==============
            N/A

            '''

            self.name = volume_group[0].strip()
            self.uuid = volume_group[1]
            self.vg_size_bytes = int(volume_group[2].replace('B', ''))
            self.vg_size_gbytes = float("{0:.2f}".format(float(self.vg_size_bytes) / 1024/1024/1024))
            self.vg_free_space_bytes = int(volume_group[3].replace('B', ''))
            self.extent_size_bytes = int(volume_group[4].replace('B', ''))
            self.extent_count = int(volume_group[5])
            self.extent_free = int(volume_group[6])
            self.pv_count = int(volume_group[7])
            self.snapshot_count = int(volume_group[9])
            self.lv_count = int(volume_group[8])
            self.lvm_format = volume_group[11]
            self.logical_volumes = []
            self.physical_volumes = []
            self.partitions = []
            self.update_vg(volume_group)

            # Following fields are not available
            # on Debian 8
            self.pv_missing_count = int(volume_group[13]) if len(volume_group) > self.OLD_DISTROS_8_FIELDS_COUNT else None

        def update_vg(self, volume_group):
            '''
            Function
            =========
            update_vg

            Description
            ==============
            Update a LVM volume group object
            with logical volume membership.

            A volume group can regroup
            several logical volumes so we add here
            the additional logical volumes the vg
            is holding.

            Arguments
            ==========
            __in <class> self
            __in <tuple> volume_group: vg object
                retrieved from vgs command

            Return Values
            ==============
            N/A

            '''

            if volume_group[10] not in self.logical_volumes:
                self.logical_volumes.append(volume_group[10])


    class LVMLogicalVolume():
        '''
        Class
        =========
        LVMLogicalVolume

        Description
        ==============
        Class used to store LVM logical volume
        information.

        Fields retrieved with lvs:
            #0 lv_name: Name. LVs created for internal use are enclosed in brackets.
            #1 lv_uuid: Unique identifier
            #2 vg_name: Name
            #3 lv_size: Size of LV in current units
            #4 lv_metadata_size: For thin and cache pools, the size of the LV that holds the metadata
            #5 seg_count: Number of segments in LV
            #6 lv_time: Creation time of the LV, if known
            #7 copy_percent: For Cache, RAID, mirrors and pvmove, current percentage in-sync
            #8 stripes: Number of stripes or mirror/raid1 legs
            #9 stripe_size: For stripes, amount of data placed on one device before switching to the next
            #10 region_size: For mirrors/raids, the unit of data per leg when synchronizing devices
            #11 devices: physical volumes

            Following fields are not available on Ubuntu 14.04:
            #12 sync_percent: For Cache, RAID, mirrors and pvmove, current percentage in-sync

            Following fields are not available on CentOS 6.5:
            #13 lv_dm_path: Internal device-mapper pathname for LV (in /dev/mapper directory)
            #14 lv_parent: For LVs that are components of another LV, the parent LV
            #15 lv_layout: LV layout

            Following fields are not available on Debian 8:
            #16 cache_policy: The cache policy (cached segments only)
            #17 cache_total_blocks: Total cache blocks

            Following fields are not available on Ubuntu 16.04
            #18 seg_le_ranges: Ranges of Logical Extents of underlying devices in command line format

            Following fields are not available on Debian 8/9:
            #19 data_stripes: Number of data stripes or mirror/raid1 legs
            #20 data_copies: Number of data copies
        '''

        UBUNTU_14_FIELDS_COUNT = 12
        CENTOS_65_FIELDS_COUNT = 12
        DEBIAN_8_FIELDS_COUNT = 16
        UBUNTU_16_FIELDS_COUNT = 18
        DEBIAN_9_FIELDS_COUNT = 19

        def __init__(self, logical_volume):
            '''
            Function
            =========
            __init__

            Description
            ==============
            Initialize a LVM logical volume object
            instance.

            Arguments
            ==========
            __in <class> self
            __in <tuple> logical_volume: lv object
                retrieved from lvs command

            Return Values
            ==============
            N/A

            '''

            self.name = logical_volume[0].strip()
            self.uuid = logical_volume[1]
            self.vg_name = logical_volume[2]
            self.device_name = '/dev/' + DiskUtil.get_device_name_from_device_mapper(self.vg_name + '-' + self.name)    # Should be: /dev/?
            self.mount_point = '' # Will be updated later on
            self.lv_size_bytes = int(logical_volume[3].replace('B', ''))
            self.lv_size_gbytes = float("{0:.2f}".format(float(self.lv_size_bytes) / 1024/1024/1024))
            self.metadata_size_bytes = int(logical_volume[4].replace('B', '')) if logical_volume[4] != '' else 0
            self.segments_count = int(logical_volume[5])
            self.creation_time = logical_volume[6].replace(' +0000', '')
            self.copy_percent = logical_volume[7]
            self.raid1_stripes_count = int(logical_volume[8])
            self.stripe_size_bytes = int(logical_volume[9].replace('B', ''))
            self.region_size_bytes = int(logical_volume[10].replace('B', ''))

            # Will be updated later on
            self.physical_volumes = []
            self.raid_type = ''

            # Following field are not availabe on
            # Ubuntu 14.04
            self.sync_percent = logical_volume[12] if len(logical_volume) > self.UBUNTU_14_FIELDS_COUNT else None

            # Following fields are not available
            # on CentOS 6.5
            self.dev_mapper_path = ''
            self.parent = ''
            if len(logical_volume) > self.CENTOS_65_FIELDS_COUNT:
                self.dev_mapper_path = logical_volume[13]   # Should be: /dev/mapper/[vg]-[lv]
                self.parent = logical_volume[14]
                self.raid_type = logical_volume[15]
            else:
                if os.path.exists('/dev/mapper/' + self.vg_name + '-' + self.name):
                    # Fallback to get device mapper full path
                    self.dev_mapper_path = '/dev/mapper/' + self.vg_name + '-' + self.name

            # Following fields are not available
            # on Debian 8
            self.cache_policy = logical_volume[16] if len(logical_volume) > self.DEBIAN_8_FIELDS_COUNT else None
            self.cache_total_blocks = (int(logical_volume[17]) if logical_volume[17] else None) if len(logical_volume) > self.DEBIAN_8_FIELDS_COUNT else None

            # Following fields are not available
            # on Debian 8/9
            self.raid1_data_stripes_count = int(logical_volume[19]) if len(logical_volume) > self.DEBIAN_9_FIELDS_COUNT else None
            self.data_copies = int(logical_volume[20]) if len(logical_volume) > self.DEBIAN_9_FIELDS_COUNT else None

            # Update lv's physical volumes
            self.update_lv(logical_volume)

        def update_lv(self, logical_volume):
            '''
            Function
            =========
            update_lv

            Description
            ==============
            Update a LVM logical volume object
            with physical volume membership.

            A logical volume can use
            several physical volumes so we add here
            the additional physical volumes the lv
            is based on.
            We also add the extents information
            for each pv.

            Arguments
            ==========
            __in <class> self
            __in <tuple> logical_volume: lv object
                retrieved from lvs command

            Return Values
            ==============
            N/A

            '''

            if len(logical_volume) <= self.UBUNTU_16_FIELDS_COUNT:
                # Use devices to identify physical volumes
                # We will have extents start only.
                devices = logical_volume[11]
                if len(devices.split(',')) > 1:
                    for dev in devices.split(','):
                        self.physical_volumes.append({'device': dev.split('(')[0], 'extent_start': dev.split('(')[1].split(')')[0], 'extent_end': ''})
                else:
                    self.physical_volumes.append({'device': devices.split('(')[0], 'extent_start': devices.split('(')[1].split('-)')[0], 'extent_end': ''})
            else:
                # Use seg_le_ranges to identify physical volumes.
                # We will have extents start and end.
                devices = logical_volume[18]
                if len(devices.split(',')) > 1:
                    for dev in devices.split(','):
                        self.physical_volumes.append({'device': dev.split(':')[0], 'extent_start': dev.split(':')[1].split('-')[0], 'extent_end': dev.split(':')[1].split('-')[1]})
                else:
                    self.physical_volumes.append({'device': devices.split(':')[0], 'extent_start': devices.split(':')[1].split('-')[0], 'extent_end': devices.split(':')[1].split('-')[1]})


    class MDADMDevice():
        '''
        Class
        =========
        MDADMDevice

        Description
        ==============
        Class used to store MDADM device
        information.

        MDADM raid types:
            [raid0] [linear] [multipath] [raid1] [raid6] [raid5] [raid4] [raid10]
        '''

        def __init__(self, mdadm_device, mdadm_device_ex):
            '''
            Function
            =========
            __init__

            Description
            ==============
            Initialize a MDADM device object
            instance.

            Arguments
            ==========
            __in <class> self
            __in <tuple> mdadm_device: MDADM device object
            __in <tuple> mdadm_device_ex: MDADM device object
                extended information

            Return Values
            ==============
            N/A

            '''

            self.name = '/dev/' + mdadm_device[0]
            self.status = mdadm_device[2]
            self.raid_type = mdadm_device[3]
            self.version = mdadm_device_ex[3]
            self.status = ''
            self.sub_status = ''
            self.uuid = ''
            self.total_devices_count = 0
            self.active_devices_count = 0
            self.spare_devices_count = 0
            self.working_devices_count = 0
            self.failed_devices_count = 0
            self.raid_devices_count = 0
            self.size_kbytes = 0
            self.chunk_size_kbytes = 0
            self.creation_time = None
            self.mount_points = []
            self.block_devices = []
            self.partitions = []
            self.failed_devices_partitions = []
            self.spare_devices_partitions = []
            # Get extended information for
            # this MDADMdevice
            self.get_mdadm_info_ex(self.name)
            self.size_gbytes = float("{0:.2f}".format(float(self.size_kbytes) /1024/1024))

        def get_mdadm_info_ex(self, mdadm_device_name):
            '''
            Function
            =========
            get_mdadm_info_ex

            Description
            ==============
            Get MDADM device object extended
            information by using mdadm -Dq command.

            Update instance variables.

            Arguments
            ==========
            __in <class> self
            __in <string> mdadm_device_name: MDADM device name
            
            Return Values
            ==============
            N/A

            '''

            subproc_output = subprocess.Popen("mdadm -Dq " + mdadm_device_name + " 2> /dev/null", stdout=subprocess.PIPE, shell=True)
            mdadm_output = subproc_output.stdout.read().decode().splitlines()

            for line in mdadm_output:
                if 'Creation Time' in line:
                    self.creation_time = line.split(' : ')[1]
                if 'Array Size' in line:
                    self.size_kbytes = int(line.split()[3])
                if 'Raid Devices' in line:
                    self.raid_devices_count = int(line.split()[3])
                if 'Total Devices' in line:
                    self.total_devices_count = int(line.split()[3])
                if 'State :' in line:
                    self.status = line.split()[2]
                if 'Active Devices' in line:
                    self.active_devices_count = int(line.split()[3])
                if 'Working Devices' in line:
                    self.working_devices_count = int(line.split()[3])
                if 'Failed Devices' in line:
                    self.failed_devices_count = int(line.split()[3])
                if 'Spare Devices' in line:
                    self.spare_devices_count = int(line.split()[3])
                if 'Chunk Size' in line:
                    self.chunk_size_kbytes = int(line.split()[3].replace('K', ''))
                if 'UUID' in line:
                    self.uuid = line.split()[2]
                if 'Status' in line:
                    self.sub_status = line.split()[2]

            self.partitions = self.get_member_partitions(mdadm_output)
            self.block_devices = self.get_member_devices(self.partitions)
            self.failed_devices_partitions = self.get_failed_devices(mdadm_output)
            self.spare_devices_partitions = self.get_spare_devices(mdadm_output)

        @staticmethod
        def get_member_partitions(mdadm_output):
            '''
            Function
            =========
            get_member_partitions

            Description
            ==============
            Get MDADM device member partitions from
            mdadm command output.

            Arguments
            ==========
            __in <class> self
            __in <tuple> mdadm_output: mdadm device object
                from mdadm command output
            
            Return Values
            ==============
            __out <list> partitions_list: list of partitions

            '''

            partitions_list = []

            line_count = 0
            for line in mdadm_output:
                # Parse mdadm command output
                if 'Major' in line:
                    # Hit the beginning of the member
                    # devices list
                    line_count += 1
                    while line_count < len(mdadm_output):
                        if mdadm_output[line_count] and '/dev/sd' in mdadm_output[line_count]:
                            if mdadm_output[line_count].split()[4] == 'spare':
                                # This is a spare device
                                partitions_list.append(mdadm_output[line_count].split()[5])
                            elif mdadm_output[line_count].split()[4] == 'faulty':
                                # This is a faulty device
                                partitions_list.append(mdadm_output[line_count].split()[6])
                            else:
                                # This is an active device
                                partitions_list.append(mdadm_output[line_count].split()[6])
                        line_count += 1
                    break
                line_count += 1

            return partitions_list

        @staticmethod
        def get_failed_devices(mdadm_output):
            '''
            Function
            =========
            get_failed_devices

            Description
            ==============
            Get MDADM device failed member partitions from
            mdadm command output.

            Arguments
            ==========
            __in <class> self
            __in <tuple> mdadm_output: mdadm device object
                from mdadm command output
            
            Return Values
            ==============
            __out <list> failed_devices_list: list of failed devices

            '''

            failed_devices_list = []

            line_count = 0
            for line in mdadm_output:
                # Parse mdadm command output
                if 'Major' in line:
                    # Hit the beginning of the member
                    # devices list
                    line_count += 1
                    while line_count < len(mdadm_output):
                        if mdadm_output[line_count] and mdadm_output[line_count].split()[4] == 'faulty' and '/dev/sd' in mdadm_output[line_count]:
                            # This is a faulty device
                            failed_devices_list.append(mdadm_output[line_count].split()[6])
                        line_count += 1
                    break
                line_count += 1

            return failed_devices_list

        @staticmethod
        def get_spare_devices(mdadm_output):
            '''
            Function
            =========
            get_spare_devices

            Description
            ==============
            Get MDADM device spare member partitions from
            mdadm command output.

            Arguments
            ==========
            __in <class> self
            __in <tuple> mdadm_output: mdadm device object
                from mdadm command output
            
            Return Values
            ==============
            __out <list> spare_devices_list: list of spare devices

            '''

            spare_devices_list = []

            line_count = 0
            for line in mdadm_output:
                # Parse mdadm command output
                if 'Major' in line:
                    # Hit the beginning of the member
                    # devices list
                    line_count += 1
                    while line_count < len(mdadm_output):
                        if mdadm_output[line_count] and mdadm_output[line_count].split()[4] == 'spare' and '/dev/sd' in mdadm_output[line_count]:
                            # This is a spare device
                            spare_devices_list.append(mdadm_output[line_count].split()[5])
                        line_count += 1
                    break
                line_count += 1

            return spare_devices_list

        @staticmethod
        def get_member_devices(partitions):
            '''
            Function
            =========
            get_member_devices

            Description
            ==============
            Get MDADM device member block devices from
            MDADM partitions members.

            As MDADM relies on partitions, we need to
            parse the list of member partitions and get
            the parent device from it.

            Arguments
            ==========
            __in <class> self
            __in <list> partitions: mdadm device object
                member partitions list
            
            Return Values
            ==============
            __out <list> devices_list: list of block devices

            '''

            devices_list = []

            for partition in partitions:
                reg_ex = re.compile('\\/dev\\/sd[a-z]*')
                if reg_ex.match(partition):
                    devices_list.append(reg_ex.search(partition).group(0))

            return devices_list


class MachineMapInfoDC(IDataCollector):
    pci_map_cached = None
    os_details_map_cached = None
    kernel_details_map_cached = None
    bios_details_map_cached = None
    cpu_details_map_cached = None
    drivers_map_cached = None
    hyperv_modules_map_cached = None
    waagent_details_map_cached = None
    gpu_map_cached = None
    gpu_info_cached = None

    def __init__(self):
        self._dc_config = None

    def __del__(self):
        pass

    def init(self, dc_config):
       self._dc_config = dc_config

    def index(self):
        if self._dc_config:
            return self._dc_config['Index']
        else:
            return None

    def name(self):
        if self._dc_config:
            return self._dc_config['Name']
        else:
            return None

    def friendly_name(self):
        if self._dc_config:
            return self._dc_config['FriendlyName']
        else:
            return None

    def to_json(self):

        if self.name():
            logger.debug(self.friendly_name())

        machineInfoDict = {}

        data_collection_start_time = common.get_current_datetime("display")

        machineInfoDict["pci"] = self.pci_map
        machineInfoDict["os"] = self.os_details_map()
        machineInfoDict["Kernel"] = self.kernel_details_map()
        machineInfoDict["BIOS"] = self.bios_details_map()
        machineInfoDict["cpuinfo"] = self.cpu_details_map()
        machineInfoDict["Drivers"] = self.drivers_map()
        machineInfoDict["waagent"] = self.waagent_details_map()
        machineInfoDict["LISDrivers"] = self.hyperv_modules_map()
        machineInfoDict["GPU"] = self.gpu_map
        machineInfoDict["ProcessorCount"] = common.physical_processors_count
        machineInfoDict["VMArchitecture"] = self.vm_architecture()
        machineInfoDict["HyperThreadingInfo"] = self.hyperthreading_map()
        data_collection_end_time = common.get_current_datetime("display")
        datacollectorJSON = DataCollectorUtil.create_json(self.name(), data_collection_start_time, data_collection_end_time, machineInfoDict)

        logger.info("Machine configuration collected.")

        if self.name():
            logger.debug(self.name() + ": to_json")

        return datacollectorJSON

    @staticmethod
    def find_file(root, name):
        """
        Unix-style find utility. Only returns first result, and throws FileNotFoundError (OSError in Python 2) if it finds no matches.
        """
        from os import walk, path
        for dirpath, _, filenames in walk(root):
            if name in filenames:
                return path.join(dirpath, name)
        if six.PY2:
            raise OSError
        else:
            raise FileNotFoundError

    @CachedProperty
    def _pci_registry(self):
        '''
        A tuple of maps, vendor_id -> vendor_name and device_id -> device_name scraped from pci.ids file.
        '''
        vendor_regex = r'^([0-9a-fA-F]{4})  (.*)$'
        device_regex = r'^\t([0-9a-fA-F]{4})  (.*)$'
        vendor_map = {}
        device_map = {}
        current_vendor = None

        try:
            with open(self.find_file('/usr/share','pci.ids')) as f:
                lines = f.readlines()
        except OSError:
            logger.warning('Could not find PCI ID list')
            lines = []

        for line in lines:
            vendor_match = re.match(vendor_regex, line)
            device_match = re.match(device_regex, line)
            if vendor_match:
                vendor_id  = vendor_match.group(1)
                vendor_name = vendor_match.group(2)

                vendor_map[vendor_id] = vendor_name
                current_vendor = vendor_id
            elif current_vendor and device_match:
                device_id = device_match.group(1)
                device_name = device_match.group(2)

                full_id = current_vendor + device_id
                device_map[full_id] = device_name

        return vendor_map, device_map

    @CachedProperty
    def _pci_devices(self):
        '''
        List of the device IDs in /proc/bus/pci/devices.
        '''
        try:
            with open('/proc/bus/pci/devices') as f:
                lines = f.readlines()
        except EnvironmentError:
            logger.warning('Could not find PCI device information')
            lines = []
        return [ line.split('\t')[1] for line in lines ]

    @CachedProperty
    def pci_map(self):
        '''
        Get the vendor and name of each installed pci device.
        '''
        if self.name():
            logger.debug(self.name() + ": pci_map")

        vendor_map,device_map = self._pci_registry
        
        pci_map = [
            {
                "vendor": vendor_map[full_id[:4]],
                "device": device_map[full_id]
            }
            for full_id in self._pci_devices
            if full_id[:4] in vendor_map and full_id in device_map
        ]

        logger.info("PCI information collected.")
        return pci_map

    def os_details_map(self):
        '''
        Get the Details of the Linux Operating system installed version
        '''
        if (MachineMapInfoDC.os_details_map_cached is not None):
            logger.debug(self.name() + ": os_details_map_cached")
            return MachineMapInfoDC.os_details_map_cached

        logger.debug("Entering os_details_map")

        os_release = {}
        os_platform = distro.linux_distribution()
        os_release['name'] = os_platform[0]
        os_release['version'] = os_platform[1] + ' ' + os_platform[2]
        os_release['version_id'] = os_platform[1]
        os_release['id'] = ' '.join(os_platform)

        os_release_file = None
        if os.path.exists('/etc/os-release'):
            os_release_file = '/etc/os-release'
        elif os.path.exists(('/etc/{0}-release').format(os_platform[0].lower())):
            os_release_file = ('/etc/{0}-release').format(os_platform[0].lower())
        elif 'ubuntu' in os_platform[0].lower():
            os_release_file = '/etc/lsb-release'

        try:
            if os_release_file is not None:
                with open(os_release_file, 'r') as f:
                    for line in f:
                        line = line.replace('"','').replace('\n','').split('=')
                        os_release[line[0].lower()] = line[1]
        except:
            logger.warning("OS information may be partially collected.")

        MachineMapInfoDC.os_details_map_cached = os_release

        return os_release

    def kernel_details_map(self):

        if (MachineMapInfoDC.kernel_details_map_cached is not None):
            logger.debug(self.name() + ": kernel_details_map_cached")
            return MachineMapInfoDC.kernel_details_map_cached

        logger.debug("Entering get_kernel_details")

        kernel_rel={}

        kernel_rel["current"]=platform.release()
        if os.path.isdir('/boot'):
            allimgs = os.listdir("/boot")
            kernel_rel["availableKernels"]=list(filter(re.compile("vmlinuz.*").match,allimgs))
            kernel_rel["initimgs"] = list(filter(re.compile("initr[ad].*").match, allimgs))
            kernel_rel["config"]=list(filter(re.compile("config.*").match,allimgs))
            kernel_rel["sys_map"]=list(filter(re.compile("System.*").match,allimgs))
        else:
            logger.warning("Could not find /boot directory. Kernel image details will be incomplete")

        logger.info("Kernel information collected.")
        MachineMapInfoDC.kernel_details_map_cached = kernel_rel
        return kernel_rel

    def bios_details_map(self):
        if (MachineMapInfoDC.bios_details_map_cached is not None):
            logger.debug(self.name() + ": bios_details_map_cached")
            return MachineMapInfoDC.bios_details_map_cached

        logger.debug("Entering get_bios_details")

        bios={}
        try:
            keys = os.listdir("/sys/devices/virtual/dmi/id")
            for key in keys:
                if os.path.isdir("/sys/devices/virtual/dmi/id/"+key):
                    continue
                else:
                    with open("/sys/devices/virtual/dmi/id/"+key) as value:
                        bios[key]=value.read().replace('\n','')

            logger.info("BIOS information collected.")
        except Exception as e:
            logger.warning("Could not collect BIOS information: %s" % (e))

        MachineMapInfoDC.bios_details_map_cached = bios

        return bios

    def vm_architecture(self):
        uname = platform.uname()
        # return processor architecture value in uname [5]
        # if unavailable, use machine hardware value [4]
        if len(uname) > 5 and len(uname[5]) > 0 and uname[5].lower() is not 'unknown':
            return uname[5]
        else:
            return uname[4]

    def hyperthreading_map(self):
        hyperthreading_data = {}
        cores_per_socket = None
        sockets = None

        try:
            lscpu = subprocess.check_output(['lscpu']).split(b'\n')

            for line in lscpu:
                if line.startswith(b'Thread(s) per core:'):
                    hyperthreading_data["ThreadsPerCore"] = int(line.split(b':')[1])
                elif line.startswith(b'CPU(s):'):
                    hyperthreading_data["LogicalProcessors"] = int(line.split(b':')[1])
                # the following fields are used to calculate Number of Cores and are not added directly to hyperthreading_data
                elif line.startswith(b'Core(s) per socket:'):
                    cores_per_socket = int(line.split(b':')[1])
                elif line.startswith(b'Socket(s):'):
                    sockets = int(line.split(b':')[1])
        except:
            logger.warning("Hyperthreading info obtained from the lscpu command may be fully or partially missing.")

        # number of cores can be calculated by multplying socket count by cores per socket
        if cores_per_socket and sockets:
            hyperthreading_data["NumberOfCores"] = cores_per_socket * sockets

        return hyperthreading_data

    def cpu_details_map(self):
        if (MachineMapInfoDC.cpu_details_map_cached is not None):
            logger.debug(self.name() + ": cpu_details_map_cached")
            return MachineMapInfoDC.cpu_details_map_cached

        logger.debug("Entering get_cpu_details")

        processor={}
        kv=[]
        cpuinfo=[]
        count=0
        with open("/proc/cpuinfo") as cpu:
            for c in cpu:
                kv = c.split(":")
                try:
                    if ("processor" in kv[0].strip())  and count==0:
                        count+=1
                    elif ("processor" in kv[0].strip())  and count!=0:
                        count+=1
                        cpuinfo.append(processor)
                        processor={}
                    processor[kv[0].replace(" ","").strip()]=kv[1].strip().replace('\n','')
                except:
                    pass
            cpuinfo.append(processor)

        logger.info("CPU information collected.")
        MachineMapInfoDC.cpu_details_map_cached = cpuinfo
        return cpuinfo

    def drivers_map(self):
        if (MachineMapInfoDC.drivers_map_cached is not None):
            logger.debug(self.name() + ": drivers_map_cached")
            return MachineMapInfoDC.drivers_map_cached

        logger.debug("Entering getModules")

        drivers = []
        path="/proc/modules"
        with open(path) as modules:
            for module in modules.readlines():
                driver={}
                module=module.strip().split(' ')
                driver["FilterName"]=module[0]
                driver["sizeInBytes"]=module[1]
                driver["NumInstances"]=module[2]
                driver["dependencies"]=module[3]
                driver["state"]=module[4]
                driver["kernelMemoryOffset"]=module[5]
                drivers.append(driver)

        logger.info("Drivers information collected.")

        MachineMapInfoDC.drivers_map_cached = drivers

        return drivers

    def waagent_details_map(self):
        '''Collects Azure VM Agent details. Calls two other functions to do so \
        - waagent_status() and waagent_binarylocation()

        Returns :
            Agent version, status, binary location.
        '''
        if (MachineMapInfoDC.waagent_details_map_cached is not None):
            logger.debug(self.name() + ": waagent_details_map_cached")
            return MachineMapInfoDC.waagent_details_map_cached

        logger.debug("Entering waagent_details")

        waagent_dict = {}

        agent = None
        try:
            path = "/usr/lib"
            for i in os.listdir(path):
                if str(i).startswith("python"):
                    #ignore false-positive linter warnings here. walk returns a 3-tuple
                    for dirroot, dirs, files in os.walk(path+"/"+i): #pylint: disable=W0612
                        if dirroot.endswith("azurelinuxagent"):
                            packagedir = dirroot.strip("azurelinuxagent")
                            sys.path.insert(0, packagedir)
                            #ignore the linter warning here. This works fine during runtime
                            import azurelinuxagent.agent as agent #pylint: disable=E0401

            if agent is not None:
                try:
                    waagent_dict = {
                            "version" : agent.AGENT_LONG_VERSION,
                            "binarypath" : agent.__file__,
                            "config" : self.waagent_config_map()
                        }
                except IOError:
                    waagent_dict = {
                        "version" : "None",
                        "binarypath" : agent.__file__,
                        "config" : self.waagent_config_map()
                    }

                logger.info("Azure guest agent information collected.")

        except:
            logger.exception("Could not load guest agent information.")

        if agent is None:
            logger.warning("VMAgent is not installed or is lower than version 2.1.0")

        MachineMapInfoDC.waagent_details_map_cached = waagent_dict

        return waagent_dict

    def waagent_config_map(self):
        ''' Collects Linux Guest Agent Configuration - contents of /etc/waagent.conf

        Returns:
            Flat dictionary of config values in /etc/waagent.conf
            e.g. ResourceDisk.EnableSwap=y -> { "ResourceDisk.EnableSwap" : "y" }
        '''
        try:
            with open('/etc/waagent.conf', 'r') as f:
                lines = f.readlines()
        except EnvironmentError:
            logger.warning("/etc/waagent.conf not found")
            return None
        
        config = lambda line : line.strip().split('=')[0]
        value = lambda line : line.strip().split('=')[1]
        is_entry = lambda line : '#' not in line and '=' in line
        return {
            config(line) : value(line) for line in lines if is_entry(line)
        } 

    def hyperv_modules_map(self):
        '''Collects Hyperv driver details - hv_balloon, hv_netvsc,
        hv_storvsc, hv_utils, hv_vmbus, hyperv_fb and hyperv_keyboard
        
        Returns:
            Hyperv driver details including - alias, dependency, signature key, version details
        '''
        if (MachineMapInfoDC.hyperv_modules_map_cached is not None):
            logger.debug(self.name() + ": hyperv_modules_map_cached")
            return MachineMapInfoDC.hyperv_modules_map_cached

        import subprocess

        logger.debug("Entering hyperv_module_collection")

        hv_list=[]
        hv_details={}
        drivers_list=['hv_vmbus','hv_balloon','hv_keyboard','hv_utils','hv_netvsc','hv_storvsc']
        for driver in drivers_list:
            try:
                output = subprocess.Popen(['modinfo',driver],stdout=subprocess.PIPE,stderr=subprocess.STDOUT).communicate()
                hv_details["name"]=driver
            except Exception as e:
                logger.warning("Could not collect Hyperv driver details for driver %s: %s" % (driver, e))
                continue
            for entry in output:
                try:
                    entry_list=entry.split(':')
                    hv_details[entry_list[0]]=entry_list[1].strip()
                except:
                    pass
            hv_list.append(hv_details)
            hv_details={}

        logger.info("LIS details collected.")

        MachineMapInfoDC.hyperv_modules_map_cached = hv_list

        return hv_list

    @CachedProperty
    def gpu_map(self):
        import xml.etree.ElementTree as ET

        logger.debug("Entering get_gpumap")

        try:
            p = subprocess.Popen(['nvidia-smi', '-q', '-x'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            nvidia_output = p.communicate()[0].decode()
            root = ET.fromstring(nvidia_output)
            gpu_map = [
                {
                    'Id': gpu.attrib['id'],
                    'Model': gpu.find('product_name').text,
                    'UUID': gpu.find('uuid').text,
                    'DriverVersion': root.find('driver_version').text,
                    'DisplayMode': gpu.find('display_mode').text,
                    'SerialNumber': gpu.find('serial').text,
                    'VBIOSVersion': gpu.find('vbios_version').text
                } for gpu in root.findall('gpu')
            ]
        except (OSError, ET.ParseError):
            gpu_map = []

        if len(gpu_map) > 0:
            logger.info("GPU information collected.")
        else:
            logger.info("No GPU information collected.")

        return gpu_map

    def waagent_status(self):
        '''Collects waagent status detail - whether it is in a Running state.

        Returns :
            waagent status.
        '''

        service_state = "Not Running"

        logger.debug("Entering waagent_status")

        try:
            import psutil

            for p in psutil.process_iter():
                cmdstring = p.get_cmdline()
                service_state = "Not Running"
                if any("waagent" in str for str in cmdstring):
                    service_state = "Running"
                    return(service_state)

            logger.info("waagent status collected.")
        except ImportError:
            logger.warning("psutil module not present. waagent status information will not be collected.")

        return(service_state)

    def waagent_binarylocation(self):
        '''Collects waagent binary path.

        Returns :
            waagent binary path.
        '''

        binary_location = "Unknown"

        logger.debug("Entering waagent_binarylocation")

        try:
            import psutil

            for p in psutil.process_iter():
                cmdstring = p.get_cmdline()
                binary_location = "Unknown"
                if any("waagent" in str for str in cmdstring):
                    binary_location = os.path.dirname(cmdstring[2])
                    return(binary_location)

            logger.info("waagent binary location collected.")
        except ImportError:
            logger.warning("psutil module not present. waagent binary information will not be collected.")

        return(binary_location)

    def get_processors_number(self):

        import multiprocessing
        return multiprocessing.cpu_count()


class NetworkMapInfoDC(IDataCollector):
    routing_table_map_cached = None
    interfaces_map_cached = None
    firewall_map_cached = None

    def __init__(self):
        self._dc_config = None

    def __del__(self):
        pass

    def init(self, dc_config):
        self._dc_config = dc_config

    def index(self):
        if self._dc_config:
            return self._dc_config['Index']
        else:
            return None

    def name(self):
        if self._dc_config:
            return self._dc_config['Name']
        else:
            return None

    def friendly_name(self):
        if self._dc_config:
            return self._dc_config['FriendlyName']
        else:
            return None

    def to_json(self):

        if self.name():
            logger.debug(self.name() + ": to_json")

        networkInfoDict={}
        data_collection_start_time = common.get_current_datetime("display")

        networkInfoDict["routing_table"] = self.routing_table_map()
        networkInfoDict["Netstat"] = self.network_stat_instantsnapshot()
        networkInfoDict["Interface"] = self.interfaces_map()
        networkInfoDict["firewalldFirewallRules"] = self.firewall_map()

        data_collection_end_time = common.get_current_datetime("display")
        datacollectorJSON = DataCollectorUtil.create_json(self.name(), data_collection_start_time, data_collection_end_time, networkInfoDict)
        #datacollectorJSON = common.create_datacollector_json("NetworkInformationCollector",networkInfoDict,data_collection_start_time,data_collection_end_time)

        logger.info("Network configuration collected.")

        if self.name():
            logger.debug(self.name() + ": to_json")

        return datacollectorJSON

    def interfaces_map(self):
        if (NetworkMapInfoDC.interfaces_map_cached is not None):
            logger.debug(self.name() + ": interfaces_map_cached")
            return NetworkMapInfoDC.interfaces_map_cached

        interfaceDict=[]

        ipv6 = self.get_ipv6()
        ifaces = os.listdir("/sys/class/net")

        for iface in ifaces:
            ip_mask = self.get_ipv4_mask(iface)
            statistics = self.get_interface_stats(iface)
            if iface in ipv6:
                interfaceDict.append({"InterfaceAlias":iface,"NetworkStatistics":statistics["completeStatistics"],"isNetworkAccelerationEnabled":statistics["isNetworkAccelerationEnabled"],"Hwdetails":self.get_mac_address(iface),"Index":self.get_iface_index(iface),"IPv4":ip_mask[0],"IPv6":ipv6[iface],"Mask":ip_mask[1],"NlMtu":self.iface_mtu(iface)})
            else:
                interfaceDict.append({"InterfaceAlias":iface,"NetworkStatistics":statistics["completeStatistics"],"isNetworkAccelerationEnabled":statistics["isNetworkAccelerationEnabled"],"Hwdetails":self.get_mac_address(iface),"Index":self.get_iface_index(iface),"IPv4":ip_mask[0],"IPv6":None,"Mask":ip_mask[1],"NlMtu":self.iface_mtu(iface)})

        NetworkMapInfoDC.interfaces_map_cached = interfaceDict

        return interfaceDict

    def routing_table_map(self):
        if (NetworkMapInfoDC.routing_table_map_cached is not None):
            logger.debug(self.name() + ": routing_table_map_cached")
            return NetworkMapInfoDC.routing_table_map_cached

        routingtbl=[]
        with open('/proc/net/route','r') as rt:
            rt_keys = rt.readline().split()
            for value in rt:
                values=value.split()
                values[1] = self.hex_to_ip(values[1])
                values[2] = self.hex_to_ip(values[2])
                values[7] = self.hex_to_ip(values[7])
                values[3] = self.flag_conversion(values[3])
                routingtbl.append(dict(zip(rt_keys,values)))

        NetworkMapInfoDC.routing_table_map_cached = routingtbl

        return routingtbl

    def firewall_map(self):
        '''Collects firewall config details from /etc/firewalld/zones directory.
        Uses the xmlparsingrecurse function to get nested json

        Returns :
            Json output with rules for each zone defined on the machine.
            Also includes the rich rules.
        '''
        if (NetworkMapInfoDC.firewall_map_cached is not None):
            logger.debug(self.name() + ": firewall_map_cached")
            return NetworkMapInfoDC.firewall_map_cached

        logger.debug("Entering firewall_config")
        firewall_dict = {}

        if os.path.isdir("/etc/firewalld/zones/"):
            path = "/etc/firewalld/zones/"
            for filename in os.listdir(path):
                if filename.endswith(".xml"):
                    zone = filename.strip(".xml")
                    tree = ET.parse( path + filename )
                    root = tree.getroot()
                    firewall_dict[zone] = self.xmlparsingrecurse(root)
            logger.info("Firewall information collected.")
        else :
            logger.warning("Failed to collect Firewall information. Folder /etc/firewalld/zones does not exist.")
            pass

        NetworkMapInfoDC.firewall_map_cached = firewall_dict

        return firewall_dict

    def network_stat_instantsnapshot(self, open_file=open):

        import pwd
        import glob
        logger.debug("Entering netstat")
        result=[]
        PROC_TCP4 = "/proc/net/tcp"
        PROC_UDP4 = "/proc/net/udp"
        PROC_TCP6 = "/proc/net/tcp6"
        PROC_UDP6 = "/proc/net/udp6"
        PROC_PACKET = "/proc/net/packet"
        TCP_STATE = {
                '01':'ESTABLISHED',
                '02':'SYN_SENT',
                '03':'SYN_RECV',
                '04':'FIN_WAIT1',
                '05':'FIN_WAIT2',
                '06':'TIME_WAIT',
                '07':'CLOSE',
                '08':'CLOSE_WAIT',
                '09':'LAST_ACK',
                '0A':'LISTEN',
                '0B':'CLOSING'
                }

        def _tcp4load():  # type: () -> List[str]
            ''' Read the table of tcp connections & remove the header  '''
            try:
                with open_file(PROC_TCP4,'r') as f:
                    content = f.readlines()
                    content.pop(0)
                return content
            except Exception as e:
                logger.warning("Unable to read tcp4 connections info", exc_info=e)
                return []

        def _tcp6load():  # type: () -> List[str]
            ''' Read the table of tcpv6 connections & remove the header'''
            try:
                with open_file(PROC_TCP6,'r') as f:
                    content = f.readlines()
                    content.pop(0)
                return content
            except Exception as e:
                logger.warning("Unable to read tcp6 connections info", exc_info=e)
                return []

        def _udp4load():  # type: () -> List[str]
            '''Read the table of udp connections & remove the header '''
            try:
                with open_file(PROC_UDP4,'r') as f:
                    content = f.readlines()
                    content.pop(0)
                return content
            except Exception as e:
                logger.warning("Unable to read udp4 info", exc_info=e)
                return []

        def _udp6load():  # type: () -> List[str]
            '''Read the table of udp connections & remove the header '''
            try:
                with open_file(PROC_UDP6,'r') as f:
                    content = f.readlines()
                    content.pop(0)
                return content
            except Exception as e:
                logger.warning("Unable to read udp6 info", exc_info=e)
                return []

        def _packetload():
            ''' Read the contents of /proc/net/packet for finding processes with packet sockets '''
            try:
                with open(PROC_PACKET,'r') as f:
                    content = f.readlines()
                    content.pop(0)
                return content
            except Exception as e:
                logger.warning("Unable to read packet info", exc_info=e)
                return []

        def _hex2dec(s):
            return str(int(s,16))

        def _ip(s):
            ip = [(_hex2dec(s[6:8])),(_hex2dec(s[4:6])),(_hex2dec(s[2:4])),(_hex2dec(s[0:2]))]
            return '.'.join(ip)

        def _ip6(s):
            #this may need to be converted to a string to work properly.
            ip = [s[6:8],s[4:6],s[2:4],s[0:2],s[12:14],s[14:16],s[10:12],s[8:10],s[22:24],s[20:22],s[18:20],s[16:18],s[30:32],s[28:30],s[26:28],s[24:26]]
            return ':'.join(ip)

        def _remove_empty(array):
            return [x for x in array if x !='']

        def _convert_ipv4_port(array):
            host,port = array.split(':')
            return _ip(host),_hex2dec(port)

        def _convert_ipv6_port(array):
            host,port = array.split(':')
            return _ip6(host),_hex2dec(port)

        def netstat_tcp4(open_file=open):
            '''
            Function to return a list with status of tcp connections on Linux systems.
            Please note that in order to return the pid of of a network process running on the
            system, this script must be ran as root.
            '''
            logger.debug("Entering netstat_tcp4")
            tcpcontent =_tcp4load()
            for line in tcpcontent:
                line_array = _remove_empty(line.split(' '))     # Split lines and remove empty spaces.
                l_host,l_port = _convert_ipv4_port(line_array[1]) # Convert ipaddress and port from hex to decimal.
                r_host,r_port = _convert_ipv4_port(line_array[2])
                #tcp_id = line_array[0]
                state = TCP_STATE[line_array[3]]
                uid = get_user(line_array[7])                   # Get user from UID.
                inode = line_array[9]                           # Need the inode to get process pid.
                pid = _get_pid_of_inode(inode)                  # Get pid prom inode.
                try:                                            # try read the process name.
                    exe = os.readlink('/proc/'+pid+'/exe')
                except:
                    exe = None

                nline = {"Protocol":"TCP", "User":uid, "LocalAddress":l_host,"LocalPort":l_port, "RemoteAddress":r_host,"RemotePort":r_port, "State":state, "PID":pid, "ProcessName":exe}
                result.append(nline)
            return result

        def netstat_tcp6(open_file=open):
            '''
            This function returns a list of tcp connections utilizing ipv6. Please note that in order to return the pid of of a
            network process running on the system, this script must be ran as root.
            '''
            logger.debug("Entering netstat_tcp6")
            tcpcontent = _tcp6load()
            for line in tcpcontent:
                line_array = _remove_empty(line.split(' '))
                l_host,l_port = _convert_ipv6_port(line_array[1])
                r_host,r_port = _convert_ipv6_port(line_array[2])
                #tcp_id = line_array[0]
                state = TCP_STATE[line_array[3]]
                uid = get_user(line_array[7])
                inode = line_array[9]
                pid = _get_pid_of_inode(inode)
                try:                                            # try read the process name.
                    exe = os.readlink('/proc/'+pid+'/exe')
                except:
                    exe = None

                nline = {"Protocol":"TCP", "User":uid, "LocalAddress":l_host,"LocalPort":l_port, "RemoteAddress":r_host,"RemotePort":r_port, "State":state, "PID":pid, "ProcessName":exe}
                result.append(nline)
            return result

        def netstat_udp4(open_file=open):
            '''
            Function to return a list with status of udp connections on Linux systems. Please note that UDP is stateless, so
            state will always be blank. Please note that in order to return the pid of of a network process running on the
            system, this script must be ran as root.
            '''
            logger.debug("Entering netstat_udp4")
            udpcontent =_udp4load()
            for line in udpcontent:
                line_array = _remove_empty(line.split(' '))
                l_host,l_port = _convert_ipv4_port(line_array[1])
                r_host,r_port = _convert_ipv4_port(line_array[2])
                #udp_id = line_array[0]
                udp_state ='Stateless' #UDP is stateless
                uid = get_user(line_array[7])
                inode = line_array[9]
                pid = _get_pid_of_inode(inode)
                try:
                    exe = os.readlink('/proc/'+pid+'/exe')
                except:
                    exe = None
                nline = {"Protocol":"UDP", "User":uid, "LocalAddress":l_host,"LocalPort":l_port, "RemoteAddress":r_host,"RemotePort":r_port, "State":udp_state, "PID":pid, "ProcessName":exe}
                result.append(nline)
            return result

        def get_user(uid):
            try:
                return pwd.getpwuid(int(uid))[0]
            except Exception as e:
                logger.warning("Unable to get to user. Defaulting to uid.", exc_info=e)
                return uid
    
        def netstat_udp6(open_file=open):
            '''
            Function to return a list of udp connection utilizing ipv6. Please note that UDP is stateless, so state will always
            be blank. Please note that in order to return the pid of of a network process running on the system, this script
            must be ran as root.
            '''
            logger.debug("Entering netstat_udp6")
            udpcontent =_udp6load()
            for line in udpcontent:
                line_array = _remove_empty(line.split(' '))
                l_host,l_port = _convert_ipv6_port(line_array[1])
                r_host,r_port = _convert_ipv6_port(line_array[2])
                #udp_id = line_array[0]
                udp_state ='Stateless' #UDP is stateless
                uid = get_user(line_array[7])
                inode = line_array[9]
                pid = _get_pid_of_inode(inode)
                try:
                    exe = os.readlink('/proc/'+pid+'/exe')
                except:
                    exe = None

                nline = {"Protocol":"UDP", "User":uid, "LocalAddress":l_host,"LocalPort":l_port, "RemoteAddress":r_host,"RemotePort":r_port, "State":udp_state, "PID":pid, "ProcessName":exe}
                result.append(nline)
            return result

        def _get_pid_of_inode(inode):
            '''
            To retrieve the process pid, check every running process and look for one using
            the given inode.
            '''
            logger.debug("Entering _get_pid_of_inode")
            for item in glob.glob('/proc/[0-9]*/fd/[0-9]*'):
                try:
                    if re.search(inode,os.readlink(item)):
                        return item.split('/')[2]
                except:
                    pass
            return None

        netstat_tcp4(open_file);netstat_tcp6(open_file);netstat_udp4(open_file);netstat_udp6(open_file)
        return result

    def xmlparsingrecurse(self, root):
        '''
        Loops over the ElemenTree object and returns child items in nested list of dict format.
        Args :
            Root object from ElementTree of the zone xml obtained in firewall_config function.
        
        Returns :
            Child items of the zone xml root in nested list of dict format.
            These are the firewall rules defined in the zone.
        '''
        list = []
        for subelem in root:
            if subelem.getchildren() == []:
                new_rule = {subelem.tag:subelem.attrib}
                list.append(new_rule)
            else:
                new_rule = {subelem.tag:self.xmlparsingrecurse(subelem)}
                list.append(new_rule)
        return list

    def get_mac_address(self, iface):
        '''
        Function
        =========
        get_mac_address.

        Description
        ==============
        Gets the MAC address for the given Interface present within the /sys/class/net.

        Arguments
        ==========
            __in <str> iface : Interface like eth0, eth1 ...etc
        
        Return Values
        ==============

            __out <str> hwaddress : MAC address of the eth0, eth1 ..etc interface.
        '''
        #import urllib2
        logger.debug("Entering get_mac_address")
        vendors = {}

        with open('/sys/class/net/'+iface+'/address','r') as hw:
            hw_addr = hw.readline().strip()
            #x=hw_addr.split("\n")
        '''
        while(True):
            try:
                req=urllib2.Request("http://macvendors.co/api/"+x[0],headers={'User-Agent':"Magic Browser"})
                jsonpage=urllib2.urlopen(req,timeout=2)
                x = jsonpage.read()
                vendors=json.loads(x)
                vendors["vendor"]=vendors.pop("result")
            except:
                continue
            break
        '''
        vendors["hwaddress"]=hw_addr
        return hw_addr

    def get_ipv6(self):  # type: () -> Dict[str, str]
        """
        Function
        =========
        get_ipv6

        Description
        =============
        Gets the IPv6 address for all the available interfaces in /proc/net

        Arguments
        ==========
        None

        Return Values
        ==============
        __out <list> ipaddr : List of interfaces and their IPv6 address
        """
        ipaddr = list()
        ipv6dict = dict()
        logger.debug("Entering get_ipv6")

        try:
            with open('/proc/net/if_inet6', 'r') as ip6:
                for line in ip6.readlines():
                    ipaddr.append(re.sub("[^\\w]", " ", line).split())
            for ip in ipaddr:
                i6 = ip[0]
                ipv6 = ':'.join([i6[i:i+4] for i in range(0, len(i6), 4)])
                ip[0] = re.sub("::+", "::", (ipv6.replace("00", "")).replace(":0", ":"))
                del ip[1:5]
                ipv6dict[ip[1]] = ip[0]
            return ipv6dict
        except Exception as e:
            logger.warning("Unable to read ipv6 interface data", exc_info=e)
            return {}

    def get_ipv4_mask(self, iface):
        '''
        Function
        =========
        get_ipv4

        Description
        =============
        Gets the IPv4 address and the Mask for the given interface 'iface' using sockets module 
        
        Arguments
        ==========
            __in <str> iface : Interface like eth0, eth1 ...etc
        
        Return Values
        ==============

            __out <list> [ipv4,mask]  : IPv4 and Mask address of the eth0, eth1 ..etc interface.    
        '''
        import socket
        import fcntl
        import struct
        logger.debug("Entering get_ipv4_mask")
        try:
            s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            stru = struct.Struct('256s')
            ipv4 = socket.inet_ntoa(fcntl.ioctl(
                s.fileno(),
                0x8915,  # SIOCGIFADDR
                stru.pack(iface[:15].encode('utf-8'))
            )[20:24])
            mask = socket.inet_ntoa(fcntl.ioctl(socket.socket(socket.AF_INET, socket.SOCK_DGRAM),
                                    35099, struct.pack('256s', iface))[20:24])
        except:
            return [None,None]
        return [ipv4,mask]

    def get_iface_index(self, iface):
        '''
        Function
        =========
        get_iface_index.

        Description
        ==============
        Gets the index for the given Interface present within the /sys/class/net.

        Arguments
        ==========
            __in <str> iface : Interface like eth0, eth1 ...etc
        
        Return Values
        ==============

            __out <int> iface_index : iface_index of the eth0, eth1 ..etc interface.
        '''
        logger.debug("Entering get_iface_index")
        with open('/sys/class/net/'+iface+'/ifindex','r') as index:
            iface_index = index.read()
        return int(iface_index)

    def hex_to_ip(self, hexaddr):
        logger.debug("Entering hex_to_ip")
        x = iter(hexaddr)
        res = [str(int(''.join(i), 16)) for i in zip(x, x)]
        return '.'.join(res[::-1])

    def flag_conversion(self, hexflag):
        '''
        #define RTF_UP          0x0001          /* route usable                 */
        #define RTF_GATEWAY     0x0002          /* destination is a gateway     */
        #define RTF_HOST        0x0004          /* host entry (net otherwise)   */
        #define RTF_REINSTATE   0x0008          /* reinstate route after tmout  */
        #define RTF_DYNAMIC     0x0010          /* created dyn. (by redirect)   */
        #define RTF_MODIFIED    0x0020          /* modified dyn. (by redirect)  */
        #define RTF_MTU         0x0040          /* specific MTU for this route  */
        #ifndef RTF_MSS
        #define RTF_MSS         RTF_MTU         /* Compatibility :-(            */
        #endif
        #define RTF_WINDOW      0x0080          /* per route window clamping    */
        #define RTF_IRTT        0x0100          /* Initial round trip time      */
        #define RTF_REJECT      0x0200          /* Reject route                 */

        '''
        logger.debug("Entering flag_conversion")
        iflag=int(hexflag,16)
        strflag=""
        if (iflag & 0x0001):
            strflag+="U"
        if (iflag & 0x0002):
            strflag+="G"
        if (iflag & 0x0004):
            strflag+="H"
        if (iflag & 0x0008):
            strflag+="R"
        if (iflag & 0x0010):
            strflag+="D"
        if (iflag & 0x0020):
            strflag+="M"
        if (iflag & 0x0040):
            strflag+="u"
        if (iflag & 0x0100):
            strflag+="i"
        if (iflag & 0x0200):
            strflag+="!"
        return strflag

    def iface_mtu(self, iface):
        logger.debug("Entering iface_mtu")
        with open('/sys/class/net/'+iface+'/mtu','r') as mtu:
            return mtu.readline().strip()

    def globalParams(self):

        logger.debug("Entering globalParams")
        path = "/proc/sys/net/ipv4"
        globalParam = {}
        d=os.listdir(path)
        for i in d:
            if os.path.isdir(path+"/"+i)==False:
                x=open(path+"/"+i,'r')
                globalParam[i]=x.readline().strip()
                x.close()
        return globalParam

    def get_interface_stats(self, iface):
        '''
        '''
        logger.debug("Entering NicStatistics")
        path = "/sys/class/net/"+iface+"/statistics/"
        compStat = dict()
        nicStat=dict()
        isAcceleration=False
        for param in os.listdir(path):
            if param.startswith("vf_"):
                isAcceleration=True
            with open(path+param) as file:
                compStat[param]=file.readline().strip()
        nicStat["isNetworkAccelerationEnabled"]=isAcceleration
        nicStat["completeStatistics"]=compStat
        return nicStat

class GuestOSLogDM(IDataMonitor):

    def __init__(self):
        self.data_monitor_config = None
        self.data_collection_start_time = None

    def __del__(self):
        pass

    def init(self, dm_config):
        self.data_monitor_config = dm_config
        self.log_files = self.data_monitor_config['EventLogsToSave']
        self.config_files = self.data_monitor_config['ConfigFilesToSave']
        self.log_folders = self.data_monitor_config['ExtensionsLogFoldersToSave']
        self.output_tar_file = self.data_monitor_config['GuestOSLogTarFileName']
        self.commands = self.data_monitor_config['CommandsOutput']

    def name(self):
        if self.data_monitor_config:
            return self.data_monitor_config['Name']
        else:
            return None

    def friendly_name(self):
        if self.data_monitor_config:
            return self.data_monitor_config['FriendlyName']
        else:
            return None

    def start(self):

        if self.name():
            logger.debug(self.name() + ": start")

        self.data_collection_start_time = common.get_current_datetime("display")

        self.collect_system_logs()
        self.collect_config_files()
        self.collect_extensions_logs()
        self.collect_command_outputs()

        if self.name():
            logger.debug(self.name() + ": start")

    def stop(self):
        '''
        '''
        if self.name():
            logger.debug(self.name() + ": stop")

        self.filter_events()

        #common.compress_file(common.shared_guestfiles_output_folder_path , common.Output_Folder_Path + "/" + self.output_tar_file)
        data_collection_end_time = common.get_current_datetime("display")

        self.cleanup()

        final_result = {}
        final_result['Type'] = self.data_monitor_config['Type']
        final_result['GuestOSFiles'] = self.output_tar_file
        final_result['AnalysisStartTime'] = self.data_collection_start_time
        final_result['AnalysisEndTime'] = data_collection_end_time
        final_result['RuleFindings'] = []
        final_result['Category'] = self.data_monitor_config['Category']
        final_result['RuleFriendlyName'] = self.data_monitor_config['FriendlyName']
        final_result['ReferenceDataCollectors'] = []

        logger.info("Logs have been collected.")

        if self.name():
            logger.debug(self.name() + ": stop")

        return final_result
    
    def filter_events(self):
        '''
        '''
        pass

    def cleanup(self):
        '''
        '''
        pass

        # If debug
        if os.path.exists(common.diag_output_folder_path + "/" + self.output_tar_file):
            if common.debug_enabled != True:
                shutil.rmtree(common.shared_guestfiles_output_folder_path, ignore_errors=True)

    def collect_system_logs(self):
        '''
        Function
        =========
        collect_system_logs

        Description
        ==============
        Collect system logs.

        Arguments
        ==========
            N/A
        
        Return Values
        ==============

            N/A
        '''

        logger.debug("Entering collect_system_logs")

        # collect common log files

        for log in self.log_files:
            if os.path.exists(log):
                folders = log.split('/')[+1:-1]
                sourcefolder = ''
                outputfolder = common.shared_guestfiles_output_folder_path
                for count in range(len(folders)):
                    if not os.path.exists(outputfolder + '/' + folders[count]):
                        os.makedirs(outputfolder + '/' + folders[count])
                    sourcefolder = sourcefolder + '/' + folders[count]
                    outputfolder = outputfolder + '/' + folders[count]
                common.copy_file_wlog(sourcefolder, log.split('/')[len(log.split('/'))-1:][0], outputfolder)

        logger.info("System logs collected.")

    def collect_config_files(self):
        '''
        Function
        =========
        collect_config_files

        Description
        ==============
        Collect system configuration files.

        Arguments
        ==========
            N/A
        
        Return Values
        ==============

            N/A
        '''

        logger.debug("Entering collect_config_files")

        for logFile in self.config_files:
            if os.path.exists(logFile):
                folders = logFile.split('/')[+1:-1]
                sourcefolder = ''
                outputfolder = common.shared_guestfiles_output_folder_path
                for count in range(len(folders)):
                    if not os.path.exists(outputfolder + '/' + folders[count]):
                        os.makedirs(outputfolder + '/' + folders[count])
                    sourcefolder = sourcefolder + '/' + folders[count]
                    outputfolder = outputfolder + '/' + folders[count]
                common.copy_file_wlog(sourcefolder, logFile.split('/')[len(logFile.split('/'))-1:][0], outputfolder)

        logger.info("System configuration files collected.")

    def collect_extensions_logs(self):
        '''
        Function
        =========
        collect_extensions_logs

        Description
        ==============
        Collect Azure extensions logs.

        Arguments
        ==========
            N/A
        
        Return Values
        ==============
            N/A
        '''

        logger.debug("Entering collect_extensions_logs")

        for logfolder in self.log_folders:
            if os.path.exists(logfolder):
                folders = logfolder.split('/')[+1:]
                sourcefolder = ''
                outputfolder = common.shared_guestfiles_output_folder_path
                for count in range(len(folders)):
                    if not os.path.exists(outputfolder + '/' + folders[count]):
                        os.makedirs(outputfolder + '/' + folders[count])
                    sourcefolder = sourcefolder + '/' + folders[count]
                    outputfolder = outputfolder + '/' + folders[count]
                common.recursivecopy_with_pattern(sourcefolder, outputfolder, '*')

        logger.info("Azure extensions logs collected.")

    def collect_command_outputs(self):
        '''
        Function
        =========
        collect_command_outputs

        Description
        ==============
        Run specified command lines and flush output
        to a file.

        Arguments
        ==========
        N/A
        
        Return Values
        ==============
        N/A
        '''

        for command_entry in self.commands:
            command_line = command_entry['Command'] + ' '
            for argument in command_entry['Arguments']:
                command_line += argument['Argument']
                if argument['ArgValue'] != '':
                    arg_value = argument['ArgValue']
                    arg_value.replace("'", "\'")
                    if argument['ArgType'] == 'code':
                        arg_value = eval(arg_value)
                    command_line += ' \"' + arg_value + '\" '
            command_line = command_line.strip()
            #output_file_name = common.shared_guestfiles_output_folder_path + '/' + command_entry['OutputFileName']

            #command_result = subprocess.Popen([command_line + ' > ' + output_file_name + '  2> /dev/null'],stdout=subprocess.PIPE,shell=True)
            try:
                logger.debug(command_line + ' succesfully executed.')
            except IndexError:
                logger.error(command_line + ' failed.')

class RunningProcessesDC(IDataCollector):

    def __init__(self):
        self._dc_config = None

    def __del__(self):
        pass

    def init(self, dc_config):
       self._dc_config = dc_config

    def index(self):
        if self._dc_config:
            return self._dc_config['Index']
        else:
            return None

    def name(self):
        if self._dc_config:
            return self._dc_config['Name']
        else:
            return None

    def friendly_name(self):
        if self._dc_config:
            return self._dc_config['FriendlyName']
        else:
            return None

    def to_json(self):

        if self.name():
            logger.debug(self.name() + ": to_json")

        result_dict = {}
        data_collection_start_time = common.get_current_datetime("display")

        result_dict['$values'] = self.get_running_processes
        #result_dict['$type'] = 'Microsoft.Azure.Performance.Diagnostics.Contracts.RunningProcessInfo[], Microsoft.Azure.Performance.Diagnostics'

        perfdiag_ouput = DataCollectorUtil.create_json(self.name(), data_collection_start_time, common.get_current_datetime("display"), result_dict)

        logger.info("Process information collected.")

        if self.name():
            logger.debug(self.name() + ": to_json")

        return perfdiag_ouput

    @CachedProperty
    def get_running_processes(self):

        result = []
        for pid in [int(x) for x in os.listdir('/proc') if x.isdigit()]:
            process_name = ProcessUtil.get_process_name(pid)
            process_status = ProcessUtil.get_process_status(pid)
            process_resident_memory = ProcessUtil.get_process_resident_memory_kbytes(pid)
            process_cpu_usage_ms = ProcessUtil.get_process_cpu_time_ms(pid)
            if process_name is not None:
                result.append({'ProcessName': process_name, 'ProcessId': pid, 'Status': process_status, 'ResidentMemoryInKB': process_resident_memory, 'CpuUsageInMs': process_cpu_usage_ms})

        return result

class MemoryUsageCollector(IDataCollector):

    def __init__(self):
        self._dc_config = None

    def __del__(self):
        pass

    def init(self, dc_config):
       self._dc_config = dc_config

    def index(self):
        if self._dc_config:
            return self._dc_config['Index']
        else:
            return None

    def name(self):
        if self._dc_config:
            return self._dc_config['Name']
        else:
            return None

    def friendly_name(self):
        if self._dc_config:
            return self._dc_config['FriendlyName']
        else:
            return None

    def to_json(self):

        if self.name():
            logger.debug(self.friendly_name())

        result_dict = {}
        data_collection_start_time = common.get_current_datetime("display")

        result_dict['$values'] = self.get_system_memory_info
        result_dict['$values']['SwapMap'] = self.get_system_swap_info

        perfdiag_ouput = DataCollectorUtil.create_json(self.name(), data_collection_start_time, common.get_current_datetime("display"), result_dict)

        logger.info("Memory information collected.")

        if self.name():
            logger.debug(self.name() + ": to_json")

        return perfdiag_ouput

    @CachedProperty
    def get_system_memory_info(self):

        logger.debug("Entering memoryusagecollector: get_system_memory_info")

        memory_dict = {}
        filename = '/proc/meminfo'
        memory_dict = {}
        total_physical_memory_kb = float(common.total_physical_memory)
        available_memory_kb = float(common.available_physical_memory)
        memory_dict['TotalMemorySizeMB'] = float("{0:.2f}".format(total_physical_memory_kb / 1024))
        memory_dict['AvailableMemorySizeMB'] = float("{0:.2f}".format(available_memory_kb / 1024))
        memory_dict['PercentAvailableMemory'] = float("{0:.2f}".format((available_memory_kb / total_physical_memory_kb) * 100))
        if os.path.exists(filename):
            with open(filename, 'r') as f:
                for line in f:
                    line = line.strip()
                    if line:
                        name, value = line.split(': ')
                        if name == 'MemFree':
                            free_memory_kb = float(value.split()[0])
                            memory_dict['FreeMemorySizeMB'] = float("{0:.2f}".format(free_memory_kb / 1024))
                            memory_dict['PercentFreeMemory'] = float("{0:.2f}".format(free_memory_kb / total_physical_memory_kb * 100))
                        if name == 'SwapTotal':
                            swap_memory_kb = float(int(value.split()[0]))
                            memory_dict['TotalSwapMemorySizeMB'] = float("{0:.2f}".format(swap_memory_kb / 1024))
                        if name == 'SwapFree':
                            free_swap_memory_kb = float(int(value.split()[0]))
                            memory_dict['FreeSwapMemorySizeMB'] = float("{0:.2f}".format(free_swap_memory_kb / 1024))
                        if name == 'PercentageFree':
                            memory_dict['PercentageFreeSwapMemory'] = float("{0:.2f}".format((swap_memory_kb / free_swap_memory_kb) * 100))
                        if name == 'Buffers':
                            buffers_used_kb = float(int(value.split()[0]))
                            memory_dict['BuffersUsedMB'] = float("{0:.2f}".format(buffers_used_kb / 1024))
                        if name == 'Cached':
                            cached_kb = float(int(value.split()[0]))
                            memory_dict['CachedMB'] = float("{0:.2f}".format(cached_kb / 1024))
                        if name == 'Shmem':
                            shared_mem_kb = float(int(value.split()[0]))
                            memory_dict['SharedMemMB'] = float("{0:.2f}".format(shared_mem_kb / 1024))
                used_mem = float(memory_dict['TotalMemorySizeMB']) - float(memory_dict['CachedMB']) - float(memory_dict['BuffersUsedMB']) - float(memory_dict['FreeMemorySizeMB'])
                memory_dict['UsedMemorySizeMB'] = float("{0:.2f}".format(used_mem))
                memory_dict['PercentUsedMemory'] = float("{0:.2f}".format(used_mem * 1024 / total_physical_memory_kb * 100))

        logger.debug("Leaving memoryusagecollector: get_system_memory_info")

        return memory_dict

    @CachedProperty
    def get_system_swap_info(self):

        logger.debug("Entering memoryusagecollector: get_system_swap_info")

        swaps = []
        filename = '/proc/swaps'
        if os.path.exists(filename):
            with open(filename, 'r') as f:
                for line in f:
                    if line.startswith("Filename"):
                        # First line is headers
                        continue
                    swap = {}

                    line = line.split()
                    if line:
                        swap['Name'] = line[0]
                        swap['Type'] = line[1]
                        swap['Size'] = float("{0:.2f}".format(float(line[2]) / 1024))
                        swap['Used'] = float("{0:.2f}".format(float(line[3]) / 1024))
                        swap['Priority'] = line[4]
                        swap['UsedPercent'] = float("{0:.2f}".format(float(swap['Used']) / float(swap['Size']) * 100))

                    swaps.append(swap)

        logger.debug("Leaving memoryusagecollector: get_system_swap_info")

        return swaps
