#-------------------------------------------------------------------------
# Copyright (c) Microsoft.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#--------------------------------------------------------------------------

from os import path, remove, rename
from subprocess import Popen, PIPE
import six
import tarfile
import re
import logging
from util.common import get_current_datetime
from ruleengine.RuleEngineCls import IDataCollector, DataCollectorUtil

logger = logging.getLogger(__name__)

class HPCDiagWrapper(IDataCollector):
    '''
    Class
    =========
    HPCDiagWrapper

    Description
    ==============
    Data collector class used acting as
    a wrapper around the hpcdiag script.
    '''
    def init(self, dc_config):
        self.index = dc_config["Index"]
    def index(self):
        return self.index

    def name(self):
        return "HPCDiagDC"

    def friendly_name(self):
        return "High Performance Computing Diagnostics Tool Wrapper"

    def to_json(self):
        from util.common import shared_guestfiles_output_folder_path
        data_collection_start_time = get_current_datetime("display")
        p = Popen(["bash", "azhpc-diagnostics/Linux/src/gather_azhpc_vm_diagnostics.sh","--offline","-d",
            shared_guestfiles_output_folder_path],stdin=PIPE,stdout=PIPE,stderr=PIPE)
        if six.PY2:
            (stdout_data, stderr_data) = p.communicate(input="y")
        elif six.PY3:
            (stdout_data, stderr_data) = p.communicate(input=bytes("y","utf-8"))
            stdout_data = stdout_data.decode()
            stderr_data = stderr_data.decode()
        else:
            raise RuntimeError("Invalid python version")

        if stderr_data:
            logger.warning("HPCDiag produced error output: %s", stderr_data)
        
        lines = [line.strip(' #') for line in stdout_data.split('\n')]

        tarball_names = [line for line in lines if line.endswith('.tar.gz')]
        if len(tarball_names) < 1:
            raise RuntimeError("Could not identify hpcdiag output")
        elif len(tarball_names) > 1:
            raise RuntimeError("Could not identify unique hpcdiag output ({0})".format(len(tarball_names)))
        else:
            tarball_name = tarball_names[0]
        
        tarball = tarfile.open(name=tarball_name, mode='r:*')
        tarball.extractall(path.dirname(tarball_name))
        remove(tarball_name)

        folder_name = path.splitext(path.splitext(tarball_name)[0])[0]
        output_path = path.join(shared_guestfiles_output_folder_path,'hpcdiag')
        rename(folder_name, output_path)

        with open(path.join(shared_guestfiles_output_folder_path,"hpcdiag.out"),"w") as outfile:
            if stdout_data:
                outfile.write(stdout_data)
        with open(path.join(shared_guestfiles_output_folder_path,"hpcdiag.err"),"w") as outfile:
            if stderr_data:
                outfile.write(stderr_data)

        def parse_selinux():
            try:
                with open(path.join(output_path, 'VM', 'selinux'), 'r') as f:
                    return f.readline().strip().split('=')[1]
            except:
                return None

        def parse_ulimit():
            with open(path.join(output_path,'Memory', 'limits.conf')) as f:
                lines = [line.strip() for line in f.readlines() ]
            limits = {
                "hard": {},
                "soft": {}
            }
            for line in lines:
                match = re.search(r'^\*\s+(hard|soft)\s+(\w+)\s+(\w+)$', line)
                if match:
                    limits[match.group(1)][match.group(2)] = match.group(3)
            return limits
        
        def parse_zone_reclaim():
            with open(path.join(output_path, 'Memory', 'zone_reclaim_mode')) as f:
                return f.readline().strip()


        with open(path.join(output_path, 'VM', 'services')) as f:
            lines = [line.strip() for line in f.readlines()]
        services = { line.split()[0] : line.split()[1] == 'active' for line in lines }

        self.infiniband = {
            'missing_pkey': 'Could not find pkey' in stdout_data,
            'missing_driver': 'No Infiniband Driver Detected' in stdout_data
        }
        self.nvidia = {
            'missing_driver': 'No Nvidia Driver Detected' in stdout_data,
            'eol_gpu_count': stdout_data.count('BAD GPU')
        }
        self.memory = {
            'limits': parse_ulimit(),
            'zone_reclaim': parse_zone_reclaim()
        }
        self.network = {
            'selinux': parse_selinux(),
            'firewalld': services['firewalld'],
            'iptables': services['iptables']
        }
        self.cpu = {
            'cpupower': services['cpupower'],
            'waagent': services['waagent'] or services['walinuxagent']
        }

        data_collection_end_time = get_current_datetime("display")

        hpcdiagDict = {
            'infiniband': self.infiniband,
            'nvidia': self.nvidia,
            'memory': self.memory,
            'network': self.network,
            'cpu': self.cpu
        }
        datacollectorJSON = DataCollectorUtil.create_json(self.name(), data_collection_start_time, data_collection_end_time, hpcdiagDict)
        return datacollectorJSON


import os
import util.common as common
from ruleengine.RuleEngineCls import RuleUtil
from ruleengine.RuleEngineCls import IStaticRule
class CPURule(IStaticRule):
    # Finding IDs
    CPUPOWER = 0
    WAAGENT = 1

    def __init__(self, hpcdiag_dc, machine_map_dc):
        self._ref_dcs = ['hpcdiag_dc', 'machine_map_dc']
        self._hpcdiag_dc = hpcdiag_dc
        self._machine_map_dc = machine_map_dc
        self._config = None

        self.analysisstart = None
        self.analysisend = None

    def init(self, rule_config):
        self._config = rule_config
    
    def id(self):
        return 'PIFL.HPC.SR0003'

    def name(self):
        return 'HPC Machine Configuration Analysis'
    
    def category(self):
        return self._config['Category']

    def run(self):
        self.analysisStart = common.get_current_datetime("display")

        finding = lambda fid,resources : RuleUtil.create_finding(self.id(), self._config['FindingTemplates'], fid, resources)

        findings = []
        hpcdiag_dc = self._hpcdiag_dc

        cpupower = {
            "ResourceName": "Service",
            "ResourceValue": "cpupower",
            "Configurations": {}
        }
        if hpcdiag_dc.cpu['cpupower']:
            findings.append(finding(self.CPUPOWER, [cpupower]))

        waagent_active = hpcdiag_dc.cpu['waagent']
        waagent_config = self._machine_map_dc.waagent_details_map_cached['config']
        desired_config = self._config['FindingTemplates'][self.WAAGENT]['Configuration']
        waagent = {
            "ResourceName": "Agent",
            "ResourceValue": "Windows Azure Guest Agent",
            "Configurations": waagent_config
        }
        if waagent_active and any(waagent_config.get(k) != v for k,v in desired_config.items()):
            findings.append(finding(self.WAAGENT, [waagent]))

        self.analysisend = common.get_current_datetime("display")
                   
        return RuleUtil.create_staticrule_result(self, findings)

class GPURule(IStaticRule):
    # Finding IDs
    MISSING_DRIVER = 0
    END_OF_LIFE = 1

    def __init__(self, hpcdiag_dc, machine_map_dc):
        self._ref_dcs = ['hpcdiag_dc','machine_map_dc']
        self._hpcdiag_dc = hpcdiag_dc
        self._machine_map_dc = machine_map_dc

    def init(self, rule_config):
        self._config = rule_config

    def id(self):
        return 'PIFL.HPC.SR0002'

    def name(self):
        return 'HPC GPU Configuration Analysis'

    def category(self):
        return self._config['Category']

    def run(self):
        self.analysisStart = common.get_current_datetime("display")
        pci_map = self._machine_map_dc.pci_map
        resource_of_device = lambda device : {
            "ResourceName": 'GPU',
            "ResourceValue": device['device'],
            "Configurations": []
        }
        is_gpu = lambda device : device['vendor'] == 'NVIDIA Corporation'
        gpus = [resource_of_device(device) for device in pci_map if is_gpu(device)]
        findings = []

        finding = lambda fid, gpus : RuleUtil.create_finding(self.id(), self._config['FindingTemplates'], fid, gpus)
        if self._hpcdiag_dc.nvidia['missing_driver']:
            findings.append(finding(self.MISSING_DRIVER, gpus))
        if self._hpcdiag_dc.nvidia['eol_gpu_count'] > 0:
            eol_gpus = [gpus[0]] * self._hpcdiag_dc.nvidia['eol_gpu_count'] # all gpus on a single system are the same
            findings.append(finding(self.END_OF_LIFE, eol_gpus))

        self.analysisend = common.get_current_datetime("display")
        return RuleUtil.create_staticrule_result(self, findings)

class MemoryRule(IStaticRule):
    # Finding IDs
    MEMORY_LIMIT = 0
    ZONE_RECLAIM = 1

    def __init__(self, hpcdiag_dc):
        self._ref_dcs = ['hpcdiag_dc']
        self._hpcdiag_dc = hpcdiag_dc

    def init(self, rule_config):
        self._config = rule_config

    def  id(self):
        return 'PIFL.HPC.SR0003'

    def name(self):
        return 'HPC Memory Configuration Analysis'

    def category(self):
        return self._config['Category']

    def run(self):
        self.analysisStart = common.get_current_datetime("display")

        hpcdiag_dc = self._hpcdiag_dc
        findings = []

        finding = lambda fid, resources : RuleUtil.create_finding(self.id(), self._config['FindingTemplates'], fid, resources)
        ulimits = hpcdiag_dc.memory['limits']
        ulimits_resource = {
            "ResourceName": "Policy",
            "ResourceValue": "User Resource Limits",
            "Configurations": ulimits
        }
        desired_config = self._config['FindingTemplates'][self.MEMORY_LIMIT]['Configuration']
        if any(ulimits["hard"].get(k) != v for k,v in desired_config["hard"].items()) or any(ulimits["soft"].get(k) != v for k,v in desired_config["soft"].items()):
            findings.append(finding(self.MEMORY_LIMIT, [ulimits_resource]))

        numa_policy = {
            "ResourceName": "Policy",
            "ResourceValue": "NUMA Zone Reclamation Policy",
            "Configurations": {
                'zone_reclaim': hpcdiag_dc.memory['zone_reclaim']
            }
        }

        if hpcdiag_dc.memory['zone_reclaim'] != "1":
            findings.append(finding(self.ZONE_RECLAIM, [numa_policy]))

        self.analysisend = common.get_current_datetime("display")
        return RuleUtil.create_staticrule_result(self, findings)