# -------------------------------------------------------------------------
# Copyright (c) Microsoft.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

import datetime
import logging
import time
from operator import itemgetter

import six

from ruleengine.RuleEngineCls import IResourceMonitor, MonitorTimer
from util.SystemUtil import CpuUtil, ProcessUtil, OSProcReader, SystemUtil

logger = logging.getLogger(__name__)

if six.PY3:
    long = int


class ResourceMonitor(IResourceMonitor):
    '''
    Class
    =========
    ResourceMonitor

    Description
    ==============
    Class used to manage the whole data collection.
    Starts the data collection asynchronously by running a new MonitorTimer.
    '''

    def __init__(self, duration_seconds, interval_seconds, rule_name, rule_definition, proc_reader):
        '''
        Function
        =========
        __init__

        Description
        ==============
        Initialize the ResourceMonitor class that will handle resource
        profiling.

        Arguments
        ==========
        __in <Class> self : class instance
        __in <int> duration_seconds: capture duration in seconds
        __in <int> interval_seconds: sampling interval in seconds
        __in <string> rule_name: this rule short name
        __in <object> rule_definition: this rule configuration

        Return Values
        ==============
        N/A
        '''

        self.duration_seconds = duration_seconds
        self.interval_seconds = interval_seconds
        self.rule_name = rule_name
        self.rule_definition = rule_definition
        self.rule_config = self.rule_definition.get('Configuration')
        self.proc_reader = proc_reader
        # Process CPU% threshold
        self.rule_cpu_percentage_long_spinner = float(self.rule_config['CpuPercentageLongSpinner'])
        # Processor CPU% threshold for all core
        self.rule_cpu_percentage_threshold_all_cores = float(self.rule_config['CpuPercentageThresholdAllCore'])
        # Processor CPU% threshold for any core
        self.rule_cpu_percentage_threshold_any_core = float(self.rule_config['CpuPercentageThresholdAnyCore'])
        # Process avg CPU% threshold
        self.rule_avg_cpu_percentage_long_spinner = float(self.rule_config['AvgCpuPercentageLongSpinner'])
        # Spinner inactivity timeout
        self.rule_long_spinner_timeout_ms = long(self.rule_config['LongSpinnerTimeoutMS'])
        # Number of top spinners to report
        self.rule_top_spinners_to_report = int(self.rule_config['TopSpinnersToReport'])
        self.monitor_timer = None
        self.processors_spin_periods = []
        self.process_spinners = {}
        self.data_collection_start = None
        self.processor_monitor = ProcessorMonitor(self.rule_cpu_percentage_threshold_any_core, self.rule_cpu_percentage_threshold_all_cores)
        self.process_monitor = ProcessMonitor()

    # TODO: Consider using __enter__ instead
    @staticmethod
    def init(duration_seconds, interval, rule_name, rule_config):
        '''
        Function
        =========
        init

        Description
        ==============
        Used by RuleEngine to instantiate this ResourceMonitor instance
        as __init__ is not able to return it.

        Arguments
        ==========
        __in <int> duration_seconds: capture duration_seconds in milliseconds
        __in <int> interval: sampling interval in milliseconds
        __in <string> rule_name: this rule short name
        __in <object> rule_definition: this rule configuration

        Return Values
        ==============
        __out <ResourceMonitor> this class instance
        '''

        return ResourceMonitor(duration_seconds, interval, rule_name, rule_config, OSProcReader())

    def performance_data_update_handler(self):
        '''
        Function
        =========
        performance_data_update_handler

        Description
        ==============
        At the specified interval, the MonitorTimer will call
        this function to collect current processors and processes metrics.

        Arguments
        ==========
        __in <Class> self : class instance

        Return Values
        ==============
        N/A
        '''

        spin_period_timed_out = False

        # Get latest processors statistics
        self.processor_monitor.get_processors_stats()

        if self.processor_monitor.samples_count > 0:
            if self.processor_monitor.is_any_core_above_threshold or \
                    self.processor_monitor.are_all_cores_above_threshold:
                # One of the threshold has been reached
                if self.processors_spin_periods:
                    # There is at least one spin period
                    # Did spin period timed out (milliseconds)
                    spin_period_timed_out = self.processor_monitor.update_timestamp_ms - self.processors_spin_periods[-1].end_timestamp_ms >= self.rule_long_spinner_timeout_ms

                if not self.processors_spin_periods or \
                        spin_period_timed_out or \
                        len(self.processors_spin_periods[-1].processors_usage) != len(self.processor_monitor.processors):
                    # Add a new spin period if:
                    # 1. It is the first spin period recognized
                    # 2. The spin period timed out
                    # 3. The count of processors has changed
                    self.processors_spin_periods.append(ProcessorsSpinPeriod(self.processor_monitor.previous_update_timestamp_ms, self.processor_monitor.update_timestamp_ms, self.processor_monitor.processors.values()))
                else:
                    # Update last spin period
                    self.processors_spin_periods[-1].update_usages(self.processor_monitor.update_timestamp_ms, self.processor_monitor.processors.values())

        # Get latest processes statistics
        self.process_monitor.get_processes_stats()

        if self.process_monitor.samples_count > 0:
            for process in self.process_monitor.processes.values():
                if process.unique_id not in self.process_spinners:
                    # A potential new spinner is recognized
                    if process.is_alive and process.delta_process_cpu_usage_percentage > self.rule_cpu_percentage_long_spinner:
                        # New process CPU usage in percent is above the
                        # threshold, it will be added as a new spinner
                        self.process_spinners[process.unique_id] = ProcessSpinner(process)
                else:
                    # Spinner already exist
                    if process.is_alive is False:
                        # Process does not exist anymore
                        if self.process_spinners[process.unique_id].active_process_spin_period:
                            # There is a spinner with an active spin period,
                            # archive spin period
                            self.process_spinners[process.unique_id].archive_spin_period()
                    else:
                        # Process was still alive when collecting metrics,
                        # update the spin period
                        self.update_spinner(process)

    def update_spinner(self, process_data):
        '''
        Function
        =========
        update_spinner

        Description
        ==============
        Update existing spinner by updating its active spin period,
        create a new spin period or archive the active spin period.

        Arguments
        ==========
        __in <Class> self : class instance
        __in <Process> process_data: current process metadata and statistics

        Return Values
        ==============
        N/A
        '''

        if self.process_spinners[process_data.unique_id].active_process_spin_period is None:
            # There is no spin period for this spinner now (i.e. the previous
            # one was archived)
            if process_data.delta_process_cpu_usage_percentage > self.rule_cpu_percentage_long_spinner:
                # A new spin period will be added to the spinner
                self.process_spinners[process_data.unique_id].new_spin_period(process_data)
        else:
            # There is an active spin period for this spinner
            if process_data.delta_process_cpu_usage_percentage > self.rule_cpu_percentage_long_spinner:
                # Process CPU% is greater than threshold, update active spin
                # period with last collected statistics
                self.process_spinners[process_data.unique_id].active_process_spin_period.update_spin_period(process_data)
                self.process_spinners[process_data.unique_id].children = process_data.children
            else:
                # Process CPU% gets lower for the active spinner, check if it
                # exceeds capture timeout or if cumulative CPU% is lower
                # than threshold

                spinner_timed_out = process_data.update_timestamp_ms - self.process_spinners[process_data.unique_id].active_process_spin_period.end_timestamp_ms >= self.rule_long_spinner_timeout_ms
                overall_under_threshold = ((self.process_spinners[process_data.unique_id].active_process_spin_period.average_cpu_percentage + process_data.delta_process_cpu_usage_percentage) / 2 > 0.0) and \
                    ((self.process_spinners[process_data.unique_id].active_process_spin_period.average_cpu_percentage + process_data.delta_process_cpu_usage_percentage) / 2 < self.rule_cpu_percentage_long_spinner)
                if spinner_timed_out or overall_under_threshold:
                    # Archives spin period and resets process metrics
                    self.process_spinners[process_data.unique_id].archive_spin_period()
                else:
                    # Spinner is below threshold but spin period is active,
                    # spin period will be updated with last collected metrics
                    self.process_spinners[process_data.unique_id].active_process_spin_period.update_spin_period(process_data)
                    self.process_spinners[process_data.unique_id].children = process_data.children

    def start(self):
        '''
        Function
        =========
        start

        Description
        ==============
        Start the timer that will trigger the performance update
        collection at the specified interval of time.
        This method is called by RuleEngine.

        Arguments
        ==========
        __in <Class> self : class instance

        Return Values
        ==============
        N/A
        '''

        # Substract interval from duration_seconds to display accurate
        # capture duration_seconds
        logger.info('Starting ' + self.rule_name + ' resource monitor (' + str(self.duration_seconds - self.interval_seconds) + ' seconds).')

        self.monitor_timer = MonitorTimer(self.duration_seconds, self.interval_seconds, self.performance_data_update_handler)
        self.monitor_timer.start()
        self.data_collection_start = self.monitor_timer.start_timer()

    def stop(self):
        '''
        Function
        =========
        stop

        Description
        ==============
        Resource Monitor stop is called by RuleEngine.
        Reconciliate all data before returning to RuleEngine.

        Arguments
        ==========
        __in <Class> self : class instance

        Return Values
        ==============
        N/A
        '''

        logger.info('Stopping ' + self.rule_name + ' resource monitor.')
        final_result = {}

        # Get processes spinners
        results = []
        for spinner in self.process_spinners:
            self.process_spinners[spinner].archive_spin_period()
            result = self.process_spinners[spinner].get_spinner_result(0)
            if result is not None:
                results.append(result)
        if self.rule_top_spinners_to_report > 0:
            # Return top N spinners
            final_result['TopLongCpuSpinners'] = sorted(results, key=lambda k: k['TotalResourceUsagePercentage'], reverse=True)[:self.rule_top_spinners_to_report]
        else:
            # Retun all spinners
            final_result['TopLongCpuSpinners'] = sorted(results, key=lambda k: k['TotalResourceUsagePercentage'], reverse=True)

        # Get processors spin periods
        results = []
        for spin_period in self.processors_spin_periods:
            result = spin_period.get_spin_period_result(self.rule_cpu_percentage_threshold_all_cores, self.rule_cpu_percentage_threshold_any_core)
            if result is not None:
                results.append(result)
        final_result['HighProcessorsUsagePeriods'] = results
        final_result['TopSpikeCpuSpinners'] = []

        return final_result

    def is_active(self):
        '''
        Function
        =========
        is_active

        Description
        ==============
        Provides a way to check if resource monitor is active or not.
        Used by the caller from RuleEngine.

        Arguments
        ==========
        __in <Class> self : class instance

        Return Values
        ==============
        __out <bool>
        '''

        return self.monitor_timer.is_alive()


# ######################### Process spinners #########################


class ProcessMonitor():
    '''
    Class
    =========
    ProcessMonitor

    Description
    ==============
    Used to reference all running processes on
    the system and update their statistics
    at Resource Monitor sampling interval.
    '''

    def __init__(self, proc_reader=OSProcReader()):
        '''
        Function
        =========
        __init__

        Description
        ==============
        Initialize an object which will host the collection of running
        processes on the system.

        Arguments
        ==========
        __in <Class> self : class instance

        Return Values
        ==============
        N/A
        '''

        # Processes dictionary
        self.processes = {}
        # samples_count is used to discard the first
        # collected sample as it might not represent
        # the sample interval conigured but the interval
        # since the start f the process.
        # Used in ResourceMonitor.performance_data_update_handler
        # to really start leveraging thresholds.
        self.samples_count = 0
        self.proc_reader = proc_reader

    def get_processes_stats(self):
        '''
        Function
        =========
        get_processes_stats

        Description
        ==============
        Get all active pids in /proc and collect latest statistics from them.
        This method adds to the dictionary self.process_monitor new processes.
        Otherwise, it will update the corresponding dictionary item with its
        last statistics.

        Arguments
        ==========
        __in <Class> self : class instance

        Return Values
        ==============
        N/A
        '''

        # Get current list of PIDs
        pids = SystemUtil.PROC_READER.get_pids()
        # Create a map of PIDs with their children
        process_map = ProcessUtil.get_pids_with_children(pids)

        for pid in pids:
            process_name = ProcessUtil.get_process_name(pid)
            if process_name:
                # Process is not None, thus alive
                if (str(pid) + '-' + process_name) in self.processes:
                    # Process already referenced
                    self.processes[str(pid) + '-' + process_name].update_process(process_map[pid])
                else:
                    # Process not yet referenced, add new process with
                    # statistics to dictionary
                    new_process = Process(pid, process_map[pid])
                    if new_process.is_alive:
                        self.processes[new_process.unique_id] = new_process

        # Get all dead processes and flag them as
        # not alive
        for process in self.processes.values():
            process_is_dead = True
            for pid in pids:
                if process.pid == pid:
                    process_is_dead = False
                    break
            if process_is_dead:
                process.is_alive = False

        self.samples_count += 1


class Process():
    '''
    Class
    =========
    Process

    Description
    ==============
    Used to hold process information and statistics.
    '''

    def __init__(self, pid, process_children):
        '''
        Function
        =========
        __init__

        Description
        ==============
        Initialize an object which will host a process instance information
        and statistics.
        Values are set to 0 as we want to collect statistics for the
        collection durations and not the process whole life.

        Arguments
        ==========
        __in <Class> self : class instance
        __in <int> pid : process PID
        __in <list> process_children : process children PIDs

        Return Values
        ==============
        N/A
        '''

        self.process_name = ProcessUtil.get_process_name(pid)
        if self.process_name:
            self.unique_id = str(pid) + '-' + self.process_name
            self.pid = pid
            self.ppid = ProcessUtil.get_parent_pid(self.pid) if self.pid > 0 else None
            self.parent_process_name = ProcessUtil.get_process_name(self.ppid) if self.ppid is not None else None
            self.children = process_children
            self.command_line = ProcessUtil.get_process_command_line(pid)

            self.process_cpu_time_ms = self.delta_process_cpu_time_ms = self.total_process_cpu_time_ms = 0
            self.process_start_timestamp_ms = ProcessUtil.get_process_start_time_ms(self.pid)
            self.start_timestamp_ms = self.update_timestamp_ms = self.previous_update_timestamp_ms = int(round(time.time() * 1000))

            self.previous_process_cpu_time_ms = ProcessUtil.get_process_cpu_time_ms(self.pid)
            self.delta_process_cpu_usage_percentage = 0.0
            self.is_alive = True
        else:
            # ResourceMonitor.performance_data_update_handler() will not
            # consider this process
            self.is_alive = False

    def update_process(self, process_children):
        '''
        Function
        =========
        update_process

        Description
        ==============
        Update individual process statistics.
        Values used by PIL are:
            * self.delta_process_cpu_time_ms: provides the CPU time used by
            the process during last sampling interval
            * self.delta_process_cpu_usage_percentage: provides the delta
            CPU % used by the process during last sampling interval

        Arguments
        ==========
        __in <Class> self : class instance
        __in <list> process_children : process children PIDs

        Return Values
        ==============
        N/A
        '''

        if ProcessUtil.is_alive(self.pid):
            self.previous_update_timestamp_ms = self.update_timestamp_ms
            self.process_cpu_time_ms = ProcessUtil.get_process_cpu_time_ms(self.pid)
            self.update_timestamp_ms = int(round(time.time() * 1000))
            self.delta_process_cpu_time_ms = self.process_cpu_time_ms - self.previous_process_cpu_time_ms
            delta_time_ms = self.update_timestamp_ms - self.previous_update_timestamp_ms

            self.total_process_cpu_time_ms += self.delta_process_cpu_time_ms

            if delta_time_ms > 0:
                # Delta CPU% for last sample interval
                # value used by PIL to trigger new spin periods
                # or update existing one
                self.delta_process_cpu_usage_percentage = 100 * float(self.delta_process_cpu_time_ms) / delta_time_ms
            else:
                # Interval was too short
                self.delta_process_cpu_usage_percentage = 0.0

            self.previous_process_cpu_time_ms = self.process_cpu_time_ms
            self.children = process_children
        else:
            # Process does not exist anymore.
            # ResourceMonitor.performance_data_update_handler() will not
            # consider this process.
            self.is_alive = False
            # if 'stress' in self.unique_id:
            print('Process is dead: ' + self.unique_id)


class ProcessSpinner():
    '''
    Class
    =========
    ProcessSpinner

    Description
    ==============
    A process spinner is an instance of a process which has had a CPU usage
    higher than PerfInsights defined threshold at least once during the
    profiling.
    A process spinner can have one or more CPU spin periods during which the
    CPU usage average was higher than PerfInsights defined threshold for the
    default tiemout defined in the configuration file.
    A spin period is active while process CPU usage is above the threshold or
    while the process average usage since profiling start is below the
    threshold but has not reached the defined PerfInsights timeout.
    A spin period is archived when process average CPU usage is below the
    threshold and timeout has been reached.
    '''

    def __init__(self, process_data):
        '''
        Function
        =========
        __init__

        Description
        ==============
        Initialize an object which will host an instance of a process spinner.

        Arguments
        ==========
        __in <Class> self : class instance
        __in <object> process_data : process metadata and statistics

        Return Values
        ==============
        N/A
        '''

        self.process_id = process_data.pid
        self.unique_id = process_data.unique_id
        self.process_name = process_data.process_name
        self.parent_process_id = process_data.ppid
        self.parent_process_name = process_data.parent_process_name
        self.children = process_data.children
        self.command_line = process_data.command_line
        # Active spin period, initialize with process collection start time
        self.active_process_spin_period = ProcessSpinPeriod(process_data)
        # Archived spin periods
        self.process_spin_periods = []

    def new_spin_period(self, process_data):
        '''
        Function
        =========
        new_spin_period

        Description
        ==============
        A new spin period has been detected so the active spin period will be
        used to store data.
        Initialized with process collection last update time
        (previous_update_timestamp_ms).

        Arguments
        ==========
            __in <Class> self : class instance
            __in <object> process_data : process metadata and statistics

        Return Values
        ==============
            N/A
        '''

        self.active_process_spin_period = ProcessSpinPeriod(process_data)

    def archive_spin_period(self):
        '''
        Function
        =========
        archive_spin_period

        Description
        ==============
        Archive spin period when the process average CPU usage is below the
        threshold or not active for more than the defined timeout.
        The active spin period is set to None.

        Arguments
        ==========
        __in <Class> self : class instance

        Return Values
        ==============
        N/A
        '''

        if self.active_process_spin_period is not None:
            # Active spin period is active so we can archive it
            if self.active_process_spin_period.start_timestamp_ms < self.active_process_spin_period.end_timestamp_ms:
                # Active spin period has valid samples as end time is greater
                # than start time
                self.process_spin_periods.append(self.active_process_spin_period)
        self.active_process_spin_period = None

    def get_spinner_result(self, rule_avg_cpu_percentage_long_spinner):
        '''
        Function
        =========
        get_spinner_result

        Description
        ==============
        At the end of the capture, provides a dictionary with profiling data
        that will be returned to the Resource Monitor.
        It provides the overall profiling data and profiling data per spin
        period.

        Arguments
        ==========
        __in <Class> self : class instance
        __in <float> rule_avg_cpu_percentage_long_spinner: threshold used
            to filter spinner's spin periods

        Return Values
        ==============
        __out <dict> spinner_result: spinner results
        '''

        spinner_result = {}
        spinner_result['ProcessId'] = self.process_id
        spinner_result['ProcessName'] = self.process_name
        spinner_result['ParentProcessId'] = self.parent_process_id
        spinner_result['ParentProcessName'] = self.parent_process_name
        spinner_result['Children'] = []
        for children in self.children:
            spinner_result['Children'].append({'ProcessId': children['PID'], 'ProcessName': children['ProcessName']})
        spinner_result['ServicesHosted'] = ''
        spinner_result['ImagePath'] = ''
        spinner_result['CompanyName'] = ''
        spinner_result['CommandLine'] = self.command_line
        # Calculation of the total CPU time of the process
        # for all spin periods (thus during the capture time)
        spinner_result['TotalResourceUsageMS'] = (sum(process_spin_period.process_cpu_time_ms for process_spin_period in self.process_spin_periods)) if self.process_spin_periods else 0

        spinner_result['CpuSpinPeriods'] = []
        total_cpu_percentage = 0.0
        for process_spin_period in self.process_spin_periods:
            avg_cpu_percentage = process_spin_period.average_cpu_percentage
            if avg_cpu_percentage > rule_avg_cpu_percentage_long_spinner:
                # We want to keep only spin periods for which
                # overall avg CPU% is above threshold
                total_cpu_percentage += avg_cpu_percentage
                # Workarounds for bug 4660921
                if avg_cpu_percentage > 100.00:
                    avg_cpu_percentage = 100.00
                process_cpu_time_ms = process_spin_period.process_cpu_time_ms
                if process_cpu_time_ms > (process_spin_period.end_timestamp_ms - process_spin_period.start_timestamp_ms):
                    process_cpu_time_ms = process_spin_period.end_timestamp_ms - process_spin_period.start_timestamp_ms

                start_timestamp_friendly = datetime.datetime.utcfromtimestamp(process_spin_period.start_timestamp_ms / 1000).strftime('%Y-%m-%d %H:%M:%S.%f')[:-3]
                end_timestamp_friendly = datetime.datetime.utcfromtimestamp(process_spin_period.end_timestamp_ms / 1000).strftime('%Y-%m-%d %H:%M:%S.%f')[:-3]
                spinner_result['CpuSpinPeriods'].append({'CpuUsageMS': process_cpu_time_ms, 'AvgCpuPercentage': avg_cpu_percentage, 'StartTime': start_timestamp_friendly, 'EndTime': end_timestamp_friendly})

        try:
            spinner_result['TotalResourceUsagePercentage'] = total_cpu_percentage / len(spinner_result['CpuSpinPeriods'])
        except ZeroDivisionError:
            # There was no valid spin period
            # Ignore this spinner
            return None

        return spinner_result


class ProcessSpinPeriod():
    '''
    Class
    =========
    ProcessSpinPeriod

    Description
    ==============
    A spin period is a time frame during which
    a process was above the CPU usage threshold.
    '''

    def __init__(self, process_data):
        '''
        Function
        =========
        __init__

        Description
        ==============
        Initialize an object which will host
        statistics for a process spin period.

        Arguments
        ==========
        __in <Class> self : class instance
        __in <object> process_data : process metadata and statistics

        Return Values
        ==============
        N/A
        '''

        self.start_timestamp_ms = process_data.previous_update_timestamp_ms
        self.end_timestamp_ms = process_data.update_timestamp_ms
        # Cumulative process CPU time in milliseconds
        self.process_cpu_time_ms = process_data.delta_process_cpu_time_ms
        self.average_cpu_percentage = process_data.delta_process_cpu_usage_percentage

    def update_spin_period(self, process_data):
        '''
        Function
        =========
        update_spin_period

        Description
        ==============
        Update process spin period with latests process statistics.

        Arguments
        ==========
        __in <Class> self : class instance
        __in <object> process_data : process metadata and statistics

        Return Values
        ==============
        N/A
        '''

        self.end_timestamp_ms = process_data.update_timestamp_ms
        self.process_cpu_time_ms += process_data.delta_process_cpu_time_ms
        self.average_cpu_percentage = 100 * float(self.process_cpu_time_ms) / (self.end_timestamp_ms - self.start_timestamp_ms)


# ######################### Processor spinners #########################


class ProcessorMonitor():
    '''
    Class
    =========
    ProcessorMonitor

    Description
    ==============
    Used to reference all processors and update their statistics
    at Resource Monitor sampling interval.
    '''

    def __init__(self, rule_cpu_percentage_threshold_any_core, rule_cpu_percentage_threshold_all_cores):
        '''
        Function
        =========
        __init__

        Description
        ==============
        Initialize an object which will host the collection of processors on
        the system.

        Arguments
        ==========
        __in <Class> self : class instance
        __in <float> rule_cpu_percentage_threshold_any_core: threshold used to
            identify if a single CPU is above normal
        __in <float> rule_cpu_percentage_threshold_all_cores: threshold used to
            identify if all CPUs are above normal

        Return Values
        ==============
        N/A
        '''

        self.rule_cpu_percentage_threshold_any_core = rule_cpu_percentage_threshold_any_core
        self.rule_cpu_percentage_threshold_all_cores = rule_cpu_percentage_threshold_all_cores
        self.cpu_percentage_all_cores = 0.0
        # self.update_timestamp_ms -> sample period end time stamp
        # self.previous_update_timestamp_ms -> sample period start time
        self.update_timestamp_ms = self.previous_update_timestamp_ms = int(round(time.time() * 1000))
        # Following two variables are used to simplify verification
        # of triggers in caller function
        # (ResourceMonitor.performance_data_update_handler())
        self.is_any_core_above_threshold = False
        self.are_all_cores_above_threshold = False
        self.processor_count = CpuUtil.cpu_count_logical()
        self.processors = {}    # Processor dictionary
        # samples_count is used to discard the first
        # collected sample as it might not represent
        # the sample interval conigured but the interval
        # since the start f the process.
        # Used in ResourceMonitor.performance_data_update_handler
        # to really start leveraging thresholds.
        self.samples_count = 0

    def get_processors_stats(self):
        '''
        Function
        =========
        get_processors_stats

        Description
        ==============
        Collect latest statistics from all CPUs.
        This method adds to the dictionary self.processors any new processor.
        Otherwise, it will update the corresponding dictionary item with its
        latest statistics.

        Arguments
        ==========
        __in <Class> self : class instance

        Return Values
        ==============
        N/A
        '''

        self.is_any_core_above_threshold = False
        self.are_all_cores_above_threshold = False

        processors_data = CpuUtil.get_all_processors_cpu_time_ms()
        self.previous_update_timestamp_ms = self.update_timestamp_ms
        self.update_timestamp_ms = processors_data['TimeStamp']

        # Update number of CPUs if it has changed. Used to simplify
        # verification of thresholds in the caller function
        # (ResourceMonitor.performance_data_update_handler())
        self.processor_count = len(processors_data['ProcessorsCpuUsage'])

        for processor in processors_data['ProcessorsCpuUsage']:
            if processor['CpuNbr'] in self.processors:
                # Processor already referenced, update its statistics
                self.processors[processor['CpuNbr']].update_processor(self.update_timestamp_ms, processor)
            else:
                # Processor not referenced, create a new Processor instance
                self.processors[processor['CpuNbr']] = Processor(self.update_timestamp_ms, processor)

            if self.processors[processor['CpuNbr']].delta_processor_cpu_percentage > self.rule_cpu_percentage_threshold_any_core:
                # This processor CPU % is above threshold, used to trigger a
                # new spin period or update existing one
                self.is_any_core_above_threshold = True

        # Can't have an average with no delta. Wait until we have data
        if self.samples_count > 0:
            # Update overall CPU usage percentage, used to trigger a new spin
            # period or update last one
            self.cpu_percentage_all_cores = sum(processor.delta_processor_cpu_percentage for processor in self.processors.values()) / len(processors_data['ProcessorsCpuUsage'])
            if self.cpu_percentage_all_cores > self.rule_cpu_percentage_threshold_all_cores:
                self.are_all_cores_above_threshold = True

        self.samples_count += 1


class Processor():
    '''
    Class
    =========
    Processor

    Description
    ==============
    Used to hold processor information and statistics.
    '''

    def __init__(self, update_timestamp_ms, processor_data):
        '''
        Function
        =========
        __init__

        Description
        ==============
        Collect latest statistics from all CPUs.
        This method adds to the dictionary self.processors any new processor.
        Otherwise, it will update the corresponding dictionary item with its
        latest statistics.

        References:
            https://github.com/Leo-G/DevopsWiki/wiki/How-Linux-CPU-Usage-Time-and-Percentage-is-calculated

        Arguments
        ==========
        __in <Class> self : class instance
        __in <long> update_timestamp_ms: data collection start time, provided
            by caller ProcessorMonitor.get_processors_stats()
            Timestamp is provided by caller to maintain timestamp accuracy and
            consistency accross all processors statistics
        __in <object> processor_data: processor metrics

        Return Values
        ==============
        N/A
        '''

        self.cpu_number = processor_data['CpuNbr']
        self.unique_id = 'Processor ' + str(self.cpu_number)
        self.processor_time_ms = self.processor_idle_time_ms = 0
        self.delta_processor_time_ms = self.delta_processor_idle_time_ms = 0
        self.total_processor_time_ms = self.total_processor_idle_time_ms = 0
        self.delta_processor_cpu_percentage = self.processor_cpu_idle_percentage = 0.0

        self.previous_processor_time_ms = processor_data['NonIdleTimeMs']
        self.previous_processor_idle_time_ms = processor_data['IdleTimeMs']
        self.previous_update_timestamp_ms = update_timestamp_ms

    def update_processor(self, update_timestamp_ms, processor_data):
        '''
        Function
        =========
        update_processor

        Description
        ==============
        Collect and update latest statistics for a processor.

        Arguments
        ==========

        __in <Class> self : class instance
        __in <long> update_timestamp_ms: data collection start time, provided
            by caller ProcessorMonitor.get_processors_stats()
            Timestamp is provided by caller to maintain timestamp accuracy and
            consistency accross all processors statistics
        __in <object> processor_data: processor metrics

        Return Values
        ==============
        N/A
        '''

        self.processor_time_ms = processor_data['NonIdleTimeMs']
        self.processor_idle_time_ms = processor_data['IdleTimeMs']
        delta_time = float(update_timestamp_ms - self.previous_update_timestamp_ms)

        self.delta_processor_time_ms = self.processor_time_ms - self.previous_processor_time_ms
        self.delta_processor_idle_time_ms = self.processor_idle_time_ms - self.previous_processor_idle_time_ms

        if delta_time > 0.0:
            self.delta_processor_cpu_percentage = 100 * self.delta_processor_time_ms / float(delta_time)
            self.processor_cpu_idle_percentage = 100 * self.delta_processor_idle_time_ms / float(delta_time)
        else:
            # Interval was too low
            self.delta_processor_cpu_percentage = 0.0
            self.processor_cpu_idle_percentage = 0

        self.total_processor_time_ms += self.delta_processor_time_ms
        self.total_processor_idle_time_ms += self.delta_processor_idle_time_ms

        self.previous_update_timestamp_ms = update_timestamp_ms
        self.previous_processor_time_ms = self.processor_time_ms
        self.previous_processor_idle_time_ms = self.processor_idle_time_ms



class ProcessorsSpinPeriod():
    '''
    Class
    =========
    ProcessorsSpinPeriod

    Description
    ==============
    A ProcessorsSpinPeriod represents the capture of metrics
    on all processors during the time processor CPU usage was higher than
    PerfInsights defined thresholds at least once during the profiling.
    '''

    def __init__(self, previous_update_timestamp_ms, update_timestamp_ms, processors_data):
        '''
        Function
        =========
        __init__

        Description
        ==============
        Initialize a new processor spin period.

        Arguments
        ==========
        __in <Class> self : class instance
        __in <long> previous_update_timestamp_ms: last sample start time
        __in <long> update_timestamp_ms: last sample end time
        __in <object> processors_data: processors statistics

        Return Values
        ==============
        N/A
        '''

        self.start_timestamp_ms = previous_update_timestamp_ms
        self.end_timestamp_ms = update_timestamp_ms
        self.processors_usage = []
        for processor in processors_data:
            self.processors_usage.append(ProcessorUsage(processor))

    def update_usages(self, update_timestamp_ms, processors_data):
        """
        Function
        =========
        update_usages

        Description
        ==============
        Update processors spin period with latest statistics.

        Arguments
        ==========
        __in <Class> self : class instance
        __in <long> update_timestamp_ms: last sample update time
        __in <object> processors_data: processors statistics

        Return Values
        ==============
        N/A
        """

        self.end_timestamp_ms = update_timestamp_ms
        for processor_usage in self.processors_usage:
            for processor_data in processors_data:
                if processor_data.cpu_number == processor_usage.cpu_number:
                    processor_usage.update_usage(processor_data, self.end_timestamp_ms - self.start_timestamp_ms)
                    break

    def get_spin_period_result(self, rule_cpu_percentage_threshold_all_cores, rule_cpu_percentage_threshold_any_core):
        """
        Function
        =========
        get_spin_period_result

        Description
        ==============
        At the end of the capture, provides a dictionary with
        profiling data that will be returned to the Resource Monitor.
        It provides the overall profiling data and profiling data
        per spin period.

        Arguments
        ==========
        __in <Class> self : class instance
        __in <float> rule_cpu_percentage_threshold_all_cores:
            Processor CPU% threshold for all core
        __in <float> rule_cpu_percentage_threshold_any_core:
            Processor CPU% threshold for any core

        Return Values
        ==============
        __out <dict> spin_period_result: processors statistics for the capture time
        """

        spin_period_result = {}
        spin_period_result['ProcessorsUsage'] = []

        all_processors_cpu_time_ms = 0
        all_processors_cpu_percentage = 0.0
        is_any_core_above_threshold = False
        for processor_usage in self.processors_usage:
            # Workarounds for bug 4660921
            processor_cpu_time_ms = processor_usage.processor_cpu_time_ms
            if processor_cpu_time_ms > self.end_timestamp_ms - self.start_timestamp_ms:
                processor_cpu_time_ms = self.end_timestamp_ms - self.start_timestamp_ms
            processor_cpu_percentage = processor_usage.processor_cpu_percentage
            if processor_cpu_percentage > 100.00:
                processor_cpu_percentage = 100.00

            all_processors_cpu_time_ms += processor_cpu_time_ms
            all_processors_cpu_percentage += processor_cpu_percentage
            spin_period_result['ProcessorsUsage'].append({'CpuNumber': processor_usage.cpu_number, 'ProcessorCpuTimeMS': processor_cpu_time_ms, 'ProcessorCpuPercentage': processor_cpu_percentage})
            if processor_usage.processor_cpu_percentage > rule_cpu_percentage_threshold_any_core:
                is_any_core_above_threshold = True

        average_all_processors_cpu_percentage = all_processors_cpu_percentage / len(spin_period_result['ProcessorsUsage'])
        # if average_all_processors_cpu_percentage > rule_cpu_percentage_threshold_all_cores or is_any_core_above_threshold:
        spin_period_result['StartTime'] = datetime.datetime.utcfromtimestamp(self.start_timestamp_ms / float(1000)).strftime('%Y-%m-%d %H:%M:%S.%f')[:-3]
        spin_period_result['EndTime'] = datetime.datetime.utcfromtimestamp((self.end_timestamp_ms - 1) / float(1000)).strftime('%Y-%m-%d %H:%M:%S.%f')[:-3]
        spin_period_result['AllProcessorsCpuTimeMS'] = all_processors_cpu_time_ms
        spin_period_result['AllProcessorsCpuPercentage'] = average_all_processors_cpu_percentage
        # Sort CPUs
        spin_period_result['ProcessorsUsage'] = sorted(spin_period_result['ProcessorsUsage'], key=itemgetter('CpuNumber'))
        return spin_period_result


class ProcessorUsage():
    '''
    Class
    =========
    ProcessorUsage

    Description
    ==============
    A single processor usage stores a processor time and CPU% for
    the spin period timeframe.
    '''

    def __init__(self, processor_data):
        '''
        Function
        =========
        __init__

        Description
        ==============
        Initialize a new object to hold processors CPU time and CPU%
        for a spin period.

        Arguments
        ==========
        __in <Class> self : class instance
        __in <object> processor_data: processor statistics

        Return Values
        ==============
        N/A
        '''

        self.cpu_number = processor_data.cpu_number
        self.processor_cpu_time_ms = processor_data.delta_processor_time_ms
        self.processor_cpu_percentage = processor_data.delta_processor_cpu_percentage

    def update_usage(self, processor_data, delta_time_ms):
        '''
        Function
        =========
        update_usage

        Description
        ==============
        Update statistics which hold processors CPU time and CPU%
        for a spin period.

        Arguments
        ==========
        __in <Class> self : class instance
        __in <object> processor_data: processors statistics
        __in <long> delta_time_ms: difference with last sample
            timestamp

        Return Values
        ==============
        N/A
        '''

        self.processor_cpu_time_ms += processor_data.delta_processor_time_ms
        self.processor_cpu_percentage = self.processor_cpu_time_ms / float(delta_time_ms) * 100
