# -------------------------------------------------------------------------
# Copyright (c) Microsoft.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

import os
import time
from datetime import datetime
import re
import logging
import six
import util.common as common
from util.SystemUtil import SystemUtil, DiskUtil, ProcessUtil
from ruleengine.RuleEngineCls import IResourceMonitor, MonitorTimer

logger = logging.getLogger(__name__)

if six.PY3:
    long = int

class ResourceMonitor(IResourceMonitor):
    '''
    Class
    =========
    ResourceMonitor

    Description
    ==============
    Class used to manage the whole data collection.
    Starts the data collection asynchronously by running a new MonitorTimer.
    '''

    def __init__(self, duration_seconds, interval_seconds, rule_name, rule_definition):
        '''
        Function
        =========
        __init__

        Description
        ==============
        Initialize the ResourceMonitor class that will handle resource
        profiling.

        Arguments
        ==========
        __in <Class> self : class instance
        __in <int> duration_seconds: capture duration in seconds
        __in <int> interval_seconds: sampling interval in seconds
        __in <string> rule_name: this rule short name
        __in <object> rule_definition: this rule configuration

        Return Values
        ==============
        N/A
        '''

        self.duration_seconds = duration_seconds
        self.interval_seconds = interval_seconds
        self.rule_name = rule_name
        self.rule_definition = rule_definition
        self.rule_config = self.rule_definition.get('Configuration')
        self.rule_physical_disk_busy_percentage_threshold_any_disk = float(self.rule_config['PhysicalDiskPercentageThresholdAnyDisk'])
        self.rule_physical_disk_iops_threshold_any_disk = float(self.rule_config['PhysicalDiskIopsThresholdAnyDisk'])
        self.rule_physical_disk_throughput_kbps_threshold_any_disk = float(self.rule_config['PhysicalDiskThroughputKbpsThresholdAnyDisk'])
        self.rule_disk_iops_per_process_long_spinner = float(self.rule_config['DiskIopsPerProcessLongSpinner'])
        self.rule_long_spinner_timeout_ms = int(self.rule_config['LongSpinnerTimeoutMS'])
        self.rule_top_spinners_to_report = int(self.rule_config['TopSpinnersToReport'])                                                     # Number of top spinners to report
        # Specifies which block devices to profile:
        # all or specified one (from command line or from config file).
        # During Phase 1 only = 'all'.
        self.target_devices = 'all'  # self.rule_config['BlockDevices']
        self.data_collection_start = None
        self.monitor_timer = None
        if SystemUtil.get_kernel_version() >= SystemUtil.ENHANCED_KERNEL_VERSION:
            # Init BlockDeviceMonitor class for 2.6+ kernels
            self.block_device_monitor = BlockDeviceMonitor(self.rule_physical_disk_busy_percentage_threshold_any_disk, self.rule_physical_disk_iops_threshold_any_disk, self.rule_physical_disk_throughput_kbps_threshold_any_disk)
        else:
            # Init BlockDeviceMonitor legacy class
            # Maybe not required to deal with older
            # kernels, need to be discussed in triage
            pass
        self.process_monitor = ProcessMonitor()
        self.devices_spin_periods = []
        self.process_spinners = {}

    # TODO: Consider using __enter__ instead
    @staticmethod
    def init(duration_seconds, interval_seconds, rule_name, rule_config):
        '''
        Function
        =========
        init

        Description
        ==============
        Used by RuleEngine to instantiate this ResourceMonitor instance
        as __init__ is not able to return it.
        We might consider using __enter__ instead.

        Arguments
        ==========
        __in <int> duration_seconds: capture duration_seconds in seconds
        __in <int> interval_seconds: sampling interval_seconds in seconds
        __in <string> rule_name: this rule short name
        __in <object> rule_definition: this rule configuration

        Return Values
        ==============
        __out <ResourceMonitor> this class instance
        '''

        return ResourceMonitor(duration_seconds, interval_seconds, rule_name, rule_config)

    def performance_data_update_handler(self):
        '''
        Function
        =========
        performance_data_update_handler

        Description
        ==============
        At the specified interval_seconds, the timer will call
        this function to collect current block devices metrics.

        Arguments
        ==========
        __in <Class> self : class instance

        Return Values
        ==============
        N/A
        '''

        spinner_timed_out = False

        # Get updated I/O statistics for all block devices
        devices_performance_data = self.block_device_monitor.get_block_devices_stats()

        if self.block_device_monitor.samples_count > 0:
            if self.block_device_monitor.all_devices_delta_io_count > self.rule_physical_disk_iops_threshold_any_disk and \
                (self.block_device_monitor.is_physical_disk_busy_percentage_threshold_any_disk or \
                    self.block_device_monitor.is_physical_disk_iops_threshold_any_disk or \
                    self.block_device_monitor.is_physical_disk_throughput_kbps_threshold_any_disk):
                    # We have eligible statistics to store in a spin period if:
                    # 1. There was I/O activity over IOPS threshold
                    # and
                    #   2. Delta busy time% was above threshold
                    #   3. Or delta IOPS was above threshold
                    #   4. Or delta throughput was above threshold

                if self.devices_spin_periods:
                    spinner_timed_out = (devices_performance_data['UpdateTimeSeconds'] * 1000) - (self.devices_spin_periods[-1].end_time_seconds * 1000) >= self.rule_long_spinner_timeout_ms

                if not self.devices_spin_periods or \
                    len(self.devices_spin_periods[-1].block_devices_usage) > len(devices_performance_data['Data']) or \
                        spinner_timed_out:
                        # Add a new spin period if:
                        # 1. It is the first spin period recognized
                        # 2. The count of used block devices has increased
                        # 3. The spin period timed out

                    self.devices_spin_periods.append(BlockDevicesSpinPeriods(devices_performance_data, self.rule_physical_disk_iops_threshold_any_disk))
                else:
                    # Update last spin period
                    self.devices_spin_periods[-1].update_usages(devices_performance_data)
            else:
                pass

        # Get updated I/O stats from processes
        self.process_monitor.get_processes_stats()

        if self.process_monitor.samples_count > 0:
            for process in self.process_monitor.processes.items():
                if process[1].unique_id not in self.process_spinners:
                    # A potential new spinner is recognized
                    if (process[1].delta_iops > self.rule_disk_iops_per_process_long_spinner) or (process[1].delta_throughput_kbps > self.rule_physical_disk_throughput_kbps_threshold_any_disk):
                        # New process delta iops or throughput above
                        # threshold, it will be added as a new spinner
                        self.process_spinners[process[1].unique_id] = ProcessSpinner(process[1])
                else:
                    # Spinner already exist, we will update the
                    # spin period, archive the spin period
                    # or add a new spin period
                    self.update_spinner(process[1])

    def update_spinner(self, process_data):
        '''
        Function
        =========
        update_spinner

        Description
        ==============
        Update existing spinner by updating its active spin period,
        create a new spin period or archive the active spin period.

        Arguments
        ==========
        __in <Class> self : class instance
        __in <Process> process_data: process statistics

        Return Values
        ==============
        N/A
        '''

        if self.process_spinners[process_data.unique_id].active_process_spin_period is None:
            # The spin period is not active now
            # (i.e. the previous one was archived)
            if process_data.delta_iops > self.rule_disk_iops_per_process_long_spinner:
                # A new spin period will be added to the spinner
                self.process_spinners[process_data.unique_id].new_spin_period(process_data)
        else:
            # The spin period is active
            if process_data.delta_iops > self.rule_disk_iops_per_process_long_spinner:
                # Disk IOPS is greater
                # than threshold, update active spin period
                # with last collected metrics
                self.process_spinners[process_data.unique_id].active_process_spin_period.update_spin_period(process_data)
            else:
                # Disk IOPS gets lower
                # for the active spinner.
                # Identify if spin period should be updated or archived.
                #   1. Archived if spinner has reached timeout
                #       with no valid samples
                #       or
                #       cumulative IOPS are under threshold
                #   2. Update otherwise
                spinner_timed_out = (process_data.update_time_seconds - self.process_spinners[process_data.unique_id].active_process_spin_period.end_time_seconds) * 1000 >= self.rule_long_spinner_timeout_ms
                cumulative_iops = self.process_spinners[process_data.unique_id].active_process_spin_period.total_read_count + self.process_spinners[process_data.unique_id].active_process_spin_period.total_write_count
                cumulative_iops += (process_data.delta_reads + process_data.delta_writes) * 1.0
                cumulative_iops = cumulative_iops / (process_data.update_time_seconds - self.process_spinners[process_data.unique_id].active_process_spin_period.start_time_seconds)
                cumulative_iops_under_threshold = cumulative_iops < self.rule_disk_iops_per_process_long_spinner

                if spinner_timed_out or cumulative_iops_under_threshold:
                    # Spin period timed out (milliseconds) or
                    # cumulative IOPS is lower
                    # than threshold
                    self.process_spinners[process_data.unique_id].archive_spinner()
                else:
                    # Spinner is below threshold but spin period
                    # is active, spin period will be updated
                    # with last collected metrics
                    self.process_spinners[process_data.unique_id].active_process_spin_period.update_spin_period(process_data)

    def start(self):
        '''
        Function
        =========
        start

        Description
        ==============
        Start the timer that will trigger the performance update
        collection at the specified interval of time.
        This method is called by RuleEngine.

        Arguments
        ==========
        __in <Class> self : class instance

        Return Values
        ==============
        N/A
        '''

        # Substract interval from duration_seconds to display accurate
        # capture duration_seconds
        logger.info('Starting ' + self.rule_name + ' resource monitor (' + str(self.duration_seconds - self.interval_seconds) + ' seconds).')

        self.monitor_timer = MonitorTimer(self.duration_seconds, self.interval_seconds, self.performance_data_update_handler)
        self.monitor_timer.start()

        self.data_collection_start = self.monitor_timer.start_timer()

    def stop(self):
        '''
        Function
        =========
        stop

        Description
        ==============
        Resource Monitor stop is called by RuleEngine.
        Reconciliate all data before returning to RuleEngine.

        Arguments
        ==========
        __in <Class> self : class instance

        Return Values
        ==============
        N/A
        '''

        logger.info('Stopping ' + self.rule_name + ' resource monitor.')
        resultDict = {}

        results = []
        for spinPeriod in self.devices_spin_periods:
            result = spinPeriod.get_spin_period_result(self.interval_seconds)
            if result is not None:
                results.append(result)
        resultDict['HighDisksUsagePeriods'] = results

        results = []
        for spinPeriod in self.process_spinners:
            self.process_spinners[spinPeriod].archive_spinner()
            result = self.process_spinners[spinPeriod].get_spinner_result()
            if result is not None:
                results.append(result)

        if self.rule_top_spinners_to_report > 0:
            # Return top N spinners
            resultDict['TopLongDiskSpinners'] = sorted(results, key=lambda x: x["AvgDiskIops"], reverse=True)[:self.rule_top_spinners_to_report]
        else:
            # Retun all spinners
            resultDict['TopLongDiskSpinners'] = sorted(results, key=lambda x: x["AvgDiskIops"], reverse=True)

        resultDict['TopSpikeDiskSpinners'] = []

        return resultDict

    def is_active(self):
        '''
        Function
        =========
        is_active

        Description
        ==============
        Provides a way to check if resource monitor is active or not.
        Used by the caller from RuleEngine.

        Arguments
        ==========
        __in <Class> self : class instance

        Return Values
        ==============
        __out <bool>
        '''

        return self.monitor_timer.is_alive()


# ######################### Disks spinners #########################


class BlockDeviceMonitor():
    '''
    Class
    =========
    BlockDeviceMonitor

    Description
    ==============
    Used to reference all block devices on the system and update their
    statistics at Resource Monitor sampling interval_seconds.
    '''

    # Block device types from which we don't want to collect metrics
    EXCLUDED_DEVICES = ['fd', 'sr', 'loop', 'ram']
    # Block device patterns from which we actually want to collect metrics
    VALID_DEVICES_REGEX_PATTERN = ['sd[a-z]*\\b', 'md*', 'dm-*']
    BYTES_TO_KB_CONVERT_RATIO = 0.000976563

    def __init__(self, rule_physical_disk_percentage_threshold_any_disk, rule_physical_disk_iops_threshold_any_disk, rule_physical_disk_throughput_kbps_threshold_any_disk, target_devices='all'):
        '''
        Function
        =========
        __init__

        Description
        ==============
        Initialize the BlockDeviceMonitor class that will handle I/O statistics
        profiling.

        Arguments
        ==========
        __in <Class> self : class instance
        __in <float> rule_physical_disk_percentage_threshold_any_disk: capture duration in seconds
        __in <float> rule_physical_disk_iops_threshold_any_disk: sampling interval in seconds
        __in <float> rule_physical_disk_throughput_kbps_threshold_any_disk: this rule short name
        __in <string> target_devices: block devices targeted (not yet customizable so default is all)

        Return Values
        ==============
        N/A
        '''

        self.target_devices = target_devices
        self.rule_physical_disk_busy_percentage_threshold_any_disk = rule_physical_disk_percentage_threshold_any_disk
        self.rule_physical_disk_iops_threshold_any_disk = rule_physical_disk_iops_threshold_any_disk
        self.rule_physical_disk_throughput_kbps_threshold_any_disk = rule_physical_disk_throughput_kbps_threshold_any_disk

        self.start_time_seconds = self.update_time_seconds = self.previous_update_time_seconds = time.time()
        self.all_devices_delta_io_count = 0
        self.block_devices = {}
        self.is_physical_disk_busy_percentage_threshold_any_disk = False
        self.is_physical_disk_iops_threshold_any_disk = False
        self.is_physical_disk_throughput_kbps_threshold_any_disk = False
        self.samples_count = 0

    def get_block_devices_stats(self):
        '''
        Function
        =========
        get_block_devices_stats

        Description
        ==============
        Collect block devices statistics and verify triggers are reached.

        Fields (2.6+):
            [0] major version [1] minor version [2] BlockDevice [3] reads
            [4] reads merged [5] sectors read [6] ms spent reading
            [7] writes [8] writes merged [9] sectors written
            [10] ms spent writing [11] I/O in progress
            [12] ms doing I/O [13] weighted ms doing I/Os

        References:
            https://www.kernel.org/doc/Documentation/iostats.txt

        Arguments
        ==========
        __in <Class> self : class instance

        Return Values
        ==============
        __out <dict> devices_performance_data: collection start time, sample
            end time (update time) and a dictionnary of devices information
            and statistics
        '''

        # Get diskstats data for useful devices only
        filtered_proc_diskstats = self.get_filtered_diskstats()

        self.update_time_seconds = time.time()
        # Reset I/O delta counter and triggers
        self.all_devices_delta_io_count = 0
        self.is_physical_disk_busy_percentage_threshold_any_disk = False
        self.is_physical_disk_iops_threshold_any_disk = False
        self.is_physical_disk_throughput_kbps_threshold_any_disk = False

        delta_time_seconds = float(self.update_time_seconds - self.previous_update_time_seconds)
        for device_proc_diskstats in filtered_proc_diskstats:
            current_block_device_name = device_proc_diskstats.split()[2]
            if not self.block_devices or (current_block_device_name not in self.block_devices):
                # No statistics collected at all or no statistics collected
                # yet for this block device so add to the dictionary
                self.block_devices[current_block_device_name] = BlockDevice(device_proc_diskstats.split())  # , self.valid_devices[current_block_device_name])
            else:
                # Block device already known, update block device statistics
                self.block_devices[current_block_device_name].update_block_device(device_proc_diskstats.split(), delta_time_seconds)

            delta_io_count = self.block_devices[current_block_device_name].delta_read_count + self.block_devices[current_block_device_name].delta_write_count
            self.all_devices_delta_io_count += delta_io_count
            delta_kbytes = long(self.block_devices[current_block_device_name].delta_read_bytes + self.block_devices[current_block_device_name].delta_write_bytes) / 1024
            # TODO: ensure accuracy of the below
            delta_busy_percentage = 100 * self.block_devices[current_block_device_name].delta_busy_time_ms / (delta_time_seconds * 1000)

            # Verify we reach any trigger, shortcut to
            # ResourceMonitor.performance_data_update_handler()
            # to simplify the decision to create or update a spin period as
            # triggers are defined per disk
            if delta_io_count / delta_time_seconds >= self.rule_physical_disk_iops_threshold_any_disk:
                # Delta IOPS above threshold

                self.is_physical_disk_iops_threshold_any_disk = True
                self.block_devices[current_block_device_name].has_reached_threshold = True
            if delta_kbytes  / delta_time_seconds >= self.rule_physical_disk_throughput_kbps_threshold_any_disk:
                # Delta throughput above threshold

                self.is_physical_disk_throughput_kbps_threshold_any_disk = True
                self.block_devices[current_block_device_name].has_reached_threshold = True
            if delta_busy_percentage >= self.rule_physical_disk_busy_percentage_threshold_any_disk:
                # Delta busy time % above threshold
                self.is_physical_disk_busy_percentage_threshold_any_disk = True
                self.block_devices[current_block_device_name].has_reached_threshold = True

        devices_performance_data = ({'StartTimeSeconds': self.start_time_seconds, 'UpdateTimeSeconds': self.update_time_seconds, 'PreviousUpdateTimeSeconds': self.previous_update_time_seconds, 'DeltaTimeSeconds': delta_time_seconds, 'Data': self.block_devices})
        self.previous_update_time_seconds = self.update_time_seconds
        self.samples_count += 1

        return devices_performance_data

    def get_filtered_diskstats(self):
        '''
        Function
        =========
        get_filtered_diskstats

        Description
        ==============
        Parse /proc/diskstats to return statistics from targeted *block*
        devices only, excluding unwanted devices and partitions.
        See self.EXCLUDED_DEVICES for list of exclusion.

        Arguments
        ==========
        __in <Class> self : class instance

        Return Values
        ==============
        __out <list> raw diskstats lines for each targeted block device
        '''

        results = []
        diskstats_file = open('/proc/diskstats', 'r').readlines()
        for device_stats in diskstats_file:
            for regex_pattern in self.VALID_DEVICES_REGEX_PATTERN:
                if re.compile(regex_pattern).match(device_stats.split()[2]) and \
                    (device_stats.split()[2][:-1] not in self.EXCLUDED_DEVICES) and \
                        (device_stats.split()[2][:-2] not in self.EXCLUDED_DEVICES) and \
                            (device_stats.split()[2] not in results):
                    # Block device is valid, i.e. it's a regular block device
                    # or a RAID device
                    results.append(device_stats)

        return results


class BlockDevice():
    '''
    Class
    =========
    BlockDevice

    Description
    ==============
    BlockDevice is used to measure metrics for a single block device.
    Block device iostat comes from /proc/diskstats.

    Fields (2.6+):
        [0] major version [1] minor version [2] BlockDevice
        [3] reads [4] reads merged [5] sectors read
        [6] ms spent reading
        [7] writes [8] writes merged [9] sectors written
        [10] ms spent writing
        [11] I/O in progress [12] ms doing I/O
        [13] weighted ms doing I/Os

    References:
        https://www.kernel.org/doc/Documentation/iostats.txt
        https://www.kernel.org/doc/Documentation/ABI/testing/procfs-diskstats
        http://rickardnobel.se/storage-performance-capture_duration_iops-latency-throughput/
        https://blog.docbert.org/queue-depth-capture_duration_iops-and-latency/
        https://louwrentius.com/understanding-capture_duration_iops-latency-and-storage-performance.html
        https://serverfault.com/questions/862334/interpreting-read-write-and-total-io-time-in-proc-diskstats
    '''

    def __init__(self, device_stats):
        '''
        Function
        =========
        __init__

        Description
        ==============
        Initialize the BlockDevice class that will contain single device
        information and statistics.

        Arguments
        ==========
        __in <Class> self : class instance
        __in <list> device_stats: raw diskstats lines for each targeted device

        Return Values
        ==============
        N/A
        '''

        # Block device name (/dev/[sd*|md*|dm*])
        self.device_name = device_stats[2]
        # Block device sector size in bytes, used for throughput calculation
        self.sector_size_bytes = DiskUtil.get_block_device_sector_size_bytes(self.device_name)
        # Flag to identify a device which has reached
        # a trigged. Will be update by caller.
        self.has_reached_threshold = False
        # Init metrics to 0
        self.delta_busy_time_ms = 0
        self.delta_busy_time_percentage = 0.0
        self.delta_read_count = self.delta_write_count = self.delta_read_sectors = self.delta_write_sectors = 0
        self.delta_read_bytes = self.delta_write_bytes = 0
        self.delta_ms_reading = self.delta_ms_writing = 0
        self.queue_length = 0
        self.min_queue_depth = self.max_queue_depth = 0
        '''
        # Below is commented until we will need to compute latency
        self.avg_io_latency_ms = 0.0
        self._avgRLatencyMS = self._avgWLatencyMS = 0.0
        '''
        # Get current metrics for next sample calculation
        self.previous_read_count = long(device_stats[3])
        self.previous_write_count = long(device_stats[7])
        '''
        # Below is commented until we will need to compute latency
        self.previous_reads_merged = int(device_stats[4])    # Number of reads merged (int)
        self.previous_writes_merged = int(device_stats[8])   # Number of writes merged (int)
        '''
        self.previous_read_sectors = long(device_stats[5])
        self.previous_write_sectors = long(device_stats[9])
        self.previous_ms_reading = long(device_stats[6])
        self.previous_ms_writing = long(device_stats[10])
        self.previous_busy_time_ms = long(device_stats[12])
        self.previous_queue_length = long(device_stats[11])

    def update_block_device(self, device_stats, delta_time_seconds):
        '''
        Function
        =========
        update_block_device

        Description
        ==============
        Update block device statistics.

        Arguments
        ==========
        __in <Class> self : class instance
        __in <list> device_stats: raw diskstats lines for each targeted device
        __in <float> delta_time_seconds: delta time since last sample in
            seconds

        Return Values
        ==============
        N/A
        '''

        # Set new values
        self.has_reached_threshold = False
        self.delta_read_count = long(device_stats[3]) - self.previous_read_count
        self.delta_write_count = long(device_stats[7]) - self.previous_write_count
        self.delta_read_sectors = long(device_stats[5]) - self.previous_read_sectors
        self.delta_write_sectors = long(device_stats[9]) - self.previous_write_sectors
        self.delta_ms_reading = long(device_stats[6]) - self.previous_ms_reading
        self.delta_ms_writing = long(device_stats[10]) - self.previous_ms_writing
        # Calculations
        self.delta_busy_time_ms = long(device_stats[12]) - self.previous_busy_time_ms
        self.delta_busy_time_percentage = 100 * self.delta_busy_time_ms / (delta_time_seconds * 1000)
        self.delta_read_bytes = long(self.delta_read_sectors * self.sector_size_bytes)
        self.delta_write_bytes = long(self.delta_write_sectors * self.sector_size_bytes)

        self.queue_length = int(device_stats[11])   
        self.min_queue_depth = min(self.queue_length, self.min_queue_depth)
        self.max_queue_depth = max(self.queue_length, self.max_queue_depth)

        # Save values for next sample calculation
        self.previous_read_count = int(device_stats[3])
        self.previous_write_count = int(device_stats[7])
        self.previous_reads_merged = int(device_stats[4])
        self.previous_writes_merged = int(device_stats[8])
        self.previous_read_sectors = int(device_stats[5])
        self.previous_write_sectors = int(device_stats[9])
        self.previous_ms_reading = int(device_stats[6])
        self.previous_ms_writing = int(device_stats[10])
        self.previous_busy_time_ms = int(device_stats[12])
        self.previous_queue_length = self.queue_length


class BlockDevicesSpinPeriods():
    """
    Class
    =========
    BlockDevicesSpinPeriods

    Description
    ==============
    A spin period is a time frame during which one or more block device
    statistics were above the threshold(s).

    """

    def __init__(self, devices_performance_data, rule_physical_disk_iops_threshold_any_disk):
        '''
        Function
        =========
        __init__

        Description
        ==============
        Initialize the BlockDevicesSpinPeriods class that will store statistics
        for the time period devices statistics will be above triggers.

        Arguments
        ==========
        __in <Class> self : class instance
        __in <dict> devices_performance_data: dictionnary containing all devices statistics
        and sample start and end time

        Return Values
        ==============
        N/A
        '''

        self.start_time_seconds = devices_performance_data['PreviousUpdateTimeSeconds']
        self.end_time_seconds = devices_performance_data['UpdateTimeSeconds']
        self.rule_physical_disk_iops_threshold_any_disk = rule_physical_disk_iops_threshold_any_disk

        self.block_devices_usage = {}
        for device_performance_data in devices_performance_data['Data'].values():
            # Build BlockDeviceUsage from BlockDevice data
            #if (device_performance_data.delta_read_count + device_performance_data.delta_write_count) / devices_performance_data['DeltaTimeSeconds'] > self.rule_physical_disk_iops_threshold_any_disk:
            if device_performance_data.has_reached_threshold is True:
                # This block device above any of the threshold
                self.block_devices_usage[device_performance_data.device_name] = BlockDeviceUsage(device_performance_data)

        self.samples_count = 1

    def update_usages(self, devices_performance_data):
        '''
        Function
        =========
        update_usages

        Description
        ==============
        Update the block devices dictionnary with latest block devices
        collection.

        Arguments
        ==========
        __in <Class> self : class instance
        __in <dict> devices_performance_data: dictionnary containing all devices statistics
            and sample start and end time

        Return Values
        ==============
        N/A
        '''

        self.end_time_seconds = devices_performance_data['UpdateTimeSeconds']
        for device_performance_data in devices_performance_data['Data'].values():
            if device_performance_data.has_reached_threshold is True:
            #if (device_performance_data.delta_read_count + device_performance_data.delta_write_count) / devices_performance_data['DeltaTimeSeconds'] > self.rule_physical_disk_iops_threshold_any_disk:
                # This block device above any of the thresholds
                if device_performance_data.device_name in self.block_devices_usage:
                    # Device already referenced
                    self.block_devices_usage[device_performance_data.device_name].update_usage(device_performance_data, devices_performance_data['DeltaTimeSeconds'])
                else:
                    # New block device in the spin period
                    self.block_devices_usage[device_performance_data.device_name] = BlockDeviceUsage(device_performance_data)
        self.samples_count += 1

    def get_spin_period_result(self, interval_seconds):
        '''
        Function
        =========
        get_spin_period_result

        Description
        ==============
        At the end of the capture, provides a dictionary with profiling data
        that will be returned to the Resource Monitor.
        It provides the overall profiling data and profiling data per
        spin period.

        Arguments
        ==========
        __in <Class> self : class instance

        Return Values
        ==============
        __out <dict> spinner_result: formated spin period data
        '''

        spinner_result = {}
        if self.samples_count > 1:            
            # delta_time_seconds as float as it will be used to calculate
            # averages which must be shown as float
            delta_time_seconds = float(self.end_time_seconds - self.start_time_seconds)
            spinner_result['DisksPerfData'] = [] 
            for device_name in self.block_devices_usage:
                device_result = {} 
                device_result['DiskNumber'] = device_name
                device_result['SectorSize'] = self.block_devices_usage[device_name].sector_size_bytes
                device_result['ReadIoTimeMS'] = self.block_devices_usage[device_name].cumulative_ms_reading
                device_result['WriteIoTimeMS'] = self.block_devices_usage[device_name].cumulative_ms_writing
                device_result['BusyTimeMS'] = self.block_devices_usage[device_name].cumulative_busy_time_ms
                # TODO: busy time % may not be accurate, needs to be verified.
                # As all calculations are based in I/O or throughput PER SECONDS, 
                # then busy time % which is calculated from a ms basis with a
                # timeframe expressed in seconds will not be very accurate.
                device_result['BusyTimePercentage'] = 100 * self.block_devices_usage[device_name].cumulative_busy_time_ms / (delta_time_seconds * 1000)
                device_result['ReadsCount'] = self.block_devices_usage[device_name].cumulative_read_count
                device_result['WritesCount'] = self.block_devices_usage[device_name].cumulative_write_count
                device_result['AvgIOPS'] = (device_result['ReadsCount'] + device_result['WritesCount']) / delta_time_seconds
                device_result['ReadsKB'] = self.block_devices_usage[device_name].cumulative_read_kbytes
                device_result['WritesKB'] = self.block_devices_usage[device_name].cumulative_write_kbytes
                device_result['ReadsMB'] = long(device_result['ReadsKB'] / 1024)
                device_result['WritesMB'] = long(device_result['WritesKB'] / 1024)
                device_result['AvgThroughputKBPS'] = (device_result['ReadsKB'] + device_result['WritesKB']) / delta_time_seconds
                device_result['AvgThroughputMBPS'] = (device_result['ReadsMB'] + device_result['WritesMB']) / delta_time_seconds
                device_result['MinQueueDepth'] = self.block_devices_usage[device_name].min_queue_depth
                device_result['MaxQueueDepth'] = self.block_devices_usage[device_name].max_queue_depth

                # ReadAWait and WriteAWait are derived the same way that iostat derives them
                # See https://github.com/sysstat/sysstat/blob/d8a22f2c32f5705319ff64c83c6b4cdccd3fdc27/iostat.c#L1472-L1490
                if self.block_devices_usage[device_name].cumulative_read_count > 0:
                    device_result['ReadAWait'] = self.block_devices_usage[device_name].cumulative_ms_reading / \
                                                 self.block_devices_usage[device_name].cumulative_read_count
                else:
                    device_result['ReadAWait'] = 0


                if self.block_devices_usage[device_name].cumulative_write_count > 0:
                    device_result['WriteAWait'] = self.block_devices_usage[device_name].cumulative_ms_writing / \
                                                 self.block_devices_usage[device_name].cumulative_write_count
                else:
                    device_result['WriteAWait'] = 0

                spinner_result['StartTime'] = datetime.utcfromtimestamp(self.start_time_seconds).strftime('%Y-%m-%d %H:%M:%S.%f')[:-3]
                if self.samples_count == 1:
                    # If there is a single sample we did not have time
                    # to update end time, so add capture interval to have a valid sample
                    self.start_time_seconds += interval_seconds
                spinner_result['EndTime'] = datetime.utcfromtimestamp(self.end_time_seconds).strftime('%Y-%m-%d %H:%M:%S.%f')[:-3]
                spinner_result['DisksPerfData'].append(device_result)
        else:
            spinner_result = None
            
        return spinner_result


class BlockDeviceUsage():
    '''
    Class
    =========
    BlockDeviceUsage

    Description
    ==============
    A block device usage stores block device statistics for the spin period timeframe.
    '''

    def __init__(self, device_performance_data):
        '''
       	Function
        =========
        __init__

        Description
        ==============
        Initialize a new object to hold block device statistics for a spin period.

        Arguments
        ==========
        __in <Class> self : class instance
        __in <object> device_performance_data: block device statistics
        
        Return Values
        ==============
        N/A
        '''

        self.device_name = device_performance_data.device_name
        self.sector_size_bytes = device_performance_data.sector_size_bytes
        self.cumulative_read_count = device_performance_data.delta_read_count
        self.cumulative_write_count = device_performance_data.delta_write_count
        self.cumulative_read_kbytes = long(device_performance_data.delta_read_bytes / 1024)
        self.cumulative_write_kbytes = long(device_performance_data.delta_write_bytes / 1024)
        self.cumulative_ms_reading = device_performance_data.delta_ms_reading
        self.cumulative_ms_writing = device_performance_data.delta_ms_writing
        self.cumulative_busy_time_ms = device_performance_data.delta_busy_time_ms
        self.min_queue_depth = device_performance_data.min_queue_depth
        self.max_queue_depth = device_performance_data.max_queue_depth

    def update_usage(self , device_performance_data, delta_time_seconds):
        '''
       	Function
        =========
        update_usage

        Description
        ==============
        Update statistics which hold block device statistics for a spin period.

        Arguments
        ==========
        __in <Class> self : class instance
        __in <object> device_performance_data: block device statistics
        
        Return Values
        ==============
        N/A
        '''

        self.cumulative_read_count += device_performance_data.delta_read_count
        self.cumulative_write_count += device_performance_data.delta_write_count
        self.cumulative_read_kbytes += long(device_performance_data.delta_read_bytes / 1024)
        self.cumulative_write_kbytes += long(device_performance_data.delta_write_bytes / 1024)
        self.cumulative_ms_reading += device_performance_data.delta_ms_reading
        self.cumulative_ms_writing += device_performance_data.delta_ms_writing
        self.cumulative_busy_time_ms += device_performance_data.delta_busy_time_ms
        self.min_queue_depth = device_performance_data.min_queue_depth
        self.max_queue_depth = device_performance_data.max_queue_depth

# ######################### Process spinners #########################


class ProcessMonitor():
    '''
    Class
    =========
    ProcessMonitor

    Description
    ==============
    Used to reference all processes running on the system and update their
    I/O statistics at Resource Monitor sampling interval_seconds.
    '''

    def __init__(self):
        '''
       	Function
        =========
        __init__

        Description
        ==============
        Initialize the processes class that will handle process I/O statistics
        profiling.

        Arguments
        ==========
        __in <Class> self : class instance
        __in <float> rule_physical_disk_iops_threshold_any_disk: sampling interval in seconds
        
        Return Values
        ==============
        N/A
        '''

        self.processes = {}
        self.samples_count = 0
    
    def get_processes_stats(self):
        '''
       	Function
        =========
        get_processes_stats

        Description
        ==============
        Get all existing processes and their current I/O statistics
        and add them if they are not in the dictionary.

        Arguments
        ==========
        __in <Class> self : class instance
                
        Return Values
        ==============
        N/A 
        '''
        
        for pid in [int(x) for x in os.listdir(b'/proc') if x.isdigit()]:
            process_name = ProcessUtil.get_process_name(pid)
            if process_name is not None and (str(pid) + '-' + process_name) in self.processes:
                # Process already referenced
                if self.processes[str(pid) + '-' + process_name].is_alive:
                    # Process still alive, update process statistics
                    self.processes[str(pid) + '-' + process_name].update_process()
            else:
                new_process = Process(pid)
                if new_process.is_alive:
                    self.processes[new_process.unique_id] = new_process
        
        self.samples_count += 1


class Process:
    '''
    Class
    =========
    Process

    Description
    ==============
    Used to hold a process information and I/O statistics.
    '''

    def __init__(self , pid):
        '''
       	Function
        =========
        __init__

        Description
        ==============
        Initialize the process class that will handle a process I/O statistics
        profiling.

        Arguments
        ==========
        __in <Class> self : class instance
        __in <int> pid: process pid
        
        Return Values
        ==============
        N/A
        '''

        # TODO: Note for improvment:
        # if (self.process_name = ProcessUtil.get_process_name(pid)) is not None
        # to verify process still alive instead of following condition
        self.process_name = ProcessUtil.get_process_name(pid)
        if self.process_name is not None:
            self.pid = pid
            self.unique_id = str(pid) + '-' + self.process_name
            self.process_start_time_ms = ProcessUtil.get_process_start_time_ms(self.pid)
            self.start_time_seconds = self.update_time_seconds = self.previous_update_time_seconds = time.time()

            self.delta_reads = self.delta_writes = self.delta_read_bytes = self.delta_write_bytes = 0
            self.delta_iops = self.delta_throughput_kbps = 0.0
            self.previous_reads = self.previous_writes = self.previous_read_bytes = self.previous_write_bytes = 0
            
            # TODO: Note for improvment:
            # Ensure ProcessUtil.get_process_io_stats returns none
            # if process has no io file
            process_iostat = ProcessUtil.get_process_io_stats(self.pid)
            self.previous_reads = process_iostat['reads']
            self.previous_writes = process_iostat['writes']
            self.previous_read_bytes = process_iostat['read_bytes']
            self.previous_write_bytes = process_iostat['write_bytes']
            self.samples_count = 0
            self.is_alive = True
        else:
            self.is_alive = False
    
    def update_process(self):
        '''
       	Function
        =========
        update_process

        Description
        ==============
        Update process statistics with latest performance data.

        Arguments
        ==========
        __in <Class> self : class instance
        __in <float> delta_time_seconds: delta time since last sample in seconds
        
        Return Values
        ==============
        N/A
        '''

        if ProcessUtil.is_alive(self.pid):
            self.update_time_seconds = time.time()
            process_iostat = ProcessUtil.get_process_io_stats(self.pid)
            if process_iostat is not None:
                self.delta_reads = process_iostat['reads'] - self.previous_reads
                self.delta_writes = process_iostat['writes'] - self.previous_writes
                self.delta_read_bytes = process_iostat['read_bytes'] - self.previous_read_bytes
                self.delta_write_bytes = process_iostat['write_bytes'] - self.previous_write_bytes
                self.delta_iops = float(self.delta_reads + self.delta_writes) / (self.update_time_seconds - self.previous_update_time_seconds)
                self.delta_throughput_kbps = float(self.delta_read_bytes + self.delta_write_bytes) / (self.update_time_seconds - self.previous_update_time_seconds)
                
                self.previous_reads = process_iostat['reads']
                self.previous_writes = process_iostat['writes']
                self.previous_read_bytes = process_iostat['read_bytes']
                self.previous_write_bytes = process_iostat['write_bytes']
                self.previous_update_time_seconds = self.update_time_seconds
                self.samples_count += 1
            else:
                # Process does not exist anymore
                self.is_alive = False
        else:
            # Process does not exist anymore
            self.is_alive = False


class ProcessSpinner():
    '''
    Class
    =========
    ProcessSpinner

    Description
    ==============
    A process spinner is an instance of a process which has had an I/O usage higher than PerfInsights defined threshold at least once during the profiling.
    A process spinner can have one or more I/O spin periods during which the I/O usage average was higher than PerfInsights defined threshold.
    A spin period is active while process I/O usage is above the threshold or while the process average usage since profiling start is below the threshold but has not reached the defined PerfInsights timeout.
    A spin period is archived when process average I/O usage is below the threshold and timeout has been reached.
    '''

    def __init__(self, process_data):
        '''
       	Function
        =========
        __init__

        Description
        ==============
        Initialize an object which will host an instance of a process spinner.

        Arguments
        ==========
        __in <Class> self : class instance
        __in <int> pid: process pid
        
        Return Values
        ==============
        N/A
        '''

        self.process_id = process_data.pid
        self.process_name = process_data.process_name
        self.unique_id = str(self.process_id) + '-' + self.process_name
        self.process_spin_periods = []
        self.active_process_spin_period = ProcessSpinPeriod(process_data)
        self.total_disk_io_count = 0 # process_data.delta_reads + process_data.delta_writes
        self.total_disk_io_size_kbytes = 0 #  => spinningPeriods.Count == 0 ? 0 : spinningPeriods.Sum(x => x.PerfData.BytesReadKB + x.PerfData.BytesWriteKB);
        self.avg_disk_iops = 0 # => spinningPeriods.Count == 0 ? 0.0 : spinningPeriods.Average(x => (x.PerfData.ReadCount + x.PerfData.WriteCount) * 1.0 / (x.EndTime - x.StartTime).TotalSeconds);
        
    def new_spin_period(self , process_data):
        '''
        Function
        =========
        new_spin_period
    
        Description
        ==============
        A new spin period has been detected so the active spin period will be used to store data.
        Initialized with process collection last update time (previous_update_time_seconds).

        Arguments
        ==========
            __in <Class> self : class instance
            __in <object> process_data : process metadata and statistics
    
        Return Values
        ==============
            N/A
        '''

        self.active_process_spin_period = ProcessSpinPeriod(process_data)

    def archive_spinner(self):
        '''
       	Function
        =========
        archive_spin_period

        Description
        ==============
        Archive spin period when the process average I/O usage is below the threshold or not active for more than the defined timeout.
        The active spin period is set to None.

        Arguments
        ==========
        __in <Class> self : class instance
        
        Return Values
        ==============
        N/A
        '''

        if self.active_process_spin_period is not None:
            # Active spin period is not null
            if self.active_process_spin_period.start_time_seconds < self.active_process_spin_period.end_time_seconds:
                # Active spin period has valid samples
                self.process_spin_periods.append(self.active_process_spin_period)
        self.active_process_spin_period = None

    def get_spinner_result(self):
        '''
       	Function
        =========
        get_spinner_result

        Description
        ==============
        At the end of the capture, provides a dictionary with profiling data that will be returned to the Resource Monitor.
        It provides the overall profiling data and profiling data per spin period.

        Arguments
        ==========
        __in <Class> self : class instance
        
        Return Values
        ==============
        __out <dict> spinner_result: spinner results
        '''

        if len(self.process_spin_periods) == 0:
            return None

        spinner_result = {}
        spinner_result['ProcessName'] = self.process_name
        spinner_result['ProcessId'] = self.process_id
        spinner_result['ServicesHosted'] = ''
        spinner_result['ImagePath'] = ''
        spinner_result['CompanyName'] = ''
        spinner_result['TotalDiskIoCount'] = 0
        spinner_result['TotalDiskIoSizeKB'] = 0
        spinner_result['AvgDiskIops'] = 0

        spinner_result['DiskSpinPeriods'] = [] 
        for process_spin_period in self.process_spin_periods:
            spinner_result['TotalDiskIoCount'] += process_spin_period.total_read_count + process_spin_period.total_write_count
            spinner_result['TotalDiskIoSizeKB'] = process_spin_period.total_bytes_read_kbytes + process_spin_period.total_bytes_write_kbytes
            spinner_result['DiskSpinPeriods'].append({'PerfData':{'ReadCount': process_spin_period.total_read_count, 'BytesReadKB': process_spin_period.total_bytes_read_kbytes, 'WriteCount': process_spin_period.total_write_count, 'BytesWriteKB': process_spin_period.total_bytes_write_kbytes}, \
            'StartTime': datetime.utcfromtimestamp(process_spin_period.start_time_seconds).strftime('%Y-%m-%d %H:%M:%S.%f')[:-3], \
            'EndTime': datetime.utcfromtimestamp(process_spin_period.end_time_seconds).strftime('%Y-%m-%d %H:%M:%S.%f')[:-3]})
            spinner_result['AvgDiskIops'] = spinner_result['TotalDiskIoCount'] * 1.0 / (process_spin_period.end_time_seconds- process_spin_period.start_time_seconds)

        return spinner_result


class ProcessSpinPeriod():
    '''
    Class
    =========
    ProcessorSpinPeriod

    Description
    ==============
    A processor spin period represents the capture of metrics during the time disk usage was higher than PerfInsights defined thresholds at least once during the profiling.
    '''

    def __init__(self, process_performance_data):
        '''
       	Function
        =========
        __init__

        Description
        ==============
        Initialize a spin period for a process spinner.

        Arguments
        ==========
        __in <Class> self : class instance
        __in <Process> process_performance_data: an instance of process with statistics
        
        Return Values
        ==============
        N/A
        '''

        self.start_time_seconds = process_performance_data.update_time_seconds
        self.end_time_seconds = process_performance_data.update_time_seconds
        self.total_read_count = process_performance_data.delta_reads
        self.total_write_count = process_performance_data.delta_writes
        self.total_bytes_read_kbytes = long(process_performance_data.delta_read_bytes / 1024)
        self.total_bytes_write_kbytes = long(process_performance_data.delta_write_bytes / 1024)
    
    def update_spin_period(self, process_performance_data):
        '''
       	Function
        =========
        update_spin_period

        Description
        ==============
        Update a spin period for a process spinner with latest process statistics.

        Arguments
        ==========
        __in <Class> self : class instance
        __in <Process> process_performance_data: an instance of process with statistics
        
        Return Values
        ==============
        N/A
        '''

        

        self.end_time_seconds = process_performance_data.update_time_seconds
        self.total_read_count += process_performance_data.delta_reads
        self.total_write_count += process_performance_data.delta_writes
        self.total_bytes_read_kbytes += long(process_performance_data.delta_read_bytes / 1024)
        self.total_bytes_write_kbytes += long(process_performance_data.delta_write_bytes / 1024)
