# -------------------------------------------------------------------------
# Copyright (c) Microsoft.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

import os
import platform
import time
import shutil
import datetime
import time
from collections import namedtuple
import warnings
import errno
import sys
import re
import subprocess
import logging
import six

import util.common as common
from ruleengine import DataCollector, RuleEngineCls
from util.SystemUtil import SystemUtil, CpuUtil, ProcessUtil, MemoryUtil
from ruleengine.RuleEngineCls import IAnalyticRule, IResourceMonitor, MonitorTimer

logger = logging.getLogger(__name__)

if six.PY3:
    long = int

class ResourceMonitor(IResourceMonitor):
    """
    Class
    =========
    ResourceMonitor

    Description
    ==============
    Class used to manage the whole data collection.
    Starts the data collection asynchronously by running a new MonitorTimer
    """

    def __init__(self, duration_seconds, interval_seconds, rule_name, rule_definition):
        '''
        Function
        =========
        __init__

        Description
        ==============
        Initialize the ResourceMonitor class that will handle resource
        profiling.

        Arguments
        ==========
        __in <Class> self : class instance
        __in <int> duration_seconds: capture duration in seconds
        __in <int> interval_seconds: sampling interval in seconds
        __in <string> rule_name: this rule short name
        __in <object> rule_definition: this rule configuration

        Return Values
        ==============
        N/A
        '''

        # Profiling duration_seconds in seconds
        # (adds interval to the duration_seconds as the first sample will be
        # inconsistent)
        self.duration_seconds = duration_seconds
        self.interval_seconds = interval_seconds
        self.rule_name = rule_name
        self.rule_definition = rule_definition 
        self.rule_config = self.rule_definition.get('Configuration')
        self.rule_available_physical_memory_percentage_threshold = float(self.rule_config['AvailablePhysicalMemoryPercentageThreshold'])    # Free memory % threshold 
        self.rule_rss_memory_in_use_percentage_long_spinner_threshold = float(self.rule_config['RssInUsePercentageLongSpinner'])            # Linux resident memory % threshold
        self.rule_long_spinner_timeout_ms = long(self.rule_config['LongSpinnerTimeoutMS'])                                                  # Spinner timeout
        self.rule_top_spinners_to_report = int(self.rule_config['TopSpinnersToReport'])                                                     # Number of top spinners to report
        self.monitor_timer = None
        self.memory_spin_periods = []
        self.process_spinners = {}
        self.data_collection_start = None
        self.memory_monitor = MemoryMonitor()
        self.process_monitor = ProcessMonitor(self.memory_monitor.total_physical_memory_kbytes)

    # TODO: Consider using __enter__ instead
    @staticmethod
    def init(duration_seconds, interval_seconds, rule_name, rule_config):
        '''
        Function
        =========
        init

        Description
        ==============
        Used by RuleEngine to instantiate this ResourceMonitor instance
        as __init__ is not able to return it.

        Arguments
        ==========
        __in <int> duration_seconds: capture duration_seconds in seconds
        __in <int> interval: sampling interval in seconds
        __in <string> rule_name: this rule short name
        __in <object> rule_definition: this rule configuration

        Return Values
        ==============
        __out <ResourceMonitor> this class instance
        '''

        return ResourceMonitor(duration_seconds, interval_seconds, rule_name, rule_config)

    def performance_data_update_handler(self):
        '''
        Function
        =========
        performance_data_update_handler

        Description
        ==============
        At the specified interval, the MonitorTimer will call
        this function to collect current processors and processes metrics.

        Arguments
        ==========
        __in <Class> self : class instance

        Return Values
        ==============
        N/A
        '''

        # Get latest memory statistics
        self.memory_monitor.get_memory_stats()

        if self.memory_monitor.samples_count > 1:
            if self.memory_monitor.available_physical_memory_percentage < self.rule_available_physical_memory_percentage_threshold:
                # Available memory below threshold

                if not self.memory_spin_periods or (self.memory_monitor.update_time_ms) - (self.memory_spin_periods[-1].end_time_ms) >= self.rule_long_spinner_timeout_ms:
                    # Add a new spin period if:
                    # 1. It is the first spin period recognized
                    # 2. The spin period timed out
                    self.memory_spin_periods.append(MemorySpinPeriod(self.memory_monitor))
                else:
                    # Update last spin period
                    self.memory_spin_periods[-1].update_spin_period(self.memory_monitor)

        # Get latest processes statistics
        self.process_monitor.get_processes_stats()

        if self.process_monitor.samples_count > 1:
            for process in self.process_monitor.processes.items():
                if process[1].unique_id not in self.process_spinners:
                    # A potential new spinner is recognized
                    if process[1].current_process_rss_memory_usage_kbytes is not None and process[1].current_process_rss_memory_percentage_usage > self.rule_rss_memory_in_use_percentage_long_spinner_threshold:
                        # New process resident memory usage in percent is above the threshold
                        # it will be added as a new spinner
                        self.process_spinners[process[1].unique_id] = ProcessSpinner(process[1])
                else:
                    # Spinner already exist
                    if process[1].is_alive is False:
                        # Process does not exist anymore, archive spin period
                        self.process_spinners[process[1].unique_id].archive_spin_period()
                    else:
                        # Process was still alive when collecting metrics
                        # update the spin period
                        self.update_spinner(process[1])

    def update_spinner(self, process_data):
        '''
        Function
        =========
        update_spinner

        Description
        ==============
        Update existing spinner by updating its active spin period,
        create a new spin period or archive the active spin period.

        Arguments
        ==========
        __in <Class> self : class instance
        __in <Process> process_data: current process metadata and statistics

        Return Values
        ==============
        N/A
        '''

        if self.process_spinners[process_data.unique_id].active_process_spin_period is None:
            # There is no spin period for this spinner now (i.e. the previous one was archived)
            if process_data.current_process_rss_memory_percentage_usage > self.rule_rss_memory_in_use_percentage_long_spinner_threshold:
                # A new spin period will be added to the spinner
                self.process_spinners[process_data.unique_id].new_spin_period(process_data)
        else:
            # There is an active spin period for this spinner
            if process_data.current_process_rss_memory_percentage_usage > self.rule_rss_memory_in_use_percentage_long_spinner_threshold:
                # Process resident memory is greater than threshold, update active spin period with last collected statistics
                self.process_spinners[process_data.unique_id].active_process_spin_period.update_spin_period(process_data)
            else:
                # Process resident memory gets lower for the active spinner, check if it exceeds capture timeout
                # or if cumulative resident memory is lower than threshold

                spinner_timed_out = (process_data.update_time_ms) - (self.process_spinners[process_data.unique_id].active_process_spin_period.end_time_ms) >= self.rule_long_spinner_timeout_ms
                overall_under_threshold = ((self.process_spinners[process_data.unique_id].active_process_spin_period.avg_rss_memory_percentage_usage + process_data.current_process_rss_memory_usage_kbytes) / 2 > 0.0) and ((self.process_spinners[process_data.unique_id].active_process_spin_period.avg_rss_memory_percentage_usage + process_data.current_process_rss_memory_usage_kbytes) / 2 < self.rule_rss_memory_in_use_percentage_long_spinner_threshold)
                if spinner_timed_out or overall_under_threshold:
                    # Archives spin period and resets process metrics
                    self.process_spinners[process_data.unique_id].archive_spin_period()
                else:
                    # Spinner is below threshold but spin period is active, spin period will be updated with last collected metrics
                    self.process_spinners[process_data.unique_id].active_process_spin_period.update_spin_period(process_data)

    def display_process_spin_periods(self, process_spinners):
        '''
        Function
        =========
        display_process_spin_periods

        Description
        ==============
        For debug purpose.
        Display processes satistics from existing spinners
        for last capture interval.

        Arguments
        ==========
        __in <Class> self : class instance
        __in <ProcessSpinner> process_spinners: current spinners

        Return Values
        ==============
        N/A
        '''

        logger.debug('\n --- Process spinners ---')
        logger.debug('UniqueId#\t| Avg RSS%\t\t| Average Swap')
        for process in process_spinners:
            if process_spinners[process].active_process_spin_period is not None:
                logger.debug('{0}\t\t| {1}\t\t| {2}'.format(str(process_spinners[process].unique_id), str(process_spinners[process].active_process_spin_period.avg_rss_memory_percentage_usage), str(process_spinners[process].active_process_spin_period.avg_swap_usage_kbytes)))
            else:
                if not process_spinners[process].process_spin_periods:
                    logger.error('Error, process spinner ' + process + ' has no active spin period and no archived spin periods')

    def start(self):
        '''
        Function
        =========
        start

        Description
        ==============
        Start the timer that will trigger the performance update
        collection at the specified interval of time.
        This method is called by RuleEngine.

        Arguments
        ==========
        __in <Class> self : class instance

        Return Values
        ==============
        N/A
        '''

        # Substract interval from duration_seconds to display accurate
        # capture duration_seconds
        logger.info('Starting ' + self.rule_name + ' resource monitor (' + str(self.duration_seconds - self.interval_seconds) + ' seconds).')

        self.monitor_timer = MonitorTimer(self.duration_seconds, self.interval_seconds, self.performance_data_update_handler)
        self.monitor_timer.start()
        self.data_collection_start = self.monitor_timer.start_timer()

    def stop(self):
        '''
        Function
        =========
        stop

        Description
        ==============
        Resource Monitor stop is called by RuleEngine.
        Reconciliate all data before returning to RuleEngine.

        Arguments
        ==========
        __in <Class> self : class instance

        Return Values
        ==============
        N/A
        '''

        logger.info('Stopping ' + self.rule_name + ' resource monitor.')
        final_result = {}

        results = []
        for spinner in self.process_spinners:
            self.process_spinners[spinner].archive_spin_period()
            result = self.process_spinners[spinner].get_spinner_result()
            if result != None:
                results.append(result)
        if self.rule_top_spinners_to_report > 0:
            # Return top N spinners
            final_result['TopMemorySpinners'] = sorted(results, key=lambda k: k['AvgResidentMemoryKB'], reverse=True)[:self.rule_top_spinners_to_report]
        else:
            # Retun all spinners
            final_result['TopMemorySpinners'] = sorted(results, key=lambda k: k['AvgResidentMemoryKB'], reverse=True)

        results = []
        for spin_period in self.memory_spin_periods:
            result = spin_period.get_spin_period_result()
            if result is not None:
                results.append(result)
        final_result['HighMemoryUsagePeriods'] = results

        final_result['PhysicalMemoryKB'] = self.memory_monitor.total_physical_memory_kbytes
        final_result['PhysicalMemoryMB'] = '%.2f' % (float(self.memory_monitor.total_physical_memory_kbytes) / 1024)
        final_result['PhysicalMemoryGB'] = '%.2f' % (float(self.memory_monitor.total_physical_memory_kbytes) / (1024*1024))
        final_result['AvailablePhysicalMemoryKB'] = self.memory_monitor.available_physical_memory_kbytes
        final_result['AvailablePhysicalMemoryMB'] = '%.2f' % (float(self.memory_monitor.available_physical_memory_kbytes) / 1024)
        final_result['AvailablePhysicalMemoryGB'] = '%.2f' % (float(self.memory_monitor.available_physical_memory_kbytes) / (1024*1024))

        final_result['SwapMemoryKB'] = self.memory_monitor.total_swap_memory_kbytes
        final_result['SwapMemoryMB'] = '%.2f' % (float(self.memory_monitor.total_swap_memory_kbytes) / 1024)
        final_result['SwapMemoryGB'] = '%.2f' % (float(self.memory_monitor.total_swap_memory_kbytes) / (1024*1024))
        final_result['AvailableSwapMemoryKB'] = self.memory_monitor.available_swap_memory_kbytes
        final_result['AvailableSwapMemoryMB'] = '%.2f' % (float(self.memory_monitor.available_swap_memory_kbytes) / 1024)
        final_result['AvailableSwapMemoryGB'] = '%.2f' % (float(self.memory_monitor.available_swap_memory_kbytes) / (1024*1024))

        return final_result

    def is_active(self):
        '''
        Function
        =========
        is_active

        Description
        ==============
        Provides a way to check if resource monitor is active or not.
        Used by the caller from RuleEngine.

        Arguments
        ==========
        __in <Class> self : class instance

        Return Values
        ==============
        __out <bool>
        '''

        return self.monitor_timer.is_alive()


# ######################### Memory spin periods #########################


class MemoryMonitor():
    '''
    Class
    =========
    MemoryMonitor

    Description
    ==============
    Used to collect system overall memory usage
    and update statistics at Resource Monitor
    sampling interval.
    '''

    def __init__(self):
        '''
        Function
        =========
        __init__

        Description
        ==============
        Initialize an object which will host the memory usage
        on the system.

        Arguments
        ==========
        __in <Class> self : class instance

        Return Values
        ==============
        N/A
        '''

        self.update_time_ms = int(round(time.time() * 1000))
        self.total_physical_memory_kbytes = MemoryUtil.total_physical_memory_kbytes()
        self.available_physical_memory_percentage = 100 * float(MemoryUtil.available_physical_memory_kbytes()) / self.total_physical_memory_kbytes
        self.total_swap_memory_kbytes = MemoryUtil.total_swap_memory_kbytes()
        if self.total_swap_memory_kbytes > 0:
            self.available_swap_memory_percentage = 100 * float(MemoryUtil.available_swap_memory_kbytes()) / self.total_swap_memory_kbytes
        self.samples_count = 0

    def get_memory_stats(self):
        '''
        Function
        =========
        get_memory_stats

        Description
        ==============
        Get memory usage (physical, swap, ...) and define % usage
        that will be used to trigger spin periods.

        Arguments
        ==========
        __in <Class> self : class instance

        Return Values
        ==============
        N/A
        '''

        self.update_time_ms = int(round(time.time() * 1000))

        self.available_physical_memory_kbytes = MemoryUtil.available_physical_memory_kbytes()
        self.free_physical_memory_kbytes = MemoryUtil.free_physical_memory_kbytes()

        self.total_swap_memory_kbytes = MemoryUtil.total_swap_memory_kbytes()
        self.available_swap_memory_kbytes = MemoryUtil.available_swap_memory_kbytes()

        self.available_physical_memory_percentage = 100 * float(self.available_physical_memory_kbytes) / self.total_physical_memory_kbytes
        self.free_physical_memory_percentage = 100 * float(self.free_physical_memory_kbytes) / self.total_physical_memory_kbytes
        if self.total_swap_memory_kbytes > 0:
            self.available_swap_memory_percentage = 100 * float(self.available_swap_memory_kbytes) / self.total_swap_memory_kbytes
        else:
            self.available_swap_memory_percentage = 0.00
        self.samples_count += 1


class MemorySpinPeriod():
    '''
    Class
    =========
    MemorySpinPeriod

    Description
    ==============
    A spin period is a time frame during which the available system memory
    was under usage threshold.
    '''

    def __init__(self, memory_data):
        '''
        Function
        =========
        __init__

        Description
        ==============
        Initialize an object which will host statistics for memory
        usage during a spin period.

        Arguments
        ==========
        __in <Class> self : class instance
        __in <object> memory_data : system memory statistics

        Return Values
        ==============
        N/A
        '''

        self.start_time_ms = memory_data.update_time_ms
        self.end_time_ms = memory_data.update_time_ms
        self.cumulative_available_physical_memory_kbytes = memory_data.available_physical_memory_kbytes
        self.avg_available_physical_memory_kbytes = memory_data.available_physical_memory_kbytes
        self.available_physical_memory_percentage = 100 * float(self.cumulative_available_physical_memory_kbytes) / memory_data.total_physical_memory_kbytes

        self.cumulative_free_physical_memory_kbytes = memory_data.free_physical_memory_kbytes
        self.free_physical_memory_percentage = 100 * float(self.cumulative_free_physical_memory_kbytes) / memory_data.total_physical_memory_kbytes

        self.cumulative_available_swap_memory_kbytes = memory_data.available_swap_memory_kbytes
        self.avg_available_swap_memory_kbytes = memory_data.available_swap_memory_kbytes
        if memory_data.total_swap_memory_kbytes > 0:
            self.available_swap_memory_percentage = 100 * float(self.cumulative_available_swap_memory_kbytes) / memory_data.total_swap_memory_kbytes
        else:
            self.available_swap_memory_percentage = 0.00

        self.samples_count = 1

    def update_spin_period(self, memory_data):
        '''
        Function
        =========
        update_spin_period

        Description
        ==============
        Update memory spin period with latests memory usage statistics.

        Arguments
        ==========
        __in <Class> self : class instance
        __in <object> memory_data : process metadata and statistics

        Return Values
        ==============
        N/A
        '''

        self.end_time_ms = memory_data.update_time_ms
        self.cumulative_available_physical_memory_kbytes += memory_data.available_physical_memory_kbytes

        self.cumulative_free_physical_memory_kbytes += memory_data.free_physical_memory_kbytes

        self.cumulative_available_swap_memory_kbytes += memory_data.available_swap_memory_kbytes

        self.samples_count += 1

        self.avg_available_physical_memory_kbytes = self.cumulative_available_physical_memory_kbytes / self.samples_count
        self.available_physical_memory_percentage = 100 * float(self.avg_available_physical_memory_kbytes) / memory_data.total_physical_memory_kbytes

        self.avg_free_physical_memory_kbytes = self.cumulative_free_physical_memory_kbytes / self.samples_count
        self.free_physical_memory_percentage = 100 * float(self.avg_free_physical_memory_kbytes) / memory_data.total_physical_memory_kbytes

        self.avg_available_swap_memory_kbytes = self.cumulative_available_swap_memory_kbytes / self.samples_count
        if memory_data.total_swap_memory_kbytes > 0:
            self.available_swap_memory_percentage = 100 * float(self.avg_available_swap_memory_kbytes) / memory_data.total_swap_memory_kbytes

    def get_spin_period_result(self):
        '''
        Function
        =========
        get_spin_period_result

        Description
        ==============
        At the end of the capture, provides a dictionary with profiling data that will be returned to the Resource Monitor.
        It provides the overall profiling data and profiling data per spin period.

        Arguments
        ==========
        __in <Class> self : class instance

        Return Values
        ==============
        __out <dict> spinner_result: processors statistics for the capture time
        '''

        spinner_result = {}
        spinner_result['AvailablePhysicalMemoryPercentage'] = self.available_physical_memory_percentage
        spinner_result['AverageAvailableMemoryMB'] = self.avg_available_physical_memory_kbytes / 1024
        spinner_result['AvailableSwapMemoryPercentage'] = self.available_swap_memory_percentage
        spinner_result['AverageAvailableMemoryMB'] = self.avg_available_swap_memory_kbytes / 1024
        spinner_result['StartTime'] = datetime.datetime.utcfromtimestamp(self.start_time_ms / 1000).strftime('%Y-%m-%d %H:%M:%S.%f')[:-3] # Friendly spin period start time
        spinner_result['EndTime'] = datetime.datetime.utcfromtimestamp(self.end_time_ms / 1000).strftime('%Y-%m-%d %H:%M:%S.%f')[:-3] # Friendly spin period end time

        return spinner_result


# ######################### Process spinners #########################


class ProcessMonitor():
    '''
    Class
    =========
    ProcessMonitor

    Description
    ==============
    Used to reference all running processes on the system and update their statistics
    at Resource Monitor sampling interval.
    '''

    def __init__(self, total_physical_memory_kbytes):
        '''
        Function
        =========
        __init__

        Description
        ==============
        Initialize an object which will host the collection of running
        processes on the system.

        Arguments
        ==========
        __in <Class> self : class instance

        Return Values
        ==============
        N/A
        '''

        self.total_physical_memory_kbytes = total_physical_memory_kbytes
        self.processes = {}
        self.samples_count = 0

    def get_processes_stats(self):
        '''
        Function
        =========
        get_processes_stats

        Description
        ==============
        Get all active pids in /proc and collect latest statistics from them.
        This method adds to the dictionnary self.processes new processes.
        Otherwise, it will update the corresponding dictionnary item with its
        last statistics.

        Arguments
        ==========
        __in <Class> self : class instance

        Return Values
        ==============
        N/A
        '''

        for pid in [int(x) for x in os.listdir(b'/proc') if x.isdigit()]:
            process_name = ProcessUtil.get_process_name(pid)
            if process_name is not None and (str(pid) + '-' + process_name) in self.processes:
                # Process already referenced
                if self.processes[str(pid) + '-' + process_name].is_alive:
                    # Process still alive, update process statistics
                    self.processes[str(pid) + '-' + process_name].update_process(self.total_physical_memory_kbytes)
            else:
                # Process not yet referenced, add new process with statistics to dictionnary
                new_process = Process(pid, self.total_physical_memory_kbytes)
                if new_process.is_alive:
                    self.processes[new_process.unique_id] = new_process

        self.samples_count += 1

class Process():
    '''
    Class
    =========
    Process

    Description
    ==============
    Used to hold process information and statistics.
    '''

    def __init__(self , pid, total_physical_memory_kbytes):
        '''
        Function
        =========
        __init__

        Description
        ==============
        Initialize an object which will host a process instance information
        and statistics.
        Values are set to 0 as we want to collect statistics for the
        collection duration_seconds and not the process whole life.

        Note: we collect total physical memory at each interval
        to ensure we take in account memory addition or removal.

        Arguments
        ==========
        __in <Class> self : class instance
        __in <int> pid : process PID

        Return Values
        ==============
        N/A
        '''

        # TODO: Note for improvment:
        # if (self.process_name = ProcessUtil.get_process_name(pid)) is None to verify process still alive
        self.process_name = ProcessUtil.get_process_name(pid)
        if self.process_name is not None:
            self.unique_id = str(pid) + '-' + self.process_name
            self.pid = pid
            self.process_status = ProcessUtil.get_process_status(pid)       # Get process status, used for debug purpose at the moment
            self.command_line = ProcessUtil.get_process_command_line(pid)

            # self.process_start_time_ms = ProcessUtil.get_process_start_time_ms(self.pid)  # Get the process start time in seconds since epoch
            # Set the capture start time (self.start_time_ms), sample update time (self.update_time_ms)
            # and previous sample update time (self.previous_update_time_ms) to current time in seconds since epoch
            # At the end of the capture, self.update_time_ms will represent the end time of the capture
            self.start_time_ms = self.update_time_ms = self.previous_update_time_ms = int(round(time.time() * 1000))
            self.current_total_physical_memory_kbytes = total_physical_memory_kbytes

            self.current_process_rss_memory_usage_kbytes = ProcessUtil.get_process_resident_memory_kbytes(self.pid)
            self.current_process_rss_memory_percentage_usage = 0.0 if self.current_total_physical_memory_kbytes == 0 else (100 * float(self.current_process_rss_memory_usage_kbytes) / self.current_total_physical_memory_kbytes)

            self.current_process_swap_usage_kbytes = ProcessUtil.get_process_swap_memory_kbytes(self.pid)
            
            self.samples_count = 1
            self.is_alive = True
        else:
            # Process does not exist anymore, self.current_process_rss_memory_usage should be set to None
            # ResourceMonitor.performance_data_update_handler() will not consider this process
            self.current_process_rss_memory_usage_kbytes = None
            self.is_alive = False

    # TODO: refine the following to get rid of
    # process[1] syntax in performance_data_update_handler
    def __repr__(self):
        '''
       	Function
        =========
        __repr__

        Description
        ==============
        Redefines how the object is represented when
        basically printed.

        Arguments
        ==========
        __in <Class> self : class instance
        
        Return Values
        ==============
        N/A
        '''

        return '{}: {} {}'.format(self.__class__.__name__,
            self.unique_id,
            self.current_total_physical_memory_kbytes)

    def update_process(self, total_physical_memory_kbytes):
        '''
        Function
        =========
        update_process

        Description
        ==============
        Update individual process statistics.
        Values used by PIL are:
            * self.current_process_rss_memory_usage_kbytes:
                provides the process current resident memory
            * self.current_process_swap_usage_kbytes:
                provides the process current swap memory
        
        self.current_process_rss_memory_percentage_usage is used
        to verify threshold

        Note: we collect total physical memory at each interval
        to ensure we take in account memory addition or removal.

        Arguments
        ==========
        __in <Class> self : class instance
        
        Return Values
        ==============
        N/A
        '''

        if ProcessUtil.is_alive(self.pid):
            self.update_time_ms = int(round(time.time() * 1000))
            self.current_total_physical_memory_kbytes = total_physical_memory_kbytes

            self.current_process_rss_memory_usage_kbytes = ProcessUtil.get_process_resident_memory_kbytes(self.pid)
            if self.current_process_rss_memory_usage_kbytes is not None:
                pass
            self.current_process_rss_memory_percentage_usage = 0.0 if self.current_total_physical_memory_kbytes == 0 else (100 * float(self.current_process_rss_memory_usage_kbytes) / self.current_total_physical_memory_kbytes)
            self.current_process_swap_usage_kbytes = ProcessUtil.get_process_swap_memory_kbytes(self.pid)
            if self.current_process_swap_usage_kbytes is not None:
                pass

            self.previous_update_time_ms = self.update_time_ms
            self.samples_count += 1
        else:
            self.is_alive = False

class ProcessSpinner():
    '''
    Class
    =========
    ProcessSpinner

    Description
    ==============
    A process spinner is an instance of a process which has had a
    memory usage higher than PerfInsights defined threshold
    at least once during the profiling.
    A process spinner can have one or more memory spin periods
    during which the rss memory usage average was higher
    than PerfInsights defined threshold.
    A spin period is active while process memory usage is above
    the threshold or while the process average usage since profiling
    start is below the threshold but has not reached the defined
    PerfInsights timeout.
    A spin period is archived when process average memry usage is
    below the threshold and timeout has been reached.

    '''

    def __init__(self, process_data):
        '''
       	Function
        =========
        __init__

        Description
        ==============
        Initialize an object which will host an instance of a process spinner.

        Arguments
        ==========
        __in <Class> self : class instance
        __in <object> process_data : process metadata and statistics
        
        Return Values
        ==============
        N/A
        '''

        self.process_id = process_data.pid
        self.process_name = process_data.process_name
        self.unique_id = process_data.unique_id
        self.command_line = process_data.command_line
        self.active_process_spin_period = ProcessSpinPeriod(process_data)   # Active spin period 
        self.process_spin_periods = []                                      # Archived spin periods
    
    def new_spin_period(self , process_data):
        '''
        Function
        =========
        new_spin_period
    
        Description
        ==============
        A new spin period has been detected so the active spin period will be used to store data.
        Initialized with process collection last update time (previous_update_time_ms).

        Arguments
        ==========
            __in <Class> self : class instance
            __in <object> process_data : process metadata and statistics
    
        Return Values
        ==============
            N/A
        '''

        self.active_process_spin_period = ProcessSpinPeriod(process_data)

    def archive_spin_period(self):
        '''
       	Function
        =========
        archive_spin_period

        Description
        ==============
        Archive spin period when the process average memory usage is below the threshold or not active for more than the defined timeout.
        The active spin period is set to None.

        Arguments
        ==========
        __in <Class> self : class instance
        
        Return Values
        ==============
        N/A
        '''

        if self.active_process_spin_period is not None:
            # Active spin period is active so we can archive it
            if self.active_process_spin_period.start_time_ms < self.active_process_spin_period.end_time_ms:
                # Active spin period has valid samples as end time is greater than start time
                self.process_spin_periods.append(self.active_process_spin_period)

        self.active_process_spin_period = None
    
    def get_spinner_result(self):
        '''
       	Function
        =========
        get_spinner_result

        Description
        ==============
        At the end of the capture, provides a dictionary with profiling data that will be returned to the Resource Monitor.
        It provides the overall profiling data and profiling data per spin period.

        Arguments
        ==========
        __in <Class> self : class instance
        
        Return Values
        ==============
        __out <dict> spinner_result: spinner results
        '''

        spinner_result = {}
        spinner_result['ProcessName'] = self.process_name
        spinner_result['ProcessId'] = self.process_id
        spinner_result['CommandLine'] = self.command_line

        spinner_result['MemorySpinPeriods'] = []
        avg_resident_mem = 0
        avg_swap_mem = 0
        spinner_result['PeakResidentMemoryKB'] = 0
        spinner_result['PeakSwapMemoryKB'] = 0

        spinner_result['AvgResidentMemoryKB'] = 0
        spinner_result['AvgSwapMemoryKB'] = 0
        if len(self.process_spin_periods) > 0:
            for process_spin_period in self.process_spin_periods:
                avg_resident_mem += process_spin_period.avg_rss_memory_usage_kbytes
                avg_swap_mem += process_spin_period.avg_swap_usage_kbytes
                start_time = datetime.datetime.utcfromtimestamp(process_spin_period.start_time_ms / 1000).strftime('%Y-%m-%d %H:%M:%S.%f')[:-3] # Friendly spin period start time
                end_time = datetime.datetime.utcfromtimestamp(process_spin_period.end_time_ms / 1000).strftime('%Y-%m-%d %H:%M:%S.%f')[:-3] # Friendly spin period end time
                spinner_result['MemorySpinPeriods'].append({'AvgResidentMemoryKB': process_spin_period.avg_rss_memory_usage_kbytes, 'PeakResidentMemoryKB': process_spin_period.peak_rss_memory_usage_kbytes, 'AvgSwapMemoryKB': process_spin_period.avg_swap_usage_kbytes, 'PeakSwapMemoryKB': process_spin_period.peak_swap_usage_kbytes, 'CountOfDetectWindows': process_spin_period.samples_count,'StartTime': start_time, 'EndTime': end_time})
                spinner_result['PeakResidentMemoryKB'] = max(process_spin_period.peak_rss_memory_usage_kbytes, spinner_result['PeakResidentMemoryKB'])
                spinner_result['PeakSwapMemoryKB'] = max(process_spin_period.peak_swap_usage_kbytes, spinner_result['PeakSwapMemoryKB'])
            spinner_result['AvgResidentMemoryKB'] = avg_resident_mem / len(spinner_result['MemorySpinPeriods'])
            spinner_result['AvgSwapMemoryKB'] = avg_swap_mem / len(spinner_result['MemorySpinPeriods'])
        
        return spinner_result


class ProcessSpinPeriod():
    '''
    Class
    =========
    ProcessSpinPeriod

    Description
    ==============
    A spin period is a time frame during which a process was above the memory usage threshold.
    '''

    def __init__(self, process_data):
        '''
       	Function
        =========
        __init__

        Description
        ==============
        Initialize an object which will host statistics for a process spin period.

        Arguments
        ==========
        __in <Class> self : class instance
        __in <object> process_data : process metadata and statistics
        
        Return Values
        ==============
        N/A
        '''
          
        self.start_time_ms = process_data.update_time_ms
        self.end_time_ms = process_data.update_time_ms
        self.cumulative_total_physical_memory_kbytes = process_data.current_process_rss_memory_usage_kbytes

        self.cumulative_process_rss_memory_usage_kbytes = process_data.current_process_rss_memory_usage_kbytes
        self.peak_rss_memory_usage_kbytes = process_data.current_process_rss_memory_usage_kbytes
        self.avg_rss_memory_usage_kbytes = process_data.current_process_rss_memory_usage_kbytes
        self.cumulative_rss_memory_percentage_usage = process_data.current_process_rss_memory_percentage_usage
        self.avg_rss_memory_percentage_usage = process_data.current_process_rss_memory_percentage_usage

        self.cumulative_process_swap_usage_kbytes = process_data.current_process_swap_usage_kbytes
        self.avg_swap_usage_kbytes = process_data.current_process_swap_usage_kbytes
        self.peak_swap_usage_kbytes = process_data.current_process_swap_usage_kbytes
        
        self.samples_count = 1
    
    def update_spin_period(self, process_data):
        '''
       	Function
        =========
        update_spin_period

        Description
        ==============
        Update process spin period with latests process statistics.

        Arguments
        ==========
        __in <Class> self : class instance
        __in <object> process_data : process metadata and statistics
        
        Return Values
        ==============
        N/A
        '''
        
        self.end_time_ms = process_data.update_time_ms
        self.cumulative_total_physical_memory_kbytes += process_data.current_process_rss_memory_usage_kbytes

        self.cumulative_process_rss_memory_usage_kbytes += process_data.current_process_rss_memory_usage_kbytes
        if self.peak_rss_memory_usage_kbytes < process_data.current_process_rss_memory_usage_kbytes:
            self.peak_rss_memory_usage_kbytes = process_data.current_process_rss_memory_usage_kbytes
        self.cumulative_rss_memory_percentage_usage += process_data.current_process_rss_memory_percentage_usage

        self.cumulative_process_swap_usage_kbytes += process_data.current_process_swap_usage_kbytes
        
        if self.peak_swap_usage_kbytes < process_data.current_process_swap_usage_kbytes:
            self.peak_swap_usage_kbytes = process_data.current_process_swap_usage_kbytes

        self.samples_count += 1
        self.avg_rss_memory_usage_kbytes = self.cumulative_process_rss_memory_usage_kbytes / self.samples_count
        self.avg_swap_usage_kbytes = self.cumulative_process_swap_usage_kbytes / self.samples_count
        self.avg_rss_memory_percentage_usage = 100 * float(self.avg_rss_memory_usage_kbytes) / (self.cumulative_total_physical_memory_kbytes / self.samples_count)

