# -------------------------------------------------------------------------
# Copyright (c) Microsoft.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

import csv
import logging
import os
import re
import sys
import time
import six

from util import common
from util.SystemUtil import DiskUtil
from ruleengine.RuleEngineCls import IResourceMonitor, MonitorTimer


logger = logging.getLogger(__name__)

if six.PY3:
    long = int


class ResourceMonitor(IResourceMonitor):
    '''
    Class
    =========
    ResourceMonitor

    Description
    ==============
    Class used to manage the whole data collection.
    Starts the data collection asynchronously by running a new MonitorTimer.
    '''

    def __init__(self, duration_seconds, interval_seconds, rule_name, rule_definition):
        '''
        Function
        =========
        __init__

        Description
        ==============
        Initialize the ResourceMonitor class that will handle resource
        profiling.

        Arguments
        ==========
        __in <Class> self : class instance
        __in <int> duration_seconds: capture duration in seconds
        __in <int> interval_seconds: sampling interval in seconds
        __in <string> rule_name: this rule short name
        __in <object> rule_definition: this rule configuration

        Return Values
        ==============
        N/A
        '''
        self.duration_seconds = duration_seconds
        self.interval_seconds = interval_seconds
        self.rule_name = rule_name
        self.rule_definition = rule_definition
        self.rule_config = self.rule_definition.get('Configuration')

        self.disk_monitor = DiskMonitor()
        self.monitor_timer = None
        self.data_collection_start = None

        self.disk_usage_data = {}

        self.data_to_persist_count = 0
        self.disk_count = 0
        self.historical_data_files = []

    # TODO: Consider using __enter__ instead
    @staticmethod
    def init(duration_seconds, interval_seconds, rule_name, rule_config):
        '''
        Function
        =========
        init

        Description
        ==============
        Used by RuleEngine to instantiate this ResourceMonitor instance
        as __init__ is not able to return it.
        We might consider using __enter__ instead.

        Arguments
        ==========
        __in <int> duration_seconds: capture duration_seconds in seconds
        __in <int> interval_seconds: sampling interval_seconds in seconds
        __in <string> rule_name: this rule short name
        __in <object> rule_definition: this rule configuration

        Return Values
        ==============
        __out <ResourceMonitor> this class instance
        '''

        return ResourceMonitor(duration_seconds, interval_seconds, rule_name, rule_config)

    def performance_data_update_handler(self):
        devices_performance_data = self.disk_monitor.get_disk_stats()

        for disk_data in devices_performance_data.values():
            disk_name = disk_data.disk_name
            if disk_name in self.disk_usage_data:
                self.disk_usage_data[disk_name].add_data_point(disk_data)
                self.data_to_persist_count += 1
            else:
                # Only add the disk if there is activity
                if disk_data.has_io_activity():
                    self.disk_usage_data[disk_name] = DiskUsage(disk_name, self.disk_count)
                    self.disk_usage_data[disk_name].add_data_point(disk_data)
                    self.disk_count += 1
                    self.data_to_persist_count += 1



        if self.data_to_persist_count > self.rule_config['MaxDataPoints']:
            # create csv with data from backed up data_to_persist
            self.persist_data_to_csv(self.disk_usage_data)
            for d in self.disk_usage_data.values():
                # reset disk_usage_data.data_points for each disk
                d.data_points = []
            self.data_to_persist_count = 0

    def persist_data_to_csv(self, data):
        header = ['DiskName', 'DiskNumber', 'TM', 'QL', 'IC', 'KB', 'US50', 'US99']
        file_name = 'HighResolutionDiskRule.' + str(common.get_current_datetime('file')) + '.csv'
        file_path = os.path.join(common.diag_output_folder_path, file_name)
        data_count = 0 

        min_relative_timestamp = sys.maxsize
        max_relative_timestamp = 0
        with open(file_path, 'w') as file:
                writer = csv.writer(file)
                writer.writerow(header)
                for disk_data in data.values():
                    for data_point in disk_data.data_points:
                        if data_point['TM'] < min_relative_timestamp:
                            min_relative_timestamp = data_point['TM'] 

                        disk_data_for_csv = [disk_data.disk_name, disk_data.disk_number, data_point['TM'], data_point['QL'], data_point['IC'], data_point['KB'], data_point['US50'], data_point['US99']]

                        writer.writerow(disk_data_for_csv)
                        data_count += 1

                        if data_point['TM'] > max_relative_timestamp:
                            max_relative_timestamp = data_point['TM']
        
        min_data_time = self.disk_monitor.start_time_seconds + (min_relative_timestamp/1000)
        max_data_time = self.disk_monitor.start_time_seconds + (max_relative_timestamp/1000)

        file = {}
        file['DataFilePath'] = file_name
        file['MinDataTime'] = time.strftime('%Y-%m-%d %H:%M:%S', time.gmtime(min_data_time))
        file['MaxDataTime'] = time.strftime('%Y-%m-%d %H:%M:%S', time.gmtime(max_data_time))
        file['DataCount'] = data_count
        self.historical_data_files.append(file)

    def is_active(self):
        return self.monitor_timer.is_alive()

    def start(self):
        logger.info('Starting ' + self.rule_name + ' resource monitor (' + str(self.duration_seconds - self.interval_seconds) + ' seconds).')

        self.monitor_timer = MonitorTimer(self.duration_seconds, self.interval_seconds, self.performance_data_update_handler)
        self.monitor_timer.start()

        self.data_collection_start = self.monitor_timer.start_timer()

    def stop(self):
        '''
        Function
        =========
        stop

        Description
        ==============
        Resource Monitor stop is called by RuleEngine.
        Reconciliate all data before returning to RuleEngine.

        Arguments
        ==========
        __in <Class> self : class instance

        '''

        logger.info('Stopping ' + self.rule_name + ' resource monitor.')

        result_dict = {}
        perf_data_per_disk = []
        
        for data in self.disk_usage_data.values():
            disk_usage_dict = {}
            disk_usage_dict['DiskName'] = data.disk_name
            disk_usage_dict['DiskNumber'] = data.disk_number
            disk_usage_dict['DataPoints'] = data.data_points
            perf_data_per_disk.append(disk_usage_dict)

        # If data is already persisted to CSVs, remaining data points should be both saved to csv and written to the perfdiag object
        if len(self.historical_data_files) > 0:
            self.persist_data_to_csv(self.disk_usage_data.copy())

        result_dict['PerfDataPerDisk'] = perf_data_per_disk
        result_dict['HistoricalDataFiles'] = self.historical_data_files

        return result_dict

class DiskMonitor():
    '''
    Class
    =========
    DiskMonitor

    Description
    ==============
    Used to reference all disks on the system and update their
    statistics at Resource Monitor sampling interval_seconds.
    '''

    # Block device types from which we don't want to collect metrics
    EXCLUDED_DEVICES = ['fd', 'sr', 'loop', 'ram']
    # Block device patterns from which we actually want to collect metrics
    VALID_DEVICES_REGEX_PATTERN = ['sd[a-z]*\\b', 'md*', 'dm-*']
    BYTES_TO_KB_CONVERT_RATIO = 0.000976563

    def __init__(self):
        self.start_time_seconds = self.update_time_seconds = self.previous_update_time_seconds = time.time()
        self.start_time_ms = int(round(self.start_time_seconds * 1000))
        self.disks = {}

    def get_disk_stats(self):
        relative_time_stamp = int(round(time.time() * 1000)) - self.start_time_ms

        disk_stats = self.get_filtered_diskstats()

        for stat_line in disk_stats:
            current_disk_name = stat_line.split()[2]
            if not self.disks or current_disk_name not in self.disks:
                self.disks[current_disk_name] = Disk(stat_line.split())
            else:
                self.disks[current_disk_name].update_disk(stat_line.split(), relative_time_stamp)

        return self.disks

    def get_filtered_diskstats(self):
        results = []
        diskstats_file = open('/proc/diskstats', 'r').readlines()
        for device_stats in diskstats_file:
            for regex_pattern in self.VALID_DEVICES_REGEX_PATTERN:
                if re.compile(regex_pattern).match(device_stats.split()[2]) and \
                    (device_stats.split()[2][:-1] not in self.EXCLUDED_DEVICES) and \
                        (device_stats.split()[2][:-2] not in self.EXCLUDED_DEVICES) and \
                            (device_stats.split()[2] not in results):
                    # Block device is valid, i.e. it's a regular block device
                    # or a RAID device
                    results.append(device_stats)

        return results

class Disk():
    def __init__(self, disk_stats):
        '''
        Function
        =========
        __init__

        Description
        ==============
        Initialize the Disk class that will contain single device
        information and statistics.

        Arguments
        ==========
        
        __in <Class> self : class instance
        __in <list> disk_stats: raw diskstats lines for each targeted disk

        Return Values
        ==============
        N/A
        '''

        self.disk_name = disk_stats[2]
        self.sector_size_bytes = DiskUtil.get_block_device_sector_size_bytes(self.disk_name)

        #Initialize data collection points
        self.queue_length = 0
        self.delta_read_count = self.delta_write_count = self. delta_total_iops = 0 
        self.delta_read_bytes = self.delta_write_bytes = 0
        self.delta_write_sectors = self.delta_read_sectors = 0
        self.delta_read_kb = self.delta_write_kb = 0

        #Set current values for later calculations
        self.previous_read_count = self.reads_at_start = long(disk_stats[3])
        self.previous_write_count = self.writes_at_start = long(disk_stats[7])
        self.previous_read_sectors = long(disk_stats[5])
        self.previous_write_sectors = long(disk_stats[9])

    def update_disk(self, disk_stats, relative_time_stamp):
        '''
        Function
        =========
        update_disk

        Description
        ==============
        Update block device statistics.

        Arguments
        ==========
        __in <Class> self : class instance
        __in <list> disk_stats: raw diskstats lines for each targeted device
        __in <float> relative_time_stamp: delta time since start time in seconds

        Return Values
        ==============
        N/A
        '''

        #Set new values
        self.relative_time_stamp = relative_time_stamp
        self.queue_length = int(disk_stats[11]) 
        self.delta_read_count = long(disk_stats[3]) - self.previous_read_count
        self.delta_write_count = long(disk_stats[7]) - self.previous_write_count
        self.delta_read_sectors = long(disk_stats[5]) - self.previous_read_sectors
        self.delta_write_sectors = long(disk_stats[9]) - self.previous_write_sectors

        #Calculations
        self.delta_total_iops = self.delta_write_count + self.delta_read_count
        self.delta_read_bytes = long(self.delta_read_sectors * self.sector_size_bytes)
        self.delta_write_bytes = long(self.delta_write_sectors * self.sector_size_bytes)
        self.delta_read_kb = self.delta_read_bytes / 1024
        self.delta_write_kb = self.delta_write_bytes / 1024

        #Save values for next sample calculation
        self.previous_read_count = long(disk_stats[3])
        self.previous_write_count = long(disk_stats[7])
        self.previous_read_sectors = int(disk_stats[5])
        self.previous_write_sectors = int(disk_stats[9])

    def has_io_activity(self):
        return self.delta_read_count > 0 or self.delta_write_count > 0

class DiskUsage():

    def __init__(self, disk_name, disk_number):
        self.disk_name = disk_name
        self.disk_number = disk_number
        self.data_points = []

    def add_data_point(self, performance_data):
        data_point = {
            "QL": performance_data.queue_length,
            "IC": performance_data.delta_total_iops,
            "KB": performance_data.delta_read_kb + performance_data.delta_write_kb,
            "US50": 0,
            "US99": 0,
            "TM": performance_data.relative_time_stamp
        }
        self.data_points.append(data_point)
