#-------------------------------------------------------------------------
# Copyright (c) Microsoft.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#--------------------------------------------------------------------------
import time
import six

from ruleengine import DataCollector, AnalyticRule, StaticRule, HPC
import logging

logger = logging.getLogger(__name__)

from util import common
from ruleengine.RuleEngineCls import IRuleEngine, RuleUtil


class RuleEngine(IRuleEngine):

    def __init__(self, static_rules, analytic_rules, data_collectors, data_monitors):
        self._static_rules = RuleEngine._init_dict(static_rules)
        self._analytic_rules = RuleEngine._init_dict(analytic_rules)
        self._data_collectors = RuleEngine._init_dict(data_collectors)
        self._data_monitors = RuleEngine._init_dict(data_monitors)
        self._dc_results = []
        self._rule_results = []
        self._resMonitors = []
        self._rule_configuration = {}

    def start(self, config):
        # load rules configuration
        for rule_config in config['RuleConfigurations']:
            if rule_config['Enabled'] == True:
                rule_type = rule_config['Type']
                rule_cls = RuleUtil.get_class(rule_type)
                if issubclass(rule_cls, StaticRule.IStaticRule):
                    if rule_type in self._static_rules:
                        self._static_rules[rule_type].init(rule_config)
                elif issubclass(rule_cls, AnalyticRule.IAnalyticRule):
                    if rule_type in self._analytic_rules:
                        self._analytic_rules[rule_type].init(rule_config)
                        resMonitor = self._analytic_rules[rule_type].on_start()
                        self._resMonitors.append(resMonitor)
                else:
                    raise RuntimeError('Not the right rule type.')

        # load data collectors configuration
        for dc_config in config['DataCollectorConfigurations']:
            dc_type = dc_config['Type']
            dc_cls = RuleUtil.get_class(dc_type)
            if issubclass(dc_cls, DataCollector.IDataMonitor):
                self._data_monitors[dc_type].init(dc_config)
                self._data_monitors[dc_type].start()
            elif issubclass(dc_cls, DataCollector.IDataCollector):
                self._data_collectors[dc_type].init(dc_config)
            else:
                raise RuntimeError('Not the right configuration type.')

        # set configuration json
        config = {}
        # TODO: Replace this with per-rule capture interval
        config['ResourceMonitorIntervalMS'] = common.DEFAULT_CAPTURE_INTERVAL_SEC * 1000
        self._rule_configuration = config

        # Start resource monitors
        for resMon in self._resMonitors:
            resMon.start()
            # resMon._timer.join()

    def stop(self):

        for dc in six.itervalues(self._data_collectors):
            result = dc.to_json()
            self._dc_results.append(result)

        # Stop Data Monitors and get results from them
        for dm in six.itervalues(self._data_monitors):
            result = dm.stop()
            self._rule_results.append(result)

        # Run Static Rules on collected static information and get findings from them
        for sr in six.itervalues(self._static_rules):
            result = sr.run()
            self._rule_results.append(result)

        # Wait for Analytic Rules resource monitors and get findings from them
        resMonCount = len(self._resMonitors)
        while resMonCount > 0:
            for resMon in self._resMonitors:
                if resMon.is_active():
                    #resMon._timer.join()
                    time.sleep(1)
                else:
                    resMonCount = resMonCount - 1
        
        for ar in six.itervalues(self._analytic_rules):
            result = ar.on_stop()
            self._rule_results.append(result)
        
    def get_rule_results(self):
        return self._rule_results
    
    def get_rule_configuration(self):
        return self._rule_configuration

    def get_dc_results(self):
        return self._dc_results

    @staticmethod
    def __get_full_name(obj):
        return obj.__module__ + '.' + obj.__class__.__name__

    @staticmethod
    def _init_dict(objs):
        obj_dict = {}
        if objs == None:
            return obj_dict
        
        for obj in objs:
            obj_dict[RuleEngine.__get_full_name(obj)] = obj
        return obj_dict

    @staticmethod
    def build_for_quick_analysis():
        # Build RuleEngine with all static and analytic rules as well as data collectors

        instance_metadata_map_dc = DataCollector.InstanceMetadataInfoDC()   # Initiate Azure instance metadata collector
        oslog_dm = DataCollector.GuestOSLogDM()                             # Initiate OS logs monitor
        storage_map_dc = DataCollector.VolumeMapInfoCollector()             # Initiate disk configuration collector
        machine_map_dc = DataCollector.MachineMapInfoDC()                   # Initiate machine configuration collector
        network_map_dc = DataCollector.NetworkMapInfoDC()                   # Initiate network configuration collector
        tasklist_map_dc = DataCollector.RunningProcessesDC()                # Initiate running processes collector
        memory_usage_collector = DataCollector.MemoryUsageCollector()       # Initiate memory usage collector   
        
        # Initiate data collector list
        dcs = [instance_metadata_map_dc, machine_map_dc, network_map_dc, storage_map_dc, tasklist_map_dc, memory_usage_collector]    
        # Initiate datamonitor list
        dms = [oslog_dm]
        # Initiate analytic rules list
        ars = [AnalyticRule.HighCpuRuleConfiguration(), AnalyticRule.HighDiskRuleConfiguration(), AnalyticRule.HighMemoryRuleConfiguration(), AnalyticRule.HighResolutionDiskRuleConfiguration()]
        # Initiate static rules list
        srs = [StaticRule.InstanceMetadataRule(instance_metadata_map_dc), StaticRule.DiskConfigurationRule(storage_map_dc), StaticRule.MachineConfigurationRule(machine_map_dc), StaticRule.NetworkConfigurationRule(machine_map_dc, instance_metadata_map_dc), StaticRule.OOMScanRule(), StaticRule.SwapFileDiskRule(storage_map_dc, memory_usage_collector)]
        # Build rule engine
        rule_engine = RuleEngine(srs, ars, dcs, dms)

        return rule_engine
    
    @staticmethod
    def build_for_hpc_scenario():

        instance_metadata_map_dc = DataCollector.InstanceMetadataInfoDC()   # Initiate Azure instance metadata collector
        oslog_dm = DataCollector.GuestOSLogDM()                             # Initiate OS logs monitor
        storage_map_dc = DataCollector.VolumeMapInfoCollector()             # Initiate disk configuration collector
        machine_map_dc = DataCollector.MachineMapInfoDC()                   # Initiate machine configuration collector
        network_map_dc = DataCollector.NetworkMapInfoDC()                   # Initiate network configuration collector
        tasklist_map_dc = DataCollector.RunningProcessesDC()                # Initiate running processes collector
        memory_usage_collector = DataCollector.MemoryUsageCollector()       # Initiate memory usage collector   
        hpcdiag_dc = HPC.HPCDiagWrapper()

        data_collectors = [
            hpcdiag_dc,
            instance_metadata_map_dc,
            machine_map_dc,
            network_map_dc,
            storage_map_dc,
            tasklist_map_dc,
            memory_usage_collector
        ]
        data_monitors = [oslog_dm]
        analytic_rules = [
            AnalyticRule.HighCpuRuleConfiguration(),
            AnalyticRule.HighDiskRuleConfiguration(),
            AnalyticRule.HighMemoryRuleConfiguration(),
            AnalyticRule.HighResolutionDiskRuleConfiguration()
            ]
        static_rules = [
            StaticRule.InstanceMetadataRule(instance_metadata_map_dc),
            StaticRule.DiskConfigurationRule(storage_map_dc),
            StaticRule.MachineConfigurationRule(machine_map_dc),
            StaticRule.NetworkConfigurationRule(machine_map_dc, instance_metadata_map_dc, hpcdiag_dc),
            StaticRule.OOMScanRule(),
            StaticRule.SwapFileDiskRule(storage_map_dc, memory_usage_collector),
            HPC.MemoryRule(hpcdiag_dc),
            HPC.CPURule(hpcdiag_dc, machine_map_dc),
            HPC.GPURule(hpcdiag_dc, machine_map_dc)
            ]

        rule_engine = RuleEngine(static_rules, analytic_rules, data_collectors, data_monitors)

        return rule_engine

    @staticmethod
    def build_for_vmslow_analysis():
        # Build RuleEngine with all static and analytic rules as well as data collectors
        # As currently there is no difference between quick and vmslow scenarios build for quick analysis
         return RuleEngine.build_for_quick_analysis()
