#-------------------------------------------------------------------------
# Copyright (c) Microsoft.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#--------------------------------------------------------------------------
import abc
import threading
import time
import logging
import six
import sys
import util.common as common

logger = logging.getLogger()


# Abstract classes
@six.add_metaclass(abc.ABCMeta)
class IStaticRule:

    @abc.abstractmethod
    def init(self, rule_config):
        pass

    @abc.abstractproperty
    def id(self):
        pass

    @abc.abstractproperty
    def name(self):
        pass
        
    @abc.abstractmethod
    def run(self):
        pass


@six.add_metaclass(abc.ABCMeta)
class IRuleEngine:

    @abc.abstractmethod
    def start(self, config):
        pass

    @abc.abstractmethod
    def stop(self):
        pass

    @abc.abstractmethod
    def get_rule_results(self):
        pass
    
    @abc.abstractmethod
    def get_dc_results(self):
        pass

    @abc.abstractmethod
    def get_rule_configuration(self):
        pass


@six.add_metaclass(abc.ABCMeta)
class IDataCollector:

    @abc.abstractmethod
    def init(self):
        pass

    @abc.abstractproperty
    def index(self):
        pass

    @abc.abstractproperty
    def name(self):
        pass

    @abc.abstractmethod
    def friendly_name(self):
        pass

    @abc.abstractmethod
    def to_json(self):
        pass


@six.add_metaclass(abc.ABCMeta)
class IDataMonitor:

    @abc.abstractmethod
    def init(self):
        pass
        
    @abc.abstractproperty
    def name(self):
        pass

    @abc.abstractmethod
    def friendly_name(self):
        pass

    @abc.abstractmethod
    def start(self):
        pass
    
    @abc.abstractmethod
    def stop(self):
        pass


@six.add_metaclass(abc.ABCMeta)
class IAnalyticRule:

    @abc.abstractmethod
    def init(self, rule_config):
        pass

    @abc.abstractproperty
    def id(self):
        pass

    @abc.abstractproperty
    def name(self):
        pass

    @abc.abstractmethod
    def on_start(self):
        pass

    @abc.abstractmethod
    def on_stop(self):
        pass

    @abc.abstractmethod
    def get_results(self):
        pass


@six.add_metaclass(abc.ABCMeta)
class IResourceMonitor(object):
    
    @abc.abstractmethod
    def init(self, timeout, interval, topic):
        pass
    
    @abc.abstractmethod
    def performance_data_update_handler(self):
        pass

    @abc.abstractmethod
    def start(self):
        pass

    @abc.abstractmethod
    def stop(self):
        pass


@six.add_metaclass(abc.ABCMeta)
class ISpinner:
    # In progress
      
    @abc.abstractmethod
    def new_spin_period(self, performance_data):
        pass

    @abc.abstractmethod
    def archive_spin_period(self):
        pass
    
    @abc.abstractmethod
    def get_spinner_result(self):
        pass


@six.add_metaclass(abc.ABCMeta)
class ISpinPeriod:
    # In progress
      
    @abc.abstractmethod
    def update_spin_period(self, performance_data):
        pass

    @abc.abstractmethod
    def get_spin_period_result(self):
        pass


# Common classes
class MonitorTimer(threading.Thread):
    """
    Class
    =========
    MonitorTimer

    Description
    ==============
    Used to perform asynchronous tasks and respond to events.

    """

    def __init__(self, duration, interval, callbackFunction=None, *args):
        threading.Thread.__init__(self)

        self._duration = duration                           # Total duration of the thread in seconds
        self._interval = float(interval)                    # Interval at which tasks will be performed (via a call to a callback function), in seconds (will be in ms in the future)
        self.count = int(self._duration/self._interval)     # Calculate number of iterations
        self._callbackFunction = callbackFunction           # Callback function that actually collects and processes the data 
        self._callbackFunction_args = args                  # Callback function arguments if any 

        self.terminate_event = threading.Event()    # Terminate event 
        self.start_event = threading.Event()        # Start event 
        self.reset_event = threading.Event()        # Reset event  
        
    def run(self):
        """
        Function
        =========
        run

        Description
        ==============
        This function manages the monitor work.

        """

        while not self.terminate_event.is_set():
            while self.count > 0 and self.start_event.is_set():

                self._callbackFunction()   # Call the callback function 
                time.sleep(self._interval) # Sleep during the time specified by the interval 
                # if self.reset_event.is_set():
                """
                if self.reset_event.wait(self._interval):  # Wait for a small chunk of timeout
                    self.reset_event.clear()
                    self.count = self._duration/self._interval  # reset
                """
                self.count -= 1
            if self.count <= 0:
                self.start_event.clear()
                #print ('End of timeout')
                self.count = self._duration/self._interval  #reset
                self.terminate()

    def start_timer(self):
        self.start_event.set()
        return common.get_current_datetime("display")

    def stop_timer(self):
        self.start_event.clear()
        self.count = self._duration / self._interval  # reset

    def get_currentcount(self):
        return self.count

    def restart_timer(self):
        # reset only if timer is running. otherwise start timer afresh
        if self.start_event.is_set():
            self.reset_event.set()
        else:
            self.start_event.set()

    def terminate(self):
        self.terminate_event.set()

class DataCollectorUtil(object):

    @staticmethod
    def create_json(dc_name, start, end, data):

        datacollectorJSON = {}
        datacollectorJSON['DataCollectorName'] = dc_name
        datacollectorJSON['CollectedData'] = [{"StartTime": start, "EndTime": end, 'Data': data }]

        return datacollectorJSON

class RuleUtil(object):
    
    @staticmethod
    def print_progress(iteration, total, prefix='', suffix='', decimals=1, bar_length=100):
        """
        Call in a loop to create terminal progress bar

        @params:
            iteration   - Required  : current iteration (Int)
            total       - Required  : total iterations (Int)
            prefix      - Optional  : prefix string (Str)
            suffix      - Optional  : suffix string (Str)
            decimals    - Optional  : positive number of decimals in percent complete (Int)
            bar_length  - Optional  : character length of bar (Int)
        """
        print('\n')
        str_format = "{0:." + str(decimals) + "f}"
        percents = str_format.format(100 * (iteration / float(total)))
        filled_length = int(round(bar_length * iteration / float(total)))
        bar = '*' * filled_length + '-' * (bar_length - filled_length)
        sys.stdout.write('\r%s |%s| %s%s %s' % (prefix, bar, percents, '%', suffix)),

        if iteration == total:
            sys.stdout.write('\n')
        print('\n')
        sys.stdout.flush()

    @staticmethod
    def get_class(cls):
        parts = cls.split('.')
        module = '.'.join(parts[:-1])
        #print(module)
        m = __import__(module)
        
        for comp in parts[1:]:
            m = getattr(m, comp)
        return m

    @staticmethod
    def get_rule_config(finding_templates, finding_id):
        rc = {}
        for rule_config in finding_templates[finding_id]['Configuration']:
            rc = rule_config
        return rc
        
    @staticmethod
    def create_finding(rule_id, finding_templates, finding_id, impacted_resources, finding_parameters={}):
        f = {}
        f['Id'] = finding_id
        f['Finding'] = RuleUtil.create_formatted_string(finding_templates[finding_id]['FindingTemplate'], finding_parameters)
        f['Recommendation'] = RuleUtil.create_formatted_string(finding_templates[finding_id]['RecommendationTemplate'], finding_parameters)
        f['ImpactLevel'] = finding_templates[finding_id]['Level']
        f['ReferenceLinks'] = finding_templates[finding_id]['ReferenceLinks']
        f['ImpactedResources'] = impacted_resources
        return f

    @staticmethod
    def create_formatted_string(template, params):
        try:
            return template.format(**params)
        except KeyError:
            logger.warning("Unable to format string: '" + template + "'")
            return template 

    @staticmethod
    def create_rule_result(rule, **kwargs):
        rr = {}
        rr['RuleName'] = rule.name()
        rr['RuleFriendlyName'] = rule.friendly_name()
        rr['Category'] = rule.category()
        rr['RuleId'] = rule.id()
        if 'AR' in rule.id():
            rr['Type'] = 'perfinsights.rulengine.AnalyticRules.' + rule.name() + 'Result'
        else:
            rr['Type'] = 'perfinsights.rulengine.AnalyticRules.' + rule.name()
        rr['Configuration'] = rule.configuration()
        rr['RuleFindings'] = []
        rr['AnalysisStartTime'] = rule.rule_start_time
        rr['AnalysisEndTime'] = rule.rule_end_time

        for Name, Value in kwargs.items():
            rr[Name] = Value

        return rr

    @staticmethod
    def create_staticrule_result(rule, findings):
        rr = {}
        rr['RuleFriendlyName'] = rule.name()
        rr['Category'] = rule.category()
        rr['RuleId'] = rule.id()
        rr['RuleFindings'] = findings
        rr['RuleName'] = type(rule).__name__
        rr['AnalysisStartTime'] = rule.analysisStart
        rr['AnalysisEndTime'] = rule.analysisend
        rr['ReferenceDataCollectors'] = rule._ref_dcs
        return rr

    @staticmethod
    def create_json(rule_name, start, end, data):

        datacollectorJSON = {}
        datacollectorJSON[rule_name] = [{"StartTime": start, "EndTime": end, 'Data': data}]

        return datacollectorJSON
