#-------------------------------------------------------------------------
# Copyright (c) Microsoft.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#--------------------------------------------------------------------------
from __future__ import division

import os
import re

import logging
import subprocess
import util.common as common
from ruleengine.RuleEngineCls import RuleUtil
from ruleengine.RuleEngineCls import IStaticRule
from util.common import CachedProperty

logger = logging.getLogger(__name__)

class InstanceMetadataRule(IStaticRule):
    # Finding IDs
    NON_SRIOV_SIZE = 0
    NON_SRIOV_OS = 1

    def __init__(self, instance_metadata_map_dc):
        self._ref_dcs = ['instance_metadata_map_dc']
        self._instance_metadata_map_dc = instance_metadata_map_dc
        self._config = None

        self.analysisstart = None
        self.analysisend = None

    def init(self, rule_config):
        self._config = rule_config

    def id(self):
        return 'PIFL.SR0001'

    def name(self):
        return 'Instance Metadata Configuration Analysis'
    
    def category(self):
        try:
            return self._config['Category']
        except:
            return ''

    def run(self):
        self.analysisStart = common.get_current_datetime("display")

        def parse_sku(sku):
            match = re.search(r'^(\d).(\d)', sku)
            return int(match.group(1)), int(match.group(2))

        finding = lambda fid : RuleUtil.create_finding(self.id(), self._config['FindingTemplates'], fid, [{'ResourceName':'OS Image', 'ResourceValue':compute_map['offer'] + compute_map['sku'],'Configurations':[]}])

        findings = []
        compute_map = self._instance_metadata_map_dc.compute_map
        if compute_map['offer'] == 'CentOS-HPC':
            # check sr-iov versioning
            major, minor = parse_sku(compute_map['sku'])
            is_sriov_os = 10 * major + minor >= 76
            is_sriov_size = 'r' in compute_map['vmSize'].split('_')[1]

            if is_sriov_os and not is_sriov_size:
                findings.append(finding(self.NON_SRIOV_SIZE))
            elif not is_sriov_os and is_sriov_size:
                findings.append(finding(self.NON_SRIOV_OS))

        self.analysisend = common.get_current_datetime("display")

        return RuleUtil.create_staticrule_result(self, findings)

class DiskConfigurationRule(IStaticRule):

    def __init__(self, storage_map_dc):
        self._ref_dcs = ['storage_map_dc']
        self._storage_map_dc = storage_map_dc
        self._config = None

        self.analysisstart = None
        self.analysisend = None

    def init(self, rule_config):
        self._config = rule_config

    def id(self):
        return 'PIFL.SR0002'

    def name(self):
        return 'Disk Configuration Analysis'
    
    def category(self):
        return self._config['Category']

    def run(self):
        lvm_map = self._storage_map_dc.lvm_map
        mount_point_dict = self._storage_map_dc.mount_points_dict
        disk_map = self._storage_map_dc.block_device_map

        self.analysisStart = common.get_current_datetime("display")
        
        findings = []
        finding = self.check_spanned_volume(lvm_map)
        if finding:
            findings.append(finding)

        finding = self.check_disk_space(mount_point_dict)
        if finding:
            findings.append(finding)

        finding = self.check_volume_type(lvm_map, disk_map)
        if finding:
            findings.append(finding)

        self.analysisend = common.get_current_datetime("display")

        return RuleUtil.create_staticrule_result(self, findings)

    def check_spanned_volume(self, lvm_map):
        rule_id = 0
        impacted_resources = []

        if lvm_map == None:
            return None

        for vg in lvm_map:
            if vg['LogicalVolumes'] == None:
                continue
            for lv in vg['LogicalVolumes']:
                if lv['SegmentsNumber'] > 1 and (lv['RAIDType'] == 'linear' or lv['RAIDType'] == 'raid0'):
                    ir = {}
                    ir['LogicalVolumeName'] = lv['LogicalVolumeName']
                    ir['SegmentsNumber'] = lv['SegmentsNumber']
                    ir['RAIDType'] = lv['RAIDType']
                    #impacted_resources.append(ir)
                    impacted_resources.append({'ResourceName': 'LogicalVolumeName', 'ResourceValue': lv['LogicalVolumeName'], 'Configurations' : [ { 'RaidType': lv['RAIDType'], 'SegmentsNumber': lv['SegmentsNumber'] }] })
        
        if not impacted_resources:
            return None
        return RuleUtil.create_finding(self.id(), self._config['FindingTemplates'], rule_id, impacted_resources)

    def check_disk_space(self, mount_point_dict):
        logger.info("Entering check_disk_space")
        rule_id = 1
        impacted_resources = []

        if mount_point_dict == None:
            return None

        rule_config = RuleUtil.get_rule_config(
            self._config['FindingTemplates'], rule_id)
        
        for mp in mount_point_dict.values(): 
            logger.debug("Mount Point Name: " + mp.parent_device + "; Size (bytes):  " + str(mp.size_bytes) + "; Free Space (bytes): " + str(mp.free_space_bytes / 1024))
            # Size == 0 bytes is a likely indicator of bad data. In this case, skip this device. 
            if mp.size_bytes == 0:
                logger.warning("Mount Point '" + mp.parent_device + " size is 0 bytes. This could indicate a failure to retrieve the correct size data. Skipping the Disk Configuration finding for 'Volume(s) are running out of disk space' on this device.")
            else:
                used = mp.size_bytes - (mp.free_space_bytes / 1024)
                used_percent = round(100 * used/mp.size_bytes, 2)
                usedGB = round(float("{0:.2f}".format(
                    float(used) / 1024/1024)),2)
                if rule_config['MinVolumeFreeSpacePercentage'] > 100 - used_percent:
                    impacted_resources.append({'ResourceName': 'Device Name', 'ResourceValue': mp.parent_device, 'Configurations': [{'Size (GB)': str(mp.size_gbytes), 'Used (GB)': str(usedGB), 'Available (GB)': str(mp.free_space_gbytes), 'Used %': str(used_percent) + '%', 'Mounted on': mp.path}]})
        logger.debug("Leaving check_disk_space")
        if impacted_resources:
            return RuleUtil.create_finding(self.id(), self._config['FindingTemplates'], rule_id, impacted_resources, rule_config)
        else:
            return None

    def check_volume_type(self, lvm_map, disk_map):
        rule_id = 2
        impacted_resources = []

        if lvm_map == None:
            return None

        rule_config = RuleUtil.get_rule_config(
            self._config['FindingTemplates'], rule_id)

        for volume_group in lvm_map:
            # Check if volume group is only on data disks by checking IsDataDisk value of the physical volumes.
            # Only continue on to check the volume type if all physical volumes are on data disks, as logical volumes on OS disks are not expected to be striped.
            vg_on_data_disk = True
            for physical_volume in volume_group['PhysicalVolumes']:
                pv_block_device = physical_volume['BlockDevice']
                pv_partition_name = pv_block_device.split('/')[-1]
                device_found = False
                for bd in disk_map:
                    # First, check if the block device name matches the pv_block_device
                    if bd['BlockDevice'] == pv_block_device:
                        vg_on_data_disk = bd['IsDataDisk']
                        break
                    # Then, check if any of the partitions match the pv_block_device
                    for part in bd['Partitions']:
                        if part['Partition'] == pv_partition_name:
                            vg_on_data_disk = bd['IsDataDisk']
                            device_found = True
                            break
                    if device_found:
                        break
                # If any of the physical volumes are not on a data disk, stop and do not check the volume type, as logical volumes on OS disks are not expected to be striped.
                if not vg_on_data_disk:
                    break

            if vg_on_data_disk and volume_group['LogicalVolumes']:
                for lv in volume_group['LogicalVolumes']:
                    if lv['RAIDType'] != rule_config['RecommendedVolumeType']:
                        impacted_resources.append(
                            {'ResourceName': 'Logical volume', 'ResourceValue': lv['LogicalVolumeName'], 'Configurations': [{'Type': lv['RAIDType']}]})

        if impacted_resources:
            return RuleUtil.create_finding(self.id(), self._config['FindingTemplates'], rule_id, impacted_resources, rule_config)
        else:
            return None

class MachineConfigurationRule(IStaticRule):

    def __init__(self, machine_map_dc):
        self._ref_dcs = ['machine_map_dc']
        self._machine_map_dc = machine_map_dc
        self._config = None

        self.analysisstart = None
        self.analysisend = None

    def init(self, rule_config):
        self._config = rule_config
    
    def id(self):
        return 'PIFL.SR0003'

    def name(self):
        return 'Machine Configuration Analysis'
    
    def category(self):
        return self._config['Category']

    def run(self):
        _os_details_map = self._machine_map_dc.os_details_map()
        findings = []

        self.analysisStart = common.get_current_datetime("display")

        #finding = self.check_endorsed_distro(_os_details_map)
        #if finding:
        #    findings.append(finding)

        #finding = self.check_tainted_kernel()
        #if finding:
        #    findings.append(finding)

        finding = self.check_eol_distro(_os_details_map)
        if finding:
            findings.append(finding)

        finding = self.check_for_sar()
        if finding:
            findings.append(finding)

        self.analysisend = common.get_current_datetime("display")

                   
        return RuleUtil.create_staticrule_result(self, findings)

    def check_endorsed_distro(self, os_details_map):
        rule_id = 0
        impacted_resources = []

        rule_config = RuleUtil.get_rule_config(self._config['FindingTemplates'], rule_id)
        
        supported = False
        
        distrib = ''
        if 'distrib_id' in os_details_map:
            distrib = os_details_map['distrib_id']
        if os_details_map['name']:
            distrib = os_details_map['name']
        distribversion = ''
        if 'distrib_release' in os_details_map:
            distribversion = os_details_map['distrib_release']
        if os_details_map['version']:
            distribversion = os_details_map['version']

        for config in rule_config['EndorsedDistributions']:
            if config['Distribution'] == distrib:
                for version in config['Versions']:
                    if version == distribversion:
                        supported = True
                        break

        if supported == False:
            impacted_resources.append({'ResourceName': 'Distribution', 'ResourceValue': distrib, 'Configurations' : [ { 'Version': distribversion }] })
            return RuleUtil.create_finding(self.id(), self._config['FindingTemplates'], rule_id, impacted_resources)
        else:
            return None

    def check_tainted_kernel(self):
        rule_id = 1
        impacted_resources = []
        
        with open("/proc/sys/kernel/tainted",'r') as tainted:
            if tainted.readline().strip()=="0":
                return None
            else:
                return RuleUtil.create_finding(self.id(), self._config['FindingTemplates'], rule_id, impacted_resources)
    
    def check_eol_distro(self, os_details_map):
        rule_id = 2
        impacted_resources = []

        rule_config = RuleUtil.get_rule_config(self._config['FindingTemplates'], rule_id)
        
        eol = False
        
        distrib = ''
        if 'distrib_id' in os_details_map:
            distrib = os_details_map['distrib_id']
        if os_details_map['name']:
            distrib = os_details_map['name']
        distribversion = ''
        if 'distrib_release' in os_details_map:
            distribversion = os_details_map['distrib_release']
        if os_details_map['version']:
            distribversion = os_details_map['version']

        for config in rule_config['EOLDistributions']:
            if config['Distribution'] == distrib:
                for version in config['Versions']:
                    if version == distribversion:
                        eol = True
                        break

        if eol == True:
            impacted_resources.append({'ResourceName': 'Distribution', 'ResourceValue': distrib, 'Configurations' : [ { 'Version': distribversion }] })
            return RuleUtil.create_finding(self.id(), self._config['FindingTemplates'], rule_id, impacted_resources)
        else:
            return None

    def check_for_sar(self):
        has_sysstat_installed = False
        has_sysstat_enabled = False 

        if os.path.isdir('/var/log/sysstat'):
            has_sysstat_installed = True
        elif os.path.isdir('/var/log/sa'):
            has_sysstat_installed = True

        # If has_sysstat_installed, confirm it's enabled
        if has_sysstat_installed:
            try:
                systemctl_output = subprocess.check_output("systemctl show -p ActiveState sysstat", shell=True);
                active_state_str = 'ActiveState='
                if isinstance(systemctl_output, bytes):
                    systemctl_output = systemctl_output.decode('utf-8')
                if active_state_str in systemctl_output:
                    active_state = systemctl_output.split(active_state_str)[1].strip()
                    has_sysstat_enabled = active_state == 'active'
            except Exception as e:
                # If sysstat is installed, but we can't determine whether sysstat it is enabled, we will assume it is and set the value of has_sysstat_enabled to True
                has_sysstat_enabled = True 
                logger.error("Unable to determine if sysstat is enabled: {e}. PerfInsights will assume sysstat is enabled and attempt to collect SAR files.")


        if not has_sysstat_installed or not has_sysstat_enabled:
            # Likely newer fedora/cent based
            if os.path.exists("/usr/bin/dnf") or os.path.exists("/bin/dnf") or os.path.exists("/sbin/dnf"):
                return RuleUtil.create_finding(self.id(), self._config['FindingTemplates'], 3, [])
            # Likely older fedora/cent based
            elif os.path.exists("/usr/bin/yum") or os.path.exists("/bin/yum") or os.path.exists("/sbin/yum"):
                return RuleUtil.create_finding(self.id(), self._config['FindingTemplates'], 4, [])
            # Likely debian based
            elif os.path.exists("/usr/bin/apt") or os.path.exists("/bin/apt") or os.path.exists("/sbin/apt"):
                return RuleUtil.create_finding(self.id(), self._config['FindingTemplates'], 5, [])
            # Likely suse based
            elif os.path.exists("/usr/bin/zypper") or os.path.exists("/bin/zypper") or os.path.exists("/sbin/zypper"):
                return RuleUtil.create_finding(self.id(), self._config['FindingTemplates'], 6, [])
            else:
                return RuleUtil.create_finding(self.id(), self._config['FindingTemplates'], 7, [])

        return None

    """
    # Sample method with impacted resources
    def check_endorsed_distro(self, pci_map):
        rule_id = 2
        if pci_map == None:
            return None
        impacted_resources = []

        for device in pci_map:
            if device['vendor'] != 'NVIDIA Corporation':
                ir = {}
                ir['PCIDevice'] = device['device']
                ir['Vendor'] = device['vendor']
                impacted_resources.append(ir)
        
        if not impacted_resources:
            return None
        return RuleUtil.create_finding(self.id(), self._config['FindingTemplates'], rule_id, impacted_resources)
    """

class NetworkConfigurationRule(IStaticRule):
    # Finding IDs
    RDMA_UNNECESSARY = 0
    RDMA_DISABLED = 1
    FIREWALL_ACTIVE = 2
    SELINUX_ACTIVE = 3
    MISSING_PKEY = 4
    MISSING_IB_DRIVER = 5

    def __init__(self, machine_map_dc, instance_metadata_map_dc, hpcdiag_dc=None):
        self._ref_dcs = ['machine_map_dc', 'instance_metadata_map_dc']
        self._machine_map_dc = machine_map_dc
        self._instance_metadata_map_dc = instance_metadata_map_dc
        self._config = None
        if common.perfinsights_scenario == "hpc":
            self._ref_dcs.append("hpcdiag_dc")
            if hpcdiag_dc == None:
                raise "Scenario is hpc, but no hpcdiag_dc provided"
            self._hpcdiag_dc = hpcdiag_dc

        self.analysisstart = None
        self.analysisend = None

    def init(self, rule_config):
        self._config = rule_config

    def id(self):
        return 'PIFL.SR0004'

    def name(self):
        return 'Network Configuration Analysis'

    def category(self):
        return self._config['Category']

    def create_finding(self, finding_id, resources):
        return RuleUtil.create_finding(self.id(), self._config['FindingTemplates'], finding_id, resources)

    def run(self):
        compute_map = self._instance_metadata_map_dc.compute_map
        waagent_map = self._machine_map_dc.waagent_details_map_cached

        self.analysisStart = common.get_current_datetime("display")

        findings = []
        findings += self.check_rdma_waagent(compute_map, waagent_map)

        if common.perfinsights_scenario == 'hpc':
            findings += self.hpc_checks()

        self.analysisend = common.get_current_datetime("display")

        return RuleUtil.create_staticrule_result(self, findings)

    def hpc_checks(self):
        hpcdiag_dc = self._hpcdiag_dc
        pci_map = self._machine_map_dc.pci_map

        firewall = {
            'ResourceName': "Service",
            'ResourceValue': "Firewall",
            "Configurations": {
                "firewalld_active": hpcdiag_dc.network.get('firewalld'),
                "iptables_active": hpcdiag_dc.network.get('iptables')
            }
        }
        selinux = {
            "ResourceName": "Kernel Module",
            "ResourceValue": "SELinux",
            "Configurations": {
                "state": hpcdiag_dc.network['selinux']
            }
        }

        resource_of_device = lambda device : {
            "ResourceName": 'Infiniband Adapter',
            "ResourceValue": device['device'],
            "Configurations": {}
        }
        is_ib = lambda device : device['vendor'] == 'Mellanox Technologies'
        ib_adapters = [ resource_of_device(device) for device in pci_map if is_ib(device) ]

        hpc_findings = []
        if hpcdiag_dc.network.get('firewalld') or hpcdiag_dc.network.get('iptables'):
            hpc_findings.append(self.create_finding(self.FIREWALL_ACTIVE, [firewall]))
        if hpcdiag_dc.network['selinux'] == 'enforcing':
            hpc_findings.append(self.create_finding(self.SELINUX_ACTIVE, [selinux]))
        if hpcdiag_dc.infiniband['missing_pkey']:
            hpc_findings.append(self.create_finding(self.MISSING_PKEY, ib_adapters))
        if hpcdiag_dc.infiniband['missing_driver']:
            hpc_findings.append(self.create_finding(self.MISSING_IB_DRIVER, ib_adapters))

        return hpc_findings

    def check_rdma_waagent(self, compute_map, waagent_map):
        waagent_config = waagent_map.get("config")
        if waagent_config == None:
            logger.warning('missing waagent_config, cannot check rdma config')
            return []

        rdma_configured = waagent_config.get('OS.EnableRDMA') == 'y'
        is_rdma_size = 'r' in compute_map['vmSize'].split('_')[1]

        waagent = {
            "ResourceName": "Agent",
            "ResourceValue": "Windows Azure Guest Agent",
            "Configurations": {
                "OS.EnableRDMA": waagent_config.get('OS.EnableRDMA')
            }
        }

        if common.perfinsights_scenario == 'hpc' and is_rdma_size and not rdma_configured:
            return [self.create_finding(self.RDMA_DISABLED, [waagent])]
        elif not is_rdma_size and rdma_configured:
            return [self.create_finding(self.RDMA_UNNECESSARY, [waagent])]
        else:
            return []

class OOMScanRule(IStaticRule):
    def __init__(self):
        self._ref_dcs = []
        self._config = None

    def init(self, rule_config):
        self._config = rule_config

    def id(self):
        return 'PIFL.SR0005'

    def name(self):
        return 'Out of Memory (OOM) Scan'

    def category(self):
        return self._config['Category']

    def run(self):
        findings = []
        self.analysisStart = common.get_current_datetime("display")
        for finding_template in self._config['FindingTemplates']:
            config = finding_template['Configuration']
            if config:
                files = config['Files']
                matches = []
                if files:
                    regex = re.compile(config['Pattern'])
                    for file in files:
                        if os.path.exists(file):
                            try:
                                with open(file, 'r') as f:
                                    for line in f:
                                        if regex.search(line):
                                            matches.append({'ResourceName': 'Log', 'ResourceValue': file, 'Configurations': [{'Message': line}]})
                            except UnicodeDecodeError:
                                logger.warning("Failed with encoding utf-8, switch to ISO-8859-1")
                                try:
                                    with open(file, 'r', encoding="ISO-8859-1") as f:
                                        for line in f:
                                            if regex.search(line):
                                                matches.append({'ResourceName': 'Log', 'ResourceValue': file, 'Configurations': [{'Message': line}]})
                                except Exception as e:
                                    logger.warning("Unable to read file " + file + " for Static Rule " + self.name(), exc_info=e)
                            except Exception as e:
                                logger.warning("Unable to read file " + file + " for Static Rule " + self.name(), exc_info=e)
                    if len(matches) > 0:
                        finding = RuleUtil.create_finding(self.id(), self._config['FindingTemplates'], finding_template['Id'], matches)
                        findings.append(finding)
                else:
                    logger.warning("Misconfigured log scan rule, expected 'Files' in FindingTemplates[].Configuration")

        self.analysisend = common.get_current_datetime("display")
        return RuleUtil.create_staticrule_result(self, findings)


class SwapFileDiskRule(IStaticRule):
    def __init__(self, storage_map_dc, memory_usage_collector):
        self._ref_dcs = ['storage_map_dc', 'memory_usage_collector']
        self._storage_map_dc = storage_map_dc
        self._memory_usage_collector = memory_usage_collector
        self._config = None

        self.analysisstart = None
        self.analysisend = None

    def init(self, rule_config):
        self._config = rule_config

    def id(self):
        return 'PIFL.SR0006'

    def name(self):
        return 'Swap File Disk Configuration Analysis'

    def category(self):
        return self._config['Category']

    def run(self):
        findings = []
        self.analysisStart = common.get_current_datetime("display")
        swap_map = self._memory_usage_collector.get_system_swap_info
        
        if swap_map:
            # Check block device map for whether a temp disk exists. If not, this rule is not applicable. 
            block_device_map = self._storage_map_dc.block_device_map
            temp_disk_exists = False
            for block_device in block_device_map:
                if block_device['IsTempDisk']:
                    temp_disk_exists = True
                    break

            if temp_disk_exists:
                mount_point_map = self._storage_map_dc.mount_point_map
                if mount_point_map:
                    finding = self.check_swap_location(mount_point_map, swap_map)
                    if finding:
                        findings.append(finding)

        self.analysisend = common.get_current_datetime("display")
        return RuleUtil.create_staticrule_result(self, findings)

    def check_swap_location(self, mount_point_map, swap_map):
        rule_id = 0
        impacted_resources = []
        
        for swap in swap_map:
            swap_on_temp_disk = False
            for mount in mount_point_map:
                if mount['IsTempDisk'] and swap['Name'].startswith(mount['Mountpoint']):
                    swap_on_temp_disk = True
                    break

            if not swap_on_temp_disk:
                impacted_resources.append({'ResourceName': 'Name', 'ResourceValue': swap['Name'], 'Configurations': [{'Type': swap['Type'], 'Size (MB)': swap['Size'], 'Used (MB)': swap['Used'], 'Priority': swap['Priority']}]})

        if impacted_resources:
            return RuleUtil.create_finding(self.id(), self._config['FindingTemplates'], rule_id, impacted_resources)
        else:
            return None

