# -------------------------------------------------------------------------
# Copyright (c) Microsoft.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

import logging
import sys
import os
import platform
import re
import errno
import glob
import subprocess
import collections
import time
import six

import util.common as common

logger = logging.getLogger(__name__)

if six.PY3:
    long = int

MYPY = False
if MYPY:
    from typing import Dict, Optional

class OSProcReader:
    def get_pids(self):
        return [int(x) for x in os.listdir('/proc') if x.isdigit()]

    def get_process_stat(self, pid):
        f = self.open_process_file('/proc/%s/stat' % pid)
        return f.read() if f else None

    def get_system_stat(self):
        return self.yielding_file('/proc/stat')

    def yielding_file(self, file):
        if os.path.exists(file):
            with open(file, 'r') as f:
                for line in f:
                    yield line
            return
        raise FileNotFoundError('Could not find %s' % file)

    def open_process_file(self, file_name):
        if os.path.exists(file_name):
            try:
                return open(file_name, 'r')
            except IOError:
                logger.exception('Error retrieving %s' % file_name)
        
        return None



class SystemUtil:
    '''
    Class
    =========
    SystemUtil

    Description
    ==============
    Class holding all methods used to get operating system information.
    '''

    ENHANCED_KERNEL_VERSION = 2.6
    CLOCK_TICKS = os.sysconf('SC_CLK_TCK')
    PAGE_SIZE = int(os.sysconf('SC_PAGE_SIZE'))  # memory page size in KBytes
    DEFAULT_ENCODING = sys.getdefaultencoding()
    PROC_READER = OSProcReader()

    @staticmethod
    def get_kernel_version():
        '''
        Function
        =========
        get_kernel_version

        Description
        ==============
        Retrieve the kernel version.

        Arguments
        ==========
        N/A

        Return Values
        ==============
        __out <float> kernel version
        '''

        result = platform.release().replace('"', '')
        result = float(result.split('.')[0] + '.' + result.split('.')[1])
        return result

    @staticmethod
    def get_kernel_details():
        '''
        Function
        =========
        get_kernel_details

        Description
        ==============
        Retrieve kernel information:
            * Current kernel version
            * Available kernels
            * Kernel configurations
            * Kernel initrd images
            * Kernel system map

        Arguments
        ==========
        N/A

        Return Values
        ==============
        __out <dict> kernel_details
        '''

        kernel_details = {}

        kernel_details["current"] = platform.release()

        if os.path.isdir('/boot'):
            all_images = os.listdir("/boot")
            kernel_details["availableKernels"] = filter(re.compile("vmlinuz.*").match, all_images)
            kernel_details["initimgs"] = filter(re.compile("initr[ad].*").match, all_images)
            kernel_details["config"] = filter(re.compile("config.*").match, all_images)
            kernel_details["sys_map"] = filter(re.compile("System.*").match, all_images)
        else:
            logger.warning("Could not find /boot directory. Kernel image details will be incomplete")

        return kernel_details

    @staticmethod
    def get_system_uptime_ms():
        '''
        Function
        =========
        get_system_uptime_ms

        Description
        ==============
        Return the system uptime expressed
        in milliseconds since the epoch.

        Arguments
        ==========
        N/A

        Return Values
        ==============
        __out <long> system uptime
        '''

        for line in SystemUtil.PROC_READER.get_system_stat():
            if line.startswith('btime'):
                return long(float(line.strip().split()[1]) * 1000)

    @staticmethod
    def is_hv_vm():
        '''
        Function
        =========
        is_hv_vm

        Description
        ==============
        Checks if the guest operating system is running on
        Hyper-V.

        Arguments
        ==========
        N/A

        Return Values
        ==============
        __out <bool>
        '''

        try:
            result = subprocess.Popen(['/usr/sbin/dmidecode -s system-manufacturer 2> /dev/null'], stdout=subprocess.PIPE, shell=True)
            if 'Microsoft Corporation' in str(result.stdout.read()):
                return True
        except IndexError:
            pass

        return False

    @staticmethod
    def is_azure_vm():
        '''
        Function
        =========
        is_azure_vm

        Description
        ==============
        Checks if the guest operating system is running on
        Azure.
        We assume if the waagent.log exists, the VM
        runs on Azure.

        Arguments
        ==========
        N/A

        Return Values
        ==============
        __out <bool>
        '''

        if SystemUtil.is_hv_vm():
            if os.path.exists('/var/log/waagent.log'):
                return True

        return False


class CpuUtil():
    '''
    Class
    =========
    CpuUtil

    Description
    ==============
    Class holding all methods used to get processors information and statistics.
    '''

    @staticmethod
    def get_processor_cpu_time_ms(cpu_number):
        '''
        Function
        =========
        get_processor_cpu_time

        Description
        ==============
        Collect cpu time in millisecond
        for a single processor.

        Linux provides CPU time in seconds, so
        it requires to be converted to milliseconds.

        user   (1) Time spent in user mode.
        nice   (2) Time spent in user mode with low
            priority (nice).
        system (3) Time spent in system mode.
        idle   (4) Time spent in the idle task.
            This value should be USER_HZ times the
            second entry in the /proc/uptime pseudo-file.
        iowait (since Linux 2.5.41) (5) Time waiting for
            I/O to complete.  This value is not reliable
            for the following reasons:
                1. The CPU will not wait for I/O to
                complete; iowait is the time that a task
                is waiting for I/O to complete.
                When a CPU goes into idle state for
                outstanding task I/O, another task will
                be scheduled on this CPU.
                2. On a multi-core CPU, the task waiting
                for I/O to complete is not running on
                any CPU, so the iowait of each CPU is
                difficult to calculate.
                3. The value in this field may decrease
                in certain conditions.
        irq (since Linux 2.6.0-test4)   (6)
            Time servicing interrupts.
        softirq (since Linux 2.6.0-test4)   (7)
            Time servicing softirqs.
        steal (since Linux 2.6.11)  (8)
            Stolen time, which is the time spent in
            other operating systems when running in
            a virtualized environment.
        guest (since Linux 2.6.24)  (9)
            Time spent running a virtual CPU for guest
            operating systems under the control of the
            Linux kernel.
        guest_nice (since Linux 2.6.33) (10)
            Time spent running a niced guest (virtual
            CPU for guest operating systems under the
            control of the Linux kernel).

        We don't count iowait and idle in CPU time.

        Reference:
            https://www.kernel.org/doc/Documentation/filesystems/proc.txt
            https://www.idnt.net/en-US/kb/941772
            https://github.com/Leo-G/DevopsWiki/wiki/How-Linux-CPU-Usage-Time-and-Percentage-is-calculated
            https://stackoverflow.com/questions/23367857/accurate-calculation-of-cpu-usage-given-in-percentage-in-linux

        Arguments
        ==========
        __in <cpu_number> processor number

        Return Values
        ==============
        __out <list>: processor CPU non idle time, processor idle time
        '''

        result = None
        for line in SystemUtil.PROC_READER.get_system_stat():
            if line.startswith('cpu' + str(cpu_number)):
                proc_stat = line.split()
                cpu_non_idle_time = (long(proc_stat[1]) + long(proc_stat[2]) + long(proc_stat[3]) + long(proc_stat[4]) + long(proc_stat[6]) + long(proc_stat[7])) / float(SystemUtil.CLOCK_TICKS)
                cpu_idle_time = long(proc_stat[4]) / float(SystemUtil.CLOCK_TICKS)
                result = [long(cpu_non_idle_time * 1000), long(cpu_idle_time * 1000)]
                break

        return result

    @staticmethod
    def get_all_processors_cpu_time_ms():
        '''
        Function
        =========
        get_all_processors_cpu_time_ms

        Description
        ==============
        Collect cpu time in millisecond
        for all processors.

        Linux provides CPU time in seconds, so
        it requires to be converted to milliseconds.

        user   (1) Time spent in user mode.
        nice   (2) Time spent in user mode with low
            priority (nice).
        system (3) Time spent in system mode.
        idle   (4) Time spent in the idle task.
            This value should be USER_HZ times the
            second entry in the /proc/uptime pseudo-file.
        iowait (since Linux 2.5.41) (5) Time waiting for
            I/O to complete.  This value is not reliable
            for the following reasons:
                1. The CPU will not wait for I/O to
                complete; iowait is the time that a task
                is waiting for I/O to complete.
                When a CPU goes into idle state for
                outstanding task I/O, another task will
                be scheduled on this CPU.
                2. On a multi-core CPU, the task waiting
                for I/O to complete is not running on
                any CPU, so the iowait of each CPU is
                difficult to calculate.
                3. The value in this field may decrease
                in certain conditions.
        irq (since Linux 2.6.0-test4)   (6)
            Time servicing interrupts.
        softirq (since Linux 2.6.0-test4)   (7)
            Time servicing softirqs.
        steal (since Linux 2.6.11)  (8)
            Stolen time, which is the time spent in
            other operating systems when running in
            a virtualized environment.
        guest (since Linux 2.6.24)  (9)
            Time spent running a virtual CPU for guest
            operating systems under the control of the
            Linux kernel.
        guest_nice (since Linux 2.6.33) (10)
            Time spent running a niced guest (virtual
            CPU for guest operating systems under the
            control of the Linux kernel).

        We don't count iowait and idle in CPU time.

        Reference:
            https://www.kernel.org/doc/Documentation/filesystems/proc.txt
            https://www.idnt.net/en-US/kb/941772
            https://github.com/Leo-G/DevopsWiki/wiki/How-Linux-CPU-Usage-Time-and-Percentage-is-calculated
            https://stackoverflow.com/questions/23367857/accurate-calculation-of-cpu-usage-given-in-percentage-in-linux

        Arguments
        ==========
        __in <cpu_number> processor number

        Return Values
        ==============
        __out <list>: processor CPU non idle time, processor idle time
        '''

        result = {}

        proc_stat = [line.split() for line in SystemUtil.PROC_READER.get_system_stat()][1:]
        result['TimeStamp'] = int(round(time.time() * 1000))

        processors_data = []
        for line in proc_stat:
            if line[0].startswith('cpu'):
                idle_jiffies = long(line[4]) + long(line[5])
                idle_ms = long(idle_jiffies / float(SystemUtil.CLOCK_TICKS) * 1000)
                busy_jiffies = long(line[1]) + long(line[2]) + long(line[3]) + long(line[6]) + long(line[7]) + long(
                    line[8])
                busy_ms = long(busy_jiffies / float(SystemUtil.CLOCK_TICKS) * 1000)
                processors_data.append({'CpuNbr': int(line[0].replace('cpu', '')), 'NonIdleTimeMs': busy_ms, 'IdleTimeMs': idle_ms})

        result['ProcessorsCpuUsage'] = processors_data

        return result

    @staticmethod
    def cpu_count_logical():
        '''
        Function
        =========
        cpu_count_logical

        Description
        ==============
        Return the number of logical CPUs
        in the system.

        Reference:
            https://github.com/giampaolo/psutil/blob/master/psutil/_pslinux.py#L599

        Arguments
        ==========
        N/A

        Return Values
        ==============
        __out <int> logical_cpu_count
        '''

        try:
            return os.sysconf("SC_NPROCESSORS_ONLN")
        except ValueError:
            # As a second fallback we try to parse /proc/cpuinfo
            logical_cpu_count = 0
            with open('/proc/cpuinfo', 'r') as f:
                for line in f:
                    if line.lower().startswith('processor'):
                        logical_cpu_count += 1

            # Unknown format (e.g. amrel/sparc architectures), see:
            # https://github.com/giampaolo/psutil/issues/200
            # Try to parse /proc/stat as a last resort
            if logical_cpu_count == 0:
                search = re.compile('cpu\\d')
                with open('/proc/stat', 'rt') as f:
                    for line in f:
                        line = line.split(' ')[0]
                        if search.match(line):
                            logical_cpu_count += 1

            if logical_cpu_count == 0:
                # Mimic os.cpu_count()
                return None

            return logical_cpu_count

    @staticmethod
    def cpu_count_physical():
        '''
        Function
        =========
        cpu_count_physical

        Description
        ==============
        Return the number of physical CPUs
        in the system.

        Arguments
        ==========
        N/A

        Return Values
        ==============
        __out <int> physical_cpu_count
        '''

        with open('/proc/cpuinfo', 'r') as f:
            physical_cpu_count = set()
            for line in f:
                if line.lower().startswith('physical id'):
                    physical_cpu_count.add(line.strip())

        # mimic os.cpu_count()
        return len(physical_cpu_count) if physical_cpu_count else None


class ProcessUtil:
    '''
    Class
    =========
    ProcessUtil

    Description
    ==============
    Class holding all methods used to get process information and statistics.
    '''

    STATUS_RUNNING = "running"
    STATUS_SLEEPING = "sleeping"
    STATUS_DISK_SLEEP = "disk-sleep"
    STATUS_STOPPED = "stopped"
    STATUS_TRACING_STOP = "tracing-stop"
    STATUS_ZOMBIE = "zombie"
    STATUS_DEAD = "dead"
    STATUS_WAKE_KILL = "wake-kill"
    STATUS_WAKING = "waking"
    STATUS_IDLE = "idle"  # BSD
    STATUS_LOCKED = "locked"  # BSD
    STATUS_WAITING = "waiting"  # BSD

    PROC_STATUSES = {
        "R": STATUS_RUNNING,
        "S": STATUS_SLEEPING,
        "D": STATUS_DISK_SLEEP,
        "T": STATUS_STOPPED,
        "t": STATUS_TRACING_STOP,
        "Z": STATUS_ZOMBIE,
        "X": STATUS_DEAD,
        "x": STATUS_DEAD,
        "K": STATUS_WAKE_KILL,
        "W": STATUS_WAKING
    }

    @staticmethod
    def get_pids_with_children(pid_map=None):
        '''
        Function
        =========
        get_pids_with_children

        Description
        ==============
        Retrieve all processes with their children.
        First parse all PIDs and get their PPID
        then reverse to get all PIDS with their child PIDs. 

                                USER VIEW
        <-- PID 43 --> <----------------- PID 42 ----------------->
                             +---------+
                             | process |
                            _| pid=42  |_
                          _/ | tgid=42 | \_ (new thread) _
               _ (fork) _/   +---------+                  \
            /                                        +---------+
        +---------+                                    | process |
        | process |                                    | pid=44  |
        | pid=43  |                                    | tgid=42 |
        | tgid=43 |                                    +---------+
        +---------+
        <-- PID 43 --> <--------- PID 42 --------> <--- PID 44 --->
                             KERNEL VIEW

        tgid: thread group id, i.e. parent process id

        Reference:
            man proc
            https://www.kernel.org/doc/Documentation/filesystems/proc.txt

        Arguments
        ==========
        __in N/A

        Return Values
        ==============
        __out <dict> pids with their children PIDS
        '''

        if pid_map is None:
            pid_map = SystemUtil.PROC_READER.get_pids()

        ppid_map = {}
        for pid in pid_map:
            data = SystemUtil.PROC_READER.get_process_stat(pid)
            if data is not None:
                rpar = data.rfind(')')
                dset = data[rpar + 2:].split()
                ppid = int(dset[1])
                ppid_map[pid] = ppid

        process_map = collections.defaultdict(list)
        for pid, ppid in ppid_map.items():
            process_map[ppid].append({'PID': pid, 'ProcessName': ProcessUtil.get_process_name(pid)})

        return process_map

    @staticmethod
    def get_parent_pid(pid):
        '''
        Function
        =========
        get_parent_pid

        Description
        ==============
        Retrieve the process parent pid.
        First checks the process is still alive then
        open /proc/PID/status.

        Reference:
            man proc
            https://www.kernel.org/doc/Documentation/filesystems/proc.txt

        Arguments
        ==========
        __in <int> pid : process PID

        Return Values
        ==============
        __out <int> parent process PID
        '''

        f = SystemUtil.PROC_READER.open_process_file('/proc/%s/status' % pid)
        if f != None: 
            for line in f.readlines():
                line = line.strip().split(':')
                if 'PPid' in line[0] and line[1].strip() != pid:
                    return int(line[1].strip())
        
        return None

    @staticmethod
    def get_process_name(pid):
        '''
        Function
        =========
        get_process_name

        Description
        ==============
        Retrieve the process name.
        First checks the process is still alive then
        open /proc/PID/stat.

        Reference:
            man proc
            https://www.kernel.org/doc/Documentation/filesystems/proc.txt

        Arguments
        ==========
        __in <int> pid : process PID

        Return Values
        ==============
        __out <string> process name
        '''

        f = SystemUtil.PROC_READER.get_process_stat(pid)
        if f != None:
            return f.split(' ')[1].replace('(', '').replace(')', '')
        else:
            return None 

    @staticmethod
    def get_process_status(pid):
        '''
        Function
        =========
        get_process_status

        Description
        ==============
        Retrieve the process status.
        First checks the process is still alive then
        open /proc/PID/stat.

        Reference:
            man proc
            https://www.kernel.org/doc/Documentation/filesystems/proc.txt

        Arguments
        ==========
        __in <int> pid : process PID

        Return Values
        ==============
        __out <string> process status
        '''
        
        f = SystemUtil.PROC_READER.get_process_stat(pid)
        return f.split(' ')[2] if f else None

    @staticmethod
    def get_process_start_time_ms(pid):
        '''
        Function
        =========
        get_process_start_time_ms

        Description
        ==============
        Retrieve the process start time in milliseconds.
        First checks the process is still alive then
        open /proc/PID/stat.

        Reference:
            man proc
            https://www.kernel.org/doc/Documentation/filesystems/proc.txt

        Arguments
        ==========
        __in <int> pid : process PID

        Return Values
        ==============
        __out <long> process start time
        '''

        f = SystemUtil.PROC_READER.get_process_stat(pid)
        if f != None: 
            cpu_start_time = int(f.split(' ')[21]) / float(SystemUtil.CLOCK_TICKS)
            return long(cpu_start_time * 1000) + SystemUtil.get_system_uptime_ms()
        else:
            return None

    @staticmethod
    def get_process_command_line(pid):
        '''
        Function
        =========
        get_process_command_line

        Description
        ==============
        Retrieve the process start command line.
        First checks the process is still alive then
        open /proc/PID/cmdline.

        Reference:
            man proc
            https://www.kernel.org/doc/Documentation/filesystems/proc.txt

        Arguments
        ==========
        __in <int> pid : process PID

        Return Values
        ==============
        __out <string> process command line
        '''
        f = SystemUtil.PROC_READER.open_process_file('/proc/%s/cmdline' % pid)
        return f.read().replace('\x00', ' ') if f else None

    @staticmethod
    def get_process_cpu_time_ms(pid):
        '''
        Function
        =========
        get_process_cpu_time_ms

        Description
        ==============
        Retrieve the process CPU time in milliseconds.
        We don't capture child data as it would bring
        confusion in the report (Cpu% way above 100%).
        So we only useuser + system (user's children
        + system's children still on on hold for validation).
        Process iowait is field 42 but not captured as it
        is considered as idle time for process.

        First checks the process is still alive then
        open /proc/PID/stat.

        We need to convert to milliseconds as Linux
        provides CPU time in seconds.

        Reference:
            man proc
            https://www.kernel.org/doc/Documentation/filesystems/proc.txt
            https://stackoverflow.com/questions/16726779/how-do-i-get-the-total-cpu-usage-of-an-application-from-proc-pid-stat
            https://unix.stackexchange.com/questions/58539/top-and-ps-not-showing-the-same-cpu-result/58541#58541

        Arguments
        ==========
        __in <int> pid : process PID

        Return Values
        ==============
        __out <long> process CPU time
        '''

        f = SystemUtil.PROC_READER.get_process_stat(pid)
        if f != None: 
            proc_stat = f.split(' ')
            # cpu_time = (long(proc_stat[13]) + long(proc_stat[14]) + long(proc_stat[15]) + long(proc_stat[16])) / float(SystemUtil.CLOCK_TICKS)
            cpu_time = (long(proc_stat[13]) + long(proc_stat[14])) / float(SystemUtil.CLOCK_TICKS)
            return long(cpu_time * 1000)
        else:
            return None

    @staticmethod
    def get_process_cpu_times_seconds(pid):
        '''
        Function
        =========
        get_process_cpu_times_seconds

        Description
        ==============
        Retrieve the process CPU times in seconds
        for user and system.
        (user + user's children and system + system's children)
        Process iowait is field 42 but not captured as it
        is considered as idle time for process

        First checks the process is still alive then
        open /proc/PID/stat.

        Reference:
            man proc
            https://www.kernel.org/doc/Documentation/filesystems/proc.txt
            https://stackoverflow.com/questions/16726779/how-do-i-get-the-total-cpu-usage-of-an-application-from-proc-pid-stat
            https://unix.stackexchange.com/questions/58539/top-and-ps-not-showing-the-same-cpu-result/58541#58541

        Arguments
        ==========
        __in <int> pid : process PID

        Return Values
        ==============
        __out <array[float]> process user and system CPU time in seconds
        '''

        f = SystemUtil.PROC_READER.get_process_stat(pid)
        if f != None:
            proc_stat = f.split(' ')
            return (long(proc_stat[13]) + long(proc_stat[14])) / SystemUtil.CLOCK_TICKS, (long(proc_stat[15]) + long(proc_stat[16])) / SystemUtil.CLOCK_TICKS
        else:
            # Process does not exist anymore
            return None

    @staticmethod
    def get_process_resident_memory_kbytes(pid):
        '''
        Function
        =========
        get_process_resident_memory_kbytes

        Description
        ==============
        Returns the Linux Resident memory of the process gathered
        from /proc/pid/statm.

        First checks the process is still alive then
        open /proc/PID/statm.

        /proc/[pid]/statm
        Provides information about memory usage, measured in pages.
        The columns are:
            size       (1) total program size
                        (same as VmSize in /proc/[pid]/status)
            resident   (2) resident set size
                        (same as VmRSS in /proc/[pid]/status)
            shared     (3) number of resident shared pages
                        (i.e., backed by a file)
                        (same as RssFile+RssShmem in /proc/[pid]/status)
            text       (4) text (code)
            lib        (5) library (unused since Linux 2.6; always 0)
            data       (6) data + stack
            dt         (7) dirty pages (unused since Linux 2.6; always 0)

        Reference:
            man proc
            https://www.kernel.org/doc/Documentation/filesystems/proc.txt

        Arguments
        ==========
        __in <int> pid : process PID

        Return Values
        ==============
        __out <int> resident memory used by the process in KBytes
        '''
        f = SystemUtil.PROC_READER.open_process_file('/proc/%s/statm' % pid)
        if f != None:
            return int(f.read().split()[1]) * (SystemUtil.PAGE_SIZE / 1024)
        else:
            return None

    @staticmethod
    def get_process_swap_memory_kbytes(pid):
        '''
        Function
        =========
        get_process_swap_memory_kbytes

        Description
        ==============
        Returns the swap memory used by the process.
        First checks the process is still alive then
        open /proc/PID/status.

        Reference:
            https://www.kernel.org/doc/Documentation/filesystems/proc.txt

        Arguments
        ==========
        __in <int> pid : process PID

        Return Values
        ==============
        __out <long> swap memory used by the process in KBytes.
        '''

        f = SystemUtil.PROC_READER.open_process_file('/proc/%s/status' % pid)
        if f != None:
            for line in f:
                if line.startswith("VmSwap:"):
                    return int(line.split()[1])
            return 0
        else:
            return None

    @staticmethod
    def get_process_io_stats(pid):  # type: (int) -> Optional[Dict[str, int]]
        '''
        Function
        =========
        get_process_io_stats

        Description
        ==============
        Get current I/O stats from /proc/pid/io.
        All values from file are cumulative.

        /proc/<pid>/io - Display the IO accounting fields
            rchar: The number of bytes which this task has caused to be read
            from storage.
            This is simply the sum of bytes which this process passed to read()
            and pread().
            It includes things like tty IO and it is unaffected by whether or
            not actual physical disk IO was required (the read might have been
            satisfied from pagecache)

            wchar: The number of bytes which this task has caused, or shall
            cause to be written to disk. Similar caveats apply here as with
            rchar.

            syscr: Attempt to count the number of read I/O operations,
            i.e. syscalls like read() and pread().

            syscw: Attempt to count the number of write I/O operations,
            i.e. syscalls like write() and pwrite().

            read_bytes: Attempt to count the number of bytes which this
            process really did cause to be fetched from the storage layer.
            Done at the submit_bio() level, so it is accurate for block-backed
            filesystems.

            write_bytes: Attempt to count the number of bytes which this
            process caused to be sent to the storage layer.
            This is done at page-dirtying time.

            cancelled_write_bytes: The big inaccuracy here is truncate. If a
            process writes 1MB to a file and then deletes the file, it will
            in fact perform no writeout.
            But it will have been accounted as having caused 1MB of write.
            In other words: The number of bytes which this process caused to
            not happen, by truncating pagecache.
            A task can cause "negative" IO too.

            If this task truncates some dirty pagecache, some IO which another
            task has been accounted for (in its write_bytes) will not be
            happening. We _could_ just subtract that from the truncating task's
            write_bytes, but there is informationloss in doing that.

        Reference:
            https://www.kernel.org/doc/Documentation/filesystems/proc.txt

        Arguments
        ==========
        __in <int> pid : process PID

        Return Values
        ==============
        __out <dict> process_iostat: Dictionnary with current process reads,
                                     writes, read bytes and write bytes.
                                     Returns None if process does not exist
                                     anymore.
        '''

        if os.path.exists('/proc/%s/io' % pid):
            process_iostat = {}
            try:
                for line in open('/proc/%s/io' % pid, 'r').readlines():
                    split = line.strip().split(': ')
                    # number of read I/O operations, i.e. syscalls like read()
                    # and pread()
                    if split[0] == 'syscr':
                        process_iostat['reads'] = int(split[1])
                    # number of write I/O operations, i.e. syscalls like write()
                    # and pwrite()
                    elif split[0] == 'syscw':
                        process_iostat['writes'] = int(split[1])
                    # number of bytes which this task has caused to be read
                    # from storage
                    elif split[0] == 'rchar':
                        process_iostat['read_bytes'] = int(split[1])
                    # number of bytes which this task has caused, or shall cause
                    # to be written to disk
                    elif split[0] == 'wchar':
                        process_iostat['write_bytes'] = int(split[1])
                return process_iostat
            except Exception as e:
                logger.exception("Error opening io file for pid {}".format(pid), exc_info=e)
                return None
        else:
            # Process does not exist anymore
            return None

    @staticmethod
    def exists(pid):
        '''
        Function
        =========
        exists

        Description
        ==============
        Checks if the provided PID exists
        (process is then alive) by looking into /proc.

        Arguments
        ==========
        __in <int> pid : process PID

        Return Values
        ==============
        __out <bool> is_alive
        '''

        alive = False
        if os.path.exists('/proc/%s' % pid):
            alive = True
        return alive

    @staticmethod
    def is_alive(pid):
        '''
        Function
        =========
        is_process_alive

        Description
        ==============
        Checks if the provided PID exists
        (process is then alive) by using kill.

        Exceptions:
            EINVAL An invalid signal was specified.
            EPERM The process does not have permission
                to send the signal to any of the target processes.
            ESRCH The pid or process group does not exist.

        Arguments
        ==========
        __in <int> pid : process PID

        Return Values
        ==============
        __out <bool>
        '''

        if pid == 0:
            # According to "man 2 kill" PID 0 has a special meaning:
            # it refers to <<every process in the process group of the calling
            # process>> so we don't want to go any further.
            # If we get here it means this UNIX platform *does* have a process
            # with id 0.
            return True
        try:
            os.kill(pid, 0)
        except OSError as err:
            if err.errno == errno.ESRCH:
                # ESRCH == No such process
                return False
            elif err.errno == errno.EPERM:
                return True
            else:
                # According to "man 2 kill" possible error values are (EINVAL,
                # EPERM, ESRCH) therefore we should never get here.
                # If we do let's be explicit in considering this an error.
                raise err
        else:
            return True


class DiskUtil():
    '''
    Class
    =========
    DiskUtil

    Description
    ==============
    Class holding all methods used to get block devices
    information and statistics.
    '''

    DEFAUL_SECTOR_SIZE = 512
    DEFAULT_PARTITION_BLOCK_SIZE = 1024
    # VALID_BLOCK_DEVICES_PATTERN
    # On Azure we should not have /dev/hd? (HDD)
    # only SCSI and RAID devices
    VALID_BLOCK_DEVICES_PATTERN = ['sd[a-z]*\\b', 'md*', 'dm*']
    SCSI_REGEX_PATTERN = '\\d+:\\d+:\\d+:\\d+'
    OS_DISK_SCSI_PATTERN = ['2:0:0:0', '0:0:0:0']
    TEMP_DISK_SCSI_PATTERN = ['3:0:1:0', '1:0:1:0', '4:0:1:0']
    DATA_DISK_SCSI_PATTERN = [':::']
    MDADM_DEVICE_PATTERN = 'md\\d+'

    @staticmethod
    def get_block_devices():
        '''
        Function
        =========
        get_block_devices

        Description
        ==============
        Get all block devices available
        in the system.
        Filters to only get *disk* block
        devices.

        For each block device, returns:
            * name (/dev/[sd|md*|dm*])
            * SCSI Id
            * sector size
            * Size in GiB

        Arguments
        ==========
        N/A

        Return Values
        ==============
        __out <dict> block_devices
        '''

        block_devices = {}
        for device in glob.glob('/sys/block/*'):
            for pattern in DiskUtil.VALID_BLOCK_DEVICES_PATTERN:
                device_name = os.path.basename(device)
                if re.compile(pattern).match(device_name):
                    sector_size = DiskUtil.get_block_device_sector_size_bytes(device_name)
                    block_devices[device_name] = {'SCSIId': DiskUtil.get_device_scsi_id(device_name), 'SizeGB': DiskUtil.get_block_device_size_gigabytes(device_name), 'SectorSize': sector_size}

        return block_devices

    @staticmethod
    def get_device_scsi_id(device_short_name):
        '''
        Function
        =========
        get_device_scsi_id

        Description
        ==============
        Get the SCSI Id for a block device.
            * Bus
            * Port
            * TargetId
            * Lun

        Arguments
        ==========
        __in <string> device_short_name: device name without '/dev/'

        Return Values
        ==============
        __out <dict> scsi_id: SCSI Id
        '''

        if os.path.exists('/sys/block/' + device_short_name + '/device'):
            try: 
                storage_port_id = os.readlink(glob.glob('/sys/block/' + device_short_name + '/device')[0]).split('/')[3]
                if re.compile(DiskUtil.SCSI_REGEX_PATTERN).match(storage_port_id):
                    # This is a SCSI id
                    storage_port_id = storage_port_id.split(':')
                    return {'SCSIBus': storage_port_id[0], 'SCSIPort': storage_port_id[1], 'SCSITargetId': storage_port_id[2], 'SCSILun': storage_port_id[3]}
            except:
                logger.exception('Error retrieving device scsi id for %s' % device_short_name)

        return {'SCSIBus': '', 'SCSIPort': '', 'SCSITargetId': '', 'SCSILun': ''}

    @staticmethod
    def get_block_device_sector_size_bytes(device_short_name):
        '''
        Function
        =========
        get_block_device_sector_size_bytes

        Description
        ==============
        Get sector size for a block device in bytes.

        Reference:
            https://www.kernel.org/doc/Documentation/block/queue-sysfs.txt

        hw_sector_size (RO): this is the hardware sector
            size of the device, in bytes.
        physical_block_size (RO): this is the physical
            block size of device, in bytes.

        Arguments
        ==========
        __in <string> device_short_name: device name without '/dev/'

        Return Values
        ==============
        __out <int> device sector size or default value (DEFAUL_SECTOR_SIZE)
        '''

        try:
            with open('/sys/block/%s/queue/hw_sector_size' % device_short_name) as f:
                return int(f.read().rstrip('\n'))
        except (IOError, ValueError):
            return DiskUtil.DEFAUL_SECTOR_SIZE

    @staticmethod
    def get_block_device_size_gigabytes(device_short_name):
        '''
        Function
        =========
        get_block_device_size_gigabytes

        Description
        ==============
        Get size of a block device in GiB.

        Reference:
            https://www.kernel.org/doc/Documentation/block/queue-sysfs.txt

        hw_sector_size (RO): this is the hardware sector
            size of the device, in bytes.
        physical_block_size (RO): this is the physical
            block size of device, in bytes.

        Arguments
        ==========
        __in <string> device_short_name: device name without '/dev/'

        Return Values
        ==============
        __out <float> device size in GiB
        '''

        with open('/sys/block/%s/size' % device_short_name) as f:
            device_size = int(f.read().rstrip('\n'))
        sector_size = DiskUtil.get_block_device_sector_size_bytes(device_short_name)

        return float("{0:.2f}".format((device_size * sector_size) / float(1024 * 1024 * 1024)))

    @staticmethod
    def get_device_mapper_name_from_device(device_short_name):
        '''
        Function
        =========
        get_device_mapper_name_from_device

        Description
        ==============
        Get device mapper name of a block device.

        Arguments
        ==========
        __in <string> device_short_name: device name without '/dev/'

        Return Values
        ==============
        __out <string> device mapper name
        '''

        for sym_link in glob.glob('/dev/mapper/*'):
            try:
                if device_short_name == os.readlink(sym_link).split('/')[1]:
                    return sym_link.replace('/dev/mapper/', '')
            except OSError:
                # We may have read a non-symbolic link
                pass

        return ''

    @staticmethod
    def get_device_name_from_device_mapper(mapper_device_short_name):
        '''
        Function
        =========
        get_device_name_from_device_mapper

        Description
        ==============
        Get the device name of a device mapper device.

        Arguments
        ==========
        __in <string> mapper_device_short_name: device name without
            '/dev/mapper'

        Return Values
        ==============
        __out <string> device short name
        '''

        for sym_link in glob.glob('/dev/mapper/*'):
            try:
                if mapper_device_short_name == sym_link.replace('/dev/mapper/', ''):
                    return os.readlink(sym_link).split('/')[1]
            except OSError:
                # We may have read a non-symbolic link
                pass

        return ''

    @staticmethod
    def is_lvm_device(device_short_name):

        for sym_link in glob.glob('/dev/mapper/*'):
            try:
                if device_short_name == os.readlink(sym_link).split('/')[1]:
                    return True
            except OSError:
                # We may have read a non-symbolic link
                pass

        return False

    @staticmethod
    def is_mdadm_device(device_short_name):

        if re.compile(DiskUtil.MDADM_DEVICE_PATTERN).match(device_short_name):
            return True
        return False

    @staticmethod
    def is_azure_os_disk(scsi_bus, scsi_port, scsi_target_id, scsi_lun):
        '''
        Function
        =========
        is_azure_os_disk

        Description
        ==============
        Check if the block device is the OS disk from
        the SCSI id pattern.

        Arguments
        ==========
        __in <class> self: class instance

        Return Values
        ==============
        __out <bool>
        '''

        scsi_id = str(scsi_bus) + ':' + str(scsi_port) + ':' + str(scsi_target_id) + ':' + str(scsi_lun)
        if scsi_id in DiskUtil.OS_DISK_SCSI_PATTERN:
            return True
        return False

    @staticmethod
    def is_azure_data_disk(scsi_bus, scsi_port, scsi_target_id, scsi_lun):
        '''
        Function
        =========
        is_azure_data_disk

        Description
        ==============
        Check if the block device is a data disk from
        the SCSI id pattern.

        Arguments
        ==========
        __in <class> self: class instance

        Return Values
        ==============
        __out <bool>
        '''

        scsi_id = str(scsi_bus) + ':' + str(scsi_port) + ':' + str(scsi_target_id) + ':' + str(scsi_lun)
        if (scsi_id not in DiskUtil.OS_DISK_SCSI_PATTERN \
            and scsi_id not in DiskUtil.TEMP_DISK_SCSI_PATTERN) \
                or scsi_id in DiskUtil.DATA_DISK_SCSI_PATTERN: # or \
                    # self.name.startswith('sd'):
                    return True

        # Empty SCSI ID means lvm/mdadm device
        return False

    @staticmethod
    def is_azure_temp_disk(scsi_bus, scsi_port, scsi_target_id, scsi_lun):
        '''
        Function
        =========
        is_azure_temp_disk

        Description
        ==============
        Check if the block device is the temporary
        disk from the SCSI id pattern.

        Arguments
        ==========
        __in <class> self: class instance

        Return Values
        ==============
        __out <bool>
        '''

        scsi_id = str(scsi_bus) + ':' + str(scsi_port) + ':' + str(scsi_target_id) + ':' + str(scsi_lun)
        if scsi_id in DiskUtil.TEMP_DISK_SCSI_PATTERN:
            return True
        return False

    @staticmethod
    def get_azure_temporary_drive():
        '''
        Function
        =========
        get_temporary_drive

        Description
        ==============
        Get the current Azure temporary drive.

        Arguments
        ==========
        N/A

        Return Values
        ==============
        __out <string> device name
        '''

        for device in glob.glob('/sys/block/*'):
            for pattern in DiskUtil.VALID_BLOCK_DEVICES_PATTERN:
                if re.compile(pattern).match(os.path.basename(device)):
                    scsi_id = DiskUtil.get_device_scsi_id(os.path.basename(device))
                    if DiskUtil.is_azure_temp_disk(scsi_id['SCSIBus'], scsi_id['SCSIPort'], scsi_id['SCSITargetId'],scsi_id['SCSILun'] ):
                        return '/dev/%s' % os.path.basename(device)

        return None


class PartitionUtil():
    '''
    Class
    =========
    PartitionUtil

    Description
    ==============
    Class holding all methods used to get partitions
    information and statistics.
    '''

    @staticmethod
    def get_partition_parent_block_device_name(partition_short_name):
        '''
        Function
        =========
        get_partition_parent_block_device_name

        Description
        ==============
        Get the partition parent block device short name

        Arguments
        ==========
        __in <string> partition_short_name: partition full path (/dev/[*]) without '/dev/'

        Return Values
        ==============
        __out <string> block device short name
        '''

        if 'md' in partition_short_name:
            return '/dev/' + partition_short_name
        elif 'dm-' in partition_short_name:
            return '/dev/' + partition_short_name
        else:
            for line in open('/proc/partitions', 'r').readlines():
                current_device = (line.replace('\f', '').replace('\r', '').replace('\n', '')).split()
                if (len(current_device) >= 4) and current_device[0] != 'major':
                    if int(current_device[0]) in [8, 65, 66, 67, 68, 69, 70, 71, 128, 129, 130, 131, 132, 133, 134, 135] and \
                        current_device[3] in partition_short_name and \
                        current_device[3] != partition_short_name:
                            return '/dev/' + current_device[3]

    @staticmethod
    def get_partition_fstype(partition_short_name):
        '''
        Function
        =========
        get_partition_fstype

        Description
        ==============
        Get partition file system.

        Arguments
        ==========
        __in <string> partition_path: partition short name
            (without /dev/)

        Return Values
        ==============
        __out <string> partition file system
        '''

        # TODO: Test with more file systems
        # ext4 tested only so far
        for fs in glob.glob('/proc/fs/*'):
            file_system = os.path.basename(fs)
            for dev in glob.glob('/proc/fs/' + file_system + '/*'):
                device = os.path.basename(dev)
                if device == partition_short_name:
                    return file_system

        return ''

    @staticmethod
    def get_partition_uuid(partition_short_name):
        '''
        Function
        =========
        get_partition_uuid

        Description
        ==============
        Get partition UUID.

        Arguments
        ==========
        __in <string> partition_short_name: partition short name
            without /dev/

        Return Values
        ==============
        __out <string> UUID
        '''

        for sym_link in glob.glob('/dev/disk/by-uuid/*'):
            if partition_short_name == os.readlink(sym_link).split('/')[2]:
                return sym_link.replace('/dev/disk/by-uuid/', '')

        return ''

    @staticmethod
    def get_partition_mount_point(partition_short_name, device_mapper_name=''):
        '''
        Function
        =========
        get_partition_mount_point

        Description
        ==============
        Get partition mount point.

        Arguments
        ==========
        __in <string> partition_short_name: partition short name
            without /dev/

        Return Values
        ==============
        __out <string> mount point
        '''

        partition_path = '/dev/' + partition_short_name
        if device_mapper_name != '':
            partition_path = '/dev/mapper/' + device_mapper_name

        if os.path.exists(partition_path):
            f = SystemUtil.PROC_READER.open_process_file('/proc/mounts')
            if f != None: 
                for line in f.readlines():
                    current_device = (line.replace('\f', '').replace('\r', '').replace('\n', '')).split()
                    if partition_path == current_device[0]:
                        return current_device[1]
            
        return ''

    @staticmethod
    def get_partition_block_size_bytes(partition_path):
        '''
        Function
        =========
        get_partition_available_space_bytes

        Description
        ==============
        Get partition block size in bytes.

        Arguments
        ==========
        __in <string> partition_path: partition full path (/dev/[*])

        Return Values
        ==============
        __out <int> partition block size in bytes
        '''

        if os.path.exists(partition_path):
            try:
                stat_vfs = os.statvfs(partition_path)
                return int(stat_vfs.f_bsize)
            except (IOError, ValueError):
                logger.exception('Exception while getting partition bock size. Return default block size.')
                return DiskUtil.DEFAULT_PARTITION_BLOCK_SIZE

        return 0

    @staticmethod
    def get_partition_size_bytes(partition_short_name):
        '''
        Function
        =========
        get_partition_size_bytes

        Description
        ==============
        Get partition size in bytes.

        /proc/partitions
              Contains the major and minor numbers of each partition as well
              as the number of 1024-byte blocks and the partition name.

        Reference:
            http://man7.org/linux/man-pages/man5/proc.5.html

        Arguments
        ==========
        __in <string> partition_short_name: partition short name
            without /dev/

        Return Values
        ==============
        __out <int> partition size in bytes
        '''

        partitions_file = [line.split() for line in open('/proc/partitions', 'r').readlines()][2:]
        for line in partitions_file:
            if line[3] == partition_short_name:
                return int(line[2])

        return 0

    @staticmethod
    def get_partition_available_space_bytes(mount_point_path):
        '''
        Function
        =========
        get_partition_available_space_bytes

        Description
        ==============
        Get partition available space in bytes.
        That excludes reserved space.

        f_bsize;    /* Filesystem block size */
        f_bavail;   /* Number of free blocks for
                        unprivileged users */

        Reference:
            man statvfs

        Arguments
        ==========
        __in <string> mount_point_path: moint pount full path

        Return Values
        ==============
        __out <int> partition availabe size in bytes
        '''

        if os.path.exists(mount_point_path):
            try:
                if common.PY3:
                    stat_vfs = os.statvfs(mount_point_path)
                else:
                    stat_vfs = os.statvfs(mount_point_path.encode(sys.getfilesystemencoding()))
                return int(stat_vfs.f_bsize * stat_vfs.f_bavail)
            except (IOError, ValueError):
                logger.exception("Exception while getting partition available space.")

        return 0

    @staticmethod
    def get_partition_free_space_bytes(mount_point_path):
        '''
        Function
        =========
        get_partition_free_space_bytes

        Description
        ==============
        Get partition free space in bytes.
        Includes reserved space.

        f_bsize;    /* Filesystem block size */
        f_bfree;    /* Number of free blocks */

        Reference:
            man statvfs

        Arguments
        ==========
        __in <string> mount_point_path: moint pount full path

        Return Values
        ==============
        __out <int> partition free size in bytes
        '''

        if os.path.exists(mount_point_path):
            try:
                if common.PY3:
                    stat_vfs = os.statvfs(mount_point_path)
                else:
                    stat_vfs = os.statvfs(mount_point_path.encode(sys.getfilesystemencoding()))
                return int(stat_vfs.f_bsize * stat_vfs.f_bfree)
            except (IOError, ValueError):
                logger.exception("Exception while getting partition free space.")

        return 0

    @staticmethod
    def get_partition_number_of_files(mount_point_path):
        '''
        Function
        =========
        get_partition_number_of_files

        Description
        ==============
        Get number of files located on the partition.

        Arguments
        ==========
        __in <string> mount_point_path: moint pount full path

        Return Values
        ==============
        __out <int> number of files
        '''

        if os.path.exists(mount_point_path):
            try:
                if common.PY3:
                    stat_vfs = os.statvfs(mount_point_path)
                else:
                    stat_vfs = os.statvfs(mount_point_path.encode(sys.getfilesystemencoding()))
                return int(stat_vfs.f_files)
            except (IOError, ValueError):
                logger.exception("Exception while getting partition number of files.")

        return 0

    @staticmethod
    def get_azure_temporary_drive():
        '''
        Function
        =========
        get_temporary_drive

        Description
        ==============
        Get temporary disk.

        Arguments
        ==========
        N/A

        Return Values
        ==============
        __out <string> temporary_disk: full block device name (/dev/sd?)
        '''

        temporary_disk = None

        for device in glob.glob('/sys/block/*'):
            for pattern in DiskUtil.VALID_BLOCK_DEVICES_PATTERN:
                if re.compile(pattern).match(os.path.basename(device)):
                    scsi_id = DiskUtil.get_device_scsi_id(os.path.basename(device))
                    if str(scsi_id['SCSIBus'])+':'+str(scsi_id['SCSIPort'])+':'+str(scsi_id['SCSITargetId'])+':'+str(scsi_id['SCSILun']) == '3:0:1:0' or str(scsi_id['SCSIBus'])+':'+str(scsi_id['SCSIPort'])+':'+str(scsi_id['SCSITargetId'])+':'+str(scsi_id['SCSILun']) == '1:0:1:0' or str(scsi_id['SCSIBus'])+':'+str(scsi_id['SCSIPort'])+':'+str(scsi_id['SCSITargetId'])+':'+str(scsi_id['SCSILun']) == '4:0:1:0':
                        temporary_disk = '/dev/%s' % os.path.basename(device)
                        break

        return temporary_disk


class MemoryUtil():
    '''
    Class
    =========
    MemoryUtil

    Description
    ==============
    Class holding all methods used to get memory
    information and statistics.
    '''

    @staticmethod
    def total_physical_memory_kbytes():
        '''
        Function
        =========
        total_physical_memory_kbytes

        Description
        ==============
        Returns the total physical memory available in the
        system in Kilobytes.
        Should be a cached property in the future.

        Arguments
        ==========
        N/A

        Return Values
        ==============
        __out <long> total physical memory seen by the system
        '''

        f = SystemUtil.PROC_READER.open_process_file('/proc/meminfo')
        if f != None:
            for line in f:
                if line and line.strip().split(': ')[0] == 'MemTotal':
                    return int(line.split()[1])
        
        return 0

    @staticmethod
    def total_swap_memory_kbytes():
        '''
        Function
        =========
        total_swap_memory_kbytes

        Description
        ==============
        Returns the total swap memory in kilobytes.
        Should be a cached property in the future.

        Arguments
        ==========
        N/A

        Return Values
        ==============
        __out <long> total virtual memory configured in the system
        '''

        cumulative_swap_size = 0

        f = SystemUtil.PROC_READER.open_process_file('/proc/swaps')
        if f != None: 
            for line in f:
                if line.split()[1] in ['file', 'partition']:
                    cumulative_swap_size += int(line.split()[2])

        return cumulative_swap_size

    @staticmethod
    def total_physical_and_virtual_memory_kbytes():
        '''
        Function
        =========
        total_physical_and_virtual_memory_kbytes

        Description
        ==============
        Returns the total physical + total
        swap size in kilobytes.
        Should be a cached property in the future.

        Arguments
        ==========
        N/A

        Return Values
        ==============
        __out <long> total physical and virtual
                     memory available in the system
        '''

        return MemoryUtil.total_physical_memory_kbytes() + MemoryUtil.total_swap_memory_kbytes()

    @staticmethod
    def available_physical_memory_kbytes():
        '''
        Function
        =========
        available_physical_memory_kbytes

        Description
        ==============
        Returns the available physical memory
        available in the system in kilobytes.

        Fr systems on which MemAVailable is not available
        in /proc/meminfo:
        MemAvailable = MemFree - LowWaterMark
            + (PageCache - min(PageCache / 2, LowWaterMark))

        Reference:
            https://git.kernel.org/pub/scm/linux/kernel/git/torvalds/linux.git/commit/?id=34e431b0ae398fc54ea69ff85ec700722c9da773

        Arguments
        ==========
        N/A

        Return Values
        ==============
        __out <long> available physical memory seen
                     by the system
        '''

        mem_free = 0
        page_cache = 0
        sreclaimable = 0

        meminfo = SystemUtil.PROC_READER.open_process_file('/proc/meminfo')
        if meminfo != None:
            for line in meminfo:
                if line and line.strip().split(': ')[0] == 'MemAvailable':
                    return int(line.split()[1])
                if line and line.strip().split(': ')[0] == 'MemFree':
                    mem_free = int(line.split()[1])
                if line and line.strip().split(': ')[0] == 'Active(file)':
                    page_cache += int(line.split()[1])
                if line and line.strip().split(': ')[0] == 'Inactive(file)':
                    page_cache += int(line.split()[1])
                if line and line.strip().split(': ')[0] == 'SReclaimable':
                    sreclaimable = int(line.split()[1])

        # Fallback to calculate available memory for systems
        # with no MemAvailable in /proc/meminfo
        low_watermark = 0
        zoneinfo = SystemUtil.PROC_READER.open_process_file('/proc/zoneinfo')
        if zoneinfo != None:
            for line in zoneinfo:
                if line and line.strip().split()[0] == 'low':
                    low_watermark += int(line.strip().split()[1])

        available_memory = mem_free - low_watermark
        available_memory += page_cache - min(page_cache / 2, low_watermark)
        available_memory += sreclaimable - min(sreclaimable / 2, low_watermark)

        if available_memory < 0:
            available_memory = 0

        return available_memory

    @staticmethod
    def free_physical_memory_kbytes():
        '''
        Function
        =========
        free_physical_memory_kbytes

        Description
        ==============
        Returns the total physical memory free in the
        system.

        Arguments
        ==========
        N/A

        Return Values
        ==============
        __out <long> total free physical memory seen
                     by the system
        '''

        f = SystemUtil.PROC_READER.open_process_file('/proc/meminfo')
        if f != None: 
            for line in f:
                if line and line.strip().split(': ')[0] == 'MemFree':
                    return int(line.split()[1])

        return 0

    @staticmethod
    def available_swap_memory_kbytes():
        '''
        Function
        =========
        available_swap_memory_kbytes

        Description
        ==============
        Returns the total swap memory available
        in kilobytes.

        Arguments
        ==========
        N/A

        Return Values
        ==============
        __out <long> total swap memory
                     available in the system
        '''

        cumulative_swap_free = 0

        f = SystemUtil.PROC_READER.open_process_file('/proc/swaps')
        if f != None: 
            for line in f:
                if line.split()[1] in ['file', 'partition']:
                    cumulative_swap_free += int(line.split()[2]) - int(line.split()[3])

        return cumulative_swap_free
