import logging
import os
import shutil
import subprocess

import six
import sys

import util.common as common

logger = logging.getLogger()

_download_directory = os.path.join(common.current_working_folder, 'perfInsightsDownloads')
_download_file_path = os.path.join(_download_directory, 'perfinsightslatest.tar.gz')
_pil_latest_dir = os.path.join(_download_directory, 'perfinsightslatest')


def update_perf_insights():
    logger.info('Entering auto update.')
    download_url = 'https://aka.ms/perfinsightslinuxlatest'
    ret = False
    try:
        full_download_url = _get_full_download_url(download_url)

        if full_download_url and _newer_version_available(full_download_url):
            logger.info('Starting update to latest version.')
            _download_latest(full_download_url, _download_file_path)
            _uncompress_latest()
            ret = _run_perfinsights_latest()
    except:
        logger.exception('An error was encountered updating to the latest version of PerfInsights.')

    logger.info('Leaving auto update')
    return ret


def _get_full_download_url(download_url):
    try:
        full_url = six.moves.urllib.request.urlopen(download_url).geturl()
        logger.info('Retrieved full download url from download center: %s' % full_url)
        return full_url
    except:
        logger.exception(
            'An exception occurred while retrieving the download url for the latest version of PerfInsights. Will continue running existing version %s' % common.SCRIPT_VERSION)
        return None


def _newer_version_available(download_url):
    new_version_available = False
    if download_url.__contains__('version='):
        latest_version = download_url.split('version=')[-1]
    else:
        logger.warning('The url for downloading the latest version of PerfInsights has unexpected formatting. Will continue running existing version %s' % common.SCRIPT_VERSION)
        return new_version_available

    new_version_available = latest_version > common.SCRIPT_VERSION
    if new_version_available:
        logger.info("A new version of PerfInsights is available. Updating to version %s" % latest_version)
    else:
        logger.info('The latest version of PerfInsights is running.')
    return new_version_available


def _download_latest(download_url, download_file_path):
    logger.info('Downloading the latest version of PerfInsights')
    if os.path.isdir(_download_directory):
        shutil.rmtree(_download_directory)

    os.mkdir(_download_directory)
    six.moves.urllib.request.urlretrieve(download_url, download_file_path)


def _uncompress_latest():
    logger.info("Extracting the latest version of PerfInsights")
    try:
        if os.path.isdir(_pil_latest_dir):
            shutil.rmtree(_pil_latest_dir)

        os.mkdir(_pil_latest_dir)

        extraction_success_status = common.uncompress_file(_download_file_path, _pil_latest_dir, 'Extracting latest version of perfinsights')
        if extraction_success_status == 0:
            return
        else:
            raise RuntimeError('Error encountered extracting %s' % _pil_latest_dir)
    except:
        shutil.rmtree(_pil_latest_dir)
        six.reraise(*sys.exc_info())


def _run_perfinsights_latest():
    logger.info('Running the latest version of PerfInsights from %s' % _pil_latest_dir)
    try:
        # Create arguments list for running the new perfinsights.py
        proc_args_list = [sys.executable, 'perfinsights.py', '-t', common.storage_account, '-k', common.storage_account_key, '-i', common.resource_id, '--skip-auto-update']
        # Skip key in script_arguments as it will overwrite the actual key with **REDACTED**
        skip_args_list = [u'-k', u'**REDACTED**']

        # Add any exisitng arguments to the argument list
        for a in common.script_arguments.split(' '):
            if a in skip_args_list:
                continue
            proc_args_list.append(a)

        p = subprocess.Popen(proc_args_list, stdin=subprocess.PIPE, stdout=subprocess.PIPE, universal_newlines=True, cwd=_pil_latest_dir)
        logger.info(p.communicate()[0])
        ret = p.returncode
        if ret is None or ret == 0:
            return True
        else:
            logger.warning('The latest version of PerfInsights exited with a code of %s' % ret)
            return False
    except:
        logger.exception('An error was encountered running the latest version of PerfInsights')
        return False

def recursive_overwrite(src, dest, ignore=None):
    # Simple function to recursively overwrite the src directory to dest directory to used instead of shutil.copytree
    # Optional ignore argument should be a callable that takes in src directory and list of files to generate a set of ignored files
    # similar to how ignore argument of shutil.copytree works (Search for examples of shutil.copytree for more information)
    if os.path.isdir(src):
        if not os.path.isdir(dest):
            os.makedirs(dest)

        files = os.listdir(src)

        ignored = set()
        if ignore is not None:
            ignored = ignore(src, files)

        for f in files:
            if f not in ignored:
                recursive_overwrite(os.path.join(src, f), os.path.join(dest, f), ignore)
    else:
        shutil.copyfile(src, dest)

def replace_perfinsights_version():
    # back up existing PerfInsights version
    parent_dir = os.path.dirname(common.current_working_folder)
    backup_dir = os.path.join(parent_dir, 'perfInsightsBackup')
    logger.info('The current version of PerfInsights will be backed up to %s' % backup_dir)
    try:
        if os.path.isdir(backup_dir):
            shutil.rmtree(backup_dir)

        shutil.copytree(common.current_working_folder, backup_dir)
    except:
        logger.exception('An error occurred while backing up the currently installed version of PerfInsights, and the currently installed version will not be replaced by the latest version.')

    # replace existing version with latest downloaded version
    logger.info('Replacing the current directory with %s' % _pil_latest_dir)
    try:
        #we cannot use shutil.copytree here because it requires the dest directory to not exist before copy (dirs_exist_ok parameter is not added until 3.8)
        recursive_overwrite(_pil_latest_dir, common.current_working_folder)
    except:
        logger.exception(
            'Failed to update binaries under ' + common.current_working_folder + '. Please extract binaries from ' + backup_dir + ' and use them in next run.')

