#-------------------------------------------------------------------------
# Copyright (c) Microsoft.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#--------------------------------------------------------------------------
import sys
import os
import json
import datetime
import tarfile
import shutil
import time
import platform
import re
import logging

logger = logging.getLogger(__name__)

from lib import applicationinsights

# STATIC VARIABLES
SCRIPT_VERSION = "0.3.4.32"

GREEN = '\033[32m'  #green color
YELLOW = '\033[33m' #yellow color
RED = '\033[31m'    #red color
RESET = '\033[m'    #reset color

PY3 = sys.version_info[0] == 3
MINIMUM_SUPPORTED_PYTHON_VERSION_FOR_TELEMETRY = 2.7
MINIMUM_SUPPORTED_PYTHON_VERSION = 2.6
PIL_SUPPORTED_LINUX_DISTRIBUTIONS = ['CentOS', 'CentOS Linux', 'Ubuntu', 'Red Hat Enterprise Linux', 'Red Hat Enterprise Linux Server', 'Oracle Linux', 'Oracle Linux Server', 'Oracle Linux Server ', 'Debian', 'Debian GNU/Linux', 'SLES', 'SUSE Linux Enterprise', 'AlmaLinux', 'Common Base Linux Mariner']

PIFL_LOG_FILE_NAME = "perfinsights.log"
PIFL_VM_SIZE_LOOKUP_FILE_NAME = 'vmlookup.csv'
PIFL_RULE_CONFIG_FILE = 'RuleEngineConfig.json'
PIFL_DIAG_OUTPUT_FOLDER_NAME = "/output"
PIFL_LOG_FOLDER_NAME = "/logs"
PIFL_REPORT_COMPRESSED_FILENAME = "PerfInsights_Report.tar.gz"
PIFL_PERFDIAG_FILENAME = "perfdiag.js"
PIFL_GUEST_LOG_FOLDER_NAME = "/guestosfiles"

RULEENGINE_RULES = None
DEFAULT_RUN_DURATION_SEC = 60
DEFAULT_CAPTURE_INTERVAL_SEC = 5

POSSIBLE_SCENARIOS = [ \
    {'Name': 'quick','DisplayName': 'Quick Performance Analysis', 'Implemented': 'Available'},
    {'Name': 'vmslow','DisplayName': 'Slow VM Performance Analysis', 'Implemented': 'Available'},
    {'Name': 'hpc','DisplayName': 'HPC Performance Analysis', 'Implemented': 'Available'},
    ]
SCRIPT_SWITCHES =[\
   {'Short': '-h', 'Long': '--help', 'Text': 'Show help' , 'Type': 'Optional', 'Sample': ''}, \
   {'Short': '-s', 'Long': '--support-request', 'Text': 'Microsoft support request #.' , 'Type': 'Optional', 'Sample': '[Microsoft support case #]'}, \
   {'Short': '-r', 'Long': '--run', 'Text': 'Scenario to run.' , 'Type': 'Required', 'Sample': POSSIBLE_SCENARIOS}, \
   {'Short': '-o', 'Long': '--output-folder', 'Text': 'Folder in which temporary data and report will be generated (default is current path).'  , 'Type': 'Optional', 'Sample': '[Path to output]'}, \
   {'Short': '-d', 'Long': '--duration', 'Text': 'Capture duration. Can be in hours (H), minutes (M) or seconds (S).' , 'Type': 'Optional', 'Sample': '[xH|xM|xS]'}, \
   {'Short': '-a', 'Long': '--accept-disclaimer-and-share-diagnostics', 'Text': 'Accept EULA, bypassing prompt.' , 'Type': 'Optional', 'Sample': ''}, \
   {'Short': '-q', 'Long': '--bypass-questions', 'Text': 'Answer \'yes\' to all questions.' , 'Type': 'Optional', 'Sample': ''}, \
   {'Short': '-t', 'Long': '--storage-account', 'Text': 'Name of the storage account.' , 'Type': 'Optional', 'Sample': ''}, \
   {'Short': '-k', 'Long': '--storage-account-key', 'Text': 'Key for the storage account' , 'Type': 'Optional', 'Sample': ''}, \
   {'Short': '-i', 'Long': '--resource-id', 'Text': 'Full Resource Uri for the VM' , 'Type': 'Optional', 'Sample': '/subscriptions/<subcriptionid>/resourceGroups/<resourceGroupName>/providers/Microsoft.Compute/virtualMachines/<vmName>'}, \
   {'Short': '-u', 'Long': '--skip-auto-update', 'Text': 'Skip auto update and run the currently installed version of PerfInsights', 'Type': 'Optional', 'Sample': ''}
 ] 
GPU_VM_SIZES = [ \
    {'Name': 'Standard_NC6'}, {'Name': 'Standard_NC12'}, {'Name': 'Standard_NC24'}, {'Name': 'Standard_NC24r'}, \
    {'Name': 'Standard_NC6s_v2'}, {'Name': 'Standard_NC12s_v2'}, {'Name': 'Standard_NC24s_v2'}, {'Name': 'Standard_NC24rs_v2'}, \
    {'Name': 'Standard_NC6s_v3'}, {'Name': 'Standard_NC12s_v3'}, {'Name': 'Standard_NC24s_v3'}, {'Name': 'Standard_NC24rs_v3'}, \
    {'Name': 'Standard_ND6s'}, {'Name': 'Standard_ND12s'}, {'Name': 'Standard_ND24s'}, {'Name': 'Standard_ND24rs'}, \
    {'Name': 'Standard_NV6'}, {'Name': 'Standard_NV12'}, {'Name': 'Standard_NV24'} \
    ]

REPORT_ID = ""
PUBLIC_APPINSIGHTS_KEY = "9561b95d-d813-42d5-98a0-3177b3290094"
PUBLIC_APPINSIGHTS_ENDPOINT = 'https://dc.services.visualstudio.com/v2/track'
PUBLIC_DEBUG_APPINSIGHTS_KEY = "3faa390c-b75b-46a4-b31a-f36e5ebb6719" #Telemetry key for debug scenario

GOV_CLOUD_APPINSIGHTS_KEY = "467f482c-20f9-fa17-99c2-5251d19f6114"
GOVCLOUD_APPINSIGHTS_ENDPOINT = "https://dc.applicationinsights.us/v2/track"

USNAT_CLOUD_APPINSIGHTS_KEY = "a24fac76-a4cc-4435-9de4-0a212da7a8bd"
USNAT_APPINSIGHTS_ENDPOINT = "https://dc.applicationinsights.azure.eaglex.ic.gov/v2/track"

USSEC_CLOUD_APPINSIGHTS_KEY = "ea44baa5-8065-47be-91a6-f82415f7046b"
USSEC_APPINSIGHTS_ENDPOINT = "https://dc.applicationinsights.azure.microsoft.scloud/v2/track"

# Storage end points
PUBLIC_CLOUD_SERVICE_HOST_BASE = '.core.windows.net'  #AzurePublicCloud
GOV_CLOUD_SERVICE_HOST_BASE = '.core.usgovcloudapi.net'  #AzureUSGovernmentCloud
GERMAN_CLOUD_SERVICE_HOST_BASE = '.core.microsoftazure.de'  #AzureGermanCloud
CHINA_CLOUD__SERVICE_HOST_BASE = '.core.chinacloudapi.cn'  #AzureChinaCloud
USSEC_CLOUD_SERVICE_HOST_BASE = '.core.microsoft.scloud'  #USSec
USNAT_CLOUD_SERVICE_HOST_BASE = '.core.eaglex.ic.gov'  #USNat

# MODULES SHARED VARIABLES
current_working_folder = ''
perfinsights_script_folder = ''
vm_sizes_lookup_file_path = ''
output_folder = ""
diag_output_folder_path = ""
shared_guestfiles_output_folder_path = ""
guestlogs_output_folder_path = ""
log_folder_path = ""
pifl_log_file_path = ""

user_agrees_to_share_data = False
debug_enabled = False
feature_debug = None
collect_gpu = False
bypass_agreements = False
bypass_questions = None

script_arguments = ""
support_request_number = ""
perfinsights_scenario = ""
capture_duration_sec = DEFAULT_RUN_DURATION_SEC
capture_interval = DEFAULT_CAPTURE_INTERVAL_SEC
cpu_stat_capture_interval_secs = DEFAULT_CAPTURE_INTERVAL_SEC
memory_stat_capture_interval_secs = DEFAULT_CAPTURE_INTERVAL_SEC
disk_stat_capture_interval_secs = DEFAULT_CAPTURE_INTERVAL_SEC

thisazurevm = None
linux_distribution = ""
linux_version = ""
linux_version_id = ""
linux_id_like = ""
linux_id = ""
execution_start_time = None
execution_end_time = None
perfinsights_execution_time = None

temp_drive = ''
total_physical_memory = 0
available_physical_memory = 0
free_physical_memory = 0
physical_processors_count = 0

# Azure Storage
EXTENSION_RESULT_CONTAINER_NAME = "azdiagextnresults"
EXTENSION_RESULT_RUNS_TABLE_NAME = "AzPerfDiagRunsTable"
EXTENSION_RESULT_FINDINGS_TABLE_NAME = "AzPerfDiagFindingsTable"
EXTENSION_RESULT_IMPACTED_RESOURCES_TABLE_NAME = "AzPerfDiagImpactedResourcesTable"
storage_account = ""
storage_account_key = ""
resource_id = ""
result_blob_filename = ""
table_service = None
block_blob_service = None
upload_to_storage = False

# Auto Update
auto_update = True

_missing = object()

class CachedProperty(property):

    def __init__(self, func, name = None, doc = None):
        self.__name__ = name or func.__name__
        self.__module__ = func.__module__
        self.__doc__ = doc or func.__doc__
        self._func = func

    def __set__(self, obj, value):
        obj.__dict__[self.__name__] = value

    def __get__(self, obj, cls = None):
        if obj is None:
            return self

        value = obj.__dict__.get(self.__name__, _missing)
        if value is _missing:
            value = self._func(obj)
            obj.__dict__[self.__name__] = value
        return value
        
def show_help():
    """
    Function
    =========
    ShowHelp

    Description
    ==============
    Display PerfInsights script help and usage.

    Arguments
    ==========
        N/A
    
    Return Values
    ==============

        N/A
    """

    print('')
    allCommands = 'perfinsights'
    for scrSwitch in SCRIPT_SWITCHES:
        text = ''
        if scrSwitch['Sample'] != '' and type(scrSwitch['Sample']) is list:
            for sample in scrSwitch['Sample']:
                text = text + sample['Name'] + '|'
            text = text[:-1] 
        else:
            text = scrSwitch['Sample']
        allCommands = allCommands + ' ' + scrSwitch['Short'] + '|' + scrSwitch['Long']
        if text != '':
            allCommands = allCommands + ' ' + text
        if scrSwitch['Sample'] == '':
            allCommands = allCommands + ' ' + scrSwitch['Sample']

    print(allCommands)
    print('')
    print('Version: ' + str(SCRIPT_VERSION))
    print('')
    print('Collect general information and resources statistics on Linux systems.')
    print('')
    print ('Scenarios:')
    print ('')
    for scenario in POSSIBLE_SCENARIOS:
        if scenario['Implemented'] == 'Available':
            print(' * ' + scenario['DisplayName'] + ' (' + scenario['Name'] + ')')
        else:
            print(' * ' + scenario['DisplayName'] + ' (' + scenario['Name'] + ') - ' + scenario['Implemented'])
    print('')
    print('Options:')
    print('')
    for scrSwitch in SCRIPT_SWITCHES:
        print(' ' + scrSwitch['Short'] + '/' + scrSwitch['Long'] + ' : ' + scrSwitch['Text'] )
    print('')
    
    print ("Examples:")
    print ("python perfinsights.py -r quick -d 10s -a")
    
def get_current_datetime(arg_purpose):
    """
    Function
    =========
    GetCurrentDateTime

    Description
    ==============
    Get the current date and time.

    Arguments
    ==========
        __in <str> arg_purpose : display type: 'display' for screen outout, 'file' for file naming usage (without special characters)
    
    Return Values
    ==============

        __out <str> _dateTime : Date and time formated string
    """
    
    _dateTime = ""

    if arg_purpose == "epoch":
        _dateTime = get_epoch_from_tme(datetime.datetime.utcnow().strftime('%m/%d/%Y %H:%M:%S.%f') , '%m/%d/%Y %H:%M:%S.%f')
    if arg_purpose == "display":
        _dateTime = datetime.datetime.utcnow().strftime('%m/%d/%Y %H:%M:%S')
    if arg_purpose == "timestamp":
        _dateTime = datetime.datetime.utcnow().strftime('%m/%d/%Y %H:%M:%S.%f')
    if arg_purpose == "file":
        _dateTime = datetime.datetime.utcnow().strftime('%Y-%m-%d_%H-%M-%S-%f')[:-3]
    if arg_purpose == "charting":
        _dateTime = datetime.datetime.utcnow().strftime('%Y-%m-%d %H-%M-%S-%f')[:-3]

    return _dateTime

def open_as_textfile(arg_fileName, **kwargs):
    """
    Function
    =========
    OpenTextFileRealTime

    Description
    ==============
    On Python 3 opens a file in text mode by using fs encoding and a proper en/decoding errors handler.
    On Python 2 this is just an alias for open(name, 'rt').

    Arguments
    ==========
        __in <str> arg_fileName : full file path to read
        __in <str> kwargs : default arguments
    
    Return Values
    ==============

        __out <file> : an open file
    """

    if PY3:
        # See:
        # https://github.com/giampaolo/psutil/issues/675
        # https://github.com/giampaolo/psutil/pull/733
        ENCODING = sys.getfilesystemencoding()
        kwargs.setdefault('encoding', ENCODING)
        kwargs.setdefault('errors', get_encoding_errors_handler())

    return open(arg_fileName, "rt", **kwargs)

def get_encoding_errors_handler():
    return sys.modules['psutil'].ENCODING_ERRORS_HANDLER

def open_as_binaryfile(fname, **kwargs):

    return open(fname, "rb", **kwargs)

def open_as_csvfile(arg_fileName, arg_delimiter):
    """
    Function
    =========
    OpenCSVFile

    Description
    ==============
    Read a file with delimiters and return a string ready to be used as an array.

    Arguments
    ==========
        __in <str> arg_fileName : full file path to read
        __in <str> arg_delimiter : delimiter to use to split each line
    
    Return Values
    ==============

        __out <str> _csvContent : string with parsed text
    """

    _csvContent = None

    try:
        with open(arg_fileName, "r") as _csvFile:
            _csvContent = [_line.split(arg_delimiter) for _line in _csvFile]

        return _csvContent

    except IOError as e:
        print("I/O error({0}) opening " + arg_fileName + ": {0}-{1}".format(e.errno, e.strerror))
        
def output_findings_json(arg_jsonContent, arg_outputFilePath, arg_BOF, arg_EOF):
    """
    Function
    =========
    OutputFindingsJS

    Description
    ==============
    Output JSON-like content to final output file.

    Arguments
    ==========
        __in <str> arg_jsonContent : JSON-like content
        __in <str> arg_outputFilePath : full file path tp the output file
        __in <bool> arg_BOF : beginning of the file (first content added)
        __in <bool> arg_EOF : end of the file (last content added)
    
    Return Values
    ==============

        N/A
    """

    if arg_BOF == True:
        if arg_jsonContent != "":
            _outputfile = open(arg_outputFilePath , "w")
            _outputfile.write("var perfDiagObj={")
            _outputfile.write(arg_jsonContent)
    elif arg_EOF == True:
        _outputfile = open(arg_outputFilePath , "a")
        if arg_jsonContent != "":
            _outputfile.write(",\n")
            _outputfile.write(arg_jsonContent)
            _outputfile.write("};")
        else:
            _outputfile.write("};")
    else:
        if arg_jsonContent != "":
            _outputfile = open(arg_outputFilePath , "a")
            _outputfile.write(",")
            _outputfile.write(arg_jsonContent)
           
    _outputfile.close()

def create_datacollector_json(argdatacollectorname, argjsondict, argstarttime, argendtime):
    """
    Function
    =========
    create_datacollector_json

    Description
    ==============
    Used to encapsulate Data Collectors data in their own JSON schema.

    Arguments
    ==========
        __in <str> argdatacollectorname : Data collector name
        __in <dict> argjsondict : JSON dictionnay containing the DC data
        __in <str> argstarttime : collection start time
        __in <str> argendtime : collection end time
    
    Return Values
    ==============
        __out <dict> datacollectorjson : Data collector well formatted data
    """

    datacollectorjson = {}

    create_json(datacollectorjson, "DataCollectorName", argdatacollectorname)
    create_json(datacollectorjson, "CollectedData", [{"StartTime": argstarttime, "EndTime": argendtime, "Data": argjsondict}])

    return datacollectorjson

def create_json(arg_jsondict, _key=None, _value=None):
    """
    """

    try:
        arg_jsondict[_key] = _value
    except Exception as e:
        return e.message

def create_final_json(arg_jsondict):
    try:
        return json.dumps(arg_jsondict,skipkeys=False,indent=4,sort_keys=True)
    except UnicodeDecodeError:
        return json.dumps(arg_jsondict,skipkeys=False,indent=4,sort_keys=True,ensure_ascii=False)

def compress_file(arg_folderPath, arg_fileName, arg_comment="Files have been compressed."):
    """
    Function
    =========
    CompressFile

    Description
    ==============
    Compress a given folder to a given file. In case an exception is encountered, will retry two times.

    Arguments
    ==========
        __in <str> arg_folderPath : full path to the folder to compress
        __in <str> _arg_fileName : full path to the compressed file
    
    Return Values
    ==============

        N/A
    """

    _errExit = 0

    logger.debug("Entering compress_file")
    logger.debug("Compressing " + arg_folderPath + " to " + os.path.basename(arg_fileName) + "...")

    if os.path.exists(arg_folderPath):
        attempts_left = 2 # file compression attempts will allow one initial attempt, plus two retries
        while (attempts_left >= 0):
            try:
                _tar = tarfile.open(arg_fileName, "w:gz")
                _tar.add(arg_folderPath, arcname=os.path.sep)
                _tar.close()

                logger.info(arg_comment)
                break
            except Exception:
                logger.exception('Exception encountered while compressing report output file.')
                logger.info('Report file compression retries left: %s' % attempts_left)
                if attempts_left > 0 and os.path.exists(arg_fileName):
                    os.remove(arg_fileName)
                    time.sleep(1)
                    logger.info('Retrying report file compression.')
            attempts_left-=1
        else: 
            logger.error("Compressing " + arg_folderPath + " to " + os.path.basename(arg_fileName) + " failed.")
            _errExit = 2
    else:
        logger.error("Folder" + arg_folderPath + " does not exists.")
        _errExit = 2

    if(_errExit == 2):
        logger.error("Script execution will continue but there may have been some unexpected errors.")
    
    logger.debug("Leaving compress_file")

    return _errExit

def uncompress_file(arg_compressedFileName, arg_targetFolder, arg_comment):

    """
    Function
    =========
    UncompressFile

    Description
    ==============
    Uncompress a given tar.gz file to a given directory.

    Arguments
    ==========
        __in <str> arg_compressedFileName : full path to the compressed file
        __in <str> arg_targetFolder : full path to the folder in which file will be uncompressed
        __in <str> arg_comment : comment to display with LogOutput
    
    Return Values
    ==============

        N/A
    """

    _errExit = 0
    _textToDisplay = arg_compressedFileName

    if arg_comment != "":
        _textToDisplay = arg_comment

    logger.debug("Entering uncompress_file")
    logger.debug("Uncompressing " + os.path.basename(arg_compressedFileName) + " to " + arg_targetFolder + "...")

    if os.path.exists(arg_compressedFileName):
        if os.path.exists(arg_targetFolder):
            _tar = tarfile.open(arg_compressedFileName)
            _tar.extractall(arg_targetFolder)
            _tar.close()

            logger.debug(_textToDisplay)
        else:
            logger.error("Target folder" + arg_targetFolder + " does not exists.")
            logger.error("Script execution will continue but there may have been some unexpected errors.")
            _errExit = 2
    else:
        logger.error(arg_compressedFileName + " does not exists.")
        logger.error("Script execution will continue but there may have been some unexpected errors.")
        _errExit = 2

    logger.debug("Leaving UncompressFile")

    return _errExit

def get_scenario_displayname(arg_scenario):
    """
    Function
    =========
    get_scenario_displayname

    Description
    ==============
    Return the display name of a scenario.

    Arguments
    ==========
        __in <str> arg_scenario : short scenario name
    
    Return Values
    ==============

        __out <str> arg_scenario : scenario display name
    """

    logger.debug("Entering get_scenario_displayname")

    _scenarioDisplayName = "Undefined"

    for _possibleScenario in POSSIBLE_SCENARIOS:
        if arg_scenario in _possibleScenario['Name']:
            _scenarioDisplayName = _possibleScenario['DisplayName']
            break

    logger.debug("Leaving get_scenario_displayname")

    return _scenarioDisplayName

def copy_file_wlog(argsourcefolder, argsourcefile,argdestinationfolder):
    """
    Function
    =========
    copy_file_wlog

    Description
    ==============
    Copy file and log result.

    Arguments
    ==========
        __in <str> argsourcefolder : source folder path
        __in <str> argsourcefile : source file name
        __in <str> argdestinationfolder : destination folder path

    Return Values
    ==============

        __out <int> result :
                            0 = success
                            2 = source folder not found
                            2 = source file not found
                            3 = destination folder not found
                            4 = copy failed
    """

    result = 0

    if os.path.exists(argsourcefolder):
        if os.path.exists(argsourcefolder + "/" + argsourcefile):
            if os.path.exists(argdestinationfolder):
                try:
                    shutil.copyfile(argsourcefolder + "/" + argsourcefile, argdestinationfolder + "/" + argsourcefile)
                except IOError:
                    pass
                if not os.path.exists(argdestinationfolder + "/" + argsourcefile):
                    logger.warning("Failed to copy file" + argsourcefile)
                    result = 4
                else:
                    logger.debug("Successfully copied file: " + argsourcefile)
            else:
                result = 3
                logger.warning("Folder " + argdestinationfolder + " not found. Copy failed.")
        else:
            result = 2
            logger.warning("File " + argsourcefile + " not found. Copy failed.")
    else:
        result = 1
        logger.warning("Folder " + argsourcefolder + " not found. Copy failed.")

    return result

def recursivecopy_with_pattern(arg_sourcefolder, arg_destinationfolder, arg_pattern):

    if os.path.isdir(arg_sourcefolder):
        listdir=os.listdir(arg_sourcefolder)
        for item in listdir:
            if not os.path.isdir(arg_sourcefolder+'/'+item):
                if arg_pattern in item or arg_pattern == '*':
                    try:
                        shutil.copy(arg_sourcefolder+'/'+item,arg_destinationfolder+'/'+item)
                    except OSError as e:
                        logger.exception(e.__str__())
            else:
                if not os.path.exists(arg_destinationfolder+'/'+item):
                    os.makedirs(arg_destinationfolder+'/'+item)
                    recursivecopy_with_pattern(arg_sourcefolder+'/'+item,arg_destinationfolder+'/'+item, arg_pattern)
                recursivecopy_with_pattern(arg_sourcefolder+'/'+item,arg_destinationfolder+'/'+item, arg_pattern)

def is_gpu_vmsize(arg_vmsize):

    for size in GPU_VM_SIZES:
        if arg_vmsize == size['Name']:
            return True
    
    return False

def get_epoch_from_tme(strTime, strFormat):

    return (int(datetime.datetime.strptime(strTime, strFormat).strftime("%s")) * 1000)

def get_storage_endpoint_info(storage_account_name, azEnvironment):
    
    # Try parsing the storage endpoint info if passed as fully qualified storage account name.
    trimmed_storage_account_name = storage_account_name.split('.',1)[0]
    storage_base_endpoint = ("." + storage_account_name.split('.',1)[1]).replace(".blob.", ".").replace(".table.",".") if storage_account_name.find('.') != -1 else ""

    if storage_base_endpoint == "":
        if azEnvironment == "AzureUSGovernmentCloud":
            storage_base_endpoint = GOV_CLOUD_SERVICE_HOST_BASE
        elif azEnvironment == "AzureGermanCloud":
            storage_base_endpoint = GERMAN_CLOUD_SERVICE_HOST_BASE
        elif azEnvironment == "AzureChinaCloud":
            storage_base_endpoint = CHINA_CLOUD__SERVICE_HOST_BASE
        elif azEnvironment == "USSec":
            storage_base_endpoint = USSEC_CLOUD_SERVICE_HOST_BASE
        elif azEnvironment == "USNat":
            storage_base_endpoint = USNAT_CLOUD_SERVICE_HOST_BASE
        else:
            storage_base_endpoint = PUBLIC_CLOUD_SERVICE_HOST_BASE
    
    return trimmed_storage_account_name, storage_base_endpoint

def get_applicationinsights_info(azEnvironment):

    application_insights_key = ''
    application_insights_endpoint = ''

    if azEnvironment == "AzureUSGovernmentCloud":
        application_insights_key = GOV_CLOUD_APPINSIGHTS_KEY
        application_insights_endpoint = GOVCLOUD_APPINSIGHTS_ENDPOINT
    elif azEnvironment == "USNat":
        application_insights_key = USNAT_CLOUD_APPINSIGHTS_KEY
        application_insights_endpoint = USNAT_APPINSIGHTS_ENDPOINT
    elif azEnvironment == "USSec":
        application_insights_key = USSEC_CLOUD_APPINSIGHTS_KEY
        application_insights_endpoint = USSEC_APPINSIGHTS_ENDPOINT
    else:
        if debug_enabled:
            application_insights_key = PUBLIC_DEBUG_APPINSIGHTS_KEY
        else:
            application_insights_key = PUBLIC_APPINSIGHTS_KEY
        application_insights_endpoint = PUBLIC_APPINSIGHTS_ENDPOINT

    logger.info('Application Insights: '+ application_insights_key + ", Endpoint:" + application_insights_endpoint)
    return application_insights_key, application_insights_endpoint
