#-------------------------------------------------------------------------
# Copyright (c) Microsoft.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#--------------------------------------------------------------------------
import json
import base64, hashlib, hmac
import urllib
from datetime import datetime as dt
from datetime import timedelta
import sys

from util import common

class RunJson():
    def __init__(self, perf_insights_result, sr_number):
        self.partition_key = encode_partition_key(perf_insights_result['ResourceId']).lower() or ''
        self.report_id = perf_insights_result['ReportId'] or ''
        self.row_key = self.report_id.lower() or ''
        self.sr_number = sr_number or ''
        self.version = perf_insights_result['ScriptVersion'] or ''
        self.scenario = perf_insights_result['Scenario'] or ''
        self.start_time_utc = perf_insights_result['StartTimeUTC'] or ''
        self.end_time_utc = perf_insights_result['EndTimeUTC'] or ''
        self.command_line = perf_insights_result['CommandLine'] or ''
        self.blob_url = perf_insights_result['BlobUrlWithSASToken'] or (perf_insights_result['BlobUrl'] or 'N/A')

        self.perf_insights_result = perf_insights_result
        impact_map = self.gather_impact_findings()

        self.low_impact_findings = impact_map['Low']
        self.medium_impact_findings = impact_map['Medium']
        self.high_impact_findings = impact_map['High']
        
    # Go through rule results, search for impact findings, and
    # increment counters for serparate levels.
    # Return in a dict with low, medium, & high keys set to values
    def gather_impact_findings(self):
        impact_map = {
            'Low': 0,
            'Medium': 0,
            'High': 0
        }

        key = 'RuleResults'
        if key in self.perf_insights_result:
            rule_results = self.perf_insights_result[key]
            for rule_result in rule_results:
                key = 'RuleFindings'
                if key in rule_result.keys():
                    rule_findings = rule_result[key]
                    if len(rule_findings) > 0:
                        for rule_finding in rule_findings:
                            key = 'ImpactLevel'
                            if key in rule_finding.keys():
                                impact_level = rule_finding[key]
                                if impact_level in impact_map.keys():
                                    impact_map[impact_level] = impact_map[impact_level] + 1

        return impact_map

    def as_entity(self):
        entity = {}

        entity['PartitionKey'] = self.partition_key.lower()
        entity['RowKey'] = self.row_key.lower()
        entity['BlobUri'] = self.blob_url
        entity['CommandLine'] = self.command_line
        entity['EndTimeUtc'] = self.end_time_utc
        entity['HighImpactFindings'] = self.high_impact_findings
        entity['LowImpactFindings'] = self.low_impact_findings
        entity['MediumImpactFindings'] = self.medium_impact_findings
        entity['PerfDiagAssemblyFileVersion'] = ''
        entity['PerfInsightsExeFileVersion'] = ''
        entity['ResetCachedDataCollectors'] = False
        entity['ResourceUri'] = common.resource_id
        entity['Scenario'] = self.scenario
        entity['StartTimeUtc'] = self.start_time_utc
        entity['Version'] = self.version
        
        return entity

class FindingsJson():
    def __init__(self, perf_insights_result):
        self.perf_insights_result = perf_insights_result
        self.findings_json_result = self.gather_findings()

    def gather_findings(self):
        findings_json = []

        key = 'RuleResults'
        if key in self.perf_insights_result:
            rule_results = self.perf_insights_result[key]
            for rule_result in rule_results:
                key = 'RuleFindings'
                if key in rule_result.keys():
                    rule_findings = rule_result[key]
                    for rule_finding in rule_findings:
                        finding_json = FindingJson(self.perf_insights_result['ReportId'], rule_result, rule_finding)
                        findings_json.append(finding_json)

        return findings_json

    def as_entity_list(self):
        entities = []

        for finding_json in self.findings_json_result:
            entity = finding_json.as_entity()
            entities.append(entity)

        return entities

class FindingJson():
    def __init__(self, report_id, rule_result, rule_finding):
        self.report_id = report_id
        self.partition_key = self.report_id.lower()
        self.rule_id = rule_result['RuleId'] or ''
        self.id = rule_finding['Id'] or 0
        self.row_key = (self.rule_id + '-' + str(self.id)).lower() or ''
        self.rule_friendly_name = rule_result['RuleFriendlyName'] or ''
        self.rule_category = rule_result['Category'] or ''
        self.recommendation = rule_finding['Recommendation'] or ''
        self.reference_links = json.dumps(rule_finding['ReferenceLinks'])
        self.impact_level = rule_finding['ImpactLevel'] or ''
        self.finding = rule_finding['Finding'] or ''

    def as_entity(self):
        entity = {}

        entity['PartitionKey'] = self.partition_key.lower()
        entity['RowKey'] = self.row_key
        entity['Id'] = self.row_key
        entity['ImpactLevel'] = self.impact_level
        entity['RuleCategory'] = self.rule_category
        entity['RuleFriendlyName'] = self.rule_friendly_name
        entity['Finding'] = self.finding
        entity['Recommendation'] = self.recommendation
        entity['ReferenceLinks'] = self.reference_links

        return entity

class ImpactedResourcesJson():
    def __init__(self, perf_insights_result):
        self.perf_insights_result = perf_insights_result
        self.impacted_resources_result = self.gather_impacted_resources()

    def gather_impacted_resources(self):
        impacted_resources_json = []

        key = 'RuleResults'
        rule_id_key = 'RuleId'
        sequence_number = 0
        
        if key in self.perf_insights_result:
            rule_results = self.perf_insights_result[key]
            for rule_result in rule_results:
                rule_Id = ''
                if rule_id_key in rule_result:
                    rule_Id = rule_result[rule_id_key]
                key = 'RuleFindings'
                if key in rule_result.keys():
                    rule_findings = rule_result[key]
                    for rule_finding in rule_findings:
                        key = 'ImpactedResources'                        
                        if key in rule_finding.keys():
                            for impacted_resource in rule_finding[key]:
                                impacted_resource_json = ImpactedResourceJson(
                                    self.perf_insights_result['ReportId'],
                                    rule_Id,
                                    rule_finding['Id'],
                                    sequence_number,
                                    impacted_resource)                                            

                                impacted_resources_json.append(impacted_resource_json)
                                sequence_number += 1

        return impacted_resources_json

    def as_entity_list(self):

        entities = []

        for impacted_resource in self.impacted_resources_result:
            entity = impacted_resource.as_entity()
            entities.append(entity)

        return entities

class ImpactedResourceJson():
    def __init__(self, report_id, rule_id, finding_id, sequence_number, impacted_resource):
        self.partition_key = report_id.lower() + '.' + str(rule_id).lower() + '-' + str(finding_id).lower()
        self.row_key = str(sequence_number).lower()

        self.resource_name = impacted_resource['ResourceName'] or ''
        self.resource_value = impacted_resource['ResourceValue'] or ''
        
        configurations = impacted_resource['Configurations'] or []
        self.configurations = json.dumps(configurations)

    def as_entity(self):
        entity = {}

        entity['PartitionKey'] = self.partition_key
        entity['RowKey'] = self.row_key
        entity['ResourceName'] = self.resource_name
        entity['ResourceValue'] = self.resource_value
        entity['Configurations'] = self.configurations

        return entity

def batch_insert_storage_entities(table_service, table_name, entities):
    entity_count = 0

    while entity_count < len(entities):
        table_service.insert_entity(table_name, entities[entity_count])
        entity_count += 1

def encode_partition_key(partition_key):
    encoded_partition_key = ""
    if common.PY3:
        # pylint: disable=no-member
        encoded_partition_key = urllib.parse.quote(partition_key, safe='').replace("%", "-~-")
    else:
        encoded_partition_key = urllib.quote(partition_key, safe='').replace("%", "-~-")
    
    return encoded_partition_key